/* eslint-disable no-param-reassign */
import { ref, Ref } from '@vue/composition-api';
import Vue from 'vue';

import Track, { PoseAttributes } from 'vue-media-annotator/track';
import Recipe, { UpdateResponse } from 'vue-media-annotator/recipe';
import { EditAnnotationTypes } from 'vue-media-annotator/layers';
import { Mousetrap } from 'vue-media-annotator/types';

const EmptyResponse: UpdateResponse = { data: {}, union: [], unionWithoutBounds: [] };

export default class Pose implements Recipe {
  active: Ref<boolean>;

  name: string;

  bus: Vue;

  toggleable: Ref<boolean>;

  icon: Ref<string>;

  constructor() {
    this.bus = new Vue();
    this.active = ref(false);
    this.name = 'Pose';
    this.toggleable = ref(true);
    this.icon = ref('mdi-vector-line');
  }


  /**
   * findBounds computes a padding polygon around the linestring given paddingVector
   * @param ls list of pose points
   * @param paddingVector polypoints in terms of C and CPerp
   */
  private static findBounds(
    ls: Record<string, GeoJSON.Position>,
  ): GeoJSON.Polygon[] {
    // Coords = [ Vec A, Vec B ]
    const coords: GeoJSON.Position[] = [];
    Object.values(ls).forEach((item) => {
      coords.push(item);
    });
    // If only 1 point is available so far
    return [{
      type: 'Polygon',
      coordinates: coords.map((p) => ([
        p.map((c) => c),
        p.map((c) => c),
      ])),
    }];
  }

  update(
    mode: 'in-progress' | 'editing',
    frameNum: number,
    track: Track,
    data: GeoJSON.Feature<GeoJSON.LineString | GeoJSON.Polygon | GeoJSON.Point>[],
    key?: string,
  ) {
    const linestrings = data.filter((d) => d.geometry.type === 'Point');
    if (linestrings.length && key !== '') {
      const linestring = linestrings[0] as GeoJSON.Feature<GeoJSON.Point>;
      if (this.active.value && mode === 'in-progress') {
        /**
       * IF the recipe is active, we are creating a new headtail
       */
        if (key) {
          const [feature] = track.getFeature(frameNum);
          if (feature && feature.attributes && feature.attributes.poses) {
            // eslint-disable-next-line no-param-reassign
            (feature.attributes.poses as PoseAttributes)
              .points[key] = linestring.geometry.coordinates;
            track.setFeatureAttribute(frameNum, 'poses', feature.attributes.poses);
          } else {
            const poses = {
              points: {
              },
              lines: [[]],
            };
            poses.points[key] = linestring.geometry.coordinates;
            track.setFeatureAttribute(frameNum, 'poses', poses);
          }
        }
        return {
          ...EmptyResponse,
          done: true,
          newSelectedKey: key,
          union: Pose.findBounds((track.attributes.poses as PoseAttributes)
            .points),

        };
      }
      if (mode === 'editing') {
      /**
       * IF recipe isn't active, but the key matches, we are editing
       */
        if (key) {
          const [feature] = track.getFeature(frameNum);
          if (feature && feature.attributes && feature.attributes.poses) {
            // eslint-disable-next-line no-param-reassign
            (feature.attributes.poses as PoseAttributes)
              .points[key] = linestring.geometry.coordinates;
            track.setFeatureAttribute(frameNum, 'poses', feature.attributes.poses);
          } else {
            const poses = {
              points: {
              },
              lines: [[]],
            };
            poses.points[key] = linestring.geometry.coordinates;
            track.setFeatureAttribute(frameNum, 'poses', poses);
          }
        }

        const [feature] = track.getFeature(frameNum);
        if (feature && feature.attributes && feature.attributes.poses) {
          return {
            ...EmptyResponse,
            done: true,
            newSelectedKey: key,
            union: feature.attributes.diva_id ? [] : Pose.findBounds(
              (feature.attributes.poses as PoseAttributes).points,
            ),

          };
        }
      }
    }
    return EmptyResponse;
  }

  // eslint-disable-next-line class-methods-use-this
  delete(frame: number, track: Track, key: string) {
    const [feature] = track.getFeature(frame);
    if (feature && feature.attributes && feature.attributes.poses) {
      // eslint-disable-next-line no-param-reassign
      delete feature.attributes.poses;
    }
    if (key === 'rightShoulder') {
      this.selectKey('leftShoulder');
    } else {
      this.selectKey('rightShoulder');
    }
  }

  // eslint-disable-next-line class-methods-use-this
  deletePoint(frame: number, track: Track, idx: number, key: string) {
    const [feature] = track.getFeature(frame);
    if (feature && feature.attributes && feature.attributes.poses) {
      // eslint-disable-next-line no-param-reassign
      delete (feature.attributes.poses as PoseAttributes).points[key];
      track.setFeatureAttribute(frame, 'poses', feature.attributes.poses);
    }
  }

  selectKey(key: string) {
    this.bus.$emit('activate', {
      editing: 'Point' as EditAnnotationTypes,
      key,
      recipeName: this.name,
    });
  }

  activate() {
    this.active.value = true;
    this.icon.value = 'mdi-vector-line';
    this.bus.$emit('activate', {
      editing: 'Point' as EditAnnotationTypes,
      key: 'rightShoulder',
      recipeName: this.name,
    });
  }

  deactivate() {
    this.active.value = false;
  }

  mousetrap(): Mousetrap[] {
    return [
      { bind: 'escape', handler: () => this.deactivate },
    ];
  }
}
