import type { GirderModel } from '@girder/components/src';

import {
  calibrationFileTypes, fileVideoTypes, inputAnnotationFileTypes,
  inputAnnotationTypes, otherImageTypes,
  otherVideoTypes, websafeImageTypes, websafeVideoTypes, zipFileTypes,
} from 'dive-common/constants';

import {
  DatasetMetaMutable,
  Pipe, Pipelines, SaveAttributeArgs, MevaFilter, DatasetType,
} from 'dive-common/apispec';
import girderRest from '../plugins/girder';

interface HTMLFile extends File {
  webkitRelativePath?: string;
}
interface ValidationResponse {
  ok: boolean;
  type: 'video' | 'image-sequence';
  media: string[];
  annotations: string[];
  message: string;
}
export interface BrandData {
  vuetify: unknown;
  favicon: string | null;
  logo: string;
  name: string;
  loginMessage: string;
}

async function getBrandData(): Promise<BrandData> {
  const { data } = await girderRest.get<BrandData>('meva/brand_data');
  return data;
}

async function isOwnerAdmin(id: string): Promise<boolean> {
  const girderLocation = (await girderRest.get(`/folder/${id}`)).data;
  const { creatorId } = girderLocation;
  const { user } = girderRest;
  if (user && (creatorId === user._id || user.admin)) {
    return true;
  }
  return false;
}


function makeMevaFolder({
  folderId, name, fps, type,
}: {
  folderId: string;
  name: string;
  fps: number;
  type: string;
}) {
  return girderRest.post(
    '/folder',
    `metadata=${JSON.stringify({
      fps,
      type,
    })}`,
    {
      params: { parentId: folderId, name },
    },
  );
}

function deleteResources(resources: Array<GirderModel>) {
  const formData = new FormData();
  formData.set(
    'resources',
    JSON.stringify({
      folder: resources
        .filter((resource) => resource._modelType === 'folder')
        .map((resource) => resource._id),
      item: resources
        .filter((resource) => resource._modelType === 'item')
        .map((resource) => resource._id),
    }),
  );
  return girderRest.post('resource', formData, {
    headers: { 'X-HTTP-Method-Override': 'DELETE' },
  });
}

async function getPipelineList() {
  const { data } = await girderRest.get<Pipelines>('meva/pipelines');
  return data;
}

function runPipeline(itemId: string, pipeline: Pipe) {
  return girderRest.post('/meva/pipeline', null, {
    params: {
      folderId: itemId,
      pipeline,
    },
  });
}

function saveMetadata(folderId: string, metadata: DatasetMetaMutable) {
  return girderRest.put(
    `/meva/metadata/${folderId}`,
    { ...metadata },
  );
}

function saveAttributes(folderId: string, args: SaveAttributeArgs) {
  return girderRest.put('/meva/attributes', {
    upsert: args.upsert,
    delete: args.delete,
  }, {
    params: { folderId },
  });
}

async function importAnnotation(parentId: string, file: HTMLFile) {
  const resp = await girderRest.post('/file', null, {
    params: {
      parentType: 'folder',
      parentId,
      name: file.name,
      size: file.size,
      mimeType: file.type,
    },
  });
  if (resp.status === 200) {
    const uploadResponse = await girderRest.post('file/chunk', file, {
      params: {
        uploadId: resp.data._id,
        offset: 0,
      },
      headers: { 'Content-Type': 'application/octet-stream' },
    });
    if (uploadResponse.status === 200) {
      const final = await girderRest.post(`meva/postprocess/${parentId}`, null, {
        params: {
          skipJobs: true,
        },
      });
      if (final.status === 200) {
        return true;
      }
    }
  }
  return false;
}


function postProcess(folderId: string) {
  return girderRest.post(`meva/postprocess/${folderId}`);
}

function approveFolder(folderId: string) {
  return girderRest.post(`meva/extract_track_chips/${folderId}`);
}

async function validateUploadGroup(names: string[]): Promise<ValidationResponse> {
  const { data } = await girderRest.post<ValidationResponse>('meva/validate_files', names);
  return data;
}

async function getValidWebImages(folderId: string) {
  const { data } = await girderRest.get<GirderModel[]>('meva/valid_images', {
    params: { folderId },
  });
  return data;
}

async function getMevaFolders() {
  const { data } = await girderRest.get<GirderModel[]>('meva/meva_folders', {});
  return data;
}


async function getMevaData() {
  const { data } = await girderRest.get<GirderModel[]>('meva/meva_data', {});
  return data;
}

function getMevaFiltered(filters: MevaFilter) {
  return girderRest.get('/meva/meva_filter', {
    params: {
      filters,
    },
  });
}

async function openFromDisk(datasetType: DatasetType | 'calibration' | 'annotation' | 'zip' | 'batch',
  directory = false):
Promise<{ canceled: boolean; filePaths: string[]; fileList?: File[]; root?: string }> {
  const input: HTMLInputElement = document.createElement('input');
  input.type = 'file';
  const baseTypes: string[] = inputAnnotationFileTypes.map((item) => `.${item}`);
  input.multiple = true;
  if (directory) {
    // Webkit directory is necessary for loading a folder
    // eslint-disable-next-line @typescript-eslint/ban-ts-ignore
    //@ts-ignore
    input.webkitdirectory = true;
    input.multiple = false;
  }
  if (datasetType === 'image-sequence') {
    input.accept = baseTypes.concat(websafeImageTypes).concat(otherImageTypes).join(',');
  } else if (datasetType === 'video') {
    input.accept = baseTypes.concat(websafeVideoTypes).concat(otherVideoTypes)
      .concat(fileVideoTypes.map((item) => `.${item}`)).join(',');
  } else if (datasetType === 'calibration') {
    input.accept = calibrationFileTypes.map((item) => `.${item}`).join(',');
  } else if (datasetType === 'annotation') {
    input.accept = inputAnnotationTypes
      .concat(inputAnnotationFileTypes.map((item) => `.${item}`)).join(',');
  } else if (datasetType === 'zip') {
    input.accept = zipFileTypes.join(',');
  } else if (datasetType === 'batch') {
    input.accept = '.json';
  }
  return new Promise((resolve) => {
    input.onchange = (event) => {
      if (event) {
        const { files } = event.target as HTMLInputElement;
        if (files) {
          let fileList = Array.from(files) as HTMLFile[];
          let root;
          // Calculate the root and remove any recursive subdirectories
          if (fileList[0]?.webkitRelativePath !== undefined && directory) {
            root = fileList[0]?.webkitRelativePath.replace(`/${fileList[0].name}`, '');
            let filterType = ['application/octet-stream']
              .concat(inputAnnotationTypes);

            if (datasetType === 'image-sequence') {
              filterType = filterType
                .concat(websafeImageTypes)
                .concat(otherImageTypes);
            } else if (datasetType === 'video') {
              filterType = filterType
                .concat(websafeVideoTypes)
                .concat(otherVideoTypes);
            }
            fileList = fileList.filter((item) => {
              if (item.webkitRelativePath && item.webkitRelativePath.split('/').length > 2) {
                return false;
              }
              return filterType.includes(item.type);
            });
          }
          const response = {
            canceled: !files.length,
            fileList,
            filePaths: fileList.map((item) => item.name),
            root,
          };
          return resolve(response);
        }
      }
      return resolve({
        canceled: true,
        filePaths: [],
      });
    };
    input.click();
  });
}

function beginS3Import(folderId: string, folderList) {
  return girderRest.put(`/meva/s3_import/${folderId}`, folderList);
}

async function createCameraModelFolder(baseId: string) {
  const createActor = await girderRest.post(`meva/${baseId}/create_camera_model_folder`);
  if (createActor.status === 200) {
    return createActor.data;
  }
  return '';
}

async function createActorDescriptorFolder(baseId: string) {
  const createActor = await girderRest.post(`meva/${baseId}/create_actor_descriptor_folder`);
  if (createActor.status === 200) {
    return createActor.data;
  }
  return '';
}


export {
  getBrandData,
  isOwnerAdmin,
  deleteResources,
  getPipelineList,
  makeMevaFolder,
  postProcess,
  runPipeline,
  saveMetadata,
  saveAttributes,
  validateUploadGroup,
  getValidWebImages,
  getMevaData,
  getMevaFiltered,
  approveFolder,
  openFromDisk,
  importAnnotation,
  getMevaFolders,
  beginS3Import,
  createCameraModelFolder,
  createActorDescriptorFolder,
};
