import { SaveAttributeArgs, SaveDetectionsArgs } from 'dive-common/apispec';
import Track, { TrackData } from 'vue-media-annotator/track';
import { RectBounds } from 'vue-media-annotator/utils';
import girderRest from '../plugins/girder';

interface ExportUrlsResponse {
  mediaType: string;
  exportAllUrl: string;
  exportMediaUrl: string;
  exportDetectionsUrl: string;
  currentThresholds: Record<string, number>;
}

export interface ChipData {
  frame: number;
  bounds: RectBounds;
  size: number;
  score: number;
  frameId: string;
  chipId: string;
  pose: Record<string, [number, number]>;
}

interface SerializedTrackstore {
  [key: number]: TrackData;
}

async function getExportUrls(
  id: string,
  excludeBelowThreshold: boolean,
): Promise<ExportUrlsResponse> {
  return (await girderRest.get<ExportUrlsResponse>(`meva_detection/${id}/export`, {
    params: { excludeBelowThreshold },
  })).data;
}

async function loadDetections(folderId: string, formatting = 'track_json') {
  const { data } = await girderRest.get('meva_detection', {
    params: { folderId, formatting },
  });
  return data as { [key: string]: TrackData };
}

async function loadChipData(folderId: string, trackId: string, formatting = 'track_json') {
  const { data } = await girderRest.get('meva_detection/meva_chip', {
    params: { folderId, trackId, formatting },
  });
  return data as ChipData[];
}


async function loadTrackDetections(folderId: string, trackId: string, formatting = 'track_json') {
  const { data } = await girderRest.get('meva_detection/meva_track', {
    params: { folderId, trackId, formatting },
  });
  const trackJSON = data as Track;
  if (trackJSON) {
    const id = trackJSON.trackId;
    const tracks = {
    };
    tracks[id] = trackJSON;
    return tracks as { [key: string]: TrackData };
  }
  return {} as { [key: string]: TrackData};
}

async function saveDetections(folderId: string, args: SaveDetectionsArgs) {
  return girderRest.put('meva_detection', {
    upsert: args.upsert,
    delete: args.delete,
  }, {
    params: { folderId },
  });
}

async function saveAttributes(folderId: string, args: SaveAttributeArgs) {
  return girderRest.put('meva_attribute', {
    upsert: args.upsert,
    delete: args.delete,
  }, {
    params: { folderId },
  });
}


interface ClipMetaResponse {
  videoUrl: string;
}

async function getClipMeta(folderId: string) {
  const { data } = await girderRest.get('meva_detection/clip_meta', {
    params: { folderId },
  });
  return data as ClipMetaResponse;
}
interface TrackDisplayData {
  type: 'track';
  folderId: string;
  trackId: string;
}

async function getTrackImages(trackData: TrackDisplayData) {
  const { data } = await girderRest.get('meva_detection/meva_chip_images', {
    params: { folderId: trackData.folderId, trackId: trackData.trackId },
  });
  return data;
}
async function getPersonCount(folderId: string) {
  const { data } = await girderRest.get('meva_detection/person_count', {
    params: { folderId },
  });
  return data;
}

export {
  getClipMeta,
  getExportUrls,
  loadDetections,
  loadTrackDetections,
  loadChipData,
  saveDetections,
  saveAttributes,
  getTrackImages,
  getPersonCount,
};
