import { StringKeyObject } from 'vue-media-annotator/track';
import girderRest from '../plugins/girder';


const baseEndpoint = 'meva_state';
interface MevaFilter {
  date?: string;
  startTime?: string;
  location?: string;
  cameraId?: string;
  assigned?: string;
}
interface MevaFilters {
  dates: string[];
  times: string[];
  locations: string[];
  cameraIds: string[];
  users: string[];
}
interface MEVAMetadata {
    date: string;
    startTime: string;
    endTime: string;
    location: string;
    cameraId: string;
    thumbnails?: string[];
}
interface MEVAState {
    Imported: boolean;
    Cleaning: boolean;
    Reviewing: boolean;
    Linking: boolean;
    Complete: boolean;
}
interface MEVAData {
    beginTracks: number;
    endTracks: number;
    cleanTime: number;
    cleanDuration: number;
    reviewTime: number;
    reviewDuration: number;
    linkDuration: number;
    linkTime: number;
}
interface MEVAUser {
    id: string;
    name: string;
}
interface MEVAAssignment {
    Cleaner?: MEVAUser;
    Reviewer?: MEVAUser;
    Linker?: MEVAUser;
}
export interface MEVAFullState {
    folderId: string;
    owner: string;
    owner_name: string;
    locked: boolean;
    created: string;
    metadata: MEVAMetadata;
    MEVA_State: MEVAState;
    MEVA_Data: MEVAData;
    MEVA_Assignment: MEVAAssignment;
}

export interface StateSubInfo {
  'totalVideos': number;
  'totalCleanedVideos': number;
  'totalCleanedTracks': number;
  'totalVideosInCleaning': number;
  'totalVideosInReview': number;
  'totalVideosAwaitingAssignment': number;
}
export interface StateInfo {
  '2018-05': StateSubInfo;
  '2018-03': StateSubInfo;
}

export interface FullStateResults {
  pageResults: MEVAFullState[];
  totalPages: number;
}

export interface BatchJSON {
  filename: string;
  videoURL: string;
  trackURL?: string;
}
async function getMevaFiltered(
  id: string,
  filters: MevaFilter = {},
  offset = 0,
  limit = 50,
  sort = 'created',
): Promise<FullStateResults> {
  return (await girderRest.get<FullStateResults>(`${baseEndpoint}/${id}/all_meva`, {
    params: {
      filters, offset, limit, sort,
    },
  })).data;
}

async function getMevaFilters(id: string) {
  return (await girderRest.get<MevaFilters>(`${baseEndpoint}/${id}/meva_filter`, {})).data;
}

async function setMevaAssignment(
  folderId: string,
  role: string,
  userId: string,
) {
  return (girderRest.post(`${baseEndpoint}/${folderId}/set_assignment`, null, {
    params: {
      role,
      userId,
    },
  }));
}

async function setStateTransition(
  folderId: string,
  state: string,
  loggedTime: string,
) {
  return (girderRest.post(`${baseEndpoint}/${folderId}/set_state`, null, {
    params: {
      state,
      time: loggedTime,
    },
  }));
}

async function setMevaData(
  folderId: string,
  data: StringKeyObject,
) {
  return (girderRest.put(`${baseEndpoint}/${folderId}/set_mevadata`, data));
}

async function getExportVideos(
  folderId: string,
  filters: MevaFilter = {},
  offset = 0,
  limit = 50,
  sort = 'created',
) {
  return (await girderRest.post<FullStateResults>(`${baseEndpoint}/${folderId}/export_video_and_tracks`, {
    params: {
      filters, offset, limit, sort,
    },
  })).data;
}

async function getStateInfo(
  folderId: string,
) {
  return (await girderRest.get<StateInfo>(`${baseEndpoint}/${folderId}/get_state_info`, {
    params: {},
  })).data;
}

async function runStats(
  folderId: string,
  month: '2018-05' | '2018-03' | 'both',
) {
  if (month === 'both') {
    return Promise.all([
      (girderRest.post(`${baseEndpoint}/${folderId}/calculate_stats`, null, {
        params: {
          month: '2018-03',
        },
      })),
      (girderRest.post(`${baseEndpoint}/${folderId}/calculate_stats`, null, {
        params: {
          month: '2018-05',
        },
      })),
    ]);
  }
  return (girderRest.post(`${baseEndpoint}/${folderId}/calculate_stats`, null, {
    params: {
      month,
    },
  }));
}

export interface MevaStats {
  'contains_diva_count': number;
  'diva_detection_count': number;
  'diva_person_count': number;
  'ignore_below_size': number;
  'ignore_with_large_size': number;
  'non_diva_person_count': number;
  'person_track_count': number;
  'total_detection_count': number;
  'total_track_count': number;
  'unique_detection_ids': number;
  'unique_person_ids': number;
}

async function getStats(
  folderId: string,
) {
  return (await girderRest.get<Record<string, MevaStats>>(`${baseEndpoint}/${folderId}/get_stats`, {
    params: {},
  })).data;
}

async function exportTracksJSON(
  folderId: string,
) {
  return (await girderRest.post<Record<string, MevaStats>>(`${baseEndpoint}/${folderId}/export_final_tracks`, {
    params: {},
  })).data;
}

export {
  getMevaFiltered,
  getMevaFilters,
  setMevaAssignment,
  setStateTransition,
  setMevaData,
  getExportVideos,
  getStateInfo,
  runStats,
  getStats,
  exportTracksJSON,
};
