import { MevaFilter } from 'dive-common/apispec';
import girderRest from '../plugins/girder';


const baseEndpoint = 'meva_database';

export interface ActorOutfit {
  actor: string;
  outfit: string;
  gps: string;
  date: string;
  back: string;
  front: string;

}
export interface ActorFilterResults {
  actors: string[];
  dates: string[];
}
export interface ActorFilter {
  actor?: string;
  date?: string;
}

export interface LinkedTracks {
    pageResults: ActorLinks[];
    totalPages: number;
}

export interface ActorLinks {
  actorId: string;
  date: string;
  outfitId: string;
  links: LinkedTrack[];
}
export interface LinkedTrack {
    folderId: string;
    trackId: string;
    folderName: string;
    response?: number;
}

export interface TrackDisplayData {
  type: 'track';
  folderId: string;
  trackId: string;
  state?: 'positive' | 'negative' | 'unknown' | '';
  response?: number;
}
export interface ActorData {
    type: 'actor';
    actorId: string;
    date: string;
    outfitId: string;
    state?: 'positive' | 'negative' | 'unknown' | '';
    links?: TrackDisplayData[];
}


async function startTrackLinking(
  folderId?: string,
  trackId?: string,
  offset = 0,
  pageSize = 10,
): Promise<LinkedTracks> {
  return (await girderRest.get<LinkedTracks>(`${baseEndpoint}/start_track_linking`, {
    params: {
      folderId, trackId, offset, pageSize,
    },
  })).data;
}

async function getPage(
  sourceUID: string,
  offset = 0,
  pageSize = 10,
  filter: string[] = [],
): Promise<LinkedTracks> {
  return (await girderRest.get<LinkedTracks>(`${baseEndpoint}/get_page`, {
    params: {
      sourceUID, offset, pageSize, filter,
    },
  })).data;
}

async function getLinkedTracks(
  offset = 0,
  pageSize = 20,
  filters: MevaFilter,
  linkedState = 'All',
): Promise<LinkedTracks> {
  let url = `${baseEndpoint}/get_linked_tracks?offset=${offset}&pageSize=${pageSize}`;
  if (linkedState === 'Confirmed') {
    url += '&confirmed=true';
  } else if (linkedState === 'Unconfirmed') {
    url += '&confirmed=false';
  }
  return (await girderRest.get<LinkedTracks>(url, {
    params: {
      filters,
    },
  })).data;
}

async function getActorLinks(
  actorId: string,
  date: string,
  outfitId: string,
  filters?: MevaFilter,
): Promise<RandomResults> {
  return (await girderRest.get<RandomResults>(`${baseEndpoint}/get_actor_links?actorId=${actorId}&datestr=${date}&outfitId=${outfitId}`, {
    params: {
      filters,
    },
  })).data;
}


async function getLinkingInfo(
): Promise<Record<string, string>> {
  return (await girderRest.get<Record<string, string>>(`${baseEndpoint}/get_linking_info`)).data;
}

async function getuserCache(): Promise<Record<string, string>> {
  return (await girderRest.get<Record<string, string>>(`${baseEndpoint}/get_user_cache`)).data;
}

async function clearUserCache(userId: string): Promise<boolean> {
  return (await girderRest.post<boolean>(`${baseEndpoint}/force_unload_user_cache?userId=${userId}`)).data;
}

async function addTracks(
  folderId: string,
): Promise<{count: number}> {
  return ((await girderRest.put(`${baseEndpoint}/add_tracks?folderId=${folderId}`)).data);
}

async function removeLink(
  sourceUID: string,
  folderId: string,
  trackId: string,
): Promise<number> {
  const query = `sourceUID=${sourceUID}&folderId=${folderId}&trackId=${trackId}`;
  return (await girderRest.post<LinkedTracks>(`${baseEndpoint}/remove_track_link?${query}`, {
  })).status;
}

async function getActorFilter(): Promise<ActorFilterResults> {
  const results = await girderRest.get(`${baseEndpoint}/get_actor_filter`);
  return results.data;
}

async function getActorChips(
  offset = 0,
  pageSize = 10,
  filters?: ActorFilter,
): Promise<{pageResults: ActorOutfit[]; totalPages: number}> {
  return (await girderRest.get<{pageResults: ActorOutfit[]; totalPages: number}>(`${baseEndpoint}/get_actor_chips?offset=${offset}&pageSize=${pageSize}`, {
    params: {
      filters,
    },
  })).data;
}

async function getActorChip(
  actorId: string,
  datestr: string,
  outfitId: string,
): Promise<ActorOutfit> {
  return (await girderRest.get<ActorOutfit>(`${baseEndpoint}/get_actor_chip`, { params: { actorId, datestr, outfitId } })).data;
}

async function getActorImages(actorData: ActorData): Promise<string[]> {
  return (await girderRest.get<string[]>(`${baseEndpoint}/get_actor_images`, {
    params: { data: actorData },
  })).data;
}

export type LinkData = TrackDisplayData | ActorData;
export interface RandomResults {
  source: LinkData;
  results: LinkData[];
  totalPages: number;
}
async function getRandom(
  offset = 0,
  pageSize = 200,
  runReID = false,
  getAllActorsForTheDay = true,

): Promise<RandomResults> {
  return (await girderRest.get<RandomResults>(`${baseEndpoint}/get_random?offset=${offset}&pageSize=${pageSize}&runReID=${runReID}&getAllActorsForTheDay=${getAllActorsForTheDay}`, {
  })).data;
}

// eslint-disable-next-line @typescript-eslint/no-explicit-any
function isTrackData(x: any): x is TrackDisplayData {
  return x.type === 'track';
}
// eslint-disable-next-line @typescript-eslint/no-explicit-any
function isActorData(x: any): x is ActorData {
  return x.type === 'actor';
}

async function refineActorLinks(
  source: LinkData,
  refinements: LinkData[],
): Promise<RandomResults> {
  return (await girderRest.post<RandomResults>(
    `${baseEndpoint}/refine_actor_links`, {
      params: {
        source, refinements,
      },
    },
  )).data;
}

async function refine(
  source: LinkData,
  refinements: LinkData[],
  pageSize = 200,
  runReID = false,
  getAllActorsForTheDay = true,
): Promise<RandomResults> {
  return (await girderRest.post<RandomResults>(
    `${baseEndpoint}/refine?pageSize=${pageSize}&runReID=${runReID}&getAllActorsForTheDay=${getAllActorsForTheDay}`, {
      params: {
        source, refinements,
      },
    },
  )).data;
}

async function pushDown(
  source: LinkData,
  refinements: LinkData[],
  runReID = false,
  getAllActorsForTheDay = true,
): Promise<RandomResults> {
  return (await girderRest.post<RandomResults>(
    `${baseEndpoint}/push_down?runReID=${runReID}&getAllActorsForTheDay=${getAllActorsForTheDay}`, {
      params: {
        source, refinements,
      },
    },
  )).data;
}

async function importCameraModels(folderId: string) {
  const final = await girderRest.put(`${baseEndpoint}/add_camera_models/${folderId}`, null, {
  });
  if (final.status === 200) {
    return true;
  }
  return false;
}

async function importActorDescriptors(folderId: string) {
  const final = await girderRest.put(`${baseEndpoint}/add_actor_descriptors/${folderId}`, null, {
  });
  if (final.status === 200) {
    return true;
  }
  return false;
}

async function getInputStatus(
): Promise<{cameraModels: number; actorGPS: number; actorChips: number; actorDescriptors: number}> {
  return (await girderRest.get<{cameraModels: number; actorGPS: number; actorChips: number; actorDescriptors: number}>(`${baseEndpoint}/get_input_status`, {
  })).data;
}

async function exportDatabase(folderId: string) {
  return (girderRest.post(
    `${baseEndpoint}/export_database/${folderId}`, {},
  ));
}

async function getDatabaseDownload(folderId: string) {
  return (await girderRest.get<string>(
    `${baseEndpoint}/get_database_download/${folderId}`, {},
  )).data;
}

async function populateTracks(folderId: string) {
  return (girderRest.post(
    `${baseEndpoint}/populate_linked_tracks/${folderId}`, {},
  ));
}

async function populateMNLs() {
  return (girderRest.post(
    `${baseEndpoint}/populate_MNLs/`, {},
  ));
}
async function deleteActorDescriptors() {
  return (girderRest.delete(
    `${baseEndpoint}/delete_actor_descriptors/`, {},
  ));
}

async function downloadLinks(id: string, confirmed = false) {
  return (await girderRest.post(`${baseEndpoint}/export_linked_tracks/${id}?confirmed=${confirmed}`)).data;
}

export {
  startTrackLinking,
  getLinkedTracks,
  getLinkingInfo,
  getPage,
  refine,
  refineActorLinks,
  pushDown,
  removeLink,
  addTracks,
  getActorFilter,
  getActorChips,
  getActorChip,
  getActorImages,
  getActorLinks,
  getRandom,
  isTrackData,
  isActorData,
  importCameraModels,
  getInputStatus,
  exportDatabase,
  getDatabaseDownload,
  populateTracks,
  populateMNLs,
  importActorDescriptors,
  deleteActorDescriptors,
  downloadLinks,
  getuserCache,
  clearUserCache,
};
