/* eslint-disable class-methods-use-this */
import geo, { GeoEvent } from 'geojs';

import BaseLayer, { BaseLayerParams, LayerStyle } from '../BaseLayer';
import { FrameDataTrack } from '../LayerTypes';

interface PointGeoJSData {
    trackId: string;
    selected: boolean;
    editing: boolean | string;
    trackType: [string, number] | null;
    feature: string;
    radius?: number;
    x: number;
    y: number;
}

export default class PointLayer extends BaseLayer<PointGeoJSData> {
  editPointMode: boolean;


  constructor(params: BaseLayerParams) {
    super(params);
    this.editPointMode = true;
    this.initialize();
  }

  initialize() {
    const layer = this.annotator.geoViewerRef.value.createLayer('feature', {
      features: ['point'],
    });
    this.featureLayer = layer.createFeature('point')
      .geoOn(geo.event.feature.mouseclick, (e: GeoEvent) => {
      /**
       * Handle clicking on individual annotations, if DrawingOther is true we use the
       * Rectangle type if only the polygon is visible we use the polygon bounds
       * */
        if (e.mouse.buttonsDown.right) {
          if ((e.data.editing && this.editPointMode && !e.data.radius)) {
            this.bus.$emit('point-clicked', e.data.feature, false);
          }
        }
      });
    this.featureLayer.geoOn(geo.event.feature.mouseover,
      (e: GeoEvent) => this.hoverEditHandle(e));

    this.featureLayer.geoOn(geo.event.zoom,
      (e: GeoEvent) => this.redraw());

    super.initialize();
  }

  hoverEditHandle(e: GeoEvent) {
    if (e.data.editing && this.editPointMode) {
      this.annotator.setCursor('pointer');
    } else {
      this.annotator.setCursor('default');
    }
  }

  // eslint-disable-next-line class-methods-use-this
  formatData(frameData: FrameDataTrack[]): PointGeoJSData[] {
    const arr: PointGeoJSData[] = []; //(this.checkHeadTail(frameData));
    frameData.forEach((track: FrameDataTrack) => {
      if (track.features && track.features.bounds) {
        if (track.features.geometry?.features?.[0]) {
          track.features.geometry.features.forEach((feature) => {
            if (feature.geometry && feature.geometry.type === 'Point') {
              const [x, y] = feature.geometry.coordinates;
              let key = 'point';
              let radius: number | undefined;
              if (feature.properties && feature.properties.key) {
                key = feature.properties.key;
                radius = feature.properties.radius;
              }
              const annotation: PointGeoJSData = {
                trackId: track.trackId,
                selected: track.selected,
                editing: track.editing,
                trackType: track.trackType,
                feature: key,
                radius,
                x,
                y,
              };
              arr.push(annotation);
            }
          });
        }
      }
    });
    return arr;
  }

  pointRadiustoScreen(radius) {
    const base = this.annotator.geoViewerRef.value.gcsToDisplay(
      { x: 0, y: 0 },
    );
    const val = this.annotator.geoViewerRef.value.gcsToDisplay(
      { x: radius, y: radius },
    );
    return Math.sqrt((val.x - base.x, 2) ** 2 + (val.y - base.y) ** 2);
  }

  createStyle(): LayerStyle<PointGeoJSData> {
    return {
      ...super.createStyle(),
      fill: true,
      fillColor: (data: PointGeoJSData) => {
        if (data.radius && data.trackType) {
          return this.typeStyling.value.color(data.trackType[0]);
        }
        return this.typeStyling.value.color(data.feature);
      },
      radius: (data: PointGeoJSData) => {
        if (data.radius) {
          return this.pointRadiustoScreen(data.radius);
        }

        if (data.selected) {
          return this.stateStyling.selected.strokeWidth * 1.25;
        }
        if (data.trackType) {
          return this.typeStyling.value.strokeWidth(data.trackType[0]) * 1;
        }
        return this.stateStyling.standard.strokeWidth * 2;
      },
    };
  }

  redraw(): null {
    return this.featureLayer.data(this.formattedData).draw();
  }

  disable() {
    this.featureLayer
      .data([])
      .draw();
  }
}
