import BaseLayer, { BaseLayerParams, LayerStyle } from './BaseLayer';
import { FrameDataTrack } from './LayerTypes';

export interface TextData {
  selected: boolean;
  editing: boolean | string;
  type: string;
  confidence: number;
  text: string;
  x: number;
  y: number;
  offsetY?: number;
  offsetX?: number;
  currentPair: boolean;
  DIVA_Track?: boolean;
}

export type FormatTextRow =
(track: FrameDataTrack, maxPairs?: number, lineHeight?: number) => TextData[] | null;

interface TextLayerParams {
  formatter?: FormatTextRow;
}

/**
 * @param track - standard frameDataTrack info
 * @param maxPairs - maximum number of lines to show
 * @param lineHeight - height of each text line
 * @returns value or null.  null indicates that the text should not be displayed.
 */
function defaultFormatter(track: FrameDataTrack, maxPairs = 1): TextData[] | null {
  if (track.features && track.features.bounds) {
    const { bounds } = track.features;
    if (bounds && track.confidencePairs !== null) {
      const arr: TextData[] = [];
      let totalVisiblePairs = Math.min(track.confidencePairs.length, maxPairs);
      // eslint-disable-next-line dot-notation
      if (track.features?.attributes && track.features.attributes['diva_id']) {
        totalVisiblePairs += 1;
        const [type, confidence] = ['DIVA', 1.0];
        arr.push({
          selected: track.selected,
          editing: track.editing,
          type,
          confidence,
          text: `${type}`,
          x: bounds[2],
          y: -1, // updated below
          currentPair: false,
        });
      }
      for (let i = 0; i < track.confidencePairs.length; i += 1) {
        const [type, confidence] = track.confidencePairs[i];
        const isCurrentPair = (type === track.trackType[0]);
        const currentTypeIndication = (isCurrentPair && totalVisiblePairs > 1) ? '**' : '';
        arr.push({
          selected: track.selected,
          editing: track.editing,
          type,
          confidence,
          text: `${currentTypeIndication}${type}: ${confidence.toFixed(2)}`,
          x: bounds[2],
          y: -1, // updated below
          currentPair: isCurrentPair,
        });
      }
      return arr
        .sort((a, b) => (+b.currentPair) - (+a.currentPair)) // sort currentPair=true first
        .slice(0, totalVisiblePairs)
        .map((v) => ({ ...v, y: bounds[1] })); // calculate y after sort
    }
  }
  return null;
}

export default class TextLayer extends BaseLayer<TextData> {
  formatter: FormatTextRow;

  constructor(params: BaseLayerParams & TextLayerParams) {
    super(params);
    this.formatter = params.formatter || defaultFormatter;
  }

  initialize() {
    const layer = this.annotator.geoViewerRef.value.createLayer('feature', {
      features: ['text'],
    });
    this.featureLayer = layer
      .createFeature('text')
      .text((data: TextData) => data.text)
      .position((data: TextData) => ({ x: data.x, y: data.y }));
    super.initialize();
  }

  formatData(frameData: FrameDataTrack[]) {
    const arr = [] as TextData[];
    frameData.forEach((track: FrameDataTrack) => {
      const formatted = this.formatter(track, 1, 15);
      if (formatted !== null) {
        arr.push(...formatted);
      }
    });
    return arr;
  }

  redraw() {
    this.featureLayer.data(this.formattedData).draw();
    return null;
  }

  disable() {
    this.featureLayer.data([]).draw();
  }

  createStyle(): LayerStyle<TextData> {
    const baseStyle = super.createStyle();
    return {
      ...baseStyle,
      rotation: (data) => {
        if (data.type === 'DIVA') {
          return 90 * (Math.PI / 180);
        }
        return 0;
      },
      color: (data) => {
        if (data.editing || data.selected) {
          if (!data.selected && !data.currentPair) {
            if (this.stateStyling.disabled.color !== 'type') {
              return this.stateStyling.disabled.color;
            }
            return this.typeStyling.value.color(data.type);
          }
          if (data.selected && data.currentPair) {
            return this.stateStyling.selected.color;
          }
          if (data.type === 'DIVA') {
            return '#FF0000';
          }
          return this.typeStyling.value.color(data.type);
        }
        if (data.type === 'DIVA') {
          return '#BB0000';
        }
        return this.typeStyling.value.color(data.type);
      },
      textOpacity: ((data) => {
        if (data.currentPair || data.type === 'DIVA') {
          return 1.0;
        }
        return this.stateStyling.disabled.opacity;
      }),
      offset: (data) => {
        if (data.type !== 'DIVA') {
          return {
            x: data.offsetY || 3,
            y: data.offsetX || -8,
          };
        }
        return {
          x: data.offsetX || 2,
          y: data.offsetY || 25,
        };
      },
    };
  }
}
