import * as posenet from '@tensorflow-models/posenet';
import '@tensorflow/tfjs-backend-webgl';
import { TrackSupportedFeature } from 'vue-media-annotator/track';
import { RectBounds } from 'vue-media-annotator/utils';

export interface PoseNetSettings {
    show: boolean;
    advanced: boolean;
    input: {
    architecture: 'ResNet50' | 'MobileNetV1';
    outputStride: 32 | 16 | 8;
    inputResolution: number | { width: number; height: number};
    multiplier: 0.50 | 0.75 | 1.0;
    quantBytes: 1 | 2 | 4;
    };
    multiPoseDetection: {
        maxPoseDetections: number;
        minPoseConfidence: number;
        minPartConfidence: number;
        nmsRadius: number;
    };
    jointFilter: number;
}

export interface PosesType {
    bounds: RectBounds;
    features: GeoJSON.FeatureCollection<TrackSupportedFeature>;
    score: number;
    points?: Record<string, [number, number]>;
}


export default function PoseNetController() {
  const defaultSettings: PoseNetSettings = {
    show: false,
    advanced: false,
    input: {
      architecture: 'ResNet50',
      outputStride: 16,
      inputResolution: 200,
      multiplier: 1.00,
      quantBytes: 2,
    },
    multiPoseDetection: {
      maxPoseDetections: 15,
      minPoseConfidence: 0.3,
      minPartConfidence: 0.3,
      nmsRadius: 20,
    },
    jointFilter: 0.0,
  };

  let net: posenet.PoseNet;

  const canvas = document.createElement('CANVAS') as HTMLCanvasElement;

  const loadPosenet = async (settings?: PoseNetSettings) => {
    let loadSettings = defaultSettings;
    if (settings !== undefined) {
      loadSettings = settings;
    }
    net = await posenet.load({
      architecture: loadSettings.input.architecture,
      outputStride: loadSettings.input.outputStride,
      inputResolution:
      loadSettings.input.inputResolution,
      multiplier: loadSettings.input.multiplier,
      quantBytes: loadSettings.input.quantBytes,
    });
  };

  const convertPoseToGeoJSON = (
    pose: posenet.Pose,
    settings: PoseNetSettings,
    baseBounds?: RectBounds,
  ) => {
    const bbox = posenet.getBoundingBox(pose.keypoints);
    let basedBounds = [0, 0, 0, 0];
    if (baseBounds) {
      basedBounds = baseBounds;
    }
    const bounds = [
      bbox.minX + basedBounds[0],
      bbox.minY + basedBounds[1],
      bbox.maxX + basedBounds[0],
      bbox.maxY + basedBounds[1],
    ] as RectBounds;
    const adjacentKeyPoints = posenet.getAdjacentKeyPoints(
      pose.keypoints,
      settings.multiPoseDetection.minPartConfidence,
    );

    const features: GeoJSON.FeatureCollection<TrackSupportedFeature> = {
      type: 'FeatureCollection',
      features: [],
    };
    //Now we need to draw multiple geoJSON Line geometries using this
    adjacentKeyPoints.forEach((element) => {
      const feature: GeoJSON.Feature<GeoJSON.LineString> = {
        type: 'Feature',
        properties: {},
        geometry: {
          type: 'LineString',
          coordinates: [
            [element[0].position.x + basedBounds[0], element[0].position.y + basedBounds[1]],
            [element[1].position.x + basedBounds[0], element[1].position.y + basedBounds[1]],
          ],
        },
      };
      features.features.push(feature);
    });
    //Adding in points
    const points = {};
    pose.keypoints.forEach((element) => {
      if (element.score > settings.multiPoseDetection.minPartConfidence) {
        points[element.part] = [
          element.position.x + basedBounds[0],
          element.position.y + basedBounds[1],
        ];
        const feature: GeoJSON.Feature<GeoJSON.Point> = {
          type: 'Feature',
          properties: { key: element.part },
          geometry: {
            type: 'Point',
            coordinates:
                [element.position.x + basedBounds[0], element.position.y + basedBounds[1]]
            ,
          },
        };
        features.features.push(feature);
      }
    });

    return {
      bounds,
      features,
      score: pose.score,
      points,
    };
  };

  const convertToGeoJSON = (poses: posenet.Pose[], settings: PoseNetSettings) => {
    const posesGeo: {
        bounds: RectBounds;
        features: GeoJSON.FeatureCollection<TrackSupportedFeature>;
        score: number;
        points: Record<string, [number, number]>;
    }[] = [];
    poses.forEach((pose) => {
      if (pose.score > settings.multiPoseDetection.minPoseConfidence) {
        const geoJSON = convertPoseToGeoJSON(pose, settings);
        posesGeo.push(geoJSON);
      }
    });
    return posesGeo;
  };

  const convertSinglePoseToGeoJSON = (pose: posenet.Pose,
    settings: PoseNetSettings, bounds: RectBounds) => {
    const posesGeo: {
        bounds: RectBounds;
        features: GeoJSON.FeatureCollection<TrackSupportedFeature>;
        score: number;
        points: Record<string, [number, number]>;
    }[] = [];

    if (pose.score > settings.multiPoseDetection.minPoseConfidence) {
      const geoJSON = convertPoseToGeoJSON(pose, settings, bounds);
      posesGeo.push(geoJSON);
    }
    return posesGeo;
  };

  function sourceToHiddenCanvas(source, x, y, width, height) {
    const hiddenCanvas = canvas;
    const hiddenCtx = hiddenCanvas.getContext('2d');
    if (hiddenCtx) {
      // hiddenCanvas.imageSmoothingEnabled = false;
      hiddenCanvas.width = width;
      hiddenCanvas.height = height;
      hiddenCtx.drawImage(source, x, y, width, height, 0, 0, width, height);
    }
    return hiddenCanvas;
  }


  const calculatePoses = async (
    video: ImageData | HTMLImageElement | HTMLCanvasElement | HTMLVideoElement,
    settings?: PoseNetSettings,
    bounds?: RectBounds,
  ) => {
    let loadSettings = defaultSettings;
    if (settings !== undefined) {
      loadSettings = settings;
    }
    if (!net || (video as HTMLMediaElement).readyState <= 1) {
      return false;
    }
    let el: HTMLVideoElement | HTMLCanvasElement = video as HTMLVideoElement;

    if (!bounds) {
      const allPoses = await net.estimatePoses(el, {
        flipHorizontal: false,
        decodingMethod: 'multi-person',
        maxDetections: loadSettings.multiPoseDetection.maxPoseDetections,
        scoreThreshold: loadSettings.multiPoseDetection.minPartConfidence,
        nmsRadius: loadSettings.multiPoseDetection.nmsRadius,
      });
      return convertToGeoJSON(allPoses, loadSettings);
    }
    if (bounds) {
      const x = bounds[0];
      const y = bounds[1];
      const width = bounds[2] - bounds[0];
      const height = bounds[3] - bounds[1];
      el = sourceToHiddenCanvas(video, x, y, width, height);
    }
    const singlePose = await net.estimateSinglePose(el, {
      flipHorizontal: false,
    });
    return convertSinglePoseToGeoJSON(singlePose, loadSettings, bounds);
  };

  return {
    loadPosenet,
    calculatePoses,
  };
}
