import IntervalTree from '@flatten-js/interval-tree';
import {
  provide, inject, ref, Ref,
} from '@vue/composition-api';

import { CustomStyle, StateStyles, TypeStyling } from './use/useStyling';
import { EditAnnotationTypes } from './layers/EditAnnotationLayer';
import Track, { TrackId } from './track';
import { VisibleAnnotationTypes } from './layers';
import { RectBounds } from './utils';
import { Attribute } from './use/useAttributes';
import { TrackWithContext } from './use/useTrackFilters';
import { PosesType } from './posenet/posenet';

/**
 * Type definitions are read only because injectors may mutate internal state,
 * but should never overwrite or delete the injected object.
 */

const AttributesSymbol = Symbol('attributes');
type AttributesType = Readonly<Ref<Attribute[]>>;

const AllTypesSymbol = Symbol('allTypes');
type AllTypesType = Readonly<Ref<readonly string[]>>;

const DatasetIdSymbol = Symbol('datasetID');
type DatasetIdType = Readonly<Ref<string>>;

const UsedTypesSymbol = Symbol('usedTypes');
type UsedTypesType = Readonly<Ref<readonly string[]>>;

const CheckedTrackIdsSymbol = Symbol('checkedTrackIds');
type CheckedTrackIdsType = Readonly<Ref<readonly TrackId[]>>;

const CheckedTypesSymbol = Symbol('checkedTypes');
type CheckedTypesType = Readonly<Ref<readonly string[]>>;

const EnabledTracksSymbol = Symbol('enabledTracks');
type EnabledTracksType = Readonly<Ref<readonly TrackWithContext[]>>;

const EditingModeSymbol = Symbol('editingMode');
type EditingModeType = Readonly<Ref<false | EditAnnotationTypes>>;

const FrameSymbol = Symbol('frame');
type FrameType = Readonly<Ref<number>>;

const MergeListSymbol = Symbol('mergeList');
type MergeList = Readonly<Ref<readonly TrackId[]>>;


const IntervalTreeSymbol = Symbol('intervalTree');
type IntervalTreeType = Readonly<IntervalTree>;

const TrackMapSymbol = Symbol('trackMap');
type TrackMapType = Readonly<Map<TrackId, Track>>;

const TracksSymbol = Symbol('tracks');
type FilteredTracksType = Readonly<Ref<readonly TrackWithContext[]>>;

const TypeStylingSymbol = Symbol('typeStyling');
type TypeStylingType = Readonly<Ref<TypeStyling>>;

const SelectedKeySymbol = Symbol('selectedKey');
type SelectedKeyType = Readonly<Ref<string>>;

const SelectedTrackIdSymbol = Symbol('selectedTrackId');
type SelectedTrackIdType = Readonly<Ref<TrackId | null>>;

const StateStylesSymbol = Symbol('stateStyles');
type StateStylesType = Readonly<StateStyles>;

const VisibleModesSymbol = Symbol('visibleModes');
type VisibleModesType = Readonly<Ref<readonly VisibleAnnotationTypes[]>>;

const PosesSymbol = Symbol('poses');
type PosesStateType = Readonly<Ref<readonly PosesType[]>>;

const ShowPosesSymbol = Symbol('showPoses');
type ShowPosesType = Readonly<Ref<boolean>>;

const JointFilterSymbol = Symbol('jointFilter');
type JointFilterType = Readonly<Ref<number>>;

/**
 * Handler interface describes all global events mutations
 * for above state
 */
export interface Handler {
  /* Select and seek to track */
  trackSeek(trackId: TrackId): void;
  /* Toggle editing mode for track */
  trackEdit(trackId: TrackId): void;
  /* set checked track ids */
  trackEnable(trackId: TrackId, value: boolean): void;
  /* toggle selection mode for track */
  trackSelect(trackId: TrackId | null, edit: boolean): void;
  /* select next track in the list */
  trackSelectNext(delta: number): void;
  /* split track */
  trackSplit(trackId: TrackId | null, frame: number): void;
  /* Change tracks difinitive type */
  trackTypeChange(trackId: TrackId | null, value: string): void;
  /* Add new empty track and select it */
  trackAdd(): TrackId;
  /* update Rectangle bounds for track */
  updateRectBounds(
    frameNum: number,
    bounds: RectBounds,
  ): void;
  /* update geojson for track */
  updateGeoJSON(
    eventType: 'in-progress' | 'editing',
    frameNum: number,
    data: GeoJSON.Feature,
    key?: string,
    preventInterrupt?: () => void,
  ): void;
  /* Remove a whole track */
  removeTrack(trackIds: TrackId[]): void;
  /* Remove a single point from selected track's geometry by selected index */
  removePoint(): void;
  /* Remove an entire annotation from selected track by selected key */
  removeAnnotation(): void;
  /* set selectFeatureHandle and selectedKey */
  selectFeatureHandle(i: number, key: string): void;
  /* set checked type strings */
  setCheckedTypes(types: string[]): void;
  /* set checked type strings */
  removeTypeTracks(types: string[]): void;
  /* removes an individual type */
  deleteType(types: string): void;
  /* Change type name */
  selectKey(key: string): void;
  /* Change type name */
  updateTypeName({ currentType, newType }: { currentType: string; newType: string }): void;
  /* change styles */
  updateTypeStyle(args: {
    type: string;
    value: CustomStyle;
  }): void;
  /* set an Attribute in the metaData */
  setAttribute({ data, oldAttribute }:
    {data: Attribute; oldAttribute?: Attribute }, updateAllTracks?: boolean): void;
  /* delete an Attribute in the metaData */
  deleteAttribute({ data }: { data: Attribute }, removeFromTracks?: boolean): void;
  /* Commit the staged merge tracks */
  commitMerge(): void;
  /* Turn merge mode on and off */
  toggleMerge(): TrackId[];
  /* Remove trackIds from merge */
  unstageFromMerge(ids: TrackId[]): void;
}
const HandlerSymbol = Symbol('handler');


/**
 * Make a trivial noop handler. Useful if you only intend to
 * override some small number of values.
 * @param handle callbacl for all handler methods
 */
function dummyHandler(handle: (name: string, args: unknown[]) => void): Handler {
  return {
    trackSeek(...args) { handle('trackSeek', args); },
    trackEdit(...args) { handle('trackEdit', args); },
    trackEnable(...args) { handle('trackEnable', args); },
    trackSelect(...args) { handle('trackSelect', args); },
    trackSelectNext(...args) { handle('trackSelectNext', args); },
    trackSplit(...args) { handle('trackSplit', args); },
    trackTypeChange(...args) { handle('trackTypeChange', args); },
    trackAdd(...args) { handle('trackAdd', args); return '0'; },
    updateRectBounds(...args) { handle('updateRectBounds', args); },
    updateGeoJSON(...args) { handle('updateGeoJSON', args); },
    removeTrack(...args) { handle('removeTrack', args); },
    removePoint(...args) { handle('removePoint', args); },
    removeAnnotation(...args) { handle('removeAnnotation', args); },
    selectFeatureHandle(...args) { handle('selectFeatureHandle', args); },
    setCheckedTypes(...args) { handle('setCheckedTypes', args); },
    removeTypeTracks(...args) { handle('removeTypeTracks', args); },
    deleteType(...args) { handle('deleteType', args); },
    selectKey(...args) { handle('selectKey', args); },
    updateTypeName(...args) { handle('updateTypeName', args); },
    updateTypeStyle(...args) { handle('updateTypeStyle', args); },
    setAttribute(...args) { handle('setAttribute', args); },
    deleteAttribute(...args) { handle('deleteAttribute', args); },
    toggleMerge(...args) { handle('toggleMerge', args); return []; },
    commitMerge(...args) { handle('commitMerge', args); },
    unstageFromMerge(...args) { handle('unstageFromMerge', args); },

  };
}

/**
 * State interface describes the reactive properties that should be provided
 * by a vue-media-annotator consumer.
 *
 * Implementations of most of this state is provided via src/use, but some
 * you will need to construct on your own.
 */
export interface State {
  attributes: AttributesType;
  allTypes: AllTypesType;
  datasetId: DatasetIdType;
  usedTypes: UsedTypesType;
  checkedTrackIds: CheckedTrackIdsType;
  checkedTypes: CheckedTypesType;
  editingMode: EditingModeType;
  enabledTracks: EnabledTracksType;
  frame: FrameType;
  intervalTree: IntervalTreeType;
  mergeList: MergeList;
  trackMap: TrackMapType;
  filteredTracks: FilteredTracksType;
  typeStyling: TypeStylingType;
  selectedKey: SelectedKeyType;
  selectedTrackId: SelectedTrackIdType;
  stateStyles: StateStylesType;
  visibleModes: VisibleModesType;
  poses: PosesStateType;
  showPoses: ShowPosesType;
  jointFilter: JointFilterType;
}

/**
 * make a trivial state store. Useful if you only
 * intend to override some small number of values.
 */
function dummyState(): State {
  const style = {
    strokeWidth: 2,
    opacity: 1,
    color: 'white',
    fill: false,
  };
  return {
    attributes: ref([]),
    allTypes: ref([]),
    datasetId: ref(''),
    usedTypes: ref([]),
    checkedTrackIds: ref([]),
    checkedTypes: ref([]),
    editingMode: ref(false),
    enabledTracks: ref([]),
    frame: ref(0),
    intervalTree: new IntervalTree(),
    mergeList: ref([]),
    trackMap: new Map<TrackId, Track>(),
    filteredTracks: ref([]),
    typeStyling: ref({
      color() { return style.color; },
      strokeWidth() { return style.strokeWidth; },
      opacity() { return style.opacity; },
      fill() { return style.fill; },
    }),
    selectedKey: ref(''),
    selectedTrackId: ref(null),
    stateStyles: {
      disabled: style,
      selected: style,
      standard: style,
    },
    visibleModes: ref(['rectangle', 'text'] as VisibleAnnotationTypes[]),
    poses: ref([]),
    showPoses: ref(false),
    jointFilter: ref(0.0),
  };
}

/**
 * Provide global state and handler for a single instance
 * of vue-media-annotator.  Multiple annotator windows
 * are currently not supported.
 *
 * @param {State} state
 * @param {Hander} handler
 */
function provideAnnotator(state: State, handler: Handler) {
  provide(AttributesSymbol, state.attributes);
  provide(AllTypesSymbol, state.allTypes);
  provide(DatasetIdSymbol, state.datasetId);
  provide(UsedTypesSymbol, state.usedTypes);
  provide(CheckedTrackIdsSymbol, state.checkedTrackIds);
  provide(CheckedTypesSymbol, state.checkedTypes);
  provide(EnabledTracksSymbol, state.enabledTracks);
  provide(EditingModeSymbol, state.editingMode);
  provide(FrameSymbol, state.frame);
  provide(IntervalTreeSymbol, state.intervalTree);
  provide(MergeListSymbol, state.mergeList);
  provide(TrackMapSymbol, state.trackMap);
  provide(TracksSymbol, state.filteredTracks);
  provide(TypeStylingSymbol, state.typeStyling);
  provide(SelectedKeySymbol, state.selectedKey);
  provide(SelectedTrackIdSymbol, state.selectedTrackId);
  provide(StateStylesSymbol, state.stateStyles);
  provide(VisibleModesSymbol, state.visibleModes);
  provide(HandlerSymbol, handler);
  provide(PosesSymbol, state.poses);
  provide(ShowPosesSymbol, state.showPoses);
  provide(JointFilterSymbol, state.jointFilter);
}

function _handleMissing(s: symbol): Error {
  return new Error(`Missing provided object for symbol ${s.toString()}: must provideAnnotator()`);
}

function use<T>(s: symbol) {
  const v = inject<T>(s);
  if (v === undefined) {
    throw _handleMissing(s);
  }
  return v;
}

function useAttributes() {
  return use<AttributesType>(AttributesSymbol);
}

function useAllTypes() {
  return use<AllTypesType>(AllTypesSymbol);
}
function useDatasetId() {
  return use<DatasetIdType>(DatasetIdSymbol);
}
function useUsedTypes() {
  return use<UsedTypesType>(UsedTypesSymbol);
}

function useCheckedTrackIds() {
  return use<CheckedTrackIdsType>(CheckedTrackIdsSymbol);
}

function useCheckedTypes() {
  return use<CheckedTypesType>(CheckedTypesSymbol);
}

function useEnabledTracks() {
  return use<EnabledTracksType>(EnabledTracksSymbol);
}

function useEditingMode() {
  return use<EditingModeType>(EditingModeSymbol);
}

function useFrame() {
  return use<FrameType>(FrameSymbol);
}

function useHandler() {
  return use<Handler>(HandlerSymbol);
}

function useIntervalTree() {
  return use<IntervalTreeType>(IntervalTreeSymbol);
}
function useMergeList() {
  return use<MergeList>(MergeListSymbol);
}

function useTrackMap() {
  return use<TrackMapType>(TrackMapSymbol);
}

function useFilteredTracks() {
  return use<FilteredTracksType>(TracksSymbol);
}

function useTypeStyling() {
  return use<TypeStylingType>(TypeStylingSymbol);
}

function useSelectedKey() {
  return use<SelectedKeyType>(SelectedKeySymbol);
}

function useSelectedTrackId() {
  return use<SelectedTrackIdType>(SelectedTrackIdSymbol);
}

function useStateStyles() {
  return use<StateStylesType>(StateStylesSymbol);
}

function useVisibleModes() {
  return use<VisibleModesType>(VisibleModesSymbol);
}

function usePoses() {
  return use<PosesStateType>(PosesSymbol);
}

function useShowPoses() {
  return use<ShowPosesType>(ShowPosesSymbol);
}
function useJointFilter() {
  return use<JointFilterType>(JointFilterSymbol);
}

export {
  dummyHandler,
  dummyState,
  provideAnnotator,
  use,
  useAttributes,
  useAllTypes,
  useDatasetId,
  useUsedTypes,
  useCheckedTrackIds,
  useCheckedTypes,
  useEnabledTracks,
  useEditingMode,
  useHandler,
  useIntervalTree,
  useMergeList,
  useFrame,
  useTrackMap,
  useFilteredTracks,
  useTypeStyling,
  useSelectedKey,
  useSelectedTrackId,
  useStateStyles,
  useVisibleModes,
  usePoses,
  useShowPoses,
  useJointFilter,
};
