
import { StringTuple } from '@tensorflow-models/posenet/dist/keypoints';
import {
  ref, Ref, watch,
} from '@vue/composition-api';
import { MediaController } from 'vue-media-annotator/components/annotators/mediaControllerType';
import PoseNetController, { PoseNetSettings, PosesType } from 'vue-media-annotator/posenet/posenet';
import Track, { TrackId, TrackSupportedFeature } from 'vue-media-annotator/track';
import { RectBounds } from 'vue-media-annotator/utils';
import { getTrack } from './useTrackStore';


const connectedPartNames: StringTuple[] = [
  ['leftHip', 'leftShoulder'], ['leftElbow', 'leftShoulder'],
  ['leftElbow', 'leftWrist'], ['leftHip', 'leftKnee'],
  ['leftKnee', 'leftAnkle'], ['rightHip', 'rightShoulder'],
  ['rightElbow', 'rightShoulder'], ['rightElbow', 'rightWrist'],
  ['rightHip', 'rightKnee'], ['rightKnee', 'rightAnkle'],
  ['leftShoulder', 'rightShoulder'], ['leftHip', 'rightHip'],
];
const defaultBounds = [857, 516, 946, 702];
const defaultPose = {
  points: {
    nose: [898.677496438016, 522.4187767148804],
    leftEye: [903, 516],
    rightEye: [894, 516],
    leftEar: [910, 518],
    rightEar: [890, 518],
    leftShoulder: [919, 543],
    rightShoulder: [876, 543],
    leftElbow: [934.1117397468458, 569.6466108146686],
    rightElbow: [867.6294200843847, 568.2846262641954],
    leftWrist: [945.8406337617176, 601.977395045413],
    rightWrist: [856.5694971835799, 606.6580509160614],
    leftHip: [917, 610],
    rightHip: [887.9800562828648, 609.1720787960107],
    leftKnee: [926, 657],
    rightKnee: [890, 660],
    leftAnkle: [935, 698],
    rightAnkle: [892, 702],
  },
  score: 1.0,
};

export default function usePosenet({
  annotator,
  selectedTrackIds,
  trackMap,
}: {
  annotator: Ref<MediaController>;
  selectedTrackIds: Ref<TrackId[]>;
  trackMap: Map<TrackId, Track>;
}) {
  const poseNet = PoseNetController();
  poseNet.loadPosenet();

  const showPoses: Ref<boolean> = ref(false);
  const jointFilter: Ref<number> = ref(0.0);
  const poses: Ref<PosesType[]> = ref([]);
  const loading = ref(false);

  const videoRef: Ref<HTMLVideoElement | null> = ref(null);

  let poseNetSettings: PoseNetSettings = {
    show: false,
    advanced: false,
    input: {
      architecture: 'ResNet50',
      outputStride: 8,
      inputResolution: 200,
      multiplier: 1.00,
      quantBytes: 2,
    },
    multiPoseDetection: {
      maxPoseDetections: 15,
      minPoseConfidence: 0,
      minPartConfidence: 0,
      nmsRadius: 0,
    },
    jointFilter: 0.0,
  };


  const getBounds = () => {
    if (selectedTrackIds.value.length && annotator?.value?.frame.value !== undefined) {
      const track = getTrack(trackMap, selectedTrackIds.value[0]);
      const [feature] = track.getFeature(annotator.value.frame.value);
      return feature?.bounds;
    }
    return [];
  };

  function clearPose() {
    loading.value = true;
    poses.value = [];
    showPoses.value = false;
    loading.value = false;
  }

  function updateFilterSettings(value: number) {
    jointFilter.value = value;
  }
  function displayDefaultPose() {
    //We need the current bounds and then modify the default pose points to fit within those bounds
    if (selectedTrackIds.value.length && annotator?.value?.frame.value !== undefined) {
      const track = getTrack(trackMap, selectedTrackIds.value[0]);
      const [feature] = track.getFeature(annotator.value.frame.value);
      if (feature && feature.bounds) {
        const baseBounds = feature.bounds;
        //Ratio the points in the default bounds to create new points
        const { points } = defaultPose;
        const scaledPoints = { };
        const width = defaultBounds[2] - defaultBounds[0];
        const height = defaultBounds[3] - defaultBounds[1];
        const baseWidth = baseBounds[2] - baseBounds[0];
        const baseHeight = baseBounds[3] - baseBounds[1];
        Object.entries(points).forEach(([key, arr]) => {
          const x = arr[0] - defaultBounds[0];
          const y = arr[1] - defaultBounds[1];
          const scaleX = x / width;
          const scaleY = y / height;
          scaledPoints[key] = [
            baseBounds[0] + scaleX * baseWidth, baseBounds[1] + scaleY * baseHeight,
          ];
        });
        const features: GeoJSON.FeatureCollection<TrackSupportedFeature> = {
          type: 'FeatureCollection',
          features: [],
        };
        //Now we need to create geoJSON for the points
        Object.entries(scaledPoints as Record<string, [number, number]>)
          .forEach(([key, item]) => {
            const pointFeature: GeoJSON.Feature<GeoJSON.Point> = {
              type: 'Feature',
              properties: { key },
              geometry: {
                type: 'Point',
                coordinates:
                      item,
              },
            };
            // eslint-disable-next-line no-unused-expressions
            features.features.push(pointFeature);
          });

        connectedPartNames.forEach((val) => {
          if (scaledPoints[val[0]] && scaledPoints[val[1]]) {
            const lineFeature: GeoJSON.Feature<GeoJSON.LineString> = {
              type: 'Feature',
              properties: { key: `${val[0]}_${val[1]}` },
              geometry: {
                type: 'LineString',
                coordinates: [
                  scaledPoints[val[0]], scaledPoints[val[1]],
                ],
              },
            };
            features.features.push(lineFeature);
          }
        });
        poses.value = [{
          bounds: baseBounds,
          features,
          points: scaledPoints,
          score: 1.0,
        }];
        showPoses.value = true;
        loading.value = false;
      }
    }
  }

  function lockPose() {
    loading.value = true;
    if (selectedTrackIds.value.length && annotator?.value?.frame.value !== undefined) {
      const track = getTrack(trackMap, selectedTrackIds.value[0]);
      const [feature] = track.getFeature(annotator.value.frame.value);
      if (feature && poses.value[0]) {
        if (!feature.attributes?.diva_id) {
          feature.bounds = poses.value[0].bounds;
        }
        if (!feature.attributes) {
          feature.attributes = {};
        }
        if (poses.value[0].points) {
          feature.attributes.poses = {
            points: poses.value[0].points,
            score: poses.value[0].score,
          };
        }
        track.setFeature(feature);
        poses.value = [];
        showPoses.value = false;
      }
    }
    loading.value = false;
  }

  async function updatePoseSettings(settings: PoseNetSettings) {
    jointFilter.value = settings.jointFilter;
    if (videoRef.value) {
      poses.value = [];
      poseNetSettings = settings;
      showPoses.value = settings.show;
      /*
      const inputSettings = { ...poseNetSettings };
      if (typeof (inputSettings.input.inputResolution) === 'number') {
        const scale = inputSettings.input.inputResolution as number;
        inputSettings.input.inputResolution = {
          width: vidHeight * scale, height: vidWidth * scale,
        };
      }
      */
      poseNetSettings = settings;
      showPoses.value = true;
      loading.value = true;
      await poseNet.loadPosenet(poseNetSettings);
      //poseNetSettings = inputSettings;
      const bounds = getBounds() as RectBounds;
      const result = await poseNet.calculatePoses(
        videoRef.value, poseNetSettings, bounds as RectBounds,
      );
      if (result) {
        poses.value = result;
      } else {
        poses.value = [];
      }
      loading.value = false;
    }
  }

  watch(annotator, (updated) => {
    if (updated.videoRef) {
      videoRef.value = updated.videoRef.value;
      if (videoRef && videoRef.value) {
        videoRef.value.addEventListener('seeked', async () => {
          if (videoRef.value && showPoses.value) {
            poses.value = [];
            showPoses.value = false;
          }
        });
      }
    }
  });

  return {
    updatePoseSettings,
    updateFilterSettings,
    displayDefaultPose,
    lockPose,
    clearPose,
    poses,
    showPoses,
    loading,
    jointFilter,
  };
}
