import sys
# sys.path.append('/home/local/KHQ/daniel.davila/projects/FairMOT/src/')

import cv2
import torch
import numpy as np
import scriptconfig as scfg

import _init_paths
from opts import opts
import datasets.dataset.jde as datasets
from tracker.multitracker import JDETracker
from tracking_utils.timer import Timer
from mmpose.apis import init_pose_model, vis_pose_result
from mmpose.apis.inference import LoadImage, _box2cs


class Config(scfg.Config):
    default = {
        # System Options
        'src_vid' : scfg.Path(None, help='path to the src video path'),

        'tracker_model_path': scfg.Path(None, help='path to tracker feature extractor weights'), 
        
        'save_dir': scfg.Path('.', help='path where this script can dump stuff'),
        
        'use_cuda': scfg.Path('True', help='boolean, whether to attempt to use cuda'),
        
        'tracker_det_thresh': scfg.Value(0.1, help='person detector threshold in tracker'),
        
        'frame_rate': scfg.Value(30, help='Framerate of video being processed'),

        'detector_skip' : scfg.Value(1, help='Number of frames to skip in raw video between running detector/tracker. The 0th frame will be the first processed in video'),
        
        'save_skip' : scfg.Value(1, help='Number of frames in raw video to skip between saving track states. The 0th frame will be the first processed in video'),

        'save_images' : scfg.Value(False, help='boolean, whether to save video frames to disc'),

        'extract_pose' : scfg.Value(True, help='boolean, whether to extract pose for each track state'),        

        'max_frame_count' : scfg.Value(None, help='Optional input for stopping early. For debugging.'),

    }


def letterbox(img, height=608, width=1088,
              color=(127.5, 127.5, 127.5)):  # resize a rectangular image to a padded rectangular
    shape = img.shape[:2]  # shape = [height, width]
    ratio = min(float(height) / shape[0], float(width) / shape[1])
    new_shape = (round(shape[1] * ratio), round(shape[0] * ratio))  # new_shape = [width, height]
    dw = (width - new_shape[0]) / 2  # width padding
    dh = (height - new_shape[1]) / 2  # height padding
    top, bottom = round(dh - 0.1), round(dh + 0.1)
    left, right = round(dw - 0.1), round(dw + 0.1)
    img = cv2.resize(img, new_shape, interpolation=cv2.INTER_AREA)  # resized, no border
    img = cv2.copyMakeBorder(img, top, bottom, left, 
            right, cv2.BORDER_CONSTANT, value=color)  # padded rectangular
    return img, ratio, dw, dh

class LoadVideo:  # for inference
    def __init__(self, path, img_size=(1088, 608)):
        self.cap = cv2.VideoCapture(path)
        self.frame_rate = int(round(self.cap.get(cv2.CAP_PROP_FPS)))
        self.vw = int(self.cap.get(cv2.CAP_PROP_FRAME_WIDTH))
        self.vh = int(self.cap.get(cv2.CAP_PROP_FRAME_HEIGHT))
        self.vn = int(self.cap.get(cv2.CAP_PROP_FRAME_COUNT))

        self.width = img_size[0]
        self.height = img_size[1]
        self.count = 0

        self.w, self.h = 1920, 1080
        print('Lenth of the video: {:d} frames'.format(self.vn))

    def get_size(self, vw, vh, dw, dh):
        wa, ha = float(dw) / vw, float(dh) / vh
        a = min(wa, ha)
        return int(vw * a), int(vh * a)

    def __iter__(self):
        self.count = -1
        return self

    def __next__(self):
        self.count += 1
        if self.count == len(self):
            raise StopIteration
        # Read image
        res, img0 = self.cap.read()  # BGR
        assert img0 is not None, 'Failed to load frame {:d}'.format(self.count)
        #img0 = cv2.resize(img0, (self.w, self.h))

        # Padded resize
        img, _, _, _ = letterbox(img0, height=self.height, width=self.width)

        # Normalize RGB
        img = img[:, :, ::-1].transpose(2, 0, 1)
        img = np.ascontiguousarray(img, dtype=np.float32)
        img /= 255.0

        # cv2.imwrite(img_path + '.letterbox.jpg', 255 * img.transpose((1, 2, 0))[:, :, ::-1])  # save letterbox image
        return self.count, img, img0

    def __len__(self):
        return self.vn  # number of files
    
def write_results(filename, results, data_type):
    if data_type == 'mot':
        save_format = '{frame},{id},{x1},{y1},{w},{h},1,-1,-1,-1\n'
    elif data_type == 'kitti':
        save_format = '{frame} {id} pedestrian 0 0 -10 {x1} {y1} {x2} {y2} -10 -10 -10 -1000 -1000 -1000 -10\n'
    else:
        raise ValueError(data_type)

    with open(filename, 'w') as f:
        for frame_id, tlwhs, track_ids in results:
            if data_type == 'kitti':
                frame_id -= 1
            for tlwh, track_id in zip(tlwhs, track_ids):
                if track_id < 0:
                    continue
                x1, y1, w, h = tlwh
                x2, y2 = x1 + w, y1 + h
                line = save_format.format(frame=frame_id, id=track_id, x1=x1, y1=y1, x2=x2, y2=y2, w=w, h=h)
                f.write(line)
    logger.info('save results to {}'.format(filename))


from mmpose.apis import init_pose_model, vis_pose_result
from mmpose.apis.inference import LoadImage, _box2cs

# Pose utilities
def affine_transform(pt, t):
    new_pt = np.array([pt[0], pt[1], 1.]).T
    new_pt = np.dot(t, new_pt)
    return new_pt[:2]

def get_3rd_point(a, b):
    direct = a - b
    return b + np.array([-direct[1], direct[0]], dtype=np.float32)

def get_dir(src_point, rot_rad):
    sn, cs = np.sin(rot_rad), np.cos(rot_rad)

    src_result = [0, 0]
    src_result[0] = src_point[0] * cs - src_point[1] * sn
    src_result[1] = src_point[0] * sn + src_point[1] * cs

    return src_result

def get_affine_transform(
        center, scale, rot, output_size,
        shift=np.array([0, 0], dtype=np.float32), inv=0
):
    if not isinstance(scale, np.ndarray) and not isinstance(scale, list):
        print(scale)
        scale = np.array([scale, scale])

    scale_tmp = scale * 200.0
    src_w = scale_tmp[0]
    dst_w = output_size[0]
    dst_h = output_size[1]

    rot_rad = np.pi * rot / 180
    src_dir = get_dir([0, src_w * -0.5], rot_rad)
    dst_dir = np.array([0, dst_w * -0.5], np.float32)

    src = np.zeros((3, 2), dtype=np.float32)
    dst = np.zeros((3, 2), dtype=np.float32)
    src[0, :] = center + scale_tmp * shift
    src[1, :] = center + src_dir + scale_tmp * shift
    dst[0, :] = [dst_w * 0.5, dst_h * 0.5]
    dst[1, :] = np.array([dst_w * 0.5, dst_h * 0.5]) + dst_dir

    src[2:, :] = get_3rd_point(src[0, :], src[1, :])
    dst[2:, :] = get_3rd_point(dst[0, :], dst[1, :])

    if inv:
        trans = cv2.getAffineTransform(np.float32(dst), np.float32(src))
    else:
        trans = cv2.getAffineTransform(np.float32(src), np.float32(dst))

    return trans


def round_str_num(input_str):
    if '.' in input_str:
        wholes, decimals = str(input_str).split('.')
        return '.'.join([wholes, decimals[:2]])
    else:
        return input_str


def box_to_center_scale(x1, y1, x2, y2, model_image_width, model_image_height):
        """convert a box to center,scale information required for pose transformation
        Parameters
        ----------
        box : list of tuple
            list of length 2 with two tuples of floats representing
            bottom left and top right corner of a box
        model_image_width : int
        model_image_height : int
        Returns
        -------
        (numpy array, numpy array)
            Two numpy arrays, coordinates for the center of the box and the scale of the box
        """
        center = np.zeros((2), dtype=np.float32)

        box_width = abs(x2 - x1)
        box_height = abs(y2 - y1)
        center[0] = x1 + box_width * 0.5
        center[1] = y1 + box_height * 0.5

        import xdev
        xdev.embed()

        aspect_ratio = model_image_width * 1.0 / model_image_height
        pixel_std = 200

        if box_width > aspect_ratio * box_height:
            box_height = box_width * 1.0 / aspect_ratio
        elif box_width < aspect_ratio * box_height:
            box_width = box_height * aspect_ratio
        scale = np.array(
            [box_width * 1.0 / pixel_std, box_height * 1.0 / pixel_std],
            dtype=np.float32)
        if center[0] != -1:
            scale = scale * 1.25

        return center, scale


def run(cmdline=True, **kw):

    # Init config from cli
    config = Config(default=kw, cmdline=cmdline)

    use_cuda=config['use_cuda'] == 'True'
    if not use_cuda:
        print('Not configured to use cuda, executing on CPU')

    if config['save_images'] and config['extract_pose']:
        from tracking_utils import visualization as vis

    tracker_args  = ['mot',
                    '--load_model' , config['tracker_model_path'],
                    '--conf_thres' , int(config['tracker_det_thresh']),
                    ]

    if not use_cuda:
        tracker_args.extend(['--gpus', '-1'])

    save_dir = config['save_dir']
    input_video_path = config['src_vid']
    frame_rate = config['frame_rate']
    extract_pose = config['extract_pose']
    save_images = config['save_images']
    detector_skip = config['detector_skip']
    save_skip = config['save_skip']
    max_frame_count = int(config['max_frame_count']) if config['max_frame_count'] else None

    # HRNet default paths for mmpose trained model. Should remain hardcoded for now
    hrnet_config_path = '/opt/mevaTracker/artifacts/hrnet_w32_coco_256x192.py'
    checkpoint_path = '/opt/mevaTracker/artifacts/hrnet_w32_coco_256x192-c78dce93_20200708.pth'

    # Initialize FairMOT tracker
    opt = opts().init(args=tracker_args)
    #TODO make thse input sizes configurable
    #dataloader = LoadVideo(input_video_path, img_size=(1088*2, 608*2))
    dataloader = LoadVideo(input_video_path, img_size=(1088, 608))
    tracker = JDETracker(opt, frame_rate=frame_rate)

    # Init tracker loop accumulators and helpers
    timer = Timer()
    tracker_timer = Timer()
    pose_timer = Timer()
    results = {}
    frame_id = 0 

    # Initialize hrnet pose extraction model
    pose_model = init_pose_model(hrnet_config_path,
                        checkpoint_path)
    pose_model.test_cfg['flip_test'] = False
    pose_model.test_cfg['shift_heatmap'] = False
    _ = pose_model.eval()
    if use_cuda:
        _ = pose_model.cuda()
    else:
        _ = pose_model.cpu()


    for i, (path, img, img0) in enumerate(dataloader):

        if i % detector_skip == 0:
            # run tracking
            timer.tic()
            tracker_timer.tic()
            if use_cuda:
                blob = torch.from_numpy(img).cuda().unsqueeze(0)
            else:
                blob = torch.from_numpy(img).unsqueeze(0)

            online_targets = tracker.update(blob, img0)
            tracker_timer.toc()
            
            online_tlwhs = []
            online_ids = []
            online_scores = []
            online_features = []
            online_poses = []
            for t in online_targets:
                pose_timer.tic()
                tlwh = t.tlwh
                tid = t.track_id
                score = t.score
                features = t.features
                vertical = tlwh[2] / tlwh[3] > 1.6
                if tlwh[2] * tlwh[3] > opt.min_box_area and not vertical:
                    online_tlwhs.append(tlwh)
                    online_ids.append(tid)
                    online_scores.append(score)
                    online_features.append(features)

                    if score > config['tracker_det_thresh']:
                        #Pose Extraction 
                        if extract_pose:
                            rotation = 0
                            center, scale = _box2cs(pose_model.cfg, tlwh)
                            # Working
                            trans = get_affine_transform(center, 
                                                        scale, 
                                                        rotation, 
                                                        (192,256))
                            chip = cv2.warpAffine(
                                img0,
                                trans,
                                (int(192), int(256)),
                                flags=cv2.INTER_LINEAR)

                            display_chip = chip.copy()

                            chip = (chip - chip.max()) / (chip.max() - chip.min())
                            chip = chip - np.array([0.485, 0.456, 0.406])
                            chip = chip / np.array([0.229, 0.224, 0.225])

                            model_in = torch.FloatTensor(chip.transpose([2,0,1])).unsqueeze(0)
                            if use_cuda:
                                model_in = model_in.cuda()

                            img_metas = [{
                                    'image_file' : 'test.tiff',
                                    'flip_pairs' : None,
                                    'center'     : center,
                                    'rotation'   : 0,
                                    'bbox_score' : score,
                                    'scale'      : scale,
                                    }]

                            with torch.no_grad():
                                all_preds, _, _, heatmap = pose_model(
                                    img=model_in,
                                    img_metas=img_metas,
                                    return_loss=False,
                                    return_heatmap=True)

                            keypoints = all_preds[0]

                            pose_results = [{'bbox' : [int(a) for a in [tlwh[0],tlwh[1],tlwh[0]+tlwh[2],tlwh[1]+tlwh[3]]],
                               'keypoints' : keypoints}]

                            if save_images:
                                img0 = vis_pose_result(pose_model, 
                                                    img0, 
                                                    pose_results, 
                                                    dataset='TopDownCocoDataset')

                            online_poses.append(keypoints)
                    else:
                        online_poses.append([])

                torch.cuda.synchronize()
                pose_timer.toc()
            if i % save_skip == 0:
                print('Saving frame {}'.format(i))
                # Append frame results to output structure
                results[frame_id] = []
                for bbox, tid, score, feat, keypoints in zip(online_tlwhs, 
                                                        online_ids, 
                                                        online_scores, 
                                                        online_features, 
                                                        online_poses):
                    if score > config['tracker_det_thresh']:
                        results[frame_id].append({
                                    'bbox'      :   [round_str_num(str(a)) for a in bbox],
                                    #list([str(a)[:6] for a in bbox]),
                                    'track_id'  : str(tid),
                                    'score'     : round_str_num(str(score)),
                                    #'feat'      : [str(a) for a in feat],
                                    'pose_x'    : [round_str_num(str(a)) for a in list(keypoints[:,0])],
                                    'pose_y'    : [round_str_num(str(a)) for a in list(keypoints[:,1])],
                                    'pose_score': [round_str_num(str(a)) for a in list(keypoints[:,2])]
                                    })

            if save_images:
                online_im = vis.plot_tracking(img0, 
                                        online_tlwhs, 
                                        online_ids, 
                                        frame_id=frame_id)

        
            timer.toc()        
        # Update frame counter and timer
        frame_id += 1 

        print('Tracker FPS:', 1./tracker_timer.average_time)
        print('Pose FPS:', 1./pose_timer.average_time)        
        print('FPS:', 1./timer.average_time)        

        #TODO save periodically for error catching
        #TODO save visualizatons

        #Stop early if configured
        if max_frame_count and (frame_id + 1) >= max_frame_count:
            break

    results_filename = input_video_path.split('/')[-1]
    results_path = '{}/{}.json'.format(save_dir, results_filename)

    import json
    with open(results_path, 'w') as fp:
        json.dump(results, fp, indent=4)


if __name__ == '__main__':

    '''
    This script will create a json gt file which contains the tracks computed by the FairMOT tracker and the mmpose implementation of HRNet for pose extraction.


    Example usage:

    >> python mevaTracker.py
    >> --src_vid 2018-03-05.09-50-15.09-55-01.school.G423.avi
    >> --tracker_model_path fairmot_dla34.pth

    Output:

    save_dir/
        video_name.json
        video_name/
            image1.png
            image2.png
            ...
    '''

    run()
