import warnings
# Ignore FairMOT SyntaxWarning
warnings.filterwarnings("ignore", category=SyntaxWarning)

import sys
import os
import json

import yaml
import click
import uuid
import cv2
import torch
import numpy as np
import itertools

import _init_paths
from opts import opts
import datasets.dataset.jde as datasets
from trackers import JDETracker
#from tracker.multitracker import JDETracker
from tracking_utils.timer import Timer
from tracking_utils import visualization as vis

from mmpose.apis import init_pose_model, vis_pose_result
from mmpose.apis.inference import LoadImage, _box2cs

import logging
from tracking_utils.log import logger

import contextlib
class CaptureOutput(contextlib.AbstractContextManager):
    """
    Fake file-like stream object that redirects writes to a logger instance.
    """
    def __init__(self, log_level=logging.INFO):
        self.logger = logger
        self.log_level = log_level
        self.linebuf = ''

    def __enter__(self, *args, **kwargs):
        self.stdout, self.stderr = sys.stdout, sys.stderr
        sys.stdout, sys.stderr = self, self
        return

    def __exit__(self, *args, **kwargs):
        sys.stdout, sys.stderr = self.stdout, self.stderr

    def write(self, buf):
        temp_linebuf = self.linebuf + buf
        self.linebuf = ''
        for line in temp_linebuf.splitlines(True):
            # From the io.TextIOWrapper docs:
            #   On output, if newline is None, any '\n' characters written
            #   are translated to the system default line separator.
            # By default sys.stdout.write() expects '\n' newlines and then
            # translates them so this is still cross platform.
            if line[-1] == '\n':
                self.logger.log(self.log_level, line.rstrip())
            else:
                self.linebuf += line

    def flush(self):
        if self.linebuf != '':
            self.logger.log(self.log_level, self.linebuf.rstrip())
        self.linebuf = ''

def letterbox(img, size, color=(127.5, 127.5, 127.5)):
        shape = img.shape[:2]  # shape = [height, width]
        height, width = size
        ratio = min(float(height) / shape[0], float(width) / shape[1])
        new_shape = (round(shape[1] * ratio), round(shape[0] * ratio))  # new_shape = [width, height]
        dw = (width - new_shape[0]) / 2  # width padding
        dh = (height - new_shape[1]) / 2  # height padding
        top, bottom = round(dh - 0.1), round(dh + 0.1)
        left, right = round(dw - 0.1), round(dw + 0.1)
        img = cv2.resize(img, new_shape, interpolation=cv2.INTER_AREA)  # resized, no border
        img = cv2.copyMakeBorder(img, top, bottom, left,
                                 right, cv2.BORDER_CONSTANT, value=color)  # padded rectangular

        return img, ratio, dw, dh



class LoadVideo:  # for inference
    def __init__(self, path, img_scales=(1,1)):
        self.cap = cv2.VideoCapture(path)
        self.frame_rate = int(round(self.cap.get(cv2.CAP_PROP_FPS)))
        self.vw = int(self.cap.get(cv2.CAP_PROP_FRAME_WIDTH))
        self.vh = int(self.cap.get(cv2.CAP_PROP_FRAME_HEIGHT))
        self.vn = int(self.cap.get(cv2.CAP_PROP_FRAME_COUNT))

        self.scale_hw = img_scales 
        self.width = self.vw * self.scale_hw[1]
        self.height = self.vh * self.scale_hw[0]

        self.count = 0

        self.read_timer = Timer()

        logger.info('Lenth of the video: {:d} frames'.format(self.vn))


    def get_size(self, vw, vh, dw, dh):
        wa, ha = float(dw) / vw, float(dh) / vh
        a = min(wa, ha)
        return int(vw * a), int(vh * a)

    # resize a rectangular image to a padded rectangular
    def letterbox(self, img, color=(127.5, 127.5, 127.5)):
        shape = img.shape[:2]  # shape = [height, width]
        ratio = min(float(self.height) / shape[0], float(self.width) / shape[1])
        new_shape = (round(shape[1] * ratio), round(shape[0] * ratio))  # new_shape = [width, height]
        dw = (self.width - new_shape[0]) / 2  # width padding
        dh = (self.height - new_shape[1]) / 2  # height padding
        top, bottom = round(dh - 0.1), round(dh + 0.1)
        left, right = round(dw - 0.1), round(dw + 0.1)
        img = cv2.resize(img, new_shape, interpolation=cv2.INTER_AREA)  # resized, no border
        img = cv2.copyMakeBorder(img, top, bottom, left,
                                 right, cv2.BORDER_CONSTANT, value=color)  # padded rectangular

        return img, ratio, dw, dh


    def _read(self, idx=None):
        self.read_timer.tic()
        if idx is not None:
            self.count = idx
            self.cap.set(cv2.CAP_PROP_POS_FRAMES, idx)

        res, raw_img = self.cap.read()

        if raw_img is None:
            logger.debug(f'Error reading {self.count}')
            return self.count, None, None

        # assert img0 is not None, 'Failed to load frame {:d}'.format(self.count)
        #img0 = cv2.resize(img0, (self.w, self.h))

        # Padded resize
        img, _, _, _ = self.letterbox(raw_img)

        #Copy original image so we dont break visualization
        resized_raw_img = img.copy()

        # Normalize RGB
        img = img[:, :, ::-1].transpose(2, 0, 1)
        img = np.ascontiguousarray(img, dtype=np.float32)
        img /= 255.0

        self.read_timer.toc()
        # cv2.imwrite(img_path + '.letterbox.jpg', 255 * img.transpose((1, 2, 0))[:, :, ::-1])  # save letterbox image
        return self.count, img, resized_raw_img, raw_img

#    def __getitem__(self, idx):
#        if isinstance(idx, int):
#            return self._read(idx)
#        elif isinstance(idx, slice):
#            start, stop, step = idx.start or 0, idx.stop or self.vn, idx.step or 1
#            return [self._read(n) for n in range(start, min(stop, self.vn), step)]
#        else:
#            raise TypeError("dataloader indices should be integers or slices.")

    def __iter__(self):
        self.count = 0
        return self

    def __next__(self):
        if self.count >= len(self):
            raise StopIteration

        ret = self._read()

        self.count += 1
        return ret

    def __len__(self):
        return self.vn  # number of files
    
def write_results(filename, results, data_type):
    if data_type == 'mot':
        save_format = '{frame},{id},{x1},{y1},{w},{h},1,-1,-1,-1\n'
    elif data_type == 'kitti':
        save_format = '{frame} {id} pedestrian 0 0 -10 {x1} {y1} {x2} {y2} -10 -10 -10 -1000 -1000 -1000 -10\n'
    else:
        raise ValueError(data_type)

    with open(filename, 'w') as f:
        for frame_id, tlwhs, track_ids in results:
            if data_type == 'kitti':
                frame_id -= 1
            for tlwh, track_id in zip(tlwhs, track_ids):
                if track_id < 0:
                    continue
                x1, y1, w, h = tlwh
                x2, y2 = x1 + w, y1 + h
                line = save_format.format(frame=frame_id, id=track_id, x1=x1, y1=y1, x2=x2, y2=y2, w=w, h=h)
                f.write(line)
    logger.info('save results to {}'.format(filename))

# Pose utilities
def affine_transform(pt, t):
    new_pt = np.array([pt[0], pt[1], 1.]).T
    new_pt = np.dot(t, new_pt)
    return new_pt[:2]

def get_3rd_point(a, b):
    direct = a - b
    return b + np.array([-direct[1], direct[0]], dtype=np.float32)

def get_dir(src_point, rot_rad):
    sn, cs = np.sin(rot_rad), np.cos(rot_rad)

    src_result = [0, 0]
    src_result[0] = src_point[0] * cs - src_point[1] * sn
    src_result[1] = src_point[0] * sn + src_point[1] * cs

    return src_result

def get_affine_transform(
        center, scale, rot, output_size,
        shift=np.array([0, 0], dtype=np.float32), inv=0
):
    if not isinstance(scale, np.ndarray) and not isinstance(scale, list):
        logger.debug(f'Scale: {scale}')
        scale = np.array([scale, scale])

    scale_tmp = scale * 200.0
    src_w = scale_tmp[0]
    dst_w = output_size[0]
    dst_h = output_size[1]

    rot_rad = np.pi * rot / 180
    src_dir = get_dir([0, src_w * -0.5], rot_rad)
    dst_dir = np.array([0, dst_w * -0.5], np.float32)

    src = np.zeros((3, 2), dtype=np.float32)
    dst = np.zeros((3, 2), dtype=np.float32)
    src[0, :] = center + scale_tmp * shift
    src[1, :] = center + src_dir + scale_tmp * shift
    dst[0, :] = [dst_w * 0.5, dst_h * 0.5]
    dst[1, :] = np.array([dst_w * 0.5, dst_h * 0.5]) + dst_dir

    src[2:, :] = get_3rd_point(src[0, :], src[1, :])
    dst[2:, :] = get_3rd_point(dst[0, :], dst[1, :])

    if inv:
        trans = cv2.getAffineTransform(np.float32(dst), np.float32(src))
    else:
        trans = cv2.getAffineTransform(np.float32(src), np.float32(dst))

    return trans


def round_str_num(input_str, precision=3):
    if '.' in input_str:
        wholes, decimals = str(input_str).split('.')
        return '.'.join([wholes, decimals[:precision]])
    else:
        return input_str


def box_to_center_scale(x1, y1, x2, y2, model_image_width, model_image_height):
        """convert a box to center,scale information required for pose transformation
        Parameters
        ----------
        box : list of tuple
            list of length 2 with two tuples of floats representing
            bottom left and top right corner of a box
        model_image_width : int
        model_image_height : int
        Returns
        -------
        (numpy array, numpy array)
            Two numpy arrays, coordinates for the center of the box and the scale of the box
        """
        center = np.zeros((2), dtype=np.float32)

        box_width = abs(x2 - x1)
        box_height = abs(y2 - y1)
        center[0] = x1 + box_width * 0.5
        center[1] = y1 + box_height * 0.5

        aspect_ratio = model_image_width * 1.0 / model_image_height
        pixel_std = 200

        if box_width > aspect_ratio * box_height:
            box_height = box_width * 1.0 / aspect_ratio
        elif box_width < aspect_ratio * box_height:
            box_width = box_height * aspect_ratio
        scale = np.array(
            [box_width * 1.0 / pixel_std, box_height * 1.0 / pixel_std],
            dtype=np.float32)
        if center[0] != -1:
            scale = scale * 1.25

        return center, scale


def _get_pose_model(use_cuda=True):
    # HRNet default paths for mmpose trained model. Should remain hardcoded for now
    #
    # TODO fix these paths
    #hrnet_config_path = '/opt/mevaTracker/artifacts/hrnet_w32_coco_256x192.py'
    #checkpoint_path = '/opt/mevaTracker/artifacts/hrnet_w32_coco_256x192-c78dce93_20200708.pth'
    
    #Old mmpose
    #hrnet_config_path = '/home/local/KHQ/daniel.davila/projects/public_repos/mmpose/configs/top_down/hrnet/coco/hrnet_w32_coco_256x192.py'
    #checkpoint_path = 'https://download.openmmlab.com/mmpose/top_down/hrnet/hrnet_w32_coco_256x192-c78dce93_20200708.pth'

    # New mmpose 
    hrnet_config_path = '/opt/mevaTracker/artifacts/hrnet_w32_coco_256x192.py'
    checkpoint_path = '/opt/mevaTracker/artifacts/hrnet_w32_coco_256x192-c78dce93_20200708.pth'

    # Initialize hrnet pose extraction model
    with CaptureOutput(logging.INFO):
        pose_model = init_pose_model(hrnet_config_path,
                                     checkpoint_path)
        pose_model.test_cfg['flip_test'] = False
        pose_model.test_cfg['shift_heatmap'] = False
        pose_model.eval()

        if use_cuda:
            pose_model.cuda()
        else:
            pose_model.cpu()

        return pose_model


def _get_tracker(detector_cfg_path= '/opt/mevaTracker/artifacts/CenterNet2_DLA-BiFPN-P3_4x_ursa.yaml',
                 detector_weights_path='/opt/mevaTracker/artifacts/CenterNet2_DLA-BiFPN-P3_4x.pth',
                 dg_weights_path='/opt/mevaTracker/artifacts/market_resnet50_model_120_rank1_945.pth',
        
                 tracker_det_thresh=0.1,
                 frame_rate=30,
                 use_cuda=True):
    tracker_args  = ['mot',
                     '--conf_thres', str(tracker_det_thresh)]
    if not use_cuda:
        tracker_args.extend(['--gpus', '-1'])

    with CaptureOutput(logging.INFO):
        opt = opts().init(args=tracker_args)

        tracker = JDETracker(opt,
                detector_cfg_path,
                detector_weights_path,
                dg_weights_path)

        return tracker, opt


def _compute_intersection(bb1, bb2):
    x_left = max(bb1[0], bb2[0])
    y_top = max(bb1[1], bb2[1])
    x_right = min(bb1[2], bb2[2])
    y_bottom = min(bb1[3], bb2[3])

    return (x_right - x_left) * (y_bottom - y_top)

# Yield targets information from online_targets that are not vertical or smaller
# than min_box_area
def _get_valid_targets(online_targets, opt, img0, bbox_overflow_limit):
    for t in online_targets:
        #vertical = t.tlwh[2] / t.tlwh[3] > 1.6
        # if the detection is not vertical
        target_is_min_size = t.tlwh[2] * t.tlwh[3] > opt.min_box_area
        
        # Compute degree to which bbox is contained in image space
        bbox = [t.tlwh[0], t.tlwh[1], t.tlwh[0]+t.tlwh[2], t.tlwh[1]+t.tlwh[3]]
        img_h,img_w,_  = img0.shape
        bbox_intersection = _compute_intersection(bbox, [0,0,img_w,img_h])
        bbox_area = t.tlwh[2] * t.tlwh[3]
        bbox_overlap = bbox_intersection / bbox_area
        bbox_within_image = bbox_overlap >= bbox_overflow_limit

        #TODO maybe dont hardcode this label? Would have to get from detector
        # metadata
        #label = 'person' if t.tlwh[3] > ignore_len and t.tlwh[2] > int(ignore_len/3) else 'ignore' 

        if target_is_min_size and bbox_within_image: #and not vertical:
            yield t.tlwh, t.most_recent_raw_tlwh, t.track_id, t.score, t.features[-1]


def _pose_estimation(pose_model, img0, tlwh_batch, score_batch,
                     save_images=False, use_cuda=True):

    #import time
    #top = time.time()
    #start = time.time() 
    rotation = 0
    img_metas = []
    height, width, channels = img0.shape
    model_in = torch.zeros((len(tlwh_batch), 3, 256, 192))
    for ii, (tlwh, score) in enumerate(zip(tlwh_batch, score_batch)):
     
        x1,y1,w,h = [int(a) for a in tlwh]
        x1 = max(0,x1)
        y1 = max(0,y1)
        x2 = min(x1+w, width)
        y2 = min(y1+h, height)
        chip = img0[y1:y2, x1:x2, :]
        center, scale = _box2cs(pose_model.cfg, [0,0,x2-x1,y2-y1])

        trans = get_affine_transform(center,
                                 scale,
                                 rotation,
                                 (192,256))
        
        chip = cv2.warpAffine(
            chip,
            trans,
            (int(192), int(256)),
            flags=cv2.INTER_LINEAR)

        '''
        img_t = np.zeros(img0.shape)
        x1,y1,w,h = [int(a) for a in tlwh]
        img_t[y1:y1+h, x1:x1+w, :] = img0[y1:y1+h, x1:x1+w, :]

        center, scale = _box2cs(pose_model.cfg, tlwh)
        # Working
        trans = get_affine_transform(center,
                                 scale,
                                 rotation,
                                 (192,256))
 
        chip = cv2.warpAffine(
            img_t,
            trans,
            (int(192), int(256)),
            flags=cv2.INTER_LINEAR)
        '''

        '''
        x1,y1,w,h = [int(a) for a in tlwh]
        chip = img0[y1:y1+h, x1:x1+w, :]
        #cv2.imwrite('/results/chip.png', chip) 

        display_chip = chip.copy()
        try:
            chip = cv2.resize(chip, (192,256))
            #chip = letterbox(chip, (256, 192))[0]
        except:
            print('Found invalid box in pose extraction')
            print(tlwh)
            chip = np.zeros((256,192,3))
        '''

        chip = (chip - chip.max()) / (chip.max() - chip.min())
        chip = chip - np.array([0.485, 0.456, 0.406])
        chip = chip / np.array([0.229, 0.224, 0.225])

        img_metas.append({
            'image_file' : 'test.tiff',
            'flip_pairs' : None,
            'center'     : center,
            'rotation'   : 0,
            'bbox_score' : score,
            'scale'      : scale,
            'bbox_id'    : ii,
            })

        model_in[ii,:,:,:] = torch.FloatTensor(chip.transpose([2,0,1]))

    #model_in = torch.FloatTensor(chip.transpose([2,0,1])).unsqueeze(0)
    if use_cuda:
        model_in = model_in.cuda()

    #4662

    with torch.no_grad():
        try:
            out = pose_model(
                img=model_in,
                img_metas=img_metas,
                return_loss=False,
                return_heatmap=True)
        except:
            import xdev
            xdev.embed()

    keypoints = out['preds']
   
    # map keypoints back to image space
    for ii, ((x1,y1,w,h), kp) in enumerate(zip(tlwh_batch, keypoints)):
        keypoints[ii][:,0] += x1
        keypoints[ii][:,1] += y1

    #TODO fix this to work with batched output above
    if save_images:
        pose_results = []
        for keypoint_set, tlwh in zip(keypoints, tlwh_batch):
            pose_results.append(
                {'bbox' : [int(a) for a in
                       [tlwh[0],tlwh[1],tlwh[0]+tlwh[2],tlwh[1]+tlwh[3]]],
                'keypoints' : keypoint_set}
                )

        img0 = vis_pose_result(pose_model,
                               img0,
                               pose_results,
                               dataset='TopDownCocoDataset')

    #print('Save pose net:', 1/(time.time() - start)) 
    #print('Total:', 1/(time.time() - top))
    return keypoints, img0


def _save_results(online_results, tracker_det_thresh, scale_hw=(1,1)):
    for bbox, tid, score, features, label, diva_id, keypoints in online_results:
        if score > tracker_det_thresh:
            
            x1,y1,w,h = bbox
            x1 *= scale_hw[1]
            y1 *= scale_hw[0]
            w *= scale_hw[1]
            h *= scale_hw[0]
            bbox = [x1,y1,x1+w,y1+h]
            
            value = {
                'uuid'      : uuid.uuid4().hex,
                'bbox'      : [int(a) for a in bbox],
                'track_id'  : str(tid),
                'score'     : round_str_num(str(score)),
                'label'     : label,
                'diva_id'   : diva_id,
                'descriptors': [round_str_num(str(a), precision=4) for a in list(features)]
            }
            
            keypoints[:,0] *= scale_hw[1]
            keypoints[:,1] *= scale_hw[0]
            
            if len(keypoints):
                value.update({
                    'pose_x'    : [round_str_num(str(a)) for a in list(keypoints[:,0])],
                    'pose_y'    : [round_str_num(str(a)) for a in list(keypoints[:,1])],
                    'pose_score': [round_str_num(str(a)) for a in list(keypoints[:,2])]
                })

            yield value


def _save_images(img0, online_results, frame_id, directory):
    online_im = vis.plot_tracking(
        img0,
        [tlwh for tlwh, *_ in online_results],
        [tid for _, tid, *_ in online_results],
        scores= [score for _,_,score, *_ in online_results],
        frame_id=frame_id)
    
    for tlwh, tid, score, feat, label, *_ in online_results:
        x1,y1,w,h  = [int(a) for a in tlwh]

        if score == 1 and label != 'ignore_person':
            text = 'DIVA'
            color = (0,255,0)
        elif score == 1 and label == 'ignore_person':
            text = 'iDIVA'
            color = (0,0,255)
        elif label == 'ignore_person':
            text = 'ignore_person'
            color = (0,0,255)
        else:
            text = '{:.2f}'.format(score)
            color = (255,0,0)

        cv2.putText(online_im, text, (int(x1+(w/2)), y1-5), cv2.FONT_HERSHEY_PLAIN, 1, color,
                        thickness=2)

    try:
        os.mkdir(directory)
    except FileExistsError:
        pass
    path = os.path.join(directory, "{}.jpeg".format(str(frame_id).zfill(6)))
    cv2.imwrite(path, online_im)
    return path

def nms(dets, thresh):
        dets = np.asarray(dets)
        x1 = dets[:, 0]
        y1 = dets[:, 1]
        x2 = dets[:, 2]
        y2 = dets[:, 3]
        scores = dets[:, 4]

        areas = (x2 - x1 + 1) * (y2 - y1 + 1)
        order = scores.argsort()[::-1]

        ndets = len(dets)
        suppressed = np.zeros((ndets), dtype=np.int)

        keep = []
        for _i in range(ndets):
            i = order[_i]
            if suppressed[i] == 1:
                continue
            keep.append(i)
            ix1 = x1[i]
            iy1 = y1[i]
            ix2 = x2[i]
            iy2 = y2[i]
            iarea = areas[i]
            for _j in range(_i, ndets):
                j = order[_j]
                if scores[j] == 1:
                    continue

                if suppressed[j] == 1:
                    continue
                xx1 = max(ix1, x1[j])
                yy1 = max(iy1, y1[j])
                xx2 = min(ix2, x2[j])
                yy2 = min(iy2, y2[j])
                w = max(0.0, xx2 - xx1 + 1)
                h = max(0.0, yy2 - yy1 + 1)
                inter = w * h
                ovr = inter / (iarea + areas[j] - inter)
                if ovr >= thresh:
                    suppressed[j] = 1

        return np.array(keep)

def compute_iou(xywh1, xywh2):
    # Returns IoU between two bboxes of format (x1,y1,w,h)
    box1 = [xywh1[0], xywh1[1], xywh1[0] + xywh1[2], xywh1[1] + xywh1[3]]
    box2 = [xywh2[0], xywh2[1], xywh2[0] + xywh2[2], xywh2[1] + xywh2[3]]
    
    xx1 = max(box1[0], box2[0])
    yy1 = max(box1[1], box2[1])
    xx2 = min(box1[2], box2[2])
    yy2 = min(box1[3], box2[3])
    w = max(0.0, xx2 - xx1 + 1)
    h = max(0.0, yy2 - yy1 + 1)
    inter = w * h
    area1 = (box1[2] - box1[0] + 1) * (box1[3] - box1[1] + 1)
    area2 = (box2[2] - box2[0] + 1) * (box2[3] - box2[1] + 1)
    iou = inter / (area1 + area2 - inter)
    return iou


# TODO Better error checking on string
class SliceType(click.ParamType):
    name = "slice"
    def convert(self, value, param, ctx):
        print(f'{param}: {value}')
        t = tuple(map(lambda x: int(x.strip()) if x.strip() else None,
                      value.split(':')))

        if len(t) == 4:
            s, pose_stride = slice(*t[:3]), t[3]
        else:
            s, pose_stride = slice(*t), 1

        if pose_stride < (s.step or 1):
            raise RuntimeError(
                f"pose_stride ({pose_stride}) must be greater than "
                f"or equal to tracker_stride ({s.step or 1})")
        if pose_stride % (s.step or 1) != 0:
            raise RuntimeError(
                f"pose_stride ({pose_stride}) must be a multiple of "
                f"tracker_stride ({s.step or 1})"
            )

        return ((s.start or 0, s.stop or sys.maxsize, s.step or 1), pose_stride)

SLICE_TYPE = SliceType()


@click.group()
@click.option('--log_level', type=str, default="INFO",
              help="Log level to set the logger")
@click.pass_context
def cli(ctx, log_level):
    logger.setLevel(log_level)

    ctx.ensure_object(dict)


@cli.command()
@click.argument('input_video_path', type=click.Path(exists=True))
@click.option('--frame_slice', type=SLICE_TYPE, default="::2:30",
              help="Frame slice for tracker and pose models")
@click.pass_context
def read(ctx, frame_slice, input_video_path):
    """Read the frames from a video and log any errors."""
    (start, stop, tracker_stride), pose_stride = frame_slice
    timer = Timer()
    dataloader = LoadVideo(input_video_path, img_scales=(1,1))
    timer.tic()
    for frame_id, img, _ in dataloader:
        if frame_id == stop - 1:
            break

        if img is None:
            logger.debug(f'Could not read {frame_id}')
            if frame_id % pose_stride == 0:
                logger.warning(f'Failed to load pose frame {frame_id}')
            elif frame_id % tracker_stride == 0:
                logger.warning(f'Failed to load tracker frame {frame_id}')

        if frame_id % 500 == 0 and frame_id != 0:
            logger.info(f'Read {frame_id} frames... (FPS: {1./dataloader.read_timer.average_time if dataloader.read_timer.average_time != 0 else 0:.2f})')
    timer.toc()
    logger.info(f'Completed reading {dataloader.count} frames (FPS: {1./dataloader.read_timer.average_time if dataloader.read_timer.average_time != 0 else 0:.2f}).')
    logger.info(f'Total time: {timer.total_time:.2f} seconds')

    
@cli.command()
@click.option('--detector_cfg_path', type=click.Path(exists=True),
              default='/opt/mevaTracker/artifacts/CenterNet2_DLA-BiFPN-P3_4x_ursa.yaml',
              help='tracker detector detectron2 config`')
@click.option('--detector_weights_path', type=click.Path(exists=True),
              default='/opt/mevaTracker/artifacts/CenterNet2_DLA-BiFPN-P3_4x.pth',
              help='tracker detector weights file`')
@click.option('--dg_weights_path', type=click.Path(exists=True),
              default='/opt/mevaTracker/artifacts/market_resnet50_model_120_rank1_945.pth',
              help='tracker feature extractor weights file`')
@click.option('--save_dir', type=click.Path(exists=True), default='.',
              help='where this script can dump stuff')
@click.option('--use_cuda', type=click.BOOL, default=True,
              help='whether to attempt to use cuda')
@click.option('--bbox_overflow_limit', type=click.FLOAT, default=0.2,
              help='threshold for bbox overlap with image space')
@click.option('--tracker_det_thresh', type=click.FLOAT, default=0.1,
              help='person detector threshold in tracker')
@click.option('--save_images', type=click.BOOL, default=False,
              help='whether to save video frames to disc')
@click.option('--export_keyframes', type=click.BOOL, default=False,
              help='whether to export just the keyframes of the video')
@click.option('--keyframe_rate', type=click.INT, default=1,
              help='Rate at which to export keyframes')
@click.option('--extract_pose', type=click.BOOL, default=True,
              help='whether to extract pose for each track state')
@click.option('--frame_rate', type=click.INT, default=30,
              help='Framerate of video being processed')
@click.option('--ignore_len', type=click.INT, default=-1,
              help='Number of vertical pixels on target below which targets receive ignore label')
@click.option('--ignore_iou', type=click.FLOAT, default=0.1,
              help='Acceptable IOU with other bbox, above which targets receive ignore label')
@click.option('--frame_slice', type=SLICE_TYPE, default="::2:30",
              help="Frame slice for tracker and pose models")
@click.option('--priors_path', type=click.STRING, default=None, 
              help='Path to diva priors dir for video')
@click.argument('input_video_path', type=click.Path(exists=True))
@click.pass_context
def tracker(ctx, detector_cfg_path, detector_weights_path, dg_weights_path, 
            save_dir, use_cuda, bbox_overflow_limit,
            tracker_det_thresh, save_images, export_keyframes, keyframe_rate, extract_pose, frame_rate,
            ignore_len, ignore_iou, frame_slice, input_video_path, priors_path, **kw):
    """Run the mevaTracker on a video specified by INPUT_VIDEO_PATH

    Frame Slicing:

    What frames of the video to the run the tracker and pose estimation on are
    determiend by the --frame_slice option. This is a colon deiminated list in
    the form: [start_frame]:[end_frame]:[tracker_stride]:[pose_stride].

    \b
    Defaults:
      start_frame:    0
      end_frame:      max fames (via StopIteration)
      tracker_stride: 1
      pose_stride:    1

    \b
    Examples:
      --frame_slice ":10"    - run on each frame up to 10th frame
      --frame_slice "::2:30" - run tracker on every other frame (15Hz)
                               and pose on every 30th frame (1Hz)

    Note: pose_stride must be a larger than, and a multiple of, tracker_stride.

    """
    tracker_slice, pose_stride = frame_slice

    input_filename, _ = os.path.splitext(
        os.path.basename(input_video_path))

    if not use_cuda:
        logger.info('Not configured to use cuda, executing on CPU')

    # Initialize FairMOT tracker
    tracker, opt = _get_tracker(detector_cfg_path, detector_weights_path, dg_weights_path,
                                tracker_det_thresh=tracker_det_thresh,
                                frame_rate=frame_rate,
                                use_cuda=use_cuda)
    #TODO make thse input sizes configurable
    # NOTE Changing this size will break the diva priors without proper handling of rescale
    #dataloader = LoadVideo(input_video_path, img_size=(1920, 1080)) #This is raw shape of DIVA cams
    dataloader = LoadVideo(input_video_path, img_scales=(1,1))

    pose_model = _get_pose_model(use_cuda=use_cuda)

    if priors_path is not None:

        from pathlib import Path
        priors_dir = Path(priors_path)
        
        video_id = input_video_path.split('/')[-1][:-8]

        #with open('/data2/projects/meva/mvp_videos/diva_priors/2018-03-11.11-15-00.11-20-00.school.G299_meta/2018-03-11.11-15-00.11-20-00.school.G299.activities.yml', 'r') as yf:
        #with open('{}/{}_meta/{}.activities.yml'.format(priors_path, video_id, video_id), 'r') as yf:
        with open(list(priors_dir.rglob('*{}*activities.yml'.format(video_id)))[0]) as yf:
            activities = yaml.load(yf, Loader=yaml.CLoader)
    
        #with open('{}/{}_meta/{}.geom.yml'.format(priors_path, video_id, video_id), 'r') as yf:
        with open(list(priors_dir.rglob('*{}*geom.yml'.format(video_id)))[0]) as yf:
            geom = yaml.load(yf, Loader=yaml.CLoader)
        
        #with open('{}/{}_meta/{}.types.yml'.format(priors_path, video_id, video_id), 'r') as yf:
        with open(list(priors_dir.rglob('*{}*types.yml'.format(video_id)))[0]) as yf:
            types = yaml.load(yf, Loader=yaml.CLoader)

        # Get all gt bbox IDs for person targets
        person_ids = [tt['types']['id1'] for tt in types[2:] if 'person' in tt['types']['cset3'].keys()]
   
        # Construct dictionary of detection priors for person targets for each frame
        frame_dict = {}
        for gg in geom:
            if 'geom' in gg and gg['geom']['id1'] in person_ids:
                gg['geom']['g0'] =  [int(int(a)/dataloader.scale_hw[0]) for a in gg['geom']['g0'].split(' ')]
                frame_id = gg['geom']['ts0']
                if frame_id not in frame_dict:
                    frame_dict[frame_id] = [gg]
                else:
                    frame_dict[frame_id].append(gg)

        for fid in range(len(dataloader)):
            if fid not in frame_dict:
                frame_dict[fid] = None

    else:
        frame_dict = {fid : None for fid in range(len(dataloader))}

    # Init tracker loop accumulators and helpers
    timer = Timer()
    tracker_timer = Timer()
    pose_timer = Timer()
    results = {}
    frames_processed = 0

    video_name = input_video_path.split('/')[-1][:-4]
    save_path = '{}/{}/'.format(save_dir, video_name)

    if not os.path.exists(save_path):
        os.mkdir(save_path)

    for frame_id, img, img0, raw_img in itertools.islice(dataloader, *tracker_slice):
        
        if export_keyframes:
            if frame_id % keyframe_rate == 0:
                cv2.imwrite('{}/{}/{}_{}.png'.format(save_dir, video_name, video_name, str(frame_id).zfill(6)), raw_img)
            continue
        
        if img is None or img0 is None:
            logger.error(f'Failed to load frame {frame_id}!')
            continue

        # run tracking
        timer.tic()
        tracker_timer.tic()
        if use_cuda:
            blob = torch.from_numpy(img).cuda().unsqueeze(0)
        else:
            blob = torch.from_numpy(img).unsqueeze(0)

        logger.debug(f'Updating tracker with frame {frame_id}')

        online_targets = tracker.update(blob, frame_dict[frame_id])
        tracker_timer.toc()

        if frame_id % pose_stride == 0:
            online_results = []

            pose_timer.tic()
            tlwh_batch = []
            score_batch = []
            for tlwh, raw_tlwh, tid, score, features in _get_valid_targets(
                                                                    online_targets, 
                                                                    opt, 
                                                                    img0, 
                                                                    bbox_overflow_limit): 
                                                                    #ignore_len):
               
                # Match the track's most recent detection to the prior bboxes so we can preserve
                # DIVA ID downstream 
                if score == 1.0:
                    xyxy = [raw_tlwh[0], 
                            raw_tlwh[1], 
                            raw_tlwh[0] + raw_tlwh[2], 
                            raw_tlwh[1] + raw_tlwh[3]]
                    
                    try:
                        prior_det = [det for det in frame_dict[frame_id] if det['geom']['g0'] ==  xyxy][0]
                    except:
                        import xdev
                        xdev.embed()
                    
                    diva_id = prior_det['geom']['id1']
                else:
                    diva_id = None

                #if diva_id is None and raw_tlwh[3] < ignore_len and raw_tlwh[2] < int(ignore_len/3):
                if raw_tlwh[3] < ignore_len or raw_tlwh[2] < int(ignore_len/3):
                    label = 'ignore_person'
                else:
                    label = 'person'

                online_results.append([tlwh, tid, score, features, label, diva_id])

                tlwh_batch.append(tlwh)
                score_batch.append(score)

            for ii, (tlwh, tid, score, features, label, diva_id) in enumerate(online_results):
                #ious = [compute_iou(tlwh, box) for box,score in zip(tlwh_batch,score_batch) if np.all(tlwh != box) and score != 1.0]
                ious = [compute_iou(tlwh, box) for box,score in zip(tlwh_batch,score_batch) if np.all(tlwh != box)]
                if np.any([a > ignore_iou for a in ious]):
                    new_label = 'ignore_person'
                else:
                    new_label = label

                online_results[ii] = [tlwh, tid, score, features, new_label, diva_id]

            #Pose Extraction
            if len(tlwh_batch):
                if extract_pose:
                    keypoints, img0 = _pose_estimation(
                        pose_model, img0, tlwh_batch, score_batch,
                        save_images=save_images, use_cuda=use_cuda)
                    
                    [online_results[tt].append(keypoint) for tt, keypoint in enumerate(keypoints)]
                    #online_results[len(online_results) - 1].append(keypoints)
                else:
                    [online_results[tt].append([]) for tt in range(len(keypoints))]

            torch.cuda.synchronize()
            pose_timer.toc()

            # Append frame results to output structure
            results[frame_id] = list(_save_results(online_results, tracker_det_thresh, scale_hw=dataloader.scale_hw))

            if save_images:
                directory = os.path.join(save_dir, input_filename)
                path = _save_images(img0, online_results, frame_id, directory)
                logger.debug(f'Saved frame {frame_id} to {path}')

        timer.toc()
        # Update frame counter and timer
        if frames_processed % 5 == 0:
            logger.info(f'At frame {frame_id}: AVG Tracker, Pose, Total FPS: ('
                        f'{1./tracker_timer.average_time if tracker_timer.average_time != 0 else 0:.2f}, '
                        f'{1./pose_timer.average_time if pose_timer.average_time != 0 else 0:.2f}, '
                        f'{1./timer.average_time if timer.average_time != 0 else 0:.2f})')
        frames_processed += 1

    if export_keyframes:
        return

    logger.info('Final: AVG Tracker, Pose, Total FPS: ('
                f'{1./tracker_timer.average_time if tracker_timer.average_time != 0 else 0:.2f}, '
                f'{1./pose_timer.average_time if pose_timer.average_time != 0 else 0:.2f}, '
                f'{1./timer.average_time if timer.average_time != 0 else 0:.2f})')

    results_path = '{}/{}.json'.format(save_dir, input_filename)

    with open(results_path, 'w') as fp:
        json.dump(results, fp, indent=4)


####################################################################################
## Detection Converter
####################################################################################
from typing import Any, Dict, List, Optional, Tuple, Union

from pydantic import BaseModel, Field
from typing_extensions import Literal

# TODO These classes and constants are taken from meva_server.serializers.models
#      and should be included in some other way so as not to violate DRY
#      principals.
#
class GeoJSONGeometry(BaseModel):
    type: str
    # support point, line, or polygon,
    coordinates: Union[List[float], List[List[float]], List[List[List[float]]]]

class GeoJSONFeature(BaseModel):
    type: str
    geometry: GeoJSONGeometry
    properties: Dict[str, Union[bool, float, str]]

class GeoJSONFeatureCollection(BaseModel):
    type: str
    features: List[GeoJSONFeature]


class Feature(BaseModel):
    """Feature represents a single detection in a track."""

    frame: int
    bounds: List[int]
    attributes: Optional[Any]
    geometry: Optional[GeoJSONFeatureCollection] = None
    head: Optional[Tuple[float, float]] = None
    tail: Optional[Tuple[float, float]] = None
    fishLength: Optional[float] = None
    interpolate: Optional[bool] = False
    keyframe: Optional[bool] = True


class Track(BaseModel):
    begin: int
    end: int
    trackId: str
    features: List[Feature] = Field(default_factory=lambda: [])
    confidencePairs: List[Tuple[str, float]] = Field(default_factory=lambda: [])
    attributes: Dict[str, Any] = Field(default_factory=lambda: {})

    def exceeds_thresholds(self, thresholds: Dict[str, float]) -> bool:
        defaultThresh = thresholds.get('default', 0)
        return any(
            [
                confidence >= thresholds.get(field, defaultThresh)
                for field, confidence in self.confidencePairs
            ]
        )

COCO_PART_NAMES = [
    'nose',
    'leftEye',
    'rightEye',
    'leftEar',
    'rightEar',
    'leftShoulder',
    'rightShoulder',
    'leftElbow',
    'rightElbow',
    'leftWrist',
    'rightWrist',
    'leftHip',
    'rightHip',
    'leftKnee',
    'rightKnee',
    'leftAnkle',
    'rightAnkle',
]

# pairs = [
#     [0, 1],
#     [0, 2],
#     [1, 3],
#     [2, 4],
#     [5, 6],
#     [5, 7],
#     [7, 9],
#     [6, 8],
#     [8, 10],
#     [11, 12],
#     [11, 13],
#     [13, 15],
#     [12, 14],
#     [14, 16],
#     [6, 12],
#     [5, 11],
# ]

def convert_detections_to_tracks(json_data):
    # organize file by ID and frame Numbers
    tracks = {}
    #track_descriptors = {}
    track_count = 0
    frame_count =0
    for frame_str in json_data.keys():
        for detection in json_data[frame_str]:
            try:
                if (float(detection["score"]) < 0.0): # if we are implementing a filter based on score
                    continue
            except:
                import xdev
                xdev.embed()
            frame = int(frame_str)
            pose_x = detection["pose_x"]
            pose_y = detection["pose_y"]
            pose_scores = detection["pose_score"]
            pose = {}
            pair_count = 0
            for part in COCO_PART_NAMES:
                pose[COCO_PART_NAMES[pair_count]] = [
                    float(pose_x[pair_count]), float(pose_y[pair_count]), float(pose_scores[pair_count]),
                ]
                pair_count += 1
            score = detection['score']
            uuid = detection['uuid']
            label = detection['label']
            det_bbox = detection['bbox']
            descriptors = detection['descriptors']
            diva_id = detection['diva_id']
            trackId = track_count
            if "track_id" in detection:
                trackId = detection["track_id"]
                if trackId in tracks:
                    track = tracks[trackId]
                else:
                    track = Track(begin=frame, end=frame, trackId=trackId)
                    #track_descriptors[trackId] = {
                    #                    'trackId' : trackId,
                    #                    'features': []
                    #                    }
            else:
                track = Track(begin=frame, end=frame, trackId=trackId)
                #track_descriptors[trackId] = {
                #                        'trackId' : trackId,
                #                        'features': []
                #                        }

            #Append descriptor features
            #track_descriptors[trackId]['features'].append({
            #                                            'frame': frame, 
            #                                            'descriptor' : descriptors
            #                                            })

            if not track.confidencePairs:
                track.confidencePairs = [[f'person_{trackId}', float(detection["score"])]]
            else:
                track.confidencePairs = [[f'person_{trackId}', max(float(detection["score"]),track.confidencePairs[0][1])]]
            tracks[trackId] = track
            track.begin = min(frame, track.begin)
            track.end = max(track.end, frame)
    
            attributes = {"poses": {"points": {}, "kp_scores" : {}, "score": score},
                          "uuid" : uuid,
                          "label": label}

            if diva_id is not None:
                attributes['diva_id'] = diva_id

            for key in COCO_PART_NAMES:
                x = round(pose[key][0])
                y = round(pose[key][1])
                score = pose[key][2]
                attributes['poses']['points'][key] = [x, y]
                attributes['poses']['kp_scores'][key] = score

            # Add face centroid
            face_x = np.mean([pose[key][0] for key in list(pose.keys())[:5]])
            face_y = np.mean([pose[key][1] for key in list(pose.keys())[:5]])
            face_score = np.mean([pose[key][2] for key in list(pose.keys())[:5]])
            attributes['poses']['face_center'] = [face_x, face_y]
            attributes['poses']['face_score'] = score

            '''
            x_min = float('inf')
            x_max = float('-inf')
            y_min = float('inf')
            y_max = float('-inf')
            for key in COCO_PART_NAMES:
                x = round(pose[key][0])
                y = round(pose[key][1])
                attributes['poses']['points'][key] = [x, y]
                x_max = max(x_max, x)
                y_max = max(y_max, y)
                x_min = min(x_min, x)
                y_min = min(y_min, y)
            bounds = [x_min, y_min, x_max, y_max]
            '''

            feature: Feature = Feature(
                frame=frame, bounds=det_bbox, attributes=attributes or None
            )
            track.features.append(feature)
            track_count += 1
        frame_count += 1
        logger.debug(f'Frame: {frame_count} : frame: {frame_str} - Tracks: {track_count}')


    #tracks['1'].features[0].attributes['label']

    split_tracks = {}
    split_tracks_descs = {}
    available_track_id = 1

    # TODO need to split out the features for each track as well

    '''
    >>> track_descriptors['1']['features'][0]
    #>>> {'frame', 'descriptor'}

    '''

    for track_id, track in tracks.items():
        
        keep_track_feats = []
        ignore_track_feats = []
        #keep_descs = {}
        #ignore_descs = {}

        for state in track.features:
            desc = None
            #for track_descs in track_descriptors[track_id]['features']:
            #    if track_descs['frame'] == state.frame:
            #        desc = track_descs['descriptor']
            #desc = [track_descs['descriptor'] for track_descs in track_descriptors[track_id]['features'] if track_descs['frame'] == state.frame]
            if 'ignore' in state.attributes['label']:
                ignore_track_feats.append(state)
            #    ignore_descs[state.frame] = desc
            else:
                keep_track_feats.append(state)
            #    keep_descs[state.frame] = desc

        if keep_track_feats:
            begin = keep_track_feats[0].frame
            end = keep_track_feats[-1].frame
            keep_track = Track(begin=begin, end=end, trackId=available_track_id)
            keep_track.features = keep_track_feats
            keep_track.confidencePairs = [[f'person_{available_track_id}', track.confidencePairs[0][1]]]
            split_tracks[str(available_track_id)] = keep_track
            #split_tracks_descs[str(available_track_id)] = keep_descs
            available_track_id += 1
        if ignore_track_feats:
            begin = ignore_track_feats[0].frame
            end = ignore_track_feats[-1].frame
            ignore_track = Track(begin=begin, end=end, trackId=available_track_id)
            ignore_track.features = ignore_track_feats
            ignore_track.confidencePairs = [[f'ignore_person', track.confidencePairs[0][1]]]
            split_tracks[str(available_track_id)] = ignore_track
            #split_tracks_descs[str(available_track_id)] = ignore_descs
            available_track_id += 1

    return {trackId: track.dict(exclude_none=True) for trackId, track in split_tracks.items()}
    #return {trackId: track.dict(exclude_none=True) for trackId, track in split_tracks.items()}, split_tracks_descs


@cli.command()
@click.argument('tracker_json_file', type=click.Path(exists=True))
@click.pass_context
def convert(ctx, tracker_json_file):
    directory = os.path.dirname(tracker_json_file)
    input_filename, _ = os.path.splitext(
        os.path.basename(tracker_json_file))

    with open(tracker_json_file, 'r') as fh:
        json_data = json.loads(fh.read())

    #results, results_descriptors = convert_detections_to_tracks(json_data)
    results = convert_detections_to_tracks(json_data)

    with open(os.path.join(directory, f'{input_filename}.meva.json'), 'w') as fh:
        json.dump(results, fh)

    #with open(os.path.join(directory, f'{input_filename}.meva.descriptors.json'), 'w') as fh:
    #    json.dump(results_descriptors, fh)


if __name__ == '__main__':

    ''' This script will create a json gt file which contains the tracks computed by
    the FairMOT tracker and the mmpose implementation of HRNet for pose
    extraction.


    Example usage:

    # Run tracker at 15Hz and pose estimation at 1Hz on
    # 2018-03-05.09-50-15.09-55-01.school.G423.avi

    >> python mevaTracker.py \
       tracker \
       --frame_slice "::2:30" \
       --save_dir=/data \
       --save_images=True \
       2018-03-05.09-50-15.09-55-01.school.G423.avi

    Output:

    /data/
        2018-03-05.09-50-15.09-55-01.school.G423.json
        2018-03-05.09-50-15.09-55-01.school.G423/
            {frame_id}.png
            {frame_id}.png
            ...
    '''
    cli(obj={})
