#! /usr/bin/python3.7
"""
Copyright 2021 by Kitware, Inc.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

 * Redistributions of source code must retain the above copyright notice,
   this list of conditions and the following disclaimer.

 * Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.

 * Neither name of Kitware, Inc. nor the names of any contributors may be used
   to endorse or promote products derived from this software without specific
   prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

==============================================================================

"""
from __future__ import division, print_function
import os
import sys
import numpy as np
import argparse
from pathlib import Path
from meva_server.meva_database.database import MevaDatabase
from meva_server.meva_database.models import Detection, Camera
from meva_server.meva_database.utils import (
    geolocate_detection,
    getSource,
    cleanName,
    getCameraName,
    getCameraModelMap,
    getVecFromString,
    turnOffSqlAlchemyLogging,
    getLatLonHeightFromEastNorthUp
)

# some default values that can be overriden by the args passed in
_metadataFolder = "./meva-data-repo/metadata/"
_timezone = "US/East-Indiana"
_latitude = 39.04977294
_longitude = -85.52924953
_height = 205


if __name__ == "__main__":
    print("Reading arguments...")
    # parse the arguments - required for location of database
    parser = argparse.ArgumentParser(
        description="Script to geolocate all detections in the database"
    )
    parser.add_argument("databasePath", help="path to database file")
    parser.add_argument(
        "npyPath", help="folder containing the npy files (in folders by source name)"
    )
    parser.add_argument(
        "--metadataFolder",
        help="path to meta data: https://gitlab.kitware.com/meva/meva-data-repo/-/blob/master/metadata/",
        default=_metadataFolder,
    )
    parser.add_argument("--timezone", help="timezone for cameras", default=_timezone)
    parser.add_argument(
        "--latitude",
        help="latitude in degrees for center point",
        default=_latitude,
        type=float,
    )
    parser.add_argument(
        "--longitude",
        help="longitude in degrees for center point",
        default=_longitude,
        type=float,
    )
    parser.add_argument(
        "--height", help="meters above WGS84 ellipsoid", default=_height, type=float
    )
    args = parser.parse_args()
    if args.databasePath == "":
        print("No databasePath specified - exiting.")
        sys.exit()

    # connect to the db
    print("Setting up db...")
    database_path = Path(args.databasePath)
    npyPath = args.npyPath
    os.environ["MEVA_DATABASE_URI"] = "sqlite:///" + str(database_path)
    meva_db = MevaDatabase()
    session = meva_db.Session()

    latitude0 = args.latitude
    longitude0 = args.longitude
    height0 = args.height
    print(f"East North Up based on: {latitude0}, {longitude0}, {height0}")

    turnOffSqlAlchemyLogging()

    # grab all the detections
    print("Loading detections...")
    detections = session.query(Detection).all()

    # camera model mapping
    print("Getting camera model mapping")
    camera_model_lookup_fname = (
        args.metadataFolder + "meva-clip-camera-and-time-table.txt"
    )
    camera_model_map = getCameraModelMap(camera_model_lookup_fname)

    # run through all the detections and add location
    missing = {}
    for det in detections:
        # don't change a detection that is locked
        if det.locked is True and det.location is not None:
            continue

        # encode the detection info
        detect = {}
        detect["bounds"] = getVecFromString(det.bounding_box, dtype=int)

        # get the source for this detection
        src = getSource(session, det.source_uid)
        srcName = cleanName(src.name)
        print(f"source name = {srcName}")

        cam = getCameraName(srcName)
        # see if the camera model map has this one
        if srcName not in camera_model_map.keys():
            err = (
                "ERROR: camera model map is missing this camera: "
                + cam
                + " or this source: "
                + srcName
            )
            print(err)
            if srcName not in missing.keys():
                missing[srcName] = err
            continue

        # get the camera model
        camModelName = camera_model_map[srcName]
        print("camera model: " + camModelName)
        camModel = session.query(Camera).filter(Camera.name == camModelName).first()
        print("cam model: " + str(camModel))

        camera_model_dir = npyPath + "/" + camModelName[:-5]
        print("npy: " + camera_model_dir)
        if "no-camera-model" in camModelName:
            err = f"ERROR - missing camera model: {srcName}"
            print(err)
            if srcName not in missing.keys():
                missing[srcName] = err
            continue
        cm_xyz = np.load("%s/xyz_per_pixel.npy" % camera_model_dir)

        # get the location and covariance
        loc, cov = geolocate_detection(detect, cm_xyz)

        # convert the location to lat lon height
        loc2 = getLatLonHeightFromEastNorthUp(
            loc, latitude0, longitude0, height0)
        print(f"enu location: {loc}")
        print(f"llh location: {loc2}")
        print(f"covariance: {cov}")
        det.location = np.array2string(loc)
        det.location_cov = np.array2string(cov)

    session.commit()

    if len(missing.keys()) > 0:
        with open("./missing.txt", "w") as f:
            for miss in missing.keys():
                f.write(missing[miss] + "\n")
