#! /usr/bin/python3.7
"""
Look at different type of links and see if we have video-to-video links that
also point back to check in photos (exemplar tracks)
==============================================================================
Copyright 2021 by Kitware, Inc.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

 * Redistributions of source code must retain the above copyright notice,
   this list of conditions and the following disclaimer.

 * Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.

 * Neither name of Kitware, Inc. nor the names of any contributors may be used
   to endorse or promote products derived from this software without specific
   prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

==============================================================================

"""
from __future__ import division, print_function
import os
import argparse
import json
from pathlib import Path
from meva_server.meva_database.database import MevaDatabase
from meva_server.meva_database.models import Track, Source, Response
from meva_server.meva_database.utils import turnOffSqlAlchemyLogging

if __name__ == "__main__":

    print('Reading arguments...')
    # parse the arguments - required for location of database
    parser = argparse.ArgumentParser(
        description='Evaluate vid-to-vid linking')
    parser.add_argument('databasePath', help='path to database file')
    args = parser.parse_args()

    # connect to the db
    print('Setting up db...')
    database_path = Path(args.databasePath)
    os.environ['MEVA_DATABASE_URI'] = 'sqlite:///' + str(database_path)
    meva_db = MevaDatabase()
    session = meva_db.Session()

    turnOffSqlAlchemyLogging()

    # grab all tracks
    print('Loading all tracks...')
    allTracks = session.query(Track).all()
    print('Loading responses...')
    responses = session.query(Response).all()
    print('Loading sources...')
    sources = session.query(Source).all()

    print('Sorting source by uid...')
    sourceByUid = {}
    for src in sources:
        sourceByUid[src.uid] = src

    print('Loading exemplars...')
    exemplars = {}
    tracks = {}
    for track in allTracks:
        # make the track easily referenced by uid
        tracks[track.uid] = track
        if track.exemplar:
            exemplars[track.uid] = track

    # sort the responses
    print('Sorting responses...')
    yeses = []
    positiveLinks = {}
    for resp in responses:
        t1_t2 = resp.track1_uid + '_' + resp.track2_uid
        t2_t1 = resp.track2_uid + '_' + resp.track1_uid
        if resp.response == 1:
            if t1_t2 not in yeses and t2_t1 not in yeses:
                yeses.append(t1_t2)
                if resp.track1_uid not in positiveLinks.keys():
                    positiveLinks[resp.track1_uid] = []
                positiveLinks[resp.track1_uid].append(resp.track2_uid)

    print(f"unique Yes responses: {len(yeses)}")

    # loop through the posLinks and see which are video to video
    videoToVideo = {}
    videoToExemplar = {}
    cnt = 0
    for t1 in positiveLinks.keys():
        # see if t1 is exemplar
        if t1 not in exemplars.keys():
            for t2 in positiveLinks[t1]:
                if t2 not in exemplars.keys():
                    if t1 not in videoToVideo.keys():
                        videoToVideo[t1] = []
                    videoToVideo[t1].append(t2)
                else:
                    if t2 not in videoToExemplar.keys():
                        videoToExemplar[t2] = []
                    # put the exemplar in the key
                    videoToExemplar[t2].append(t1)
        else:
            for t2 in positiveLinks[t1]:
                if t1 not in videoToExemplar.keys():
                    videoToExemplar[t1] = []
                # put the exemplar in the key
                videoToExemplar[t1].append(t2)

    # get all exemplar
    allVidToEx = []
    for t1 in videoToExemplar.keys():
        for t2 in videoToExemplar[t1]:
            t1_t2 = t1 + '_' + t2
            if t1_t2 not in allVidToEx:
                allVidToEx.append(t1_t2)
    print(f"total video to exemplar = {len(allVidToEx)}")

    # now see if any of the video to video group references the videos
    # in the video to exemplar group
    allVidToVid = []
    for t1 in videoToVideo.keys():
        for t2 in videoToVideo[t1]:
            t1_t2 = t1 + '_' + t2
            if t1_t2 not in allVidToVid:
                allVidToVid.append(t1_t2)
    print(f"total video to video = {len(allVidToVid)}")

    indirectRef = {}
    allIndirects = []
    for ex in videoToExemplar.keys():
        for video in videoToExemplar[ex]:
            if video in allVidToVid:
                indirectRef[ex] = video
                print(f"video {video} has a indirect reference to {ex}")
                allIndirects.append(video)

    print(f"total indirect video to examplar links {len(allIndirects)}")

    # not making changes to the DB so this isn't needed:
    # session.commit()
