#! /usr/bin/python3.7
"""
Copyright 2021 by Kitware, Inc.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

 * Redistributions of source code must retain the above copyright notice,
   this list of conditions and the following disclaimer.

 * Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.

 * Neither name of Kitware, Inc. nor the names of any contributors may be used
   to endorse or promote products derived from this software without specific
   prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

==============================================================================

"""
from __future__ import division, print_function
import os
import argparse
import json
from pathlib import Path
import numpy
from meva_server.meva_database.database import MevaDatabase
from meva_server.meva_database.models import Detection, Track, Source, Response
from meva_server.meva_database.utils import turnOffSqlAlchemyLogging
from sqlalchemy import and_, case, create_engine, not_, or_

if __name__ == "__main__":

    print('Reading arguments...')
    # parse the arguments - required for location of database
    parser = argparse.ArgumentParser(
        description='Script to see how many links are confirmed')
    parser.add_argument('databasePath', help='path to database file')
    args = parser.parse_args()

    # connect to the db
    print('Setting up db...')
    database_path = Path(args.databasePath)
    os.environ['MEVA_DATABASE_URI'] = 'sqlite:///' + str(database_path)
    meva_db = MevaDatabase()
    session = meva_db.Session()

    turnOffSqlAlchemyLogging()

    # grab all tracks
    print('Loading all tracks...')
    allTracks = session.query(Track).all()
    print('Loading responses...')
    responses = session.query(Response).all()
    print('Loading sources...')
    sources = session.query(Source).all()

    print('Loading Detections...')
    detections = session.query(Detection)\
        .where(Detection.frame_number % 300 == 0)\
        .all()

    print('Grouping Detections by TrackId')
    detection_by_trackId = {}
    for detection in detections:
        if detection.track_uid not in detection_by_trackId.keys():
            detection_by_trackId[detection.track_uid] = []
        detection_by_trackId[detection.track_uid].append({
            'frame_number': detection.frame_number,
            'time': detection.time,
            'bounding_box':detection.bounding_box.replace('[','').replace(']','').split()
        })

    print(f"Length of detections: {len(detection_by_trackId.keys())}")
    print('Sorting source by uid...')
    sourceByUid = {}
    for src in sources:
        sourceByUid[src.uid] = src

    print('Loading exemplars...')
    exemplars = {}
    tracks = {}
    for track in allTracks:
        # make the track easily referenced by uid
        tracks[track.uid] = track
        if track.exemplar:
            exemplars[track.uid] = track

    # sort the responses
    print('Sorting responses...')
    yeses = []
    nos = []
    idks = []
    for resp in responses:
        if resp.response == 1:
            yeses.append(resp)
        elif resp.response == 0:
            nos.append(resp)
        else:
            idks.append(resp)

    print(f"Yes responses: {len(yeses)}")
    print(f"No responses: {len(nos)}")
    print(f"IDK responses: {len(idks)}")

    # loop through the yeses and get the details
    positiveLinks = {}
    cnt = 0
    actors = {}
    vid_to_vid = []
    vid_to_actor = {}
    
    for yes in yeses:
        src = {}
        link = {}
        t1 = yes.track1_uid
        t2 = yes.track2_uid
        src1 = sourceByUid[tracks[t1].source_uid]
        src2 = sourceByUid[tracks[t2].source_uid]
        if t1 in exemplars:
            # t1 is actor
            vid_to_actor[t2] = t1
            src['type'] = 'actor'
            src['actorId'] = src1.girder_id
            src['date'] = src1.name.split('-')[1]
            src['outfitId'] = src1.outfit_id
            # next must be video
            link['folderId'] = src2.uid
            link['name'] = src2.name
            link['trackId'] = tracks[t2].video_track_id
            link['type'] = 'track'
            if t2 in detection_by_trackId.keys():
                link['detections'] = detection_by_trackId[t2]
            else:
                continue
        elif t2 in exemplars:
            # t2 is actor
            vid_to_actor[t1] = t2
            src['type'] = 'actor'
            src['actorId'] = src2.girder_id
            src['date'] = src2.name.split('-')[1]
            src['outfitId'] = src2.outfit_id
            # next must be video
            link['folderId'] = src1.uid
            link['name'] = src1.name
            link['trackId'] = tracks[t1].video_track_id
            link['type'] = 'track'
            if t1 in detection_by_trackId.keys():
                link['detections'] = detection_by_trackId[t1]
            else:
                continue
        else:
            # neither are actors
            src['folderId'] = src1.uid
            src['trackId'] = tracks[t1].video_track_id
            src['type'] = 'track'
            link['folderId'] = src2.uid
            link['name'] = src2.name
            link['trackId'] = tracks[t2].video_track_id
            link['type'] = 'track'

        positiveLinks[cnt] = {}
        positiveLinks[cnt]['source'] = src
        if src['type'] == 'actor':
            identity = f"{src['actorId']}-{src['date']}-{src['outfitId']}"
            if identity not in actors:
                actors[identity] = {
                    "actorId": src['actorId'],
                    "date": src['date'],
                    "outfitId": src['outfitId'],
                    "type": 'actor',
                    "identity": f"{src['actorId']}-{src['date']}-{src['outfitId']}",
                    "links": []
                }
            if link not in actors[identity]['links']:
                actors[identity]['links'].append(link)
        if src['type'] == 'track' and link['type'] == 'track':
            vid_to_vid.append([t1, t2])

        positiveLinks[cnt]['link'] = link
        cnt += 1

    print(f"Length of vid to vid Links: {len(vid_to_vid)}")
    print(f"Length of vid to actor Links: {len(vid_to_actor.keys())}")

    vid_to_vid_count = 0
    vid_actor_keys = vid_to_actor.keys()
    for vid in vid_to_vid:
        if vid[0] in vid_actor_keys or vid[1] in vid_actor_keys:
            print("found Link")
            vid_to_vid_count += 1

    print(f"Found {vid_to_vid_count} vid-to-vid-to-actor links")
    actor_list = list(actors.values()) 
    total_links = len(session.query(Response).filter(Response.response == 1.0).all())/2
    links = actor_list

    actor_ids = {}
    total_outfit_links = 0
    locations_per_outfit = 0
    for outfit in links:
        if outfit["actorId"] not in actor_ids.keys():
            actor_ids[outfit["actorId"]] = True
        total_outfit_links += len(outfit["links"])
        outfit_vid_map = {}
        for outfitlinks in outfit["links"]:
            if outfitlinks['folderId'] not in outfit_vid_map:
                outfit_vid_map[outfitlinks['folderId']] = True
        locations_per_outfit += len(outfit_vid_map.keys())
        
    total_actors = len(actor_ids.keys())           

    frame_overlap_tracks = session.query(Track, Response, Detection)\
        .where(and_(Track.gps_ids == None, Track.uid == Response.track1_uid, Response.response == 1, Detection.track_uid == Response.track1_uid, Detection.frame_number % 300 == 0))\
        .group_by(Detection.track_uid)\
        .all()


    data = { 
        "Outfits Linked": len(links),
        "Total Links (Including Track to Track)": total_links,
        "Total Actors by id Linked": total_actors,
        "Total Outfit links": total_outfit_links,
        "AVG Tracks Per Outfit": total_outfit_links / len(links),
        "AVG Locations Per Outfit": locations_per_outfit / len(links),
        "AVG Locations Per PersonId": locations_per_outfit / total_actors,
        "Linked Tracks on Super Key Frames": (len(frame_overlap_tracks))
    }
    print(data)
    with open('output.json', 'w') as outfile:
        json.dump(actor_list, outfile, indent=4, sort_keys=True)

