#! /usr/bin/python3.7
"""
Copyright 2021 by Kitware, Inc.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

 * Redistributions of source code must retain the above copyright notice,
   this list of conditions and the following disclaimer.

 * Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.

 * Neither name of Kitware, Inc. nor the names of any contributors may be used
   to endorse or promote products derived from this software without specific
   prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

==============================================================================

"""
from __future__ import division, print_function
import os
import numpy as np
import argparse
import pytz
import logging
from pathlib import Path
import subprocess
from meva_server.meva_database.database import MevaDatabase
from meva_server.meva_database.models import Detection, Track, \
    Gps
from meva_server.meva_database.utils import (
    getScoreForPositions, getLocationFromString, getCovarianceFromString,
    turnOffSqlAlchemyLogging, getBoundingBoxFromString,
    getLocationAndCovariantFromGPSLog
)

# some default values that can be overriden by the args passed in
_timezone = 'US/East-Indiana'
_latitude = 39.04977294
_longitude = -85.52924953
_height = 205
_meters_err = 9
# time sync error captures uncertainty in the times assigned to
# videos with respect to other videos or gps logs and we assume that
# the true time is equal the reported time plus or minus the time sync error
_timeSyncError = 3
# threshold value for score from getScoreForPositions encoding the threshold
# for certainty of overlap of the localization ellipsoids
_scoreThreshold = 1e-4  # unitless
# don't consider the a match if it only has less than this (
# links between gps logs and detection localizations):
_minLinks = 5
# the chips shouldn't be used if they are too small:
_minChipHeight = 75
_minChipWidth = _minChipHeight/3


if __name__ == "__main__":
    # parse the arguments - required for location of database
    parser = argparse.ArgumentParser(
        description='Script to advise possible links')
    parser.add_argument('databasePath', help='path to database file')
    parser.add_argument('--timezone', help='timezone for cameras',
                        default=_timezone)
    parser.add_argument('--latitude', help='latitude in degrees for center point',
                        default=_latitude, type=float)
    parser.add_argument('--longitude', help='longitude in degrees for center point',
                        default=_longitude, type=float)
    parser.add_argument('--height', help='meters above WGS84 ellipsoid',
                        default=_height, type=float)
    parser.add_argument('--seconds', help='window of time to search in seconds',
                        default=_timeSyncError, type=float)
    args = parser.parse_args()

    timeSyncError = args.seconds
    latitude0 = args.latitude  # degrees
    longitude0 = args.longitude  # degrees
    height0 = args.height  # meters above WGS84 ellipsoid
    localTZ = pytz.timezone(args.timezone)

    # connect to the db
    print('Setting up db...')
    database_path = Path(args.databasePath)
    os.environ['MEVA_DATABASE_URI'] = 'sqlite:///' + str(database_path)
    meva_db = MevaDatabase()
    session = meva_db.Session()

    turnOffSqlAlchemyLogging()

    # grab all tracks
    print('Loading all tracks...')
    tracks = session.query(Track).all()
    print('Loading gps logs...')
    gpspoints = session.query(Gps).all()

    print('Loading detections...')
    track_detections = {}
    # For track get all the associated detections
    for track in tracks:
        track_detections[track.uid] = (
            session.query(
                Detection.frame_number,
                Detection.time,
                Detection.location,
                Detection.location_cov,
                Detection.bounding_box
            )
            .filter(Detection.track_uid == track.uid)
            .all()
        )

    # temporary: only run a few
    #tracks = tracks[:3]

    # run through all the tracks and see if any of it's detections
    # have a time that matches the gps logs - then check those for
    # location and covariance proximity to the gps log
    possibleMatchingGpsIds = {}
    print('Checking tracks from the same position...')
    for track in tracks:
        # query for this track's detections
        for det in track_detections[track.uid]:
            # skip non geolocated
            if det.location is None:
                continue

            # check the size of the chip to rule out small ones
            # since they are hard for annotators to confirm
            bb = getBoundingBoxFromString(det.bounding_box)
            if len(bb) < 4:
                continue
            # calc a size
            width = int(bb[2])-int(bb[0])
            height = int(bb[3])-int(bb[1])
            if width < _minChipWidth or height < _minChipHeight:
                print('too small: ', width, height)
                continue

            # check for a matching gps log
            gpsLogs = []
            for wp in gpspoints:
                if abs(wp.time - det.time) < timeSyncError:  # 5 minute video interval
                    gpsLogs.append(wp)

            if len(gpsLogs) == 0:
                continue
            # print('potential matches: ', len(gpsLogs))

            # convert the items into np.arrays
            loc1 = getLocationFromString(det.location)
            cov1 = getCovarianceFromString(det.location_cov)

            # check for how close they are
            for wp in gpsLogs:
                loc2, cov2 = getLocationAndCovariantFromGPSLog(wp, _meters_err,
                                                               latitude0,
                                                               longitude0,
                                                               height0)

                score = getScoreForPositions(loc1.astype('float64'), cov1,
                                             loc2, cov2)
                #print(f"score: {score}")

                if score > _scoreThreshold:
                    if track.uid not in possibleMatchingGpsIds.keys():
                        possibleMatchingGpsIds[track.uid] = {}
                    gpsID = wp.name[:4]
                    if gpsID not in possibleMatchingGpsIds[track.uid].keys():
                        possibleMatchingGpsIds[track.uid][gpsID] = 0
                    else:
                        possibleMatchingGpsIds[track.uid][gpsID] += 1

    # loop through the possible matches
    for pm in possibleMatchingGpsIds:

        for gpsID in possibleMatchingGpsIds[pm]:
            msg = 'Track [' + pm + \
                '] has ' + str(possibleMatchingGpsIds[pm][gpsID]) + \
                ' detection(s) nearby GPS ID ' + gpsID
            print(msg)

            if possibleMatchingGpsIds[pm][gpsID] > _minLinks:
                for track in tracks:
                    if track.uid == pm:
                        # save these to the track for use in suggesting
                        tracks_gps_ids = track.gps_ids
                        if tracks_gps_ids is not None:
                            gpsID = ', ' + gpsID
                        else:
                            tracks_gps_ids = ''
                        track.gps_ids = tracks_gps_ids + gpsID
