#! /usr/bin/python3.7
"""
Copyright 2021 by Kitware, Inc.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

 * Redistributions of source code must retain the above copyright notice,
   this list of conditions and the following disclaimer.

 * Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.

 * Neither name of Kitware, Inc. nor the names of any contributors may be used
   to endorse or promote products derived from this software without specific
   prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

==============================================================================

"""
from __future__ import division, print_function
import os
import argparse
from itertools import combinations
from pathlib import Path
import time
import pytz
from datetime import datetime
import pandas as pd
from multiprocessing import Pool
from meva_server.meva_database.database import MevaDatabase
from meva_server.meva_database.models import Detection, Source, Track, \
    Gps, MustNotLink, Afinity, Response
from meva_server.meva_database.utils import (
    turnOffSqlAlchemyLogging,
    isTimeOverlap,
    getLocationFromString,
    getCovarianceFromString,
    getLocationAndCovariantFromGPSLog,
    getScoreForPositions
)

# some default values that can be overriden by the args passed in
_latitude = 39.04977294
_longitude = -85.52924953
_height = 205

# time sync error captures uncertainty in the times assigned to
# videos with respect to other videos or gps logs and we assume that
# the true time is equal the reported time plus or minus the time sync error
_timeSyncError = 3  # seconds

# threshold value for score from getScoreForPositions encoding the threshold
# for certainty of overlap of the localization ellipsoids
_scoreThreshold = 1e-4  # unitless

# default start value for comparing to a minimum val
_infinity = 9999999999.0

# default value used to ref covariance in gps logs
_meters_err = 9

# variables used for saving memory in the pool for mp
_track_detections = {}
_track_gpslogs = {}
_detectMinTime = {}
_detectMaxTime = {}
_useGPS = False


def sequentialComparison(detections1, detections2,
                         t1Source, t2Source):
    # compare them sequentially since they are video to video

    # we only have something to say about affinity if there is ever
    # a pair of detections that "overlap, meaning they share states that
    # are close enough in time
    anyOverlap = False
    outerScore = _infinity

    minVal = min(len(detections1), len(detections2))
    for x in range(0, minVal):
        # these are both video comparisons so the frame rate is
        # assumed to be within the time sync error:
        det1 = detections1[x]
        det2 = detections2[x]

        # Check if tracks appear in the same frame of a video
        # only do this for video to video comparisons
        if (t1Source == t2Source):  # same video source
            if (det1.frame_number == det2.frame_number):  # same frame num
                # these can't be the same
                return 0

        # if we made it this far - there is overlap in the 2 tracks (time)
        anyOverlap = True

        # Check localization ellipsoids
        new_score = 0
        loc1 = None
        loc2 = None

        if det1.location is not None and det2.location is not None:
            loc1 = getLocationFromString(det1.location)
            cov1 = getCovarianceFromString(det1.location_cov)
            loc2 = getLocationFromString(det2.location)
            cov2 = getCovarianceFromString(det2.location_cov)
            new_score = getScoreForPositions(loc1, cov1, loc2, cov2)

        # keep the min of the scores
        outerScore = min(outerScore, new_score)
        if outerScore < _scoreThreshold and anyOverlap:
            return outerScore

    if anyOverlap:
        return outerScore
    else:
        return _infinity  # return unsure


def gpsComparison(videoDetects, gpsDetects, debugTrack):
    # assuming that detections 1 is the video and
    # detections 2 is the GPS

    # we only have something to say about affinity if there is ever
    # a pair of detections that "overlap, meaning they share states that
    # are close enough in time
    anyOverlap = False
    outerScore = _infinity

    minVal = min(len(videoDetects), len(gpsDetects))
    for x in range(0, minVal):
        det1 = gpsDetects[x]

        # get a list of all detections within this one's time sync period
        # we have to do this because of the mix in frame rate vs
        # gps log saving time
        matchingTimes = [x for x in videoDetects if abs(det1.time - x.time)
                         < timeSyncError]

        if len(matchingTimes) < 1:
            continue  # skip these that we don't have matches for

        innerScore = 0
        for det2 in matchingTimes:
            # if we made it this far - there is overlap in the 2 tracks (time)
            anyOverlap = True

            # Check localization ellipsoids
            new_score = 0
            loc1 = None
            loc2 = None

            if det2.location is not None:
                loc1 = getLocationFromString(det2.location)
                cov1 = getCovarianceFromString(det2.location_cov)

            loc2, cov2 = getLocationAndCovariantFromGPSLog(det1,
                                                           _meters_err,
                                                           _latitude,
                                                           _longitude,
                                                           _height)
            if loc1 is not None and loc2 is not None:
                new_score = getScoreForPositions(loc1, cov1, loc2, cov2)

            # store the highest match
            innerScore = max(innerScore, new_score)
        if debugTrack:
            print(f"inner: {innerScore}")
        # keep the min of the maximums
        outerScore = min(outerScore, innerScore)
        if outerScore < _scoreThreshold and anyOverlap:
            return outerScore

    if anyOverlap:
        return outerScore
    else:
        return _infinity  # return unsure


def checkForMustNotLinks(args):
    """
    Check two tracks to see if they should not be linked. Once it is determined
    the two tracks should not be linked the function returns.

    :return: (MustNotLink (True/False), Affinity score (0 to 1 float))
    """
    (track1, track2) = args
    mustNotLink = False
    detections1 = []
    detections2 = []
    gpsFlag = 0  # this indicates if any of the detections are of type gps
    t1TimeUID = track1.uid
    t2TimeUID = track2.uid

    # Two exemplars should never be linked
    if (track1.exemplar is True and track2.exemplar is True):
        mustNotLink = True
        return (mustNotLink, 0.0)
    # see if one of the two is an examplar
    elif track1.exemplar is True:
        if track1.uid in _track_gpslogs.keys():
            detections1 = _track_gpslogs[track1.uid]
        else:
            # no gps match so we need to ignore this one
            return (False, 0.0)
        detections2 = _track_detections[track2.uid]
        gpsFlag = 1  # the first item is a gps log
        t1TimeUID = detections1[0].name
    elif track2.exemplar is True:
        detections1 = _track_detections[track1.uid]
        if track2.uid in _track_gpslogs:
            detections2 = _track_gpslogs[track2.uid]
        else:
            # no gps match so we need to ignore this one
            return (False, 0.0)
        gpsFlag = 2  # the second item is a gps log
        t2TimeUID = detections2[0].name
    else:  # two video tracks
        detections1 = _track_detections[track1.uid]
        detections2 = _track_detections[track2.uid]

    win_start_time = max(_detectMinTime[t1TimeUID],
                         _detectMinTime[t2TimeUID]) - _timeSyncError
    win_end_time = min(_detectMaxTime[t1TimeUID],
                       _detectMaxTime[t2TimeUID]) + _timeSyncError
    if win_start_time >= win_end_time:
        return (False, 0.0)

    detections1 = [x for x in detections1 if x.time >=
                   win_start_time and x.time <= win_end_time]
    detections2 = [x for x in detections2 if x.time >=
                   win_start_time and x.time <= win_end_time]

    # Loop over all detections for each track
    score = 0
    debugTrack = False
    if gpsFlag == 0:
        score = sequentialComparison(detections1, detections2,
                                     track1.source_uid,
                                     track2.source_uid)
    elif gpsFlag == 1:

        if track1.uid == '388e95c766f142aa99d6c0b1dd57d1a5':
            debugTrack = True
        score = gpsComparison(detections2, detections1, debugTrack)
    else:
        if track1.uid == '388e95c766f142aa99d6c0b1dd57d1a5':
            debugTrack = True
        score = gpsComparison(detections1, detections2, debugTrack)

    # check the affinity score to see if the MustNotLink is true
    # if the score is less than the threshold then we set mnl=true
    if score < _scoreThreshold and score != _infinity:
        mustNotLink = True

    # send the final score
    return (mustNotLink, score)


if __name__ == "__main__":
    print('Reading arguments...')
    time00 = time.time()
    # parse the arguments - required for location of database
    parser = argparse.ArgumentParser(
        description='Script to determine which tracks should not be linked')
    parser.add_argument('databasePath', help='path to database file')
    parser.add_argument('--seconds', help='window of time to search in seconds',
                        default=_timeSyncError, type=float)
    parser.add_argument('--usegps', help='use gps and camera model detections to make suggested links',
                        default=_useGPS, type=bool)
    args = parser.parse_args()

    timeSyncError = args.seconds
    useGPS = args.usegps

    # connect to the db
    print('Setting up db...')
    database_path = Path(args.databasePath)
    os.environ['MEVA_DATABASE_URI'] = 'sqlite:///' + str(database_path)
    meva_db = MevaDatabase()
    session = meva_db.Session()

    turnOffSqlAlchemyLogging()

    # grab all tracks
    print('Loading all tracks...')
    allTracks = session.query(Track).all()
    print('Loading existing MustNotLinks...')
    existingMustNotLinks = session.query(MustNotLink).all()
    print('Loading existing Afinities...')
    existingAfinities = session.query(Afinity).all()
    print('Loading gps logs...')
    gpslogs = session.query(Gps).all()
    print('Loading sources...')
    sources = session.query(Source).all()
    print('Loading all detections...')
    allDetections = session.query(Detection).all()
    print('Loading responses...')
    responses = session.query(Response).all()

    # list of uids from the existing items in the db
    existMNL_t1 = [em.track1_uid for em in existingMustNotLinks]
    existMNL_t2 = [em.track2_uid for em in existingMustNotLinks]
    eMNLs = pd.DataFrame({'t1': existMNL_t1, 'value': existMNL_t2})
    existAff_t1 = [em.track1_uid for em in existingAfinities]
    existAff_t2 = [em.track2_uid for em in existingAfinities]
    eAffs = pd.DataFrame({'t1': existAff_t1, 'value': existAff_t2})
    print(f"existing mnls: {len(existMNL_t1)}")
    print(f"existing Afinities: {len(existAff_t1)}")

    # Links to be added
    mnls = []
    affinities = []

    print('Sorting source by times...')
    time1 = time.time()
    sourceTimes = {}
    for src in sources:
        if src.uid in sourceTimes.keys():
            continue
        sourceTimes[src.uid] = {}
        sourceTimes[src.uid]['start'] = src.start_time
        sourceTimes[src.uid]['end'] = src.end_time
    time2 = time.time()
    print(f"time: {time2-time1}")

    print('Sorting detections by track uid...')
    time1 = time.time()
    detects = {}
    detectMinTime = {}
    detectMaxTime = {}
    for detect in allDetections:
        if detect.track_uid not in detects.keys():
            detects[detect.track_uid] = []
        detects[detect.track_uid].append(detect)
        # grab the min/max times
        if detect.track_uid not in detectMinTime.keys():
            detectMinTime[detect.track_uid] = 99999999999999
            detectMaxTime[detect.track_uid] = 0
        detectMaxTime[detect.track_uid] = max(detectMaxTime[detect.track_uid], detect.time)
        detectMinTime[detect.track_uid] = min(detectMinTime[detect.track_uid], detect.time)
    time2 = time.time()
    print(f"time: {time2-time1}")
    # save for the pool in Multiprocessing
    _detectMinTime = detectMinTime
    _detectMaxTime = detectMaxTime

    print('Sorting gps logs by id...')
    time1 = time.time()
    gpsSorted = {}
    for gps in gpslogs:
        if gps.name not in gpsSorted.keys():
            gpsSorted[gps.name] = []
        gpsSorted[gps.name].append(gps)
        # grab the min/max times
        if gps.name not in detectMinTime.keys():
            detectMinTime[gps.name] = 99999999999999
            detectMaxTime[gps.name] = 0
        detectMaxTime[gps.name] = max(detectMaxTime[gps.name], gps.time)
        detectMinTime[gps.name] = min(detectMinTime[gps.name], gps.time)
    time2 = time.time()
    print(f"time: {time2-time1}")

    print('Loading detections...')
    time1 = time.time()
    track_detections = {}
    tracksByDay = {}
    track_gpslogs = {}
    exemplarCount = 0
    exemplars = {}
    # For track get all the associated detections
    localTZ = pytz.timezone('US/East-Indiana')
    for track in allTracks:
        if track.uid in detects.keys():
            track_detections[track.uid] = detects[track.uid]

        # separate by day
        sTime = sourceTimes[track.source_uid]['start']
        day = datetime.utcfromtimestamp(sTime).strftime("%D")
        # print(f'sTime: {sTime} day: {day}')
        # make the track easily referenced by uid
        if day not in tracksByDay.keys():
            tracksByDay[day] = {}
        tracksByDay[day][track.uid] = track

        # get the gps locations per exemplar tracks
        # print(f"is exemplar: {track.exemplar}", end='\r')
        if track.exemplar is True:
            exemplarCount += 1
            exemplars[track.uid] = track
            if track.gps_ids in gpsSorted:
                track_gpslogs[track.uid] = gpsSorted[track.gps_ids]

    time2 = time.time()
    print(f"time: {time2-time1}")
    # save for the pool in MP
    _track_detections = track_detections
    _track_gpslogs = track_gpslogs

    print(f"number of exemplar tracks: {exemplarCount}")

    print(f"tracks with detections: {len(track_detections.keys())}")
    print(f"tracks with gps: {len(track_gpslogs.keys())}")

    print('sorting responses...')
    yeses = {}
    for resp in responses:
        if resp.response == 1:
            if resp.track1_uid in exemplars.keys():
                yeses[resp.track1_uid] = resp.track2_uid
            elif resp.track2_uid in exemplars.keys():
                yeses[resp.track2_uid] = resp.track1_uid
    print(f"number of positive responses with exemplars: {len(yeses.keys())}")

    # if we have a Positive response to an exemplar than
    # make a MNL to other exemplars
    rMNL_t1 = []
    rMNL_t2 = []
    for yes in yeses:
        for exemp in exemplars:
            if yes != exemp:
                rMNL_t1.append(yeses[yes])
                rMNL_t2.append(exemp)
                rMNL_t2.append(yeses[yes])
                rMNL_t1.append(exemp)
    # create a dataframe to easily add to existing
    rMNLs = pd.DataFrame({'t1': rMNL_t1, 'value': rMNL_t2})
    print(f"response MNLs: {len(rMNL_t1)}")
    # subtract any that are already in the MNLs
    newMNLs = pd.merge(eMNLs, rMNLs, indicator=True, on=['t1', 'value'],
                       how='right').query("_merge != 'both'").drop('_merge', axis=1).reset_index(drop=True)
    print(f"subtracted existing MNLs: {len(newMNLs.index)}")
    # now add the left over
    concatFrames = [eMNLs, newMNLs]
    eMNLs = pd.concat(concatFrames)
    # add to our mnls list
    for row in newMNLs.itertuples():
        mnl = MustNotLink(track1_uid=row.t1,
                          track2_uid=row.value)
        mnls.append(mnl)
    print(f"new response MNLs: {len(mnls)}")

    newMNL_t1 = []
    newMNL_t2 = []
    newAff_t1 = []
    newAff_t2 = []
    newAff_score = []
    gpsReAdds = {}  # going to add these back in later
    for day in tracksByDay:
        print(f"day: {day}")
        dayTracks = tracksByDay[day]
        tracks = []
        for dt in dayTracks:
            tracks.append(dayTracks[dt])

        # get tracks per gps on this day
        tracksByGPS = {}
        for track in tracks:
            # store the track info by GPS ID, so these can be reduced
            # since there were multiple outfits
            if track.gps_ids not in tracksByGPS.keys():
                tracksByGPS[track.gps_ids] = []
            tracksByGPS[track.gps_ids].append(track.uid)

        # Temporary limit the number of tracks
        # tracks = list(tracks)[1800:2200]
        print(f"all tracks: {len(tracks)}")
        print(f"GPS Tracks: {len(tracksByGPS.keys())}")

        # create a DataFrame of all possible track pairs:
        t1 = []
        t2 = []
        tAll = 0
        time1 = time.time()
        print('checking for time overlaps')
        for track1, track2 in combinations(tracks, 2):
            tAll += 1
            time_overlap = isTimeOverlap(track1, track2,
                                         detects, gpsSorted,
                                         _timeSyncError)
            # look at the times to see if they overlap
            if not time_overlap:
                continue

            t1.append(track1.uid)
            t2.append(track2.uid)
        time2 = time.time()
        print(f"time: {time2-time1}")

        # create a dataframe from all possible
        allPoss = pd.DataFrame({'t1': t1, 'value': t2})

        print(f"possible combinations of tracks: {tAll}")
        print(f"combinations after looking for time overlap: {len(t1)}")

        # get the list of all possible minus already there:
        cleaned = pd.merge(allPoss, eMNLs, indicator=True, on=['t1', 'value'],
                           how='left').query("_merge != 'both'").drop('_merge', axis=1).reset_index(drop=True)
        print(f"remaining after removing existing mnls: {len(cleaned.index)}")

        # create a list of the exemplars that are repeats
        repeatExemplars = []
        for gpdsID in tracksByGPS:
            # leave off the first one so we only subtract the others
            repeatExemplars += tracksByGPS[gpdsID][1:]
            gpsReAdds[tracksByGPS[gpdsID][0]] = tracksByGPS[gpdsID][1:]
        print(f"gps repeats that we can chop {len(repeatExemplars)}")

        updates = []
        # ============= master loop over track pairs =============
        print('looping through tracks...')
        time1 = time.time()
        cnt = 0
        trackPairs = []
        for index, row in cleaned.iterrows():
            cnt += 1
            print(f"percent complete: {100*(cnt/len(cleaned.index)):.2f}",
                  end='\r')

            track1 = None
            track2 = None
            if row['t1'] in tracksByDay[day].keys():
                track1 = tracksByDay[day][row['t1']]
            else:
                print(f"no track with uid: {row['t1']}")
            if row['value'] in tracksByDay[day].keys():
                track2 = tracksByDay[day][row['value']]
            else:
                print(f"no track with uid: {row['value']}")

            if track1 is None or track2 is None:
                continue

            # save the tracks to view
            trackPairs.append((track1, track2))

        print(f"tracks to compare w/o gps repeats {len(trackPairs)}")

        if useGPS:
            # create the pooler for multi processing
            cpus = os.cpu_count()
            if cpus > 1:
                cpus -= 1  # don't use all of them if possible
            print(f"creating pool with {cpus} processes")
            chunks = int(len(trackPairs)/cpus) + 2
            print(f"chunk size: {chunks}")
            # run the pool
            results = []
            with Pool(cpus) as p:
                for i in p.imap(checkForMustNotLinks, trackPairs,
                                chunksize=chunks):
                    results.append(i)
            print('finished pooling')

            # read through results
            for tp in range(0, len(trackPairs)):
                mustNotLink = results[tp][0]
                score = results[tp][1]
                track1 = trackPairs[tp][0]
                track2 = trackPairs[tp][1]

                if (mustNotLink):
                    newMNL_t1.append(track1.uid)
                    newMNL_t2.append(track2.uid)
                    newMNL_t1.append(track2.uid)
                    newMNL_t2.append(track1.uid)
                else:
                    # don't add scores with the default value
                    if score == _infinity or score == 0:
                        continue

                    newAff_t1.append(track1.uid)
                    newAff_t2.append(track2.uid)
                    newAff_score.append(score)
                    newAff_t2.append(track1.uid)
                    newAff_t1.append(track2.uid)
                    newAff_score.append(score)

            time2 = time.time()
            print(f"time: {time2-time1}")
        else:
            print('!!!!!!!!!!!!!!!!!! not using GPS !!!!!!!!!!!!!!!!!!')
            # no need to update affinities - it will pull from tracks

    # check to see if new MNLs collide with the old ones
    rMNLs = pd.DataFrame({'t1': newMNL_t1, 'value': newMNL_t2})
    rMNLs = rMNLs.drop_duplicates()
    # subtract any that are already in the MNLs
    newMNLs = pd.merge(eMNLs, rMNLs, indicator=True, on=['t1', 'value'],
                       how='right').query("_merge != 'both'").drop('_merge', axis=1).reset_index(drop=True)
    # build the MNLs that don't collide
    for row in newMNLs.itertuples():
        mnl = MustNotLink(track1_uid=row.t1,
                          track2_uid=row.value)
        mnls.append(mnl)
    print(f"new MNLs: {len(newMNL_t1)}")

    # check to see if there are new affinities
    affSet = pd.DataFrame({'t1': newAff_t1, 'value': newAff_t2,
                          'score': newAff_score})
    newAffs = pd.merge(eAffs, affSet, indicator=True, on=['t1', 'value'],
                       how='right').query("_merge != 'both'")
    # add new ones
    for row in newAffs.itertuples():
        afinity1 = Afinity(
                        track1_uid=row.t1,
                        track2_uid=row.value,
                        score=row.score
        )
        affinities.append(afinity1)
    # updates the existing ones
    upAffs = pd.merge(eAffs, affSet, indicator=True, on=['t1', 'value'],
                      how='inner').query("_merge == 'both'")
    print(f"affinities score update: {len(upAffs.index)}")
    if len(list(upAffs.itertuples())) > 0:
        affinity_hash = {}
        for row in existingAfinities:
            affinity_hash[row.track1_uid] = row
        print(f"Length of affinity_hash: {len(affinity_hash.keys())}")

        for row in upAffs.itertuples():
            uid1 = row.t1
            uid2 = row.value
            score = row.score
            if uid1 in affinity_hash.keys() and affinity_hash[uid1].track2_uid == uid2:
                # only update if we have a big change in score
                diff_in_score = abs(affinity_hash[uid1].score - score)
                if diff_in_score > 0.2:
                    affinity_hash[uid1].score = score
                #else:
                #    print('skipping scoring')

    print(f"affinities: {len(affinities)}")

    # append to DB
    print('adding to afinity table...')
    session.add_all(affinities)
    if len(newMNL_t1) > 0:
        print('adding to MustNotLink table...')
        session.add_all(mnls)
    print('saving...')
    session.commit()
    time2 = time.time()
    print(f"Total time: {time2-time00}")
