'''
This script sets up CPU descriptor extraction with a reid net found here:
https://github.com/michuanhaohao/reid-strong-baseline/
It then compares to the existing tensors that were saved for checkin photos
'''
import time
from people_segmentation.pre_trained_models import create_model
from iglovikov_helper_functions.dl.pytorch.utils import tensor_from_rgb_image
from iglovikov_helper_functions.utils.image_utils import load_rgb, pad, unpad
import albumentations as albu
from torchvision.transforms.functional import normalize
import sys
# settings (here is the main path used for others below)
data_path = '/home/local/KHQ/joseph.vanpelt/projects/meva/'

# use the code from cloning the above repo
sys.path.append(data_path + 'reid/reid-strong-baseline')
from modeling import build_model as build_dg_model
from config import cfg as dg_cfg
import numpy as np
import glob
import torch
from PIL import Image
import matplotlib.pyplot as plt
from torchvision.transforms import ToTensor, ToPILImage
import torch.nn.functional as F
import pickle
import operator
import pathlib
import cv2

# which net?
#netname = 'market_resnet50_model_120_rank1_945'
netname = 'mcdr_resnet50_checkpoint_211376'


# note: assumes this approach for the date
test_file = data_path + 'test1_20180306.jpg'
test_date = pathlib.Path(test_file).name.split('_')[1][:-4]
config_path = data_path + 'reid/reid-strong-baseline/configs/softmax_triplet_with_center.yml'
model_path = data_path + 'reid/' + netname + '.pt'  # '.pth'
checkin_photos = data_path + 'tracks/links/all/'
#checkin_photos = data_path + 'tracks/links/masked/'
checkins_path = data_path + 'reid/checkin_desciptors_' + netname
viewIt = True
_saliency = True


def tensorletterbox(x, height=256, width=128, fillvalue=0) -> torch.Tensor:
    shape = x.shape[2:]
    ratio = min(float(height) / shape[0], float(width) / shape[1])

    new_shape = (round(shape[0] * ratio), round(shape[1] * ratio))
    dw = (width - new_shape[1]) / 2  # width padding
    dh = (height - new_shape[0]) / 2
    top, bottom = round(dh - 0.1), round(dh + 0.1)
    left, right = round(dw - 0.1), round(dw + 0.1)
    x = torch.nn.functional.interpolate(x, size=new_shape,
                                        mode='bilinear')
    x = torch.nn.functional.pad(
        x, (left, right, top, bottom), mode='constant', value=fillvalue)
    return x


def provision_model(config_path, model_path):
    dg_cfg.merge_from_file(config_path)
    # Prevent model from trying to init backbone
    dg_cfg['MODEL']['PRETRAIN_CHOICE'] = None
    dg_cfg['MODEL']['PRETRAIN_PATH'] = None
    dg_model = build_dg_model(dg_cfg, 10220)  # 751)
    checkpoint = torch.load(model_path, map_location="cpu")
    dg_model.load_state_dict(checkpoint['model'])  # .load_state_dict(checkpoint)
    return dg_model


def load_image_to_tensor(filepath) -> torch.Tensor:
    img = Image.open(filepath)
    # convert img to tensor
    return ToTensor()(img).unsqueeze(0)


def load_image_and_prepare_tensor(img, pxl_mean,
                                  pxl_std, display=False) -> torch.Tensor:
    #chip = load_image_to_tensor(filepath)
    chip = ToTensor()(img).unsqueeze(0)
    # image letterbox and resize per
    # https://gitlab.kitware.com/afrl_act3/meva-person-search/-/blob/dev/customTracker/mevaTracker/trackers.py#L502
    print(f"normalizing with mean: {pxl_mean}, and std: {pxl_std}")

    model_in = tensorletterbox(chip, height=256, width=128)

    model_in = normalize(model_in, pxl_mean, pxl_std)

    if display:
        plt.imshow(ToPILImage()(torch.squeeze(model_in)))
        plt.show()
    return model_in


def get_cv_from_tensor(tensor):
    img1 = torch.squeeze(tensor).permute(
        1, 2, 0).numpy()
    # swap red and blue?
    #red = img1[:, :, 2].copy()
    #blue = img1[:, :, 0].copy()
    #img1[:, :, 0] = red
    #img1[:, :, 2] = blue
    return img1


def loadImageAndTransform(image):
    # transform the image for the model
    transform = albu.Compose([albu.Normalize(p=1)], p=1)
    padded_image, pads = pad(image, factor=32, border=cv2.BORDER_CONSTANT)
    x = transform(image=padded_image)["image"]
    x = torch.unsqueeze(tensor_from_rgb_image(x), 0)
    return (image, x, pads)


def getMask(model_in, model, pads):
    with torch.no_grad():
        prediction = model(model_in)[0][0]
    mask = (prediction > 0).cpu().numpy().astype(np.uint8)
    mask = unpad(mask, pads)
    return mask


def hideBackground(mask, image):
    cvmask = (cv2.cvtColor(mask, cv2.COLOR_GRAY2RGB) * (255, 255, 255))\
        .astype(np.uint8)
    cvmask = cv2.bitwise_and(image, cvmask)
    return cvmask


def viewImage(image):
    plt.imshow(ToPILImage()(image))
    plt.show()


if __name__ == "__main__":

    time0 = time.time()

    # provision the segmentation model
    seg_model = create_model("Unet_2020-07-20")
    _ = seg_model.eval()

    # load the image
    image = Image.open(test_file)  # load an rgb img
    testImg = np.array(image)  # convert to numpy array format
    #viewImage(testImg)
    image, model_in, pads = loadImageAndTransform(testImg)
    mask = getMask(model_in, seg_model, pads)
    # reverse it
    #cvmask = hideBackground(mask, image)
    #savedImg = hideBackground(mask, image)
    cvmask = image  # cancel the background removal
    savedImg = image  # cancel background removal
    #viewImage(savedImg)

    # Provision the reID model
    print('model setup')
    time1 = time.time()
    dg_model = provision_model(config_path, model_path)
    _ = dg_model.eval()
    print(f"time: {time.time() - time1}")

    # load the image file
    print('loading test image and resizing')
    time1 = time.time()
    model_in = load_image_and_prepare_tensor(cvmask,
                                             dg_cfg.INPUT.PIXEL_MEAN,
                                             dg_cfg.INPUT.PIXEL_STD,
                                             False)
    print(f"time: {time.time() - time1}")

    # compute the test descriptor
    print('getting this test image desc.')
    time1 = time.time()
    if _saliency:
        tempImg = np.array(ToPILImage()(torch.squeeze(model_in)))
        for x1 in tempImg:
            for x2 in x1:
                if x2[0] == 228 and x2[1] == 249 and x2[2] == 52:
                    # set all these to zero
                    x2[0] = 0
                    x2[1] = 0
                    x2[2] = 0
        viewImage(tempImg)
        result = Image.fromarray((tempImg * 255).astype(np.uint8))
        result.save('./out.png')
        model_in.requires_grad_()
        test_tensor = dg_model(model_in)
    else:
        with torch.no_grad():
            test_tensor = dg_model(model_in)
    print(f"time: {time.time() - time1}")

    # review saliency of the ReID model
    if _saliency:
        output_idx = test_tensor.argmax()
        output_max = test_tensor[0, output_idx]
        output_max.backward()
        saliency, _ = torch.max(model_in.grad.data.abs(), dim=1)
        saliency = saliency.reshape(256, 128)
        viewImage(saliency)

    checkins = glob.glob(checkins_path + "/*.desc")

    # compare to the check in desc
    print(f"comparing to checkins with date: {test_date}")
    time1 = time.time()
    best_score = 0
    winner = ''
    scores = {}
    compImgs = []
    for check in checkins:
        # name of the checkin img
        name = check.split('/')[-1].split('.desc')[0]
        # key is the name of the img
        # TODO: limit based on the day
        check_date = check.split('-')[1]
        if test_date != check_date:
            continue

        # read in this one
        checkin = []
        with open(check, 'rb') as jfile:
            checkin = pickle.load(jfile)

        # convert this list back to tensor
        check_tensor = torch.Tensor(checkin)
        sim = F.cosine_similarity(test_tensor, check_tensor).tolist()[0]

        scores[name] = sim
        if sim > best_score:
            best_score = sim
            winner = name
    print(f"winner: {winner}")
    print(f"time: {time.time() - time1}")
    sorted_scores = sorted(scores.items(), key=operator.itemgetter(1), reverse=True)[:49]
    print("  top scores ____________  ")
    for s in sorted_scores:
        print(s)
        compImgs.append(checkin_photos + s[0])

    print(f"Total time: {time.time() - time0}")

    if viewIt:
        import cv2

        # convert to tensor
        savedImg = ToTensor()(savedImg).unsqueeze(0)
        # letterbox and re-size
        main = get_cv_from_tensor(tensorletterbox(savedImg))
        main = cv2.cvtColor(main, cv2.COLOR_BGR2RGB)
        #cv2.imshow('test', main)
        #cv2.waitKey(0)

        vis = None
        total = len(sorted_scores) + 1
        numCols = 10
        rows = round(total/numCols)
        row = 0
        col = 1
        cnt = 0
        rowMat = main
        font = cv2.FONT_HERSHEY_SIMPLEX
        for sc in sorted_scores:
            col += 1
            img2 = get_cv_from_tensor(tensorletterbox(
                load_image_to_tensor(compImgs[cnt])))
            img2 = cv2.cvtColor(img2, cv2.COLOR_BGR2RGB)
            txtColor = (0, 255, 255)
            cv2.putText(img2, str(sc[1])[:7], (12, 120),
                        font, 0.7, txtColor, 2)
            if col <= numCols:
                if rowMat is not None:
                    # if we already have stuff in the row - add to it
                    #cv2.imshow('test', rowMat)
                    #cv2.waitKey(0)
                    #cv2.imshow('test', img2)
                    #cv2.waitKey(0)
                    rowMat = cv2.hconcat([rowMat, img2])
                else:
                    # otherwise save the current one to start the row
                    rowMat = img2
            # if we have a full row of columns then save that
            if col == numCols:
                if row == 0:
                    vis = rowMat  # first row
                else:
                    # additional rows add to this one
                    vis = cv2.vconcat([vis, rowMat])

                rowMat = None  # reset the row collector
                row += 1
                col = 0
            cnt += 1
        print(row)
        #cv2.imshow('test', img2)
        #cv2.waitKey(1500)
        #vis = cv2.hconcat([main, img2])  # or use cv2.vconcat([img1, img2])
        fileOut = pathlib.Path(test_file)

        # this requires: pip install opencv-contrib-python
        cv2.imshow(fileOut.name, vis)
        cv2.waitKey(0)
        cv2.destroyAllWindows()

        # save the file out
        newfile = str(fileOut.parent) + '/' + fileOut.name + "_result.png"
        print(f"saving new file: {newfile}")
        vis = np.uint8(vis*255)
        cv2.imwrite(newfile, vis, [cv2.IMWRITE_PNG_COMPRESSION])
