import base64
from email.policy import default
from enum import auto
import io
import json
from pathlib import Path
import tempfile
from typing_extensions import Required

from numpy import append
from girder import constants
from girder.models.file import File
from girder.models.token import Token
from meva_server.meva_state.models import MEVA_State
import pymongo
from meva_server.constants import (
    ExportMarker,
    FeatureExtractionZip,
    FinalTrackDownloadFolder,
    LinkedFeaturesExportsMarker,
    LinkingFeatureBaseProperties,
    LinkingFeaturesFolderName,
    LinkingFeaturesList,
    LinkingFeaturesPropertiesName,
    StatsFolderName,
    jsonRegex,
)
from meva_server.serializers.models import Track
from re import M
import math

from girder.exceptions import RestException
from meva_server.utils import (
    detections_file,
    detections_item,
    get_filtered_folders,
    getTrackData,
)
from meva_tasks.tasks import calculate_track_stats, create_linking_list, export_feature_links
from meva_utils import FALSY_META_VALUES, TRUTHY_META_VALUES
from girder.api import access
from girder.api.describe import Description, autoDescribeRoute
from girder.api.rest import Resource
from girder.constants import AccessType
from girder.models.folder import Folder
from girder.models.item import Item
from girder.models.upload import Upload
from girder.models.user import User
from meva_server.linking_features.models import LinkingFeature
from girder.utility import ziputil
from girder.api.rest import getApiUrl


class LinkedFeature(Resource):
    def __init__(self):
        super(LinkedFeature, self).__init__()
        self.resourceName = "linked_feature"
        self.route("PUT", (":id",), self.create_linkedfeature)
        self.route("GET", (":id", "all_linkedFeatures"), self.get_all_linkedfeatures)
        self.route("POST", (":id", "init"), self.initializeFeatureFolder)
        self.route("DELETE", (":id",), self.delete_all_features)
        self.route("GET", (":id", "linking_properties"), self.get_linking_properties)
        self.route("POST", (":id", "linking_properties"), self.add_linking_property)
        self.route("POST", (":id", "linkinglist"), self.add_linking_list)
        self.route("GET", (":id", "linkinglist"), self.get_linking_list)
        self.route("GET", (":id", "linking_item"), self.get_linking_item)
        self.route("GET", (":id", "linking_list_item"), self.get_list_item)
        self.route("POST", (":id", "export_linked_features"), self.export_linked_features)
        self.route("GET", (":id", "exported_linked_features"), self.get_exported_linked_features)
        self.route("POST", (":id", "generate_linking_list"), self.generate_linking_list)
        self.route("GET", (":id", "user_progress"), self.get_user_progress)

    def _getUserFolder(self, user, folder):
        user_folder = Folder().findOne(
            {
                "meta.baseMevaFolder": str(folder["_id"]),
                "meta.userLinkingFolder": {'$in': [True, 'true', 'True']},
                "meta.userId": str(user['_id']),
            },
        )
        return user_folder

    @access.user
    @autoDescribeRoute(
        Description("Create a Linked Feature")
        .modelParam(
            "id",
            description="Base MEVA dataset Folder",
            model=Folder,
            level=AccessType.WRITE,
        )
        .jsonParam(
            "data",
            "Data compose of a linking feature",
            required=True,
            requireObject=True,
            paramType='body',
        )
    )
    def create_linkedfeature(self, folder, data):
        user = self.getCurrentUser()
        image_data = []
        print(data)
        if 'featureDrawings' in data.keys():
            featureDrawings = data['featureDrawings']
            actorIdentifier = data['linkedActor']
            sourceClipFilename = data['sourceClipFilename']
            sourceTrackId = data['sourceTrackId']
            user_folder = self._getUserFolder(user, folder)
            for item in featureDrawings:
                newResultItem = Item().createItem(
                    f"{sourceClipFilename}-{sourceTrackId}-{actorIdentifier}-{item['identifier']}.png", user, user_folder
                )
                Upload().uploadFromFile(
                    io.BytesIO(base64.b64decode((item['data']).replace('data:image/png;base64,',''))),
                    size=len(base64.b64decode((item['data']).replace('data:image/png;base64,',''))),
                    name=f"{sourceClipFilename}-{sourceTrackId}-{actorIdentifier}-{item['identifier']}.png",
                    parentType='item',
                    parent=newResultItem,
                    user=user,
                )
                image_data.append(
                    {
                        "identifier": item['identifier'],
                        'girderId': newResultItem['_id'],
                    }
                )
        # now we could be able to create the linked
        data['featureDrawings'] = image_data
        linked = LinkingFeature().createFeature(user, folder, data)
        return linked

    @access.admin
    @autoDescribeRoute(
        Description("Get All Linked Features").modelParam(
            "id",
            description="Base MEVA dataset Folder",
            model=Folder,
            level=AccessType.READ,
        )
    )
    def get_all_linkedfeatures(self, folder):
        linked_features = LinkingFeature().find({'folderId': str(folder['_id'])})
        return linked_features

    @access.user
    @autoDescribeRoute(
        Description("Initialize the linked Folder Structure").modelParam(
            "id",
            description="Base MEVA dataset Folder",
            model=Folder,
            level=AccessType.WRITE,
        )
    )
    def initializeFeatureFolder(self, folder):
        user = self.getCurrentUser()
        return self._initializeFeatureFolder(user, folder)

    def _initializeFeatureFolder(self, user, folder):
        linkingFolder = Folder().createFolder(
            folder,
            LinkingFeaturesFolderName,
            description="Linking Feature folder",
            parentType="folder",
            public=False,
            creator=user,
            reuseExisting=True,
        )
        Folder().setMetadata(
            linkingFolder,
            {"baseMevaFolder": str(folder["_id"]), "linkingFolder": True},
            allowNull=True,
        )
        item_name = LinkingFeaturesPropertiesName
        newResultItem = Item().createItem(item_name, user, linkingFolder)
        Item().setMetadata(
            newResultItem,
            {"baseMevaFolder": str(folder["_id"]), "LinkingProperties": True},
            allowNull=True,
        )
        json_structure = LinkingFeatureBaseProperties
        json_bytes = json.dumps(json_structure).encode()
        byteIO = io.BytesIO(json_bytes)
        Upload().uploadFromFile(
            byteIO,
            len(json_bytes),
            item_name,
            parentType="item",
            parent=newResultItem,
            user=user,
            mimeType="application/json",
        )
        return newResultItem

    @access.admin
    @autoDescribeRoute(
        Description("Delete All Current Linked Features").modelParam(
            "id",
            description="Base MEVA dataset Folder",
            model=Folder,
            level=AccessType.WRITE,
        )
    )
    def delete_all_features(self, folder):
        linked_features = LinkingFeature().find({'folderId': str(folder['_id'])})
        for feature in linked_features:
            LinkingFeature().remove(feature)
        return True

    @access.user
    @autoDescribeRoute(
        Description("Initialize the linked Folder Structure").modelParam(
            "id",
            description="Base MEVA dataset Folder",
            model=Folder,
            level=AccessType.READ,
            required=True,
        )
    )
    def get_linking_properties(self, folder):
        existing = Item().findWithPermissions(
            {
                '$and': [
                    {'meta.LinkingProperties': {'$in': [True, 'true', 'True']}},
                    {'meta.baseMevaFolder': str(folder['_id'])},
                ]
            },
            user=self.getCurrentUser(),
        )
        base_item = None
        if existing.count() == 0:
            base_item = self._initializeFeatureFolder(self.getCurrentUser(), folder)
        else:
            base_item = existing[0]

        propertiesJSON = Item().childFiles(base_item)[0]

        return json.loads(
            b"".join(list(File().download(propertiesJSON, headers=False)())).decode()
        )

    @access.user
    @autoDescribeRoute(
        Description("Initialize the linked Folder Structure")
        .modelParam(
            "id",
            description="Base MEVA dataset Folder",
            model=Folder,
            level=AccessType.WRITE,
        )
        .param(
            "property",
            description="Text Property to add",
            required=True,
            dataType="string",
        )
    )
    def add_linking_property(self, folder, property):
        user = self.getCurrentUser()
        existing = Item().findWithPermissions(
            {
                '$and': [
                    {'meta.LinkingProperties': {'$in': [True, 'true', 'True']}},
                    {'meta.baseMevaFolder': str(folder['_id'])},
                ]
            },
            user=user,
        )
        base_item = None
        if existing.count() == 0:
            base_item = self._initializeFeatureFolder(self.getCurrentUser(), folder)
        else:
            base_item = existing[0]

        propertiesJSON = Item().childFiles(base_item)[0]

        data = json.loads(
            b"".join(list(File().download(propertiesJSON, headers=False)())).decode()
        )
        if property not in data:
            data.append(property)
        parentId = base_item['folderId']
        parentFolder = Folder().load(
            parentId, user=self.getCurrentUser(), level=AccessType.READ
        )
        Item().remove(base_item)
        item_name = LinkingFeaturesPropertiesName
        newResultItem = Item().createItem(item_name, user, parentFolder)
        Item().setMetadata(
            newResultItem,
            {"baseMevaFolder": str(folder['_id']), "LinkingProperties": True},
            allowNull=True,
        )
        json_structure = data
        json_bytes = json.dumps(json_structure).encode()
        byteIO = io.BytesIO(json_bytes)
        Upload().uploadFromFile(
            byteIO,
            len(json_bytes),
            item_name,
            parentType="item",
            parent=newResultItem,
            user=user,
            mimeType="application/json",
        )

    @access.user
    @autoDescribeRoute(
        Description("Initialize the linked Folder Structure")
        .modelParam(
            "id",
            description="Base MEVA dataset Folder",
            model=Folder,
            level=AccessType.WRITE,
        )
        .jsonParam(
            "data",
            "Data compose of a linking feature",
            required=True,
            requireObject=True,
            paramType='body',
        )
    )
    def add_linking_list(self, folder, data):
        user = self.getCurrentUser()
        existing = Item().findWithPermissions(
            {
                '$and': [
                    {'meta.LinkingList': {'$in': [True, 'true', 'True']}},
                    {'meta.baseMevaFolder': str(folder['_id'])},
                ]
            },
            user=self.getCurrentUser(),
        )
        if existing.count() != 0:
            Item().remove(existing[0])
        linkingFolder = Folder().createFolder(
            folder,
            LinkingFeaturesFolderName,
            description="Linking Feature folder",
            parentType="folder",
            public=False,
            creator=user,
            reuseExisting=True,
        )
        item_name = LinkingFeaturesList
        newResultItem = Item().createItem(item_name, user, linkingFolder)
        Item().setMetadata(
            newResultItem,
            {"baseMevaFolder": str(folder["_id"]), "LinkingList": True},
            allowNull=True,
        )
        json_structure = data['params']
        json_bytes = json.dumps(json_structure).encode()
        byteIO = io.BytesIO(json_bytes)
        Upload().uploadFromFile(
            byteIO,
            len(json_bytes),
            item_name,
            parentType="item",
            parent=newResultItem,
            user=user,
            mimeType="application/json",
        )
        return data['params']

    @access.user
    @autoDescribeRoute(
        Description("Get the current Linking List").modelParam(
            "id",
            description="Base MEVA dataset Folder",
            model=Folder,
            level=AccessType.WRITE,
        )
    )
    def get_linking_list(self, folder):
        return self._get_linking_list(folder)

    def _get_linking_list(self, folder):
        existing = Item().findWithPermissions(
            {
                '$and': [
                    {'meta.LinkingList': {'$in': [True, 'true', 'True']}},
                    {'meta.baseMevaFolder': str(folder['_id'])},
                ]
            },
            user=self.getCurrentUser(),
        )
        if existing.count() == 0:
            return []
        propertiesJSON = Item().childFiles(existing[0])[0]

        return json.loads(
            b"".join(list(File().download(propertiesJSON, headers=False)())).decode()
        )

    @access.user
    @autoDescribeRoute(
        Description(
            "Get the next linking item based on UserId and linking List"
        ).modelParam(
            "id",
            description="Base MEVA dataset Folder",
            model=Folder,
            level=AccessType.WRITE,
        )
    )
    def get_linking_item(self, folder):
        user = self.getCurrentUser()
        linking_list = self._get_linking_list(folder)
        # Lets make sure there is a user folder for their feature uploads
        user_folder = Folder().findOne(
            {
                "meta.baseMevaFolder": str(folder["_id"]),
                "meta.userLinkingFolder": True,
                "meta.userId": str(user['_id']),
            },
        )
        if user_folder is None:
            root_folder = Folder().findOne(
                {"meta.baseMevaFolder": str(folder["_id"]), "meta.linkingFolder": True},
            )
            user_folder = Folder().createFolder(
                root_folder,
                user['login'],
                description="User Image Feature Folder",
                parentType="folder",
                public=False,
                creator=user,
                reuseExisting=True,
            )
            Folder().setMetadata(
                user_folder,
                {
                    "baseMevaFolder": str(folder["_id"]),
                    "userLinkingFolder": True,
                    "userId": str(user['_id']),
                },
                allowNull=True,
            )
        query = {'userId': str(user['_id'])}
        linked_results = LinkingFeature().find(query, user=self.getCurrentUser())
        linking_keys = []
        for item in linking_list:
            linking_keys.append(f"{item['sourceClipFilename']} {item['sourceTrackId']}")
        linked_clips = []
        for item in linked_results:
            linked_clips.append(f"{item['sourceClipFilename']} {item['sourceTrackId']}")

        counter = 0
        for key in linking_keys:
            if key not in linked_clips:
                break
            counter = counter + 1
        if counter + 1 >= len(linking_keys):
            return None  # the user is complete and no more items are left to link
        # counter should be the next item in the list
        current_item = linking_list[counter]
        # now we need to convert this structure into a RandomResult Structure
        actor_list = []
        for item in current_item['Actors']:
            item['type'] = 'actor'
            actor_list.append(item)
        # figure out folderId base on filename
        meva_state = MEVA_State().findOne({'filename': current_item['sourceClipFilename']})
        folderId = None
        if meva_state is not None:
            folderId = meva_state['folderId']
        return {
            'results' : actor_list,
            "source": {
                "folderId": folderId,
                "trackId": current_item['sourceTrackId'],
                "type": "track"
            },
            "totalPages": 1
        }

    @access.admin
    @autoDescribeRoute(
        Description(
            "Get the next linking item based on UserId and linking List"
        ).modelParam(
            "id",
            description="Base MEVA dataset Folder",
            model=Folder,
            level=AccessType.WRITE,
        ).param(
            "number",
            description="Index number in the list to grab",
            required=True,
            dataType="integer",
        )
    )
    def get_list_item(self, folder, number):
        linking_list = self._get_linking_list(folder)
        if number < len(linking_list):
            current_item = linking_list[number]
            actor_list = []
            for item in current_item['Actors']:
                item['type'] = 'actor'
                actor_list.append(item)
            # figure out folderId base on filename
            meva_state = MEVA_State().findOne({'filename': current_item['sourceClipFilename']})
            folderId = None
            if meva_state is not None:
                folderId = meva_state['folderId']
                return {
                    'results' : actor_list,
                    "source": {
                        "folderId": folderId,
                        "trackId": current_item['sourceTrackId'],
                        "type": "track"
                    },
                    "totalCount": len(linking_list),
                    "totalPages": 1
                }
        else:
            return None
        
    @access.admin
    @autoDescribeRoute(
        Description('Export Linked Features')
        .modelParam(
            "id",
            description="Base MEVA dataset Folder",
            model=Folder,
            level=AccessType.WRITE,
        )
    )
    def export_linked_features(self, folder):
        root_folder = Folder().findOne(
            {"meta.baseMevaFolder": str(folder["_id"]), "meta.linkingFolder": True},
        )
        if root_folder is None:
            raise RestException(f"No Feature Linking folder found in the baseMevaFolder {folder['_id']}")
        second_query = {
            f'meta.{FeatureExtractionZip}': {'$in': TRUTHY_META_VALUES},
        }
        export_feature_item = list(
            Item().findWithPermissions(second_query, user=self.getCurrentUser())
        )
        if len(export_feature_item) != 0:
            Item().remove(export_feature_item[0])
        user = self.getCurrentUser()
        token = Token().createToken(user=user, days=4)
        export_feature_links.delay(
            baseMevaFolder=folder,
            baseLinkingFolder=root_folder,
            girder_client_token=str(token["_id"]),
            girder_job_title=(
            "Exporting Linked Features",
            ),
        )

    @access.admin
    @autoDescribeRoute(
        Description('Download Exported Linked Features')
        .modelParam(
            "id",
            description="Base MEVA dataset Folder",
            model=Folder,
            level=AccessType.WRITE,
        )
    )
    def get_exported_linked_features(self, folder):
        root_folder = Folder().findOne(
            {"meta.baseMevaFolder": str(folder["_id"]), "meta.linkingFolder": True},
        )
        if root_folder is None:
            raise RestException(f"No Feature Linking folder found in the baseMevaFolder {folder['_id']}")
        second_query = {
            f'meta.{FeatureExtractionZip}': {'$in': TRUTHY_META_VALUES},
        }
        zip_item = Item().findOne(second_query)

        if zip_item is None:
            raise RestException(f"No Feature Linking Zip File found in the baseMevaFolder {folder['_id']}")

        url = f'/api/v1/item/{zip_item["_id"]}/download'

        return url

    @access.admin
    @autoDescribeRoute(
        Description('Download Exported Linked Features')
        .modelParam(
            "id",
            description="Base MEVA dataset Folder",
            model=Folder,
            level=AccessType.WRITE,
        ).param(
            "limit",
            description="Number of links to test",
            dataType="integer",
            default=20,
        ).param(
            "totalActors",
            description="Number of actors in each link",
            dataType="integer",
            default=20,
        )
    )
    def generate_linking_list(self, folder, limit, totalActors):
        user = self.getCurrentUser()
        token = Token().createToken(user=user, days=4)
        create_linking_list.delay(
            folder=folder,
            limit=limit,
            totalActors=totalActors,
            girder_client_token=str(token["_id"]),
            girder_job_title=(
            "Creating Linking List",
            ),
        )

    @access.admin
    @autoDescribeRoute(
        Description(
            "Get the next linking item based on UserId and linking List"
        ).modelParam(
            "id",
            description="Base MEVA dataset Folder",
            model=Folder,
            level=AccessType.WRITE,
        )
    )
    def get_user_progress(self, folder):
        linked_features = LinkingFeature().find({'folderId': str(folder['_id'])})
        linking_list = self._get_linking_list(folder)

        user_count_map = {}
        for link in linked_features:
            userId = link['userId']
            userLogin = link['login']
            if userLogin not in user_count_map.keys():
                user_count_map[userLogin] = 0
            user_count_map[userLogin] += 1
        data = {
            'users': user_count_map,
            'total': len(linking_list)
        }
        return data
