from meva_tasks.s3_import import import_meva_from_url
from meva_server.meva_state.models import MEVA_State
from typing import List, Optional
from girder.models.upload import Upload

import pymongo
import os
from girder.api import access
from girder.api.describe import Description, autoDescribeRoute, describeRoute
from girder.api.rest import Resource
from girder.constants import AccessType
from girder.exceptions import RestException
from girder.models.folder import Folder
from girder.models.item import Item
from girder.models.setting import Setting
from girder.models.token import Token
from girder.models.user import User
from girder_jobs.models.job import Job
from urllib import request

from meva_server.chip_extraction import create_extraction_data
from meva_tasks.tasks import (
    convert_images,
    convert_video,
    create_thumbnails,
    create_track_chips,
    run_pipeline,
)
from meva_utils import TRUTHY_META_VALUES, asbool, fromMeta
from meva_utils.types import (
    AvailableJobSchema,
    GirderModel,
    PipelineDescription,
    PipelineJob,
)

from .constants import (
    SETTINGS_CONST_JOBS_CONFIGS,
    ActorDescriptorFolderName,
    CameraModelFolderName,
    MEVAFolderMarker,
    alphaRegEx,
    csvRegex,
    imageRegex,
    safeImageRegex,
    videoRegex,
    ymlRegex,
)
from .pipelines import load_pipelines, verify_pipe
from .serializers import alphapose
from .serializers import meva as meva_serializer
from .serializers import models
from .training import ensure_csv_detections_file, training_output_folder
from .transforms import GetPathFromItemId
from .utils import (
    detections_file,
    detections_item,
    get_or_create_auxiliary_folder,
    get_or_create_metadata_folder,
    get_or_create_thumbnail_folder,
    getTrackData,
    move_existing_result_to_auxiliary_folder,
    process_csv,
    process_json,
    saveTracks,
)

JOBCONST_DATASET_ID = 'datset_id'
JOBCONST_TRAINING_INPUT_IDS = 'training_input_ids'
JOBCONST_TRAINING_CONFIG = 'training_config'
JOBCONST_RESULTS_FOLDER_ID = 'results_folder_id'
JOBCONST_PIPELINE_NAME = 'pipeline_name'


class Meva(Resource):
    def __init__(self):
        super(Meva, self).__init__()
        self.resourceName = "meva"
        self.route("GET", ("datasets",), self.list_datasets)
        self.route("GET", ("brand_data",), self.get_brand_data)
        self.route("GET", ("pipelines",), self.get_pipelines)
        self.route("POST", ("pipeline",), self.run_pipeline_task)
        self.route("POST", ("update_job_configs",), self.update_job_configs)
        self.route("POST", ("postprocess", ":id"), self.postprocess)
        self.route(
            "POST", ("create_meva_thumbnails", ":id"), self.create_meva_thumbnails
        )
        self.route("POST", ("extract_track_chips", ":id"), self.extract_track_chips)
        self.route("PUT", ("metadata", ":id"), self.update_metadata)
        self.route("PUT", ("attributes",), self.save_attributes)
        self.route("POST", ("validate_files",), self.validate_files)
        self.route("GET", ("valid_images",), self.get_valid_images)
        self.route("GET", ("meva_data",), self.get_meva_data)
        self.route("GET", ("meva_filter",), self.meva_filter)
        self.route("PUT", ("s3_import", ":id"), self.s3_import)
        self.route("GET", ("meva_folders",), self.get_meva_folders)
        self.route("POST", (":id", "create_camera_model_folder",), self.create_camera_model_folder)
        self.route("POST", (":id", "create_actor_descriptor_folder",), self.create_actor_descriptor_folder)

    @access.admin
    @autoDescribeRoute(
        Description("Persist discovered job configurations").jsonParam(
            "configs",
            "Replace static job configurations",
            required=True,
            requireObject=True,
            paramType='body',
        )
    )
    def update_job_configs(self, configs: AvailableJobSchema):
        Setting().set(SETTINGS_CONST_JOBS_CONFIGS, configs)

    @access.public
    @autoDescribeRoute(Description("Get custom brand data"))
    def get_brand_data(self):
        adminUserIds = [user['_id'] for user in User().getAdmins()]
        # Find an item owned by an admin with meta.brand=True
        data = Item().findOne(
            {
                'meta.brand': {'$in': [True, 'true', 'True']},
                'creatorId': {'$in': adminUserIds},
            }
        )
        if data is not None:
            return data['meta']
        return {}

    @access.user
    @describeRoute(Description("List datasets in the system").pagingParams("created"))
    def list_datasets(self, params):
        limit, offset, sort = self.getPagingParameters(params)
        query = {
            'meta.MEVA': {'$in': TRUTHY_META_VALUES},
        }
        return Folder().findWithPermissions(
            query, offset=offset, limit=limit, sort=sort, user=self.getCurrentUser()
        )

    @access.user
    @describeRoute(Description("Get available pipeline configurations"))
    def get_pipelines(self, params):
        return load_pipelines(self.getCurrentUser())

    @access.user
    @autoDescribeRoute(
        Description("Run meva pipeline")
        .modelParam(
            "folderId",
            description="Folder id of a video clip",
            model=Folder,
            paramType="query",
            required=True,
            level=AccessType.WRITE,
        )
        .jsonParam("pipeline", "The pipeline to run on the dataset", required=True)
    )
    def run_pipeline_task(self, folder, pipeline: PipelineDescription):
        """
        Run a pipeline on a dataset.

        :param folder: The girder folder containing the dataset to run on.
        :param pipeline: The pipeline to run the dataset on.
        """
        user = self.getCurrentUser()
        verify_pipe(user, pipeline)

        folder_id_str = str(folder["_id"])
        # First, verify that no other outstanding jobs are running on this dataset
        existing_jobs = Job().findOne(
            {
                JOBCONST_DATASET_ID: folder_id_str,
                'status': {
                    # Find jobs that are inactive, queued, or running
                    # https://github.com/girder/girder/blob/main/plugins/jobs/girder_jobs/constants.py
                    '$in': [0, 1, 2]
                },
            }
        )
        if existing_jobs is not None:
            raise RestException(
                (
                    f"A pipeline for {folder_id_str} is already running. "
                    "Only one outstanding job may be run at a time for "
                    "a dataset."
                )
            )

        token = Token().createToken(user=user, days=14)

        # TODO Temporary inclusion of track_user pipelines requiring input
        detection_csv: Optional[GirderModel] = None
        if 'utility' in pipeline["pipe"]:
            # Ensure detection has a csv detections item
            detection = detections_item(folder, strict=True)
            detection_csv = ensure_csv_detections_file(folder, detection, user)

        move_existing_result_to_auxiliary_folder(folder, user)

        params: PipelineJob = {
            "input_folder": folder,
            "input_type": folder["meta"]["type"],
            "output_folder": folder_id_str,
            "pipeline": pipeline,
            "pipeline_input": detection_csv,
        }
        newjob = run_pipeline.apply_async(
            queue="pipelines",
            kwargs=dict(
                params=params,
                girder_job_title=f"Running {pipeline['name']} on {str(folder['name'])}",
                girder_client_token=str(token["_id"]),
                girder_job_type="pipelines",
            ),
        )
        newjob.job[JOBCONST_DATASET_ID] = folder_id_str
        newjob.job[JOBCONST_RESULTS_FOLDER_ID] = folder_id_str
        newjob.job[JOBCONST_PIPELINE_NAME] = pipeline['name']
        # Allow any users with accecss to the input data to also
        # see and possibly manage the job
        Job().copyAccessPolicies(folder, newjob.job)
        Job().save(newjob.job)
        return newjob.job

    @access.user
    @autoDescribeRoute(
        Description("Test whether or not a set of files are safe to upload").jsonParam(
            "files", "", paramType="body"
        )
    )
    def validate_files(self, files):
        ok = True
        message = ""
        mediatype = ""
        videos = [f for f in files if videoRegex.search(f)]
        csvs = [f for f in files if csvRegex.search(f)]
        images = [f for f in files if imageRegex.search(f)]
        ymls = [f for f in files if ymlRegex.search(f)]
        if len(videos) and len(images):
            ok = False
            message = "Do not upload images and videos in the same batch."
        elif len(csvs) > 1:
            ok = False
            message = "Can only upload a single CSV Annotation per import"
        elif len(csvs) == 1 and len(ymls):
            ok = False
            message = "Cannot mix annotation import types"
        elif (not len(videos)) and (not len(images)):
            ok = False
            message = "No supported media-type files found"
        elif len(videos):
            mediatype = 'video'
        elif len(images):
            mediatype = 'image-sequence'

        return {
            "ok": ok,
            "message": message,
            "type": mediatype,
            "media": images + videos,
            "annotations": csvs + ymls,
        }

    @access.user
    @autoDescribeRoute(
        Description(
            "Updates the meva meta data for tracks and creates thumbnails if they don't exit"
        ).modelParam(
            "id",
            description="Folder containing the items to process",
            model=Folder,
            level=AccessType.WRITE,
        )
    )
    def create_meva_thumbnails(self, folder):
        """
        Updates track count and creates thumbnails for video data
        """
        user = self.getCurrentUser()
        thumbnail_folder = get_or_create_thumbnail_folder(folder, user)
        Folder().remove(thumbnail_folder)
        thumbnail_folder = get_or_create_thumbnail_folder(folder, user)
        token = Token().createToken(user=user, days=14)
        create_thumbnails.delay(
            folder=folder,
            thumbnailFolderId=thumbnail_folder["_id"],
            girder_job_title=("Getting Thumbnails for {}".format(str(folder["_id"]))),
            girder_client_token=str(token["_id"]),
        )
        return

    @access.user
    @autoDescribeRoute(
        Description(
            "Extract Track chips from the reviewed/completed folder"
        ).modelParam(
            "id",
            description="Folder containing the items to process",
            model=Folder,
            level=AccessType.WRITE,
        )
    )
    def extract_track_chips(self, folder):
        """
        Updates track count and creates thumbnails for video data
        """
        user = self.getCurrentUser()
        metadataFolderId = get_or_create_metadata_folder(folder, user)
        Folder().remove(metadataFolderId)
        metadataFolderId = get_or_create_metadata_folder(folder, user)
        token = Token().createToken(user=user, days=14)
        file = detections_file(folder)
        if file is None:
            return {}
        json_tracks = getTrackData(file)
        output_json = create_extraction_data(self, json_tracks)
        # Now we need to take this output_json and feed it into ffmpeg to do extraction of the chips/frames needed
        create_track_chips.delay(
            folder=folder,
            jsonDict=output_json,
            metadataFolderId=metadataFolderId["_id"],
            girder_job_title=("Getting Track Chips for {}".format(str(folder["_id"]))),
            girder_client_token=str(token["_id"]),
        )
        return output_json

    @access.user
    @autoDescribeRoute(
        Description("Post-processing to be run after media/annotation import")
        .modelParam(
            "id",
            description="Folder containing the items to process",
            model=Folder,
            level=AccessType.WRITE,
        )
        .param(
            "skipJobs",
            "Whether to skip processing that might dispatch worker jobs",
            paramType="formData",
            dataType="boolean",
            default=False,
            required=False,
        )
        .param(
            "forceTranscode",
            "Forcing a Transcoding of the video to fix artifacts",
            paramType="formData",
            dataType="boolean",
            default=False,
            required=False,
        )
    )
    def postprocess(self, folder, skipJobs, forceTranscode):
        """
        Post-processing to be run after media/annotation import


        When skipJobs=False, the following may run as jobs:
            Transcoding of Video
            Transcoding of Images
            Conversion of KPF annotations into track JSON

        In either case, the following may run synchronously:
            Conversion of CSV annotations into track JSON
        """
        user = self.getCurrentUser()
        auxiliary = get_or_create_auxiliary_folder(folder, user)

        # Doing a check for Activity and owner
        meva_state = MEVA_State().findOne({'folderId': str(folder['_id'])})
        if meva_state is not None:
            if meva_state['locked'] is True and forceTranscode is False:
                raise Exception(
                    f"This video folder is locked and cannot be edited/saved until it is specifically unlocked by the owner: {meva_state['owner_name']}"
                )
            meva_owner = meva_state['owner']
            if meva_owner != str(user['_id']) and forceTranscode is False:
                raise Exception(
                    f"You don't have permission to update this annotation file the owner is {meva_state['owner_name']}"
                )
        if not skipJobs:
            token = Token().createToken(user=user, days=2)
            # transcode VIDEO if necessary
            videoItems = Folder().childItems(
                folder, filters={"lowerName": {"$regex": videoRegex}}
            )

            for item in videoItems:
                meta = item.get("meta", {})
                if 'source_video' in meta.keys():
                    if meta['source_video'] is False and forceTranscode is True:
                        Item().remove(item)
                        continue
                convert_video.delay(
                    path=GetPathFromItemId(str(item["_id"])),
                    folderId=str(item["folderId"]),
                    auxiliaryFolderId=auxiliary["_id"],
                    itemId=str(item["_id"]),
                    forceTranscode=forceTranscode,
                    girder_job_title=(
                        "Converting {} to a web friendly format".format(
                            str(item["_id"])
                        )
                    ),
                    girder_client_token=str(token["_id"]),
                )

            # transcode IMAGERY if necessary
            imageItems = Folder().childItems(
                folder, filters={"lowerName": {"$regex": imageRegex}}
            )
            safeImageItems = Folder().childItems(
                folder, filters={"lowerName": {"$regex": safeImageRegex}}
            )

            if imageItems.count() > safeImageItems.count():
                convert_images.delay(
                    folderId=folder["_id"],
                    girder_client_token=str(token["_id"]),
                    girder_job_title=(
                        f"Converting {folder['_id']} to a web friendly format",
                    ),
                )

            elif imageItems.count() > 0:
                folder["meta"]["annotate"] = True

            # transform KPF if necessary
            ymlItems = Folder().childItems(
                folder, filters={"lowerName": {"$regex": ymlRegex}}
            )
            if ymlItems.count() > 0:
                # There might be up to 3 yamls
                allFiles = [Item().childFiles(item)[0] for item in ymlItems]
                saveTracks(folder, meva_serializer.load_kpf_as_tracks(allFiles), user)
                ymlItems.rewind()
                for item in ymlItems:
                    Item().move(item, auxiliary)

            Folder().save(folder)

        process_csv(folder, user)
        process_json(folder, user)
        track_dict = getTrackData(detections_file(folder))
        track_count = len(track_dict.keys())
        meva_state = MEVA_State().createState(folder, str(folder['parentId']), user)
        MEVA_State().setMevaData(str(folder['_id']), 'beginTracks', track_count)
        if track_count > 0:
            MEVA_State().setState(str(folder['_id']), 'Imported', True)

        return folder

    @access.user
    @autoDescribeRoute(
        Description("Save mutable metadata for a dataset")
        .modelParam(
            "id",
            description="datasetId or folder for the metadata",
            model=Folder,
            level=AccessType.WRITE,
        )
        .jsonParam(
            "data",
            "JSON with the metadata to set",
            requireObject=True,
            paramType="body",
        )
        .errorResponse('Using a reserved metadata key', 400)
    )
    def update_metadata(self, folder, data):
        validated = models.MetadataMutableUpdate(**data)
        for name, value in validated.dict(exclude_none=True).items():
            folder['meta'][name] = value
        Folder().save(folder)
        return folder['meta']

    @access.user
    @autoDescribeRoute(
        Description("")
        .modelParam(
            "folderId",
            description="folder id of a clip",
            model=Folder,
            paramType="query",
            required=True,
            level=AccessType.WRITE,
        )
        .jsonParam(
            "attributes",
            "upsert and delete attributes",
            paramType="body",
            requireObject=True,
        )
    )
    def save_attributes(self, folder, attributes):
        upsert = attributes.get('upsert', [])
        delete = attributes.get('delete', [])
        attributes_dict = {}
        if 'attributes' in folder['meta']:
            attributes_dict = folder['meta']['attributes']
        for attribute_id in delete:
            attributes_dict.pop(str(attribute_id), None)
        for attribute in upsert:
            validated: models.Attribute = models.Attribute(**attribute)
            attributes_dict[str(validated.key)] = validated.dict(exclude_none=True)

        upserted_len = len(upsert)
        deleted_len = len(delete)

        if upserted_len or deleted_len:
            folder['meta']['attributes'] = attributes_dict
            Folder().save(folder)

        return {
            "updated": upserted_len,
            "deleted": deleted_len,
        }

    @access.user
    @autoDescribeRoute(
        Description("").modelParam(
            "folderId",
            description="folder id of a clip",
            model=Folder,
            paramType="query",
            required=True,
            level=AccessType.READ,
        )
    )
    def get_valid_images(self, folder):
        return Folder().childItems(
            folder,
            filters={"lowerName": {"$regex": safeImageRegex}},
            sort=[("lowerName", pymongo.ASCENDING)],
        )

    @access.user
    @autoDescribeRoute(Description(""))
    def get_meva_data(
        self,
    ):
        user = self.getCurrentUser()
        meva_items = list(
            Folder().findWithPermissions(
                {'meta.MEVA': {'$in': [True, 'true', 'True']}},
                user=self.getCurrentUser(),
            )
        )
        meva_items.sort(key=lambda d: d["created"], reverse=True)
        # Compile a list of dates/ times and start/end Pairs
        results = {
            "dates": set(),
            "times": set(),
            "locations": set(),
            "cameraIds": set(),
        }
        if len(meva_items) > 0:
            for folder in meva_items:
                meta = folder.get("meta")
                if meta:
                    MEVA_Meta = meta["MEVA_Meta"]
                    results["dates"].add(MEVA_Meta["date"])
                    start = MEVA_Meta["startTime"]
                    start = start[0: str.rindex(start, '-')]
                    end = MEVA_Meta["endTime"]
                    end = end[0: str.rindex(end, '-')]
                    results["times"].add(f'{start} to {end}')
                    results["locations"].add(MEVA_Meta["location"])
                    results["cameraIds"].add(MEVA_Meta["cameraId"])

        print(results)
        return results

    @access.user
    @autoDescribeRoute(
        Description("")
        .pagingParams("created")
        .jsonParam(
            "filters",
            "JSON Settings for the filtering",
            required=True,
            requireObject=True,
        )
    )
    def meva_filter(self, filters):
        limit, offset, sort = self.getPagingParameters(filters)
        query = {
            'meta.MEVA': {'$in': TRUTHY_META_VALUES},
        }
        if filters is not None:
            query = {'$and': [query]}
            if 'date' in filters.keys():
                query["$and"].append(
                    {'meta.MEVA_Meta.date': {'$in': [filters['date']]}}
                )
            if 'location' in filters.keys():
                query["$and"].append(
                    {'meta.MEVA_Meta.location': {'$in': [filters['location']]}}
                )
            if 'cameraId' in filters.keys():
                query["$and"].append(
                    {'meta.MEVA_Meta.cameraId': {'$in': [filters['cameraId']]}}
                )
            if 'startTime' in filters.keys():
                query["$and"].append(
                    {'meta.MEVA_Meta.startTime': {'$regex': filters['startTime']}}
                )
            if 'state' in filters.keys():
                query["$and"].append(
                    {f'meta.MEVA_State.{filters["state"]}': {'$in': TRUTHY_META_VALUES}}
                )
        return Folder().findWithPermissions(
            query, offset=offset, limit=limit, sort=sort, user=self.getCurrentUser()
        )

    @access.user
    @autoDescribeRoute(
        Description("Set MEVA annotator user")
        .modelParam(
            "id",
            description="root meva project folderId",
            model=Folder,
            level=AccessType.WRITE,
        )
        .jsonParam(
            "folderList",
            "JSON Array of items for import",
            required=True,
            requireArray=True,
            paramType="body",
        )
    )
    def s3_import(self, folder, folderList):
        # Change state for cleaning tracks
        query = {"folderId": str(folder["_id"])}
        user = self.getCurrentUser()
        token = Token().createToken(user=user, days=14)
        for item in folderList:
            videoURL = item['videoURL']
            filename = item['filename']
            tracksURL = None
            descriptorURL = None
            if 'tracksURL' in item:
                tracksURL = item['tracksURL']
            if 'descriptorURL' in item:
                descriptorURL = item['descriptorURL']
            import_meva_from_url.delay(
                folder=folder,
                filename=filename,
                videoURL=videoURL,
                tracksURL=tracksURL,
                descriptorURL=descriptorURL,
                girder_job_title=(
                    f"Importing file {filename} from {videoURL}"
                ),
                girder_client_token=str(token["_id"]),
            )

    @access.user
    @describeRoute(Description("List datasets in the system").pagingParams("created"))
    def get_meva_folders(self, params):
        limit, offset, sort = self.getPagingParameters(params)
        query = {
            f'meta.{MEVAFolderMarker}': {'$in': TRUTHY_META_VALUES},
        }
        return Folder().findWithPermissions(
            query, offset=offset, limit=limit, sort=sort, user=self.getCurrentUser()
        )

    @access.user
    @autoDescribeRoute(
        Description("Creates camera model folder for uploading zip file")
        .modelParam(
            "id",
            description="Base root Folder to create camera model folder in",
            model=Folder,
            level=AccessType.WRITE,
        )
    )
    def create_camera_model_folder(self, folder):
        user = self.getCurrentUser()
        modelFolder = Folder().createFolder(
            folder,
            CameraModelFolderName,
            description="Camera Model folder",
            parentType="folder",
            public=False,
            creator=user,
            reuseExisting=True,
        )
        if modelFolder is not None:
            Folder().remove(modelFolder)
        modelFolder = Folder().createFolder(
            folder,
            CameraModelFolderName,
            description="Camera Model folder",
            parentType="folder",
            public=False,
            creator=user,
            reuseExisting=True,
        )
        Folder().setMetadata(
            modelFolder,
            {"baseMevaFolder": str(folder["_id"]), "modelFolder": True},
            allowNull=True,
        )
        return modelFolder

    @access.user
    @autoDescribeRoute(
        Description("Creates actor descriptor folder for uploading zip file")
        .modelParam(
            "id",
            description="Base root Folder to create actor descriptor folder in",
            model=Folder,
            level=AccessType.WRITE,
        )
    )
    def create_actor_descriptor_folder(self, folder):
        user = self.getCurrentUser()
        modelFolder = Folder().createFolder(
            folder,
            ActorDescriptorFolderName,
            description="Actor Descriptor folder",
            parentType="folder",
            public=False,
            creator=user,
            reuseExisting=True,
        )
        if modelFolder is not None:
            Folder().remove(modelFolder)
        modelFolder = Folder().createFolder(
            folder,
            ActorDescriptorFolderName,
            description="Actor Descriptor folder",
            parentType="folder",
            public=False,
            creator=user,
            reuseExisting=True,
        )
        Folder().setMetadata(
            modelFolder,
            {"baseMevaFolder": str(folder["_id"]), "modelFolder": True},
            allowNull=True,
        )
        return modelFolder
