import json
import io
from girder.exceptions import RestException
from girder.models.token import Token
from meva_server.utils import load_actor_json
import pymongo
from meva_tasks.tasks import import_actor_images, import_gps_tracks
from meva_server.constants import ActorFolderName, GPSFolderName, zipRegex
import random
import urllib
from typing import List

from girder.api import access
from girder.api.describe import Description, autoDescribeRoute
from girder.api.rest import Resource, setContentDisposition, setResponseHeader
from girder.constants import AccessType, TokenScope
from girder.models.file import File
from girder.models.folder import Folder
from girder.models.item import Item
from girder.models.upload import Upload

from girder.utility import ziputil


class MevaCheckIn(Resource):
    def __init__(self):
        super(MevaCheckIn, self).__init__()
        self.resourceName = "meva_checkin"
        self.route("POST", ("process_actor_zip", ":id"), self.process_actor_zip)
        self.route("POST", ("process_gps_zip", ":id"), self.process_gps_zip)
        self.route("POST", (":id", "create_actor_folder",), self.create_actor_folder)
        self.route("POST", (":id", "create_gps_folder",), self.create_gps_folder)
        self.route("POST", ("import_actors", ":id"), self.import_actors)

    @access.user
    @autoDescribeRoute(
        Description("Post-processing to be run after actor chip import").modelParam(
            "id",
            description="Folder containing the items to process",
            model=Folder,
            level=AccessType.WRITE,
        )
    )
    def process_actor_zip(self, folder):
        # so here we would start extracting the zip file and using it to determine what is happening.
        user = self.getCurrentUser()
        # find the zip file and begin extracting it
        zipItems = Folder().childItems(
            folder,
            filters={"lowerName": {"$regex": zipRegex}},
            sort=[("created", pymongo.DESCENDING)],
        )
        if zipItems.count() >= 1:
            zipFile = Item().childFiles(zipItems.next())[0]
        token = Token().createToken(user=user, days=14)
        actorJSONItems = Folder().childItems(
            folder,
            filters={'meta.actorJSON': {'$in': [True, 'true', 'True']}},
        )
        actorjsonId = None
        if actorJSONItems.count() > 0:
            actorjsonId = str(Item().childFiles(actorJSONItems.next())[0]['_id'])

        import_actor_images.delay(
            folder,
            str(zipFile["_id"]),
            actorjsonId,
            girder_job_title=("Import actor Chips {}".format(str(folder["_id"]))),
            girder_client_token=str(token["_id"]),
        )

    @access.user
    @autoDescribeRoute(
        Description("Post-processing to be run after actor chip import").modelParam(
            "id",
            description="Folder containing the items to process",
            model=Folder,
            level=AccessType.WRITE,
        )
    )
    def process_gps_zip(self, folder):
        # so here we would start extracting the zip file and using it to determine what is happening.
        user = self.getCurrentUser()
        # find the zip file and begin extracting it
        zipItems = Folder().childItems(
            folder,
            filters={"lowerName": {"$regex": zipRegex}},
            sort=[("created", pymongo.DESCENDING)],
        )
        if zipItems.count() >= 1:
            zipFile = Item().childFiles(zipItems.next())[0]
        token = Token().createToken(user=user, days=14)
        import_gps_tracks.delay(
            folder,
            str(zipFile["_id"]),
            girder_job_title=("Import GPS Tracks {}".format(str(folder["_id"]))),
            girder_client_token=str(token["_id"]),
        )

    @access.user
    @autoDescribeRoute(
        Description("Creates root actor folder for uploading zip file")
        .modelParam(
            "id",
            description="Base root Folder to create actor folder",
            model=Folder,
            level=AccessType.WRITE,
        )
    )
    def create_actor_folder(self, folder):
        user = self.getCurrentUser()
        actorFolder = Folder().createFolder(
            folder,
            ActorFolderName,
            description="Actor folder",
            parentType="folder",
            public=False,
            creator=user,
            reuseExisting=True,
        )
        # if actorFolder is not None:
        #     Folder().remove(actorFolder)
        # actorFolder = Folder().createFolder(
        #     folder,
        #     ActorFolderName,
        #     description="Actor folder",
        #     parentType="folder",
        #     public=False,
        #     creator=user,
        #     reuseExisting=True,
        # )
        Folder().setMetadata(
            actorFolder,
            {"baseMevaFolder": str(folder["_id"]), "actorFolder": True},
            allowNull=True,
        )
        return actorFolder

    @access.user
    @autoDescribeRoute(
        Description("Creates root GPS folder for uploading zip file")
        .modelParam(
            "id",
            description="Base GPS Folder to create actor folder",
            model=Folder,
            level=AccessType.WRITE,
        )
    )
    def create_gps_folder(self, folder):
        user = self.getCurrentUser()
        gpsFolder = Folder().createFolder(
            folder,
            GPSFolderName,
            description="GPS folder",
            parentType="folder",
            public=False,
            creator=user,
            reuseExisting=True,
        )
        if gpsFolder is not None:
            Folder().remove(gpsFolder)
        gpsFolder = Folder().createFolder(
            folder,
            GPSFolderName,
            description="GPS folder",
            parentType="folder",
            public=False,
            creator=user,
            reuseExisting=True,
        )
        Folder().setMetadata(
            gpsFolder,
            {"baseMevaFolder": str(folder["_id"]), "gpsFolder": True},
            allowNull=True,
        )
        return gpsFolder

    @access.user
    @autoDescribeRoute(
        Description("Post-processing to be run after chip import").modelParam(
            "id",
            description="Folder containing the items to process",
            model=Folder,
            level=AccessType.WRITE,
        )
    )
    def import_actors(self, folder):
        # we need to take the current folder and import all files and create a json file
        # function should be here that recursively grabs all items
        user = self.getCurrentUser()
        json_structure = {}
        existing_json = load_actor_json(user, folder)
        loaded_json = {}
        if 'status' in existing_json.keys():
            loaded_json = existing_json
        for (path, file) in Folder().fileList(
            folder, user=user, subpath=False, data=False
        ):
            if 'jpg' not in file["exts"]:
                continue
            name = file["name"]
            parts = name.split('-')
            if len(parts) != 6:
                continue
            actor = parts[0]
            date = parts[1]
            gps = parts[2]
            outfit = parts[3]
            key_name = f'{actor}-{date}-{gps}-{outfit}'
            if key_name in loaded_json.keys():
                continue
            front_back = str(parts[4]).lower()
            size = parts[5].replace(".jpg", "")
            itemId = file["itemId"]
            item = {
                "name": name,
                "actor": actor,
                "date": date,
                "gps": gps,
                "outfit": outfit,
                "side": front_back,
                "size": size,
                "itemId": str(itemId),
            }
            if f'{actor}-{date}-{gps}-{outfit}' not in json_structure:
                json_structure[f'{actor}-{date}-{gps}-{outfit}'] = {
                    'filename': name,
                    'actor': actor,
                    'date': date,
                    'outfit': outfit,
                    'gps': gps,
                }
            if front_back not in json_structure[f'{actor}-{date}-{gps}-{outfit}']:
                json_structure[f'{actor}-{date}-{gps}-{outfit}'][front_back] = {}
            if size not in json_structure[f'{actor}-{date}-{gps}-{outfit}'][front_back]:
                json_structure[f'{actor}-{date}-{gps}-{outfit}'][front_back][
                    size
                ] = str(itemId)
        item_name = "actors.json"
        newResultItem = Item().createItem(item_name, user, folder)
        Item().setMetadata(
            newResultItem,
            {"baseMevaFolder": str(folder["parentId"]), "actorJSON": True},
            allowNull=True,
        )

        json_bytes = json.dumps(json_structure).encode()
        byteIO = io.BytesIO(json_bytes)
        Upload().uploadFromFile(
            byteIO,
            len(json_bytes),
            item_name,
            parentType="item",
            parent=newResultItem,
            user=user,
            mimeType="application/json",
        )
        # Add the actor Chips to the database
