#! /usr/bin/python3.7
"""
Copyright 2021 by Kitware, Inc.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

 * Redistributions of source code must retain the above copyright notice,
   this list of conditions and the following disclaimer.

 * Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.

 * Neither name of Kitware, Inc. nor the names of any contributors may be used
   to endorse or promote products derived from this software without specific
   prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

==============================================================================

"""
from __future__ import division, print_function
import os
import sys
import argparse
from itertools import combinations
from pathlib import Path
import time
import math
import json
import pandas as pd
from sqlalchemy import and_, not_, func
from sqlalchemy.orm import load_only

from multiprocessing import Pool
from datetime import datetime
import glob
import pickle
import io
import cv2
import numpy as np
import matplotlib as plt

import torch
from PIL import Image
from torchvision.transforms import ToTensor, ToPILImage
import torch.nn.functional as F
from torchvision.transforms.functional import normalize
import operator
from people_segmentation.pre_trained_models import create_model
from iglovikov_helper_functions.dl.pytorch.utils import tensor_from_rgb_image
from iglovikov_helper_functions.utils.image_utils import load_rgb, pad, unpad
import albumentations as albu
from meva_server.meva_database.models import Track, MustNotLink, \
    Afinity, Response, Source, Suggested, Detection
from meva_server.meva_database.utils import turnOffSqlAlchemyLogging

# some default values that can be overriden by the args passed in
_page = 0
_amtPerPage = 200
_maxAffinitiesToLoad = 200000
_data_path = '/data/db/'
# storing in memory between executions for a speed boost
_affinities = []
_tracks = []
_sortedAffinities = []
_affinByT1 = []
_affinByT2 = []
_alreadyMatchedAtThisTime = []
_detectionsByTrackUID = {}
_areas = {}
_mnlDict = {}
_seg_model = None
_dg_model = None
_responses = {}
_lastSaveTime = 0
_netname = 'mcrv_resnet50_checkpoint_528759.pt'
_netsize = 11743
_reIDUseBGMasking = False

def getAllPossibleCombos(tracks: list) -> list:
    possibleLinks = []
    for track1, track2 in combinations(tracks, 2):
        possibleLinks.append([track1.uid, track2.uid])
    return possibleLinks


def getDFWithNoDuplicates(df1: pd.DataFrame, df2: pd.DataFrame,
                          strColumns: list) -> pd.DataFrame:
    df3 = pd.merge(df1, df2, on=strColumns,
                   how='outer',
                   indicator=True).query("_merge != 'both'")
    df3 = df3.drop(['_merge'], axis=1).reset_index(drop=True)
    return df3


def getDF1WithoutAnyInDF2(df1: pd.DataFrame, df2: pd.DataFrame,
                          strColumns: list) -> pd.DataFrame:
    df3 = pd.merge(df1, df2, on=strColumns,
                   how='left',
                   indicator=True).query("_merge != 'both'")
    df3 = df3.drop(['_merge'], axis=1).reset_index(drop=True)
    return df3


def getSettings(settingsFile) -> dict:
    settings = {}
    try:
        if not os.path.exists(settingsFile):
            print('===== settings file will be created =====')
        else:
            with open(settingsFile, 'rb') as sf:
                settings = pickle.load(sf)
    except Exception as ex:
        print(f"ERROR opening {settingsFile}: {ex}")
    finally:
        return settings


def saveSettings(settings: dict, settingsFile):
    try:
        with open(settingsFile, 'wb') as sf:
            pickle.dump(settings, sf)
        #with open(settingsFile+'.txt', 'w') as sf:
        #    json.dump(settings, sf)
        print(f"settings {settingsFile} saved")
    except Exception as ex:
        print(f"ERROR saving {settingsFile}: {ex} {ex.with_traceback}")


def getPosResponseList(uid: str, session) -> list:
    posResp = session.query(Response).\
        where(and_(Response.track1_uid == uid,
                   Response.response >= 1.0)).all()
    posRespList = [x.track2_uid for x in posResp]
    return posRespList


def getNegResponseList(uid: str, session) -> list:
    negResp = session.query(Response).\
        where(and_(Response.track1_uid == uid,
                   Response.response == 0.0)).all()
    negRespList = [x.track2_uid for x in negResp]
    return negRespList


def getMNLMatchList(uid: str, session) -> list:
    # time1 = time.time()
    matchedMNLs = session.query(MustNotLink).\
        where(MustNotLink.track1_uid == uid).all()
    mnlList = [x.track2_uid for x in matchedMNLs]
    # print(f"::: mnl query: {time.time() - time1}")
    return mnlList


def isThisAPushedDown(uid: str, session, suggested_limit = 0) -> bool:
    if suggested_limit is None:
        suggested_limit = 0
    matchSugg = session.query(Suggested).\
        where(and_(Suggested.track1_uid == uid, Suggested.suggestion_count >= suggested_limit)).all()
    if len(matchSugg) < 1:
        return False
    else:
        return True


def getFormattedOutputList(outputList, tracks, sourceByUid, exemplars) -> list:
    time1 = time.time()
    # accumulate the list
    output = []
    prev = []
    for track in outputList:
        src1 = sourceByUid[tracks[track].source_uid]
        content1 = {}
        combined = ''
        # get the exemplar link source to list in the json
        if track in exemplars:
            content1['type'] = 'actor'
            content1['actorId'] = src1.girder_id
            content1['date'] = src1.name.split('-')[1]
            content1['outfitId'] = src1.outfit_id
            combined = content1['actorId'] + "_" + \
                content1['date'] + "_" + content1['outfitId']
        else:
            content1['folderId'] = src1.uid
            content1['trackId'] = tracks[track].video_track_id
            content1['type'] = 'track'
            combined = content1['folderId'] + "_" + \
                content1['trackId']

        if combined not in prev:
            prev.append(combined)  # prevent duplicates
            # add to the list
            output.append(content1)
    print(f":: format output time: {time.time() - time1}")
    return output


def tensorletterbox(x, height=256, width=128, fillvalue=0) -> torch.Tensor:
    shape = x.shape[2:]
    ratio = min(float(height) / shape[0], float(width) / shape[1])

    new_shape = (round(shape[0] * ratio), round(shape[1] * ratio))
    dw = (width - new_shape[1]) / 2  # width padding
    dh = (height - new_shape[0]) / 2
    top, bottom = round(dh - 0.1), round(dh + 0.1)
    left, right = round(dw - 0.1), round(dw + 0.1)
    x = torch.nn.functional.interpolate(x, size=new_shape,
                                        mode='bilinear')
    x = torch.nn.functional.pad(
        x, (left, right, top, bottom), mode='constant', value=fillvalue)
    return x


def provision_model(config_path, model_path):
    dg_cfg.merge_from_file(config_path)
    # Prevent model from trying to init backbone
    dg_cfg['MODEL']['PRETRAIN_CHOICE'] = None
    dg_cfg['MODEL']['PRETRAIN_PATH'] = None
    dg_model = build_dg_model(dg_cfg, _netsize)
    checkpoint = torch.load(model_path, map_location="cpu")
    dg_model.load_state_dict(checkpoint['model'])
    return dg_model


def load_image_to_tensor(img) -> torch.Tensor:
    # convert img to tensor
    return ToTensor()(img).unsqueeze(0)


def load_image_and_prepare_tensor(image, pxl_mean,
                                  pxl_std, display=False) -> torch.Tensor:
    chip = load_image_to_tensor(image)
    # image letterbox and resize per
    # https://gitlab.kitware.com/afrl_act3/meva-person-search/-/blob/dev/customTracker/mevaTracker/trackers.py#L502
    print(f"normalizing with mean: {pxl_mean}, and std: {pxl_std}")
    chip = normalize(chip, pxl_mean, pxl_std)
    model_in = tensorletterbox(chip, height=256, width=128)
    if display:
        plt.imshow(ToPILImage()(torch.squeeze(model_in)))
        plt.show()
    return model_in


def setup_reID_imports():
    '''
    This is being done to allow for running this script locally
    with a different setup than it has in the container
    (quicker testing, etc.)
    '''
    sys.path.append(_data_path + 'reid/reid-strong-baseline')
    print(sys.path)
    global build_dg_model
    global dg_cfg
    from modeling import build_model as build_dg_model
    from config import cfg as dg_cfg


def paginateList(inList, offset, pageSize):
    time1 = time.time()
    # check the pagenation
    pages = math.ceil(len(inList)/pageSize)
    if pages < 1 and pages > 0:
        pages = 0
    print(f"pages = {pages}")
    # check for too many
    if offset > pages:
        print(f"page {offset} requested past possible amount {pages}")
        sys.exit()
    pageStart = (offset) * pageSize
    pageEnd = ((offset + 1) * pageSize)
    # page slice
    print(f"getting list indices: {pageStart}:{pageEnd}")
    outList = inList[pageStart:pageEnd]
    print(f":: paginate time: {time.time() - time1}")
    return (pages, outList)


def loadImageAndTransform(image):
    # transform the image for the model
    transform = albu.Compose([albu.Normalize(p=1)], p=1)
    padded_image, pads = pad(image, factor=32, border=cv2.BORDER_CONSTANT)
    x = transform(image=padded_image)["image"]
    x = torch.unsqueeze(tensor_from_rgb_image(x), 0)
    return (image, x, pads)


def getMask(model_in, model, pads):
    with torch.no_grad():
        prediction = model(model_in)[0][0]
    mask = (prediction > 0).cpu().numpy().astype(np.uint8)
    mask = unpad(mask, pads)
    return mask


def hideBackground(mask, image):
    if _reIDUseBGMasking:
        cvmask = (cv2.cvtColor(mask, cv2.COLOR_GRAY2RGB) * (255, 255, 255))\
            .astype(np.uint8)
        cvmask = cv2.bitwise_and(image, cvmask)
    else:
        # temp turn off masking
        cvmask = image
    return cvmask


def sortByReID(session, source, checkins, image):
    setup_reID_imports()  # these are the non pip installable imports
    saveImage = False  # flip this for testing
    if saveImage:
        imgSave = _data_path + 'image.jpg'
        with open(imgSave, 'wb') as imgFile:
            img_byte_arr = io.BytesIO()
            image.save(img_byte_arr, format='PNG')
            img_byte_arr = img_byte_arr.getvalue()
            imgFile.write(img_byte_arr)
    config_path = _data_path + \
        'reid/reid-strong-baseline/configs/softmax_triplet_with_center.yml'
    model_path = _data_path + 'reid/' + _netname
    checkins_path = _data_path + 'reid/checkin_desciptors/'

    # provision the segmentation model
    print('running person segmentation')
    time1 = time.time()
    global _seg_model
    if _seg_model is None:
        _seg_model = create_model("Unet_2020-07-20")
    _ = _seg_model.eval()
    # load the image
    image, model_in, pads = loadImageAndTransform(np.array(image))
    mask = getMask(model_in, _seg_model, pads)
    # reverse it
    cvmask = hideBackground(mask, image)
    print(f":: time: {time.time() - time1}")

    # Provision reID model
    print('model setup')
    time1 = time.time()
    global _dg_model
    if _dg_model is None:
        _dg_model = provision_model(config_path, model_path)
    _ = _dg_model.eval()
    print(f":: time: {time.time() - time1}")
    # load the image file
    print('loading test image and resizing')
    time1 = time.time()
    model_in = load_image_and_prepare_tensor(cvmask,
                                             dg_cfg.INPUT.PIXEL_MEAN,
                                             dg_cfg.INPUT.PIXEL_STD,
                                             False)
    print(f":: time: {time.time() - time1}")

    # compute the test descriptor
    print('getting this test image desc.')
    time1 = time.time()
    with torch.no_grad():
        test_tensor = _dg_model(model_in)
    print(f":: time: {time.time() - time1}")

    # in case we want all the checkins with no prior
    # exclusions
    fullDayCheckins = False
    test_date = ''
    if checkins == []:
        checkins = glob.glob(checkins_path + "*.desc")
        fullDayCheckins = True
        # get the date from the source
        sourceInfo = session.query(Source).\
            filter(Source.uid == source['folderId']).\
            first()
        test_date = sourceInfo.name.split('.')[0].replace('-', '')
        print(f"looking for all actors on {test_date}")

    time1 = time.time()
    best_score = 0
    winner = ''
    scores = {}
    removes = []
    actorCount = 0
    for check in checkins:
        descriptors = []  # in case we are using the outfitId - there could be 2
        names = []
        if fullDayCheckins:
            descriptors.append(check)
            # name of the checkin img
            names.append(check.split('/')[-1].split('.desc')[0])
            # key is the name of the img
            check_date = check.split('-')[1]
            if test_date != check_date:
                # print(f"kicking out {check_date} since not {test_date}")
                continue
            actorCount += 1
            # check for this one in the responses - to remove negatives
            searchSourceName = names[0].replace('.jpg', '').lower()
            s2 = session.query(Source).\
                filter(func.lower(Source.name).contains(searchSourceName)).\
                first()

            if s2 is not None:
                t2 = session.query(Track).\
                    filter(Track.source_uid == s2.uid).\
                    first()
                t1 = session.query(Track).\
                    filter(and_(Track.source_uid == source['folderId'],
                                Track.video_track_id == source['trackId'])).\
                    first()

                if t1 is not None and t2 is not None:
                    resp1 = session.query(Response).\
                        filter(and_(Response.track1_uid == t1.uid,
                                    Response.track2_uid == t2.uid)).\
                        first()
                    if resp1 is not None:
                        if resp1.response == 0:
                            negName = names[0].lower().replace("-f.jpg", "").\
                                replace("-b.jpg", "")
                            print(f"+++++++++ negative response {negName} ")
                            if negName not in removes:
                                removes.append(negName)
                        else:
                            print(f"+++++++++ other response {negName} ")
                else:
                    print(
                        f"can't find a track for {searchSourceName} {t1} {t2}")

        else:
            # get the name from the source
            #print("PRINTING CHECK")
            #print(check)
            if check['type'] == 'actor':
                # fix glob pattern to work with GPS ID in middle
                # actorId-date-gpsId-outfitId-[f,b].jpg.desc
                tname = f"{check['actorId']}-{check['date']}-*-{check['outfitId']}"
                list_descriptor = glob.glob(checkins_path + tname + "*.desc")
                if len(list_descriptor) > 0:
                    names = list_descriptor[0].lower().split('/')[-1].split('.desc')
                    descriptors = list_descriptor
                else:
                    continue
            else:
                continue

        # read in this one
        for de in range(0, len(descriptors)):
            with open(descriptors[de], 'rb') as jfile:
                checkin = pickle.load(jfile)

            # convert this list back to tensor
            check_tensor = torch.Tensor(checkin)
            sim = F.cosine_similarity(test_tensor, check_tensor).tolist()[0]

            scores[names[de]] = sim
            if sim > best_score:
                best_score = sim
                winner = names[de]
    print(f"winner: {winner} out of {actorCount}")
    print(f":: time: {time.time() - time1}")

    # remove actors from the list if they had the negative response
    for rem in removes:
        t1 = rem + "-f.jpg"
        if t1 in scores.keys():
            scores.pop(t1)
        t2 = rem + "-b.jpg"
        if t2 in scores.keys():
            scores.pop(t2)

    print("sorting")
    sorted_scores = sorted(
        scores.items(), key=operator.itemgetter(1), reverse=True)
    newOutList = []
    prevEntries = []
    for s in sorted_scores:
        # check to make sure we don't already have this entry
        # since we only need the front or the back - chop that off the name
        name_beginning = s[0].lower().replace("-f.jpg", "").replace("-b.jpg", "")
        if name_beginning in prevEntries:
            continue
        prevEntries.append(name_beginning)
        # get the ID from the name
        # name needs to remove the ending and use a contain instead.
        name_adjust = s[0].lower().replace(".jpg", "-full.jpg")

        sourceInfo = session.query(Source).\
            filter(func.lower(Source.name) == name_adjust).\
            first()
        # if this is empty - sometimes we have missing fronts in the DB
        if sourceInfo is None:
            # try the back image
            name_adjust = name_adjust.lower().replace("-f-full", "-b-full")
            # print(f"re-trying with {name_adjust}")
            sourceInfo = session.query(Source).\
                filter(func.lower(Source.name) == name_adjust).\
                first()

        if sourceInfo is None:
            print(f"error with this source {s[0]}")
        else:
            newOutList.append({
                'type': 'actor',
                'actorId': sourceInfo.girder_id,
                'date': sourceInfo.name.split('-')[1],
                'outfitId': sourceInfo.outfit_id
            })
            # print(f"{sourceInfo.uid} - {sourceInfo.name}")

    return (source, newOutList)


def sortExemplars(session, allTracks):
    print('Loading exemplars...')
    time1 = time.time()
    exemplars = {}
    nonExemplarTracks = {}
    checkins = session.query(Track).\
        filter(Track.exemplar == 1).\
        all()
    noncheckins = session.query(Track).\
        filter(Track.exemplar != 1).\
        all()
    for track in checkins:
        exemplars[track.uid] = track
    for track in noncheckins:
        nonExemplarTracks[track.uid] = track
    print(f"grab all checkin tracks time: {time.time() - time1}")
    tracks = {}
    trackSrcByTrackUID = {}
    trackStartTimeByTUID = {}
    trackEndTimeByTUID = {}
    tracksBySrcUID = {}
    time1 = time.time()
    for track in allTracks:
        # make the track easily referenced by uid
        tracks[track.uid] = track
        tracksBySrcUID[track.source_uid] = track
        # get the track src id from track uid
        trackSrcByTrackUID[track.uid] = track.source_uid
        trackStartTimeByTUID[track.uid] = track.start_time
        trackEndTimeByTUID[track.uid] = track.end_time
    print(f":: sort exemplars time: {time.time() - time1}")
    return (exemplars, nonExemplarTracks, tracks, tracksBySrcUID)


def sortResponses(responses, exemplars):
    print('Sorting responses by uids...')
    time1 = time.time()
    exemplarPosResponses = {}
    er1 = []
    er2 = []
    allRespT1 = []
    allRespT2 = []
    for resp in responses:
        allRespT1.append(resp.track1_uid)
        allRespT2.append(resp.track2_uid)
        allRespT2.append(resp.track1_uid)
        allRespT1.append(resp.track2_uid)
        if resp.response >= 1:
            if resp.track1_uid in exemplars.keys():
                if resp.track1_uid not in exemplarPosResponses.keys():
                    exemplarPosResponses[resp.track1_uid] = []
                exemplarPosResponses[resp.track1_uid].append(resp.track2_uid)
                er1.append(resp.track1_uid)
                er2.append(resp.track2_uid)
                er1.append(resp.track2_uid)
                er2.append(resp.track1_uid)
            elif resp.track2_uid in exemplars.keys():
                if resp.track2_uid not in exemplarPosResponses.keys():
                    exemplarPosResponses[resp.track2_uid] = []
                exemplarPosResponses[resp.track2_uid].append(resp.track1_uid)
                er1.append(resp.track1_uid)
                er2.append(resp.track2_uid)
                er1.append(resp.track2_uid)
                er2.append(resp.track1_uid)
    print(f"Exemplar Responses: {len(exemplarPosResponses.keys())}")
    exResp = pd.DataFrame({'t1': er1, 't2': er2})
    # add up all the responses
    allResponses = pd.DataFrame({'t1': allRespT1, 't2': allRespT2})

    print(f":: sort Responses time: {time.time() - time1}")
    return (exemplarPosResponses, exResp, allResponses)


def listOtherExemplarsToRemove(exemplars, exemplarPosResponses):
    print('check non related')
    time1 = time.time()
    t1 = []
    t2 = []

    for exKey in exemplarPosResponses:
        for exemp in exemplars:
            if exKey != exemp:
                for vTrack in exemplarPosResponses[exKey]:
                    t1.append(vTrack)
                    t2.append(exemp)
                    t2.append(vTrack)
                    t1.append(exemp)


    otherExemplars = pd.DataFrame({'t1': t1, 't2': t2})
    print(f"possible non related exemplars {len(otherExemplars.index)}")

    print(f":: list other exemplars time: {time.time() - time1}")
    return otherExemplars


def removeAffinitiesWithPosRespLink(affinities, otherExemplars, exResp):
    print('sorting affinities')
    time1 = time.time()
    af1 = []
    af2 = []
    affinByT1 = {}
    affinByT2 = {}
    alreadyMatchedAtThisTimeT1 = []
    alreadyMatchedAtThisTimeT2 = []
    if len(affinities) > 0:
        for aff in affinities:
            af1.append(aff.track1_uid)
            af2.append(aff.track2_uid)
            if aff.track1_uid not in affinByT1.keys():
                affinByT1[aff.track1_uid] = []
            if aff.track2_uid not in affinByT2.keys():
                affinByT2[aff.track2_uid] = []
            affinByT1[aff.track1_uid].append(aff.track2_uid)
            affinByT2[aff.track2_uid].append(aff.track1_uid)

    # add up all the matches that were already at this time
    alreadyMatchedAtThisTime = pd.DataFrame(
        {'t1': alreadyMatchedAtThisTimeT1, 't2': alreadyMatchedAtThisTimeT2})

    allAff = pd.DataFrame({'t1': af1, 't2': af2})
    print(f"all affinities loaded {len(allAff.index)}")

    # drop the duplicates (difference between the 2 lists)
    dropExResp = getDF1WithoutAnyInDF2(allAff, exResp, ['t1', 't2'])
    print(f"dropped exemplar responses {len(dropExResp.index)}")
    print(f":: time: {time.time() - time1}")

    time1 = time.time()
    # now get rid of any possible extra exemplars:
    dropOtherEx = getDF1WithoutAnyInDF2(dropExResp, otherExemplars,
                                        ['t1', 't2'])
    print(f"drop non related exemplars {len(dropOtherEx.index)}")

    print(f":: time: {time.time() - time1}")
    return (dropOtherEx, affinByT1, affinByT2, alreadyMatchedAtThisTime)


def getAllTracksWithPosResponses(session):
    # get all the tracks with positive responses
    time1 = time.time()
    vprs = session.query(Response).\
        filter(Response.response >= 1).\
        options(load_only('track1_uid')).\
        all()
    videoPositiveResponses = []
    for vp in vprs:
        videoPositiveResponses.append(vp.track1_uid)
    print(f":: get all tracks with pos time: {time.time() - time1}")
    return videoPositiveResponses


def removeVideosWithExemplarLinked(exemplarPosResponses, affinByT1, affinByT2):
    t1 = []
    t2 = []
    time1 = time.time()
    for exKey in exemplarPosResponses:
        for track in exemplarPosResponses[exKey]:
            if track in affinByT1.keys():
                for t in affinByT1[track]:
                    t1.append(track)
                    t2.append(t)
                    t2.append(track)
                    t1.append(t)
            if track in affinByT2.keys():
                for t in affinByT2[track]:
                    t2.append(track)
                    t1.append(t)
                    t1.append(track)
                    t2.append(t)

    addVideoLinks = pd.DataFrame({'t1': t1, 't2': t2})

    print(f"additional video links to a positive exemplar: {len(t1)}")
    print(f":: time: {time.time() - time1}")
    return addVideoLinks


def storeAllMNLsInMemory(session):
    allMNLs = session.query(MustNotLink).all()
    # sort them by t1
    mnlDict = {}
    for mnl in allMNLs:
        if mnl.track1_uid not in mnlDict.keys():
            mnlDict[mnl.track1_uid] = []
        mnlDict[mnl.track1_uid].append(mnl.track2_uid)
    return mnlDict


def checkCache(session, videoOrder, outputListings, mnlDict, settings):
    flagLoadCache = False
    time1 = time.time()
    removes = []
    videoOrderIsGood = False

    for vid in videoOrder:
        # check the positives
        posResp = getPosResponseList(vid, session)
        if len(posResp) > 0:
            print('_________ resetting this source _________')
            removes.append(vid)
            continue
        if vid not in outputListings.keys():
            removes.append(vid)
            continue
        # check the push downs
        if isThisAPushedDown(vid, session, settings['suggestedLimit']):
            removes.append(vid)
            continue
        outputList = outputListings[vid]
        # check for mnl and negatives to remove
        # mnlList = getMNLMatchList(vid, session)
        if vid in mnlDict.keys():
            mnlList = mnlDict[vid]
        else:
            mnlList = []
        negRespList = getNegResponseList(vid, session)
        outRemoves = []
        for outp in outputList:
            if outp in mnlList:
                outRemoves.append(outp)
            if outp in negRespList:
                outRemoves.append(outp)
        for rm in outRemoves:
            for ol in outputList:
                if ol == rm:
                    outputList.remove(ol)
        # now see if we have any left
        if len(outputList) > 0:
            videoOrderIsGood = True
    print(f"::: check loop 1: {time.time() - time1}")
    time1 = time.time()
    # get rid of those flagged
    for rm in removes:
        for vo in videoOrder:
            if vo == rm:
                videoOrder.remove(vo)
    print(f"::: remove loop: {time.time() - time1}")
    # if not then set the flag to load
    if videoOrderIsGood is False:
        flagLoadCache = True
        print('____ actually the cache is no good - reload it ____')
    print(f":: check cache time: {time.time() - time1}")
    return flagLoadCache


def createVideoOrderAndOutputLinks(affinities, exemplars,
                                   exemplarPosResponses,
                                   dropVideoWithPrevExemplar,
                                   alreadyMatchedAtThisTime,
                                   allResponses,
                                   suggestedValByUIDs):
    time1 = time.time()
    # just in case this filters to nothing
    if len(dropVideoWithPrevExemplar) > 1:
        pass4 = getDF1WithoutAnyInDF2(dropVideoWithPrevExemplar,
                                      alreadyMatchedAtThisTime,
                                      ['t1', 't2'])
        print(
            f"dropped already matched at this time: {len(alreadyMatchedAtThisTime.index)}")
        print(f"remaining: {len(pass4.index)}")

        pass5 = getDF1WithoutAnyInDF2(pass4,
                                      allResponses,
                                      ['t1', 't2'])
        print(f"removed - responses - remaining: {len(pass5.index)}")
    else:
        t1 = []
        t2 = []
        pass5 = pd.DataFrame({'t1': t1, 't2': t2})

    priorityExemplarLinks = []  # these are potential links to check in photos
    nonExemplarLinks = []  # these are potential video to video links
    outputListings = {}
    videoOrder = []  # these are the videos that we want to review next
    # ----------------------------------------------------------------------
    # this section sorts through the possible types of links
    # ----------------------------------------------------------------------
    # sort by highest affinity score
    print('sorting through afinities')
    priorityT1 = []
    priorityT2 = []

    if len(affinities) > 0:
        for afinity in pass5.itertuples():
            # orderby sorted these and we have the highest first
            if afinity.t1 in exemplars.keys():
                # track 1 is the exemplar -> 2 is a video
                # make sure we don't have a response on this video first
                skipIt = False
                for exKey in exemplarPosResponses:
                    for track in exemplarPosResponses[exKey]:
                        if afinity.t2 == track:
                            skipIt = True
                            break
                if not skipIt:
                    priorityExemplarLinks.append({'t1': afinity.t1,
                                                  't2': afinity.t2})
                    priorityT1.append(afinity.t1)
                    priorityT2.append(afinity.t2)
                    # save the video portion as the key and examplars to list
                    if afinity.t2 not in outputListings.keys():
                        outputListings[afinity.t2] = []
                    if afinity.t1 not in outputListings[afinity.t2]:
                        outputListings[afinity.t2].append(afinity.t1)
                    if afinity.t2 not in videoOrder:
                        videoOrder.append(afinity.t2)

            elif afinity.t2 in exemplars.keys():
                # track 2 is the exemplar -> 1 is a video
                # make sure we don't have a response on this video first
                skipIt = False
                for exKey in exemplarPosResponses:
                    for track in exemplarPosResponses[exKey]:
                        if afinity.t1 == track:
                            skipIt = True
                            break
                if not skipIt:
                    priorityExemplarLinks.append({'t1': afinity.t1,
                                                  't2': afinity.t2})
                    priorityT1.append(afinity.t1)
                    priorityT2.append(afinity.t2)
                    # save the video portion as the key and examplars to list
                    if afinity.t1 not in outputListings.keys():
                        outputListings[afinity.t1] = []
                    if afinity.t2 not in outputListings[afinity.t1]:
                        outputListings[afinity.t1].append(afinity.t2)
                    if afinity.t1 not in videoOrder:
                        videoOrder.append(afinity.t1)

            else:
                # these are video to video links- here we want to eliminate
                # any video that already has a positive response to an
                # exemplar (check in)
                saveIt = True
                for exKey in exemplarPosResponses:
                    for track in exemplarPosResponses[exKey]:
                        if afinity.t1 == track or afinity.t2 == track:
                            saveIt = False
                            break
                if saveIt:
                    nonExemplarLinks.append({'t1': afinity.t1,
                                             't2': afinity.t2})
    else:
        print('============ affinities are empty ============')

    print(f":: time: {time.time() - time1}")

    print(f"There are {len(priorityExemplarLinks)} links to exemplar")
    print(f"There are {len(nonExemplarLinks)} lower priority links")

    time1 = time.time()
    # see if the remaining items may be equal to suggested items
    remainingVid = len(videoOrder)
    remainingEx = len(priorityExemplarLinks)
    remainingNEx = len(nonExemplarLinks)
    for suggestion in suggestedValByUIDs.keys():
        if suggestion in videoOrder:
            remainingVid -= 1
        for ex in priorityExemplarLinks:
            if ex['t1'] == suggestion or ex['t2'] == suggestion:
                remainingEx -= 1
        for ne in nonExemplarLinks:
            if ne['t1'] == suggestion or ne['t2'] == suggestion:
                remainingNEx -= 1
    print(f":: time: {time.time() - time1}")

    print(f"video order = {len(videoOrder)}")
    print(f"There are {remainingVid} remaining in video order")
    print(f"There are {remainingEx} remaining in ex")
    print(f"There are {remainingNEx} remaining in non ex")
    nonSuggested = remainingVid + remainingEx + remainingNEx

    return (priorityExemplarLinks, nonExemplarLinks,
            outputListings, videoOrder, nonSuggested,
            remainingNEx)


def getSuperKeyFrameTracks(session, sourceByUid, tracks, sources,
                           tracksBySrcUID, videoOrder,
                           outputListings,
                           priorityExemplarLinks, settings):
    # raw SQL
    #    select d.track_uid, d.frame_number FROM detections as d
    #    where d.frame_number % 300 = 0
    #    GROUP BY d.track_uid

    mod_number = 1
    if settings['superKeyFrameMod'] is not None:
        mod_number = settings['superKeyFrameMod']

    print(f'MOD NUMBER IS: {mod_number}')
    time1 = time.time()
    start_time = datetime(2018, 5, 1)
    end_time = datetime(2018, 5, 31)
    unix_start = time.mktime(start_time.timetuple())
    unix_end = time.mktime(end_time.timetuple())

    superKeyTracks = session.query(Detection).\
        where(and_(Detection.frame_number % mod_number == 0, Detection.time > unix_start, Detection.time < unix_end)).\
        group_by(Detection.track_uid).all()
    print(f"super key frame detection == {len(superKeyTracks)}")
    print(f":: time: {time.time() - time1}")
    time1 = time.time()
    # print('checking for MNL collision with supers')
    foundVideo = False
    searchLoops = 0
    while not foundVideo and searchLoops < len(superKeyTracks):
        superVideo = superKeyTracks[searchLoops].track_uid
        # print(f"trying the first video {superVideo}")
        # first see if it was suggested:
        # check the suggested group too
        if isThisAPushedDown(superVideo, session, settings['suggestedLimit']) is False:
            # this one is good to go to the next query
            mnlList = getMNLMatchList(superVideo, session)
            # check for positive responses too:
            posResp = getPosResponseList(superVideo, session)
            if len((posResp)) > 0:
                searchLoops += 1
                continue
            negRespList = getNegResponseList(superVideo, session)

            # get the day the video happened on
            day = sourceByUid[tracks[superVideo].source_uid].name[:10]
            # print(f"checking for actors on day {day}")
            date_time = day.split('-')
            starttime = datetime(
                int(date_time[0]),
                int(date_time[1]),
                int(date_time[2]),
                0,
                0,
                1,
            )
            utime = starttime.timestamp()
            possibleSources = []
            for src in sources:
                # skip the non gps sources
                if src.gps_tracker_id is None:
                    continue
                # see if the day is the same
                if src.start_time < utime < src.end_time:
                    possibleSources.append(src.uid)

            possibleMatched = []
            for srcid in possibleSources:
                if srcid not in tracksBySrcUID.keys():
                    continue
                possibleMatched.append(tracksBySrcUID[srcid].uid)
            # print(f"possible actors {len(possibleMatched)}")

            priorityT1 = []
            priorityT2 = []
            for ex in possibleMatched:
                if ex not in mnlList and ex not in negRespList:
                    # let's add this one in
                    priorityT1.append(superVideo)
                    priorityT2.append(ex)
                    if superVideo not in videoOrder:
                        videoOrder.append(superVideo)
                    if superVideo not in outputListings.keys():
                        outputListings[superVideo] = []
                    outputListings[superVideo].append(ex)
                # else:
                    # print('actor in MNL or Pos. Resp.')

            if len(priorityT1) > 0:
                priorityExemplarLinks.append({'t1': priorityT1,
                                              't2': priorityT2})
                # print(f"video matched to {len(priorityT2)} exemplars")
                # foundVideo = True  # turning this off to fill the cache
            # else:
                # print('MNL eliminated video')
        searchLoops += 1
    print(f":: super key setup time: {time.time() - time1}")


def checkExemplar(session, priorityExemplarLinks, affinities, settings,
                  remainingNEx, nonExemplarLinks, sourceByUid, tracks, sources,
                  tracksBySrcUID, exemplars, nonExemplarTracks,
                  addVideoLinks, allResponses, suggested, nonSuggested,
                  suggestedValByUIDs, videoOrder):
    totalTime1 = time.time()
    checkingExemplar = True
    # only add these extra "left overs" in if we don't have GPS links
    # or video to video links (and "push_down" suggested would be in those)
    # at this point
    loopCount = 0
    while checkingExemplar:
        if len(priorityExemplarLinks) < 1:
            print('<< loading left over non linked tracks >>')
            # if we got here then we may need a fresh load of affinities
            # update the last score value from the affinities
            if len(affinities) > 0:
                settings['lastScore'] = affinities[-1].score
                # save the settings
                saveSettings(settings, settingsFile)

            # check to see if we have any lower priority videos
            if remainingNEx > 0 and loopCount < len(nonExemplarLinks):
                # grab a video off of the video-to-video list
                video = nonExemplarLinks[loopCount]['t1']
                loopCount += 1
                existingMNL = session.query(MustNotLink).\
                    filter(MustNotLink.track1_uid == video).\
                    all()
                # get the day the video happened on
                day = sourceByUid[tracks[video].source_uid].name[:10]
                print(f"checking video track {video} on day {day}")

                # grab actor chips for that day
                date_time = day.split('-')
                starttime = datetime(
                    int(date_time[0]),
                    int(date_time[1]),
                    int(date_time[2]),
                    0,
                    0,
                    1,
                )
                utime = starttime.timestamp()
                possibleSources = []
                for src in sources:
                    # skip the non gps sources
                    if src.gps_tracker_id is None:
                        continue
                    # see if the day is the same
                    if src.start_time < utime < src.end_time:
                        possibleSources.append(src)
                print(
                    f"has existing MNLs: {len(existingMNL)} and possible checkins {len(possibleSources)}")

                # drop the mnls
                removes = []
                for src in possibleSources:
                    if src not in tracksBySrcUID.keys():
                        removes.append(src)
                        continue
                    track = tracksBySrcUID[src.uid].uid
                    for em in existingMNL:
                        if em.track1_uid == track or \
                                em.track2_uid == track:
                            if track not in removes:
                                removes.append(track)
                # see if any remain
                if len(removes) == len(possibleSources):
                    continue
                # add these in as a possible
                for src in possibleSources:
                    if src in removes:
                        continue
                    print(
                        f"adding {video} - {tracksBySrcUID[src.uid].uid}")
                    priorityExemplarLinks.append({'t1': tracksBySrcUID[src.uid].uid,
                                                    't2': video})

            else:
                print('Loading MNLS...')
                time1 = time.time()
                existingMustNotLinks = session.query(MustNotLink).all()
                print(f":: time: {time.time() - time1}")
                print('Sorting MNL by uid...')
                # list of uids from the existing items in the db
                time1 = time.time()
                existMNL_t1 = []
                existMNL_t2 = []
                for em in existingMustNotLinks:
                    existMNL_t1.append(em.track1_uid)
                    existMNL_t2.append(em.track2_uid)
                eMNLs = pd.DataFrame(
                    {'t1': existMNL_t1, 't2': existMNL_t2})
                print(f":: time: {time.time() - time1}")

                # check exemplars for any that fell through the cracks:
                # these could be non responses and non affinities
                # i.e. the GPS was acting up so never linked and never
                # went into the MNL
                t1 = []
                t2 = []
                time1 = time.time()
                for exemp in exemplars:
                    for track in nonExemplarTracks:
                        # need to make sure that these overlap in time first
                        exStart = tracks[exemp].start_time
                        exEnd = tracks[exemp].end_time
                        tStart = tracks[track].start_time
                        tEnd = tracks[track].end_time
                        overlap = False
                        if exStart < tStart < exEnd:
                            overlap = True
                        if exStart < tEnd < exEnd:
                            overlap = True
                        if overlap:
                            t1.append(exemp)
                            t2.append(track)
                            t1.append(track)
                            t2.append(exemp)
                print(f":: time: {time.time() - time1}")

                leftOverExemplars = pd.DataFrame({'t1': t1, 't2': t2})

                leftOverMinusMNL = getDF1WithoutAnyInDF2(leftOverExemplars,
                                                         eMNLs,
                                                         ['t1', 't2'])
                # remove positive responses
                leftOver2 = leftOverMinusMNL
                if addVideoLinks.index > 0:
                    leftOver2 = getDF1WithoutAnyInDF2(leftOverMinusMNL,
                                                      addVideoLinks,
                                                      ['t1', 't2'])
                # make sure the responses are removed too
                leftOver3 = getDF1WithoutAnyInDF2(leftOver2, allResponses,
                                                  ['t1', 't2'])
            checkingExemplar = False
        else:
            time1 = time.time()
            t1 = []
            t2 = []
            leftOver3 = pd.DataFrame({'t1': t1, 't2': t2})
            # see if we need to remove the push_downs
            if len(suggested) > 0 and nonSuggested > 0:
                for suggestion in suggestedValByUIDs.keys():
                    # clean out the video order
                    if suggestion in videoOrder:
                        videoOrder.remove(suggestion)
                    # clean out the nonExemplars
                    toRemove = []
                    for ne in nonExemplarLinks:
                        if ne['t1'] == suggestion or ne['t2'] == suggestion:
                            toRemove.append(ne)
                    for rem in toRemove:
                        nonExemplarLinks.remove(rem)

                    # clean up the priority links
                    toRemove = []
                    for pe in priorityExemplarLinks:
                        if pe['t1'] == suggestion or pe['t2'] == suggestion:
                            toRemove.append(pe)
                    for rem in toRemove:
                        priorityExemplarLinks.remove(rem)
                # after removal metrics
                print(
                    f"push down removal {len(priorityExemplarLinks)} links to exemplar")
                print(
                    f"push down removal {len(nonExemplarLinks)} lower priority links")
                print(f"push down removal {len(videoOrder)} videoOrder")
            print(f"remove push_down time: {time.time() - time1}")
            if len(priorityExemplarLinks) > 0:
                checkingExemplar = False

    print(f"possible left over exemplars {len(leftOver3.index)}")
    print(f":: check exemplar time: {time.time() - totalTime1}")
    return leftOver3


def run2ndPass(leftOver3, exemplars, exemplarPosResponses,
               priorityExemplarLinks, outputListings, videoOrder):
    time1 = time.time()
    for leftOver in leftOver3.itertuples():
        if leftOver.t1 in exemplars.keys():
            # track 1 is the exemplar -> 2 is a video
            # check that we don't have something already repsonded to 1st
            saveIt = True
            for exKey in exemplarPosResponses:
                for track in exemplarPosResponses[exKey]:
                    if leftOver.t2 == track:
                        saveIt = False
                        break
            if saveIt:
                priorityExemplarLinks.append({'t1': leftOver.t1,
                                                't2': leftOver.t2})
                # save the video portion as the key and examplars to list
                if leftOver.t2 not in outputListings.keys():
                    outputListings[leftOver.t2] = []
                if leftOver.t1 not in outputListings[leftOver.t2]:
                    outputListings[leftOver.t2].append(leftOver.t1)
                if leftOver.t2 not in videoOrder:
                    videoOrder.append(leftOver.t2)

        elif leftOver.t2 in exemplars.keys():
            # track 2 is the exemplar -> 1 is a video
            # check that we don't have something already repsonded to 1st
            saveIt = True
            for exKey in exemplarPosResponses:
                for track in exemplarPosResponses[exKey]:
                    if leftOver.t1 == track:
                        saveIt = False
                        break
            if saveIt:
                priorityExemplarLinks.append({'t1': leftOver.t1,
                                              't2': leftOver.t2})
                # save the video portion as the key and examplars to list
                if leftOver.t1 not in outputListings.keys():
                    outputListings[leftOver.t1] = []
                if leftOver.t2 not in outputListings[leftOver.t1]:
                    outputListings[leftOver.t1].append(leftOver.t2)
                if leftOver.t1 not in videoOrder:
                    videoOrder.append(leftOver.t1)
    print(
        f"2nd pass returned {len(priorityExemplarLinks)} links to exemplar")
    print(f":: 2nd pass time: {time.time() - time1}")


def removePosResponses(videoOrder, videoPositiveResponses):
    time1 = time.time()
    removes = []
    print(f"videoOrder {len(videoOrder)}")
    for test in videoOrder:
        if test in videoPositiveResponses:
            removes.append(test)
    for rm in removes:
        for vo in videoOrder:
            if vo == rm:
                videoOrder.remove(vo)
    print(f"removed {len(removes)}")
    print(f"videoOrder {len(videoOrder)}")
    print(f":: remove pos repsonses time: {time.time() - time1}")


def setupSource(source, allTracks, sources, videoOrder, nonExemplarLinks):
    # list of outputs - built up from possibles
    time1 = time.time()
    outputList = []
    sourceID = None
    if source is not None:
        # get links related to this source
        if source['type'] == 'actor':
            print('sorting actors')
            actors = {}
            for src in sources:
                if src.outfit_id == '':
                    continue
                elif src.outfit_id not in actors.keys():
                    actors[src.outfit_id] = []
                actors[src.outfit_id].append(src)
            date = source['date']
            outfitId = source['outfitId']
            actorId = source['actorId']
            # get the source uid
            srcId = None
            for ac in actors[outfitId]:
                if ac.name.split('-')[1] == date and \
                        ac.girder_id == actorId:
                    srcId = ac.uid
                    break
            print(f"actor uid = {srcId}")
            # get the track uid
            for track in allTracks:
                if track.source_uid == srcId:
                    sourceID = track.uid
                    break

        else:  # type is track
            folderId = source['folderId']
            trackId = source['trackId']
            # get the track uid
            for track in allTracks:
                if track.video_track_id == trackId and \
                        track.source_uid == folderId:
                    sourceID = track.uid
                    break
    else:
        # create our own source since they didn't give us one
        if len(videoOrder) > 0:
            sourceID = videoOrder[0]
        elif len(nonExemplarLinks) > 0:
            # no exemplar type so let's get the intervideo ones
            # first sort get some videos that are similar
            sourceID = nonExemplarLinks[0]['t1']
    print(f":: setup source time: {time.time() - time1}")
    return sourceID, outputList


def reviewSourceID(session, videoOrder, outputListings, nonExemplarLinks,
                   sourceID, source, outputList):
    time1 = time.time()
    workingOnSourceID = True
    while workingOnSourceID:
        print(f"track uid = {sourceID}")
        print(f"track is in outputListings: {sourceID in outputListings}")
        # use the source uid to generate our output
        if len(videoOrder) > 0 and \
                sourceID in outputListings.keys():
            outputList = outputListings[sourceID]
            # check for positive responses to this source:
            posResp = getPosResponseList(sourceID, session)
            if len(posResp) > 0:
                print('_________ resetting this source _________')
                videoOrder.remove(sourceID)
                outputList = []  # reset if there is a positive
            # check for mnl and negatives to remove
            mnlList = getMNLMatchList(sourceID, session)
            negRespList = getNegResponseList(sourceID, session)
            removes = []
            for outp in outputList:
                if outp in mnlList:
                    removes.append(outp)
                if outp in negRespList:
                    removes.append(outp)
            for rm in removes:
                for ol in outputList:
                    if ol == rm:
                        outputList.remove(ol)
                        print(f"_________ cleaning {ol} _________")

        print(f"outputList has {len(outputList)}")
        # see if we have some items to show
        if len(outputList) > 0:
            workingOnSourceID = False
        else:
            print(f"videoOrder has {len(videoOrder)}")
            # try changing the next sourceID
            if len(videoOrder) > 0:
                # pop all the videos off of the videoOrder if
                # they are of no use to us here
                if videoOrder[0] == sourceID:
                    videoOrder.remove(sourceID)
                # get a new source ID
                sourceID = videoOrder[0]  # grabs the next one since we pop
                source = None  # must change source for this

                print(f"new sourceID {sourceID}")

            elif len(nonExemplarLinks) > 0:
                sourceID = nonExemplarLinks[0]['t1']
                source = None  # must change source for this
                # pop all non Exemplars off if not of use to us
                removes = []
                for test in nonExemplarLinks:
                    if test == sourceID:
                        break
                    removes.append(test)
                for rm in removes:
                    for vo in nonExemplarLinks:
                        if vo == rm:
                            nonExemplarLinks.remove(vo)
            else:
                workingOnSourceID = False  # fall out with blank
    print(f":: review source time {time.time() - time1}")
    return outputList


def createSource(sourceID, sourceByUid, exemplars, tracks):
    time1 = time.time()

    # get the source info
    print('setting up new source')
    source = {}
    src1 = sourceByUid[tracks[sourceID].source_uid]
    if sourceID in exemplars:
        source['type'] = 'actor'
        source['actorId'] = src1.girder_id
        source['date'] = src1.name.split('-')[1]
        source['outfitId'] = src1.outfit_id
    else:
        source['folderId'] = src1.uid
        source['trackId'] = tracks[sourceID].video_track_id
        source['type'] = 'track'
    print(f"new source: {source}")

    print(f":: create time: {time.time() - time1}")
    return source


def loadDBItems(session, lastScore, forceLoad=False, skipSourceData="201803"):
    global _tracks
    global _affinities
    global _responses
    global _lastSaveTime
    global _detectionsByTrackUID
    # grab all tracks
    time1 = time.time()
    print('Loading all tracks...')
    if len(_tracks) < 1 or forceLoad:
        if skipSourceData == '':
            allTracks = session.query(Track).all()
        else:
            start_time = datetime(2018, 5, 1)
            end_time = datetime(2018, 5, 31)
            unix_start = time.mktime(start_time.timetuple())
            unix_end = time.mktime(end_time.timetuple())
            allTracks = session.query(Track).filter(and_(
                Track.start_time > unix_start,
                Track.start_time < unix_end
            )).all()
        _tracks = allTracks
    else:
        allTracks = _tracks
    print(f":: time: {time.time() - time1}")
    time1 = time.time()

    print('Loading Afinities in desc order...')
    if len(_affinities) < 1 or forceLoad:
        affinities = session.query(Afinity).\
            filter(and_(Afinity.score < lastScore, Afinity.score >= 0.01)).\
            order_by(Afinity.score.desc()).\
            limit(_maxAffinitiesToLoad).\
            all()
        _affinities = affinities
    else:
        print('________________________________ using globals')
        affinities = _affinities
    print(f"affinities loaded: {len(affinities)}")
    print(f":: time: {time.time() - time1}")

    print('Loading responses...')
    time1 = time.time()
    if len(_responses) < 1 or forceLoad:
        responses = session.query(Response).\
            order_by(Response.saved_time.asc()).all()
        _responses = responses
        print(f"repsones: {len(responses)}")
        if len(responses) > 0:
            _lastSaveTime = responses[-1].saved_time
        else:
            _lastSaveTime = time.time()
    else:
        # check for updates first
        newResp = session.query(Response).\
            filter(Response.saved_time > _lastSaveTime).\
            order_by(Response.saved_time.asc()).all()
        if len(newResp) > 0:
            print(newResp[0].saved_time)
            _responses = _responses + newResp
            print(f"repsones: {len(_responses)}")
            if len(_responses) > 0:
                _lastSaveTime = _responses[-1].saved_time
            else:
                _lastSaveTime = time.time()
        responses = _responses
    print(f":: time: {time.time() - time1}")

    print('Loading sources...')
    time1 = time.time()
    if skipSourceData == "":
        sources = session.query(Source).all()
    else:
        skipVid = skipSourceData[:4] + "-" + skipSourceData[-2:]
        sources = session.query(Source).\
            filter(and_( not_(Source.name.contains(skipSourceData)), 
                        not_(Source.name.contains(skipVid)) ))
    print(f":: time: {time.time() - time1}")

    print('Loading previous suggestions...')
    time1 = time.time()
    suggested = session.query(Suggested).order_by(
        Suggested.suggestion_count).all()
    print(f":: time: {time.time() - time1}")

    print('Loading detections for size sorting...')
    time1 = time.time()
    if len(_detectionsByTrackUID.keys()) < 1 or forceLoad:
        start_time = datetime(2018, 5, 1)
        end_time = datetime(2018, 5, 31)
        unix_start = time.mktime(start_time.timetuple())
        unix_end = time.mktime(end_time.timetuple())
        allDets = session.query(Detection).filter(and_(
                Detection.time > unix_start,
                Detection.time < unix_end
            )).all()
        detectionsByTrackUID = {}
        for det in allDets:
            if det.track_uid not in detectionsByTrackUID.keys():
                detectionsByTrackUID[det.track_uid] = []
            detectionsByTrackUID[det.track_uid].append(det)
        _detectionsByTrackUID = detectionsByTrackUID
        setupAreas(detectionsByTrackUID)
    else:
        detectionsByTrackUID = _detectionsByTrackUID
    print(f":: time: {time.time() - time1}")

    return (allTracks, affinities, responses, sources, suggested, detectionsByTrackUID)


def sortPrevSuggested(suggested):
    print('sorting prev suggested')
    time1 = time.time()
    suggestedValByUIDs = {}
    for sug in suggested:
        name = sug.track1_uid  # we are only saving the "source" for now
        suggestedValByUIDs[name] = sug.suggestion_count
    print(f":: time: {time.time() - time1}")
    return suggestedValByUIDs


def sortSrcByUID(sources):
    print('Sorting source by uid...')
    sourceByUid = {}
    time1 = time.time()
    for src in sources:
        sourceByUid[src.uid] = src
    print(f":: time: {time.time() - time1}")
    return sourceByUid


def removeUserData(settings: dict, userID):
    if userID in settings:
        if len(settings[userID]) < 1:
            print(f"user {userID} has no data")
            return
        print(f"pulling user data from {userID}")
        reclaimed = settings[userID]
        # put these back into the pool for general use
        print(f"videoOrder before = {len(settings['videoOrder'])}")
        settings['videoOrder'] += reclaimed
        print(f"videoOrder after = {len(settings['videoOrder'])}")
        # remove this user from the active users
        if userID in settings['activeUsers']:
            settings['activeUsers'].remove(userID)
        settings[userID] = []


def setupUserData(videoOrder: list, settings: dict, userID):
    if userID not in settings.keys():
        print(f"setting up new user {userID}")
        settings[userID] = []
    if 'activeUsers' not in settings.keys():
        settings['activeUsers'] = []
    if userID not in settings['activeUsers']:
        settings['activeUsers'].append(userID)
    # pop some items off the list
    print(f"videoOrder before = {len(videoOrder)}")
    testSize = 20
    if len(videoOrder) < testSize:
        testSize = len(videoOrder)
    settings[userID] = videoOrder[:testSize]
    # remove from the original list
    if len(videoOrder) > testSize:
        videoOrder = videoOrder[testSize+1:]
    else:
        videoOrder = []
    # update the settings
    print(f"videoOrder after = {len(videoOrder)}")
    settings['videoOrder'] = videoOrder
    return (settings[userID], settings['videoOrder'])


def setupAreas(detByTrackUID):
    global _areas  # half the area since we are just adding length + width
    areas = {}
    for video in detByTrackUID:
        largestArea = 0
        for det in detByTrackUID[video]:
            preloc = det.bounding_box.replace('[', '').replace(']', '')
            box = np.array(preloc.split(), dtype=np.int32)
            #print(f" box {box}")
            area = (int(box[2])-int(box[0])) + (int(box[3])-int(box[1]))
            #print(f" area {area}")
            if area > largestArea:
                largestArea = area
        areas[video] = largestArea
        # sort and then rebuild the videoOrder
    areas = dict(sorted(areas.items(), key=lambda item: item[1], reverse=True))
    _areas = areas


def sortVideoByLargestSizeChip(videoOrder):
    time1 = time.time()
    global _areas
    newVideoOrder = []
    # get the largest chip sizes per video
    for ar in _areas:
        if ar in videoOrder:
            newVideoOrder.append(ar)
    print(f":: sortVideoByLargestSizeChip time: {time.time() - time1}")
    return newVideoOrder


def getPotentialLinks(session, settingsFile,
                      source=None, getJustSource=False,
                      forceDBReload=False, userID=None,
                      forceUnloadUserData=False):
    """
    returns potential links in order of afinity score

    Args:
        session (mevaDB session): [description]
        source (): a dict with (actor or track type)
    """
    timeo = time.time()
    turnOffSqlAlchemyLogging()
    global _sortedAffinities
    global _affinByT1
    global _affinByT2
    global _alreadyMatchedAtThisTime
    global _mnlDict

    # get the settings file
    time1 = time.time()
    settings = getSettings(settingsFile)
    if len(settings.keys()) < 1:
        # new file, so start it off
        settings['lastScore'] = 1.0
        settings['suggestedLimit'] = 1
        settings['superKeyFrameMod'] = 1
        saveSettings(settings, settingsFile)
    lastScore = settings['lastScore']
    print(f"lastScore: {lastScore}")

    # special call just to handle removing a user's held items
    if forceUnloadUserData:
        removeUserData(settings, userID)
        saveSettings(settings, settingsFile)
        return (None, [])

    # check for cache in settings
    flagLoadCache = True
    if 'videoOrder' in settings.keys():
        if len(settings['videoOrder']) > 0 and \
                (len(settings['priority']) > 0 or
                 len(settings['vid2vid']) > 0):
            flagLoadCache = False
        # don't reload cache if users have data
        if 'activeUsers' in settings.keys():
            for user in settings['activeUsers']:
                if len(settings[user]) > 0:
                    flagLoadCache = False
            print('================ checking cache ================')
    print(f":: load settings file time: {time.time() - time1}")

    # load items from the DB that we need to review
    allTracks, affinities, responses, sources, suggested, detByTrackUID = \
        loadDBItems(session, lastScore, forceDBReload)

    # sort items for easy access
    suggestedValByUIDs = sortPrevSuggested(suggested)
    sourceByUid = sortSrcByUID(sources)
    exemplars, nonExemplarTracks, tracks, tracksBySrcUID = sortExemplars(
        session, allTracks)
    exemplarPosResponses, exResp, allResponses = sortResponses(responses,
                                                               exemplars)

    # check for in memory storage of the sorted affinities
    if len(_sortedAffinities) < 1 or forceDBReload:
        # make a list of all other exemplars to remove if this one has a match
        otherExemplars = listOtherExemplarsToRemove(exemplars,
                                                    exemplarPosResponses)

        # remove extra affinities if we already had a positive response
        # on this item back to an exemplar
        dropOtherEx, affinByT1, affinByT2, alreadyMatchedAtThisTime = \
            removeAffinitiesWithPosRespLink(affinities,
                                            otherExemplars,
                                            exResp)
        print(f"drop non related exemplars {len(dropOtherEx.index)}")

        # save these for the next run
        _sortedAffinities = dropOtherEx
        _affinByT1 = affinByT1
        _affinByT2 = affinByT2
        _alreadyMatchedAtThisTime = alreadyMatchedAtThisTime
        print(
            f"<<<<<<<<<<<<<<<<<<<<<<< loading globals {len(_sortedAffinities)}")
    else:
        # run these from memory
        dropOtherEx = _sortedAffinities
        affinByT1 = _affinByT1
        affinByT2 = _affinByT2
        alreadyMatchedAtThisTime = _alreadyMatchedAtThisTime

    # get all the tracks with positive responses
    videoPositiveResponses = getAllTracksWithPosResponses(session)

    # now remove any videos that already have an exemplar linked
    # to them so that other exemplars are not sugested
    addVideoLinks = removeVideosWithExemplarLinked(
        exemplarPosResponses, affinByT1, affinByT2)

    # now remove these from the list to show
    if len(addVideoLinks.index) > 0:
        dropVideoWithPrevExemplar = getDF1WithoutAnyInDF2(dropOtherEx,
                                                          addVideoLinks,
                                                          ['t1', 't2'])
    else:
        dropVideoWithPrevExemplar = dropOtherEx
    print(f"dropped addVideoLinks {len(dropVideoWithPrevExemplar.index)}")

    # look at the annotator response matrix build a list of
    # items that shouldn't pop up again
    # (NOTE: we are not over-writing the affinity score or must not links
    # because we want those "raw" in case the annotator changes their mind)

    priorityExemplarLinks = []  # these are potential links to check in photos
    nonExemplarLinks = []  # these are potential video to video links
    outputListings = {}  # these are lists of exemplar to video possibilities
    videoOrder = []  # these are the videos that we want to review next

    # see if we need to reload the MNL dictionary
    if len(_mnlDict.keys()) < 1 or forceDBReload:
        mnlDict = storeAllMNLsInMemory(session)
        _mnlDict = mnlDict
    else:
        mnlDict = _mnlDict

    # if we are not re-Loading the cache, then get the cache and check it
    # it might be bad / not enough to work with, so we may flag it for
    # reloading
    if not flagLoadCache:
        # here we take the cache values
        videoOrder = settings['videoOrder']
        priorityExemplarLinks = settings['priority']
        nonExemplarLinks = settings['vid2vid']
        outputListings = settings['outputListings']
        # double check that we have something to use
        # sort through the video order
        flagLoadCache = checkCache(session, videoOrder,
                                   outputListings, mnlDict, settings)
        # turn this off if we have users with data
        if 'activeUsers' in settings.keys():
            for user in settings['activeUsers']:
                if len(settings[user]) > 0:
                    flagLoadCache = False

    # reload the cache since it is either no good or was missing
    if flagLoadCache:
        (priorityExemplarLinks, nonExemplarLinks,
         outputListings, videoOrder, nonSuggested,
         remainingNEx) = \
            createVideoOrderAndOutputLinks(affinities, exemplars,
                                           exemplarPosResponses,
                                           dropVideoWithPrevExemplar,
                                           alreadyMatchedAtThisTime,
                                           allResponses,
                                           suggestedValByUIDs)

        # get the amount that have the detections that match the 300
        # frame number (super key frame) that we use for scoring
        # these need to be prioritized
        getSuperKeyFrameTracks(session, sourceByUid, tracks, sources,
                               tracksBySrcUID, videoOrder,
                               outputListings,
                               priorityExemplarLinks, settings)

        leftOver3 = checkExemplar(session, priorityExemplarLinks, affinities,
                                  settings, remainingNEx,
                                  nonExemplarLinks, sourceByUid,
                                  tracks, sources,
                                  tracksBySrcUID, exemplars,
                                  nonExemplarTracks,
                                  addVideoLinks, allResponses,
                                  suggested, nonSuggested,
                                  suggestedValByUIDs, videoOrder)

        # try to add these into the exemplar responses with a really low score
        run2ndPass(leftOver3, exemplars, exemplarPosResponses,
                   priorityExemplarLinks, outputListings, videoOrder)

    print('sort videos by detection size')
    videoOrder = sortVideoByLargestSizeChip(videoOrder)

    print(f"videoOrder in non user cache {len(videoOrder)}")
    # use the stored list or make it if not there
    userVideoOrder = []
    settingUpUser = True
    setupAttempts = 0
    while settingUpUser:
        setupAttempts += 1

        if setupAttempts > 99:  # prevent infinite looping
            print("!!!!!!!!!!!!!!!!!!!!! USER SETUP ERROR !!!!!!!!!!!!!!!!!!!!!")
            break

        if userID is not None:
            if 'activeUsers' in settings.keys():
                print(f"active users: {settings['activeUsers']}")
            # see if the user is there:
            if userID in settings.keys():
                if len(settings[userID]) > 1:
                    # we have a cache to use
                    print(f"________________user {userID} has {len(settings[userID])} videos")
                    userVideoOrder = settings[userID]
                else:
                    # refill the user's data
                    userVideoOrder, videoOrder = setupUserData(videoOrder,
                                                               settings,
                                                               userID)
                    settings['videoOrder'] = videoOrder
            else:
                # setup user
                userVideoOrder, videoOrder = setupUserData(videoOrder,
                                                           settings,
                                                           userID)
                settings['videoOrder'] = videoOrder
        else:
            userVideoOrder = videoOrder  # just use the video order as is

        # one more check for user vid items
        removes = []
        for vid in userVideoOrder:
            # check the push downs
            if isThisAPushedDown(vid, session, settings['suggestedLimit']):
                removes.append(vid)
                continue
        print(f"__dropping {len(removes)} from the user cache (push down IDKs)")
        for remove in removes:
            userVideoOrder.remove(remove)
        # check to see if we have items
        if len(userVideoOrder) > 0:
            settingUpUser = False
        else:
            print('need to grab more user data - this set is empty')

    # make sure that these videos don't already have a positive response
    removePosResponses(userVideoOrder, videoPositiveResponses)
    if userID is not None:
        settings[userID] = userVideoOrder

    # ----------------------------------------------------------------------
    # This section takes the videos selected and sets up the json to return
    # ----------------------------------------------------------------------
    # resort this based on actor chips and videos
    print(f"Priority videos to review {len(userVideoOrder)}")

    sourceID, outputList = setupSource(source, allTracks, sources,
                                       userVideoOrder,
                                       nonExemplarLinks)

    outputList = reviewSourceID(session, userVideoOrder, outputListings,
                                nonExemplarLinks,
                                sourceID, source, outputList)

    if len(outputList) < 1:
        print('we seem to have run out of options to show')
        return (source, [])  # empty list of results and 0 pages

    print(f"Possible links to the source: {len(outputList)}")

    # if it isn't setup yet - make the source to return
    if source is None:
        source = createSource(sourceID, sourceByUid, exemplars, tracks)

    # if we are doing reID next then don't format the list yet
    # keep these UIDs for later
    if getJustSource:
        # getJustSource then stop here - no need to get the other info
        output = outputList
    else:
        # accumulate the list
        output = getFormattedOutputList(outputList, tracks,
                                        sourceByUid, exemplars)

    # now we need to update the settings and save
    print(f"_______ videos left in cache {len(videoOrder)} _____________")
    time1 = time.time()
    settings['videoOrder'] = videoOrder
    settings['priority'] = priorityExemplarLinks
    settings['vid2vid'] = nonExemplarLinks
    settings['outputListings'] = outputListings
    saveSettings(settings, settingsFile)
    print(f":: saving settings time: {time.time() - time1}")

    print(f":::::::::::::: full advise time: {time.time() - timeo}")
    print(f' MOD NUMBER: {settings["superKeyFrameMod"]}')
    print(f'SuggestedLimit: {settings["suggestedLimit"]}')
    return (source, output)


if __name__ == "__main__":
    # set up for running as just a script (testing)
    # this overrides other settings
    from meva_server.meva_database.database import MevaDatabase

    # settings (here is the main path used for others below)
    _data_path = '/home/local/KHQ/joseph.vanpelt/projects/meva/'
    settingsFile = _data_path + 'settings.json'

    # use the code from cloning the above repo
    sys.path.append(_data_path + 'reid/reid-strong-baseline')
    from modeling import build_model as build_dg_model
    from config import cfg as dg_cfg

    time00 = time.time()
    print('Reading arguments...')
    # parse the arguments - required for location of database
    parser = argparse.ArgumentParser(
        description='Script to advise possible links')
    parser.add_argument('databasePath', help='path to database file')
    parser.add_argument('--page', help='which page to send',
                        default=_page, type=int)
    parser.add_argument('--amtPerPage', help='items per page',
                        default=_amtPerPage, type=int)
    args = parser.parse_args()

    offset = args.page
    pageSize = args.amtPerPage

    # connect to the db
    print('Setting up db...')
    database_path = Path(args.databasePath)
    os.environ['MEVA_DATABASE_URI'] = 'sqlite:///' + str(database_path)
    meva_db = MevaDatabase()
    session = meva_db.Session()

    turnOffSqlAlchemyLogging()

    src = {
        'type': 'track',
        'folderId': '4629a0b928ab4d2babbf05b8252e3059',
        'trackId': '26'
    }
    # other example:
    src = {
        'type': 'actor',
        'actorId': '019',
        'date': '20180315',
        'outfitId': '018'
    }
    src = None

    runReID = False
    getAllActorsForTheDay = False
    userID = '3'  # '12345678'

    getJustSource = False  # get just the source video so we can run ReID
    if runReID and getAllActorsForTheDay:
        getJustSource = True  # this makes the list of results return as uids
    sourceID, results = getPotentialLinks(session=session,
                                          settingsFile=settingsFile,
                                          source=src,
                                          getJustSource=getJustSource,
                                          userID=userID,
                                          forceUnloadUserData=False)
    if sourceID is None:
        print("no source returned")

    if runReID:
        # get the image from this source
        filepath = _data_path + 'test1_20180306.jpg'
        img = Image.open(filepath)

        if getAllActorsForTheDay:
            # to run this with all checkins for the day, set
            # this to empty list
            results = []
        sourceID, results = sortByReID(session,
                                       sourceID,
                                       results,
                                       img)

    # pageinate at the end so we get all the results sent to ReID:
    totalPages, results = paginateList(results, offset, pageSize)

    print(f"results: {results}")

    print(f"totalPages: {totalPages}")
    print(f"source = {sourceID}")

    print(json.dumps(results, indent=4, sort_keys=True))
    print(f"count = {len(results)}")

    print(f"Total time: {time.time()-time00}")
