from io import BytesIO
import io
import json
from girder.models.upload import Upload
import pymongo
from meva_tasks.MustNotLinkTask import local_must_not_link_task
from girder.models.token import Token
from girder.exceptions import RestException
import gpxpy
import math
import numpy as np
import os
import stat
import random
import pytz
import uuid
import shutil
import time
import pandas as pd
from datetime import date, datetime, timedelta
from zipfile import ZipFile
from pathlib import Path
from itertools import combinations
from PIL import Image

from girder.api import access
from girder.api.describe import Description, autoDescribeRoute
from girder.api.rest import Resource, getCurrentUser, setContentDisposition, setResponseHeader
from girder.constants import AccessType, TokenScope
from girder.models.file import File
from girder.models.folder import Folder
from girder.models.user import User
from girder.models.item import Item
from girder.utility import ziputil
from sqlalchemy import and_, case, create_engine, not_, or_
from sqlalchemy.orm import (
    contains_eager,
    joinedload,
    lazyload,
    selectinload,
    sessionmaker,
    subqueryload,
)
from sqlalchemy.sql.elements import Null
from sqlalchemy.sql.expression import func, select
from sqlalchemy_pagination import paginate

from meva_server.utils import chips_file, detections_file, getChipData, getTrackData, \
    load_actor_json
from meva_server.constants import (
    CameraModelFolderName,
    DBFolderName,
    FinalTrackDownloadFolder,
    GPSFolderName,
    LinkedTracksExportFolder,
    gpxRegex,
    zipRegex
)

from meva_server.meva_database.models import (
    ActorDisplayChip,
    Base,
    Camera,
    Detection,
    Response,
    Source,
    Track,
    Gps,
    MustNotLink,
    Afinity,
    Suggested
)

from meva_server.meva_database.utils import (
    cleanName,
    geolocate_detection,
    getCameraModelMap,
    getFrameTime,
    checkForMustNotLinks,
    isTimeOverlap,
    getSource,
)
from meva_server.meva_database.advisePossibleLinks import (
    getPotentialLinks,
    getSettings,
    removeUserData,
    saveSettings,
    sortByReID,
    paginateList,
)
from meva_tasks.tasks import (
    add_linking_tracks_to_database,
    export_DB,
    export_linked_track_images_task,
    extract_actor_descriptors,
    remove_actor_images
)
from meva_utils import TRUTHY_META_VALUES

class MevaDatabase(Resource):
    def __init__(self):
        super(MevaDatabase, self).__init__()
        self.resourceName = "meva_database"
        self.databaseUri = os.environ.get('MEVA_DATABASE_URI')
        self.route("PUT", ("add_tracks",), self.add_tracks)
        self.route("PUT", ("add_actor_chips", ":id"), self.add_actor_chips)
        self.route("PUT", ("add_gps_tracks", ":id"), self.add_gps_tracks)
        self.route("PUT", ("add_camera_models", ":id"), self.add_camera_models)
        self.route("PUT", ("add_actor_descriptors", ":id"), self.add_actor_descriptors)
        self.route("PUT", ("reset_database",), self.reset_database)
        self.route("GET", ("get_source_folders",), self.get_source_folders)
        self.route("GET", ("get_page",), self.get_page)
        self.route("GET", ("get_random",), self.get_random)
        self.route("GET", ("get_actor_chips",), self.get_actor_chips)
        self.route("GET", ("get_actor_chip",), self.get_actor_chip)
        self.route("GET", ("get_actor_images",), self.get_actor_images)
        self.route("GET", ("get_actor_filter",), self.get_actor_filter)
        self.route("GET", ("get_linked_tracks",), self.get_linked_tracks)
        self.route("GET", ("get_actor_links",), self.get_actor_links)
        self.route("GET", ("get_linking_info",), self.get_linking_info)
        self.route("GET", ("get_input_status",), self.get_input_status)
        # below are 4 test methods to confirm that data is being populated
        self.route("GET", ("get_gps_tracks",), self.get_gps_tracks)
        self.route("GET", ("get_camera_models",), self.get_camera_models)
        self.route("DELETE", ("delete_gps_tracks",), self.delete_gps_tracks)
        self.route("DELETE", ("delete_camera_models",), self.delete_camera_models)
        self.route("DELETE", ("delete_linking_tables",), self.delete_linking_tables)
        self.route("DELETE", ("reset_linking_responses",), self.reset_response_tables)
        self.route("DELETE", ("delete_video_tracks",), self.delete_video_tracks)
        self.route("DELETE", ("delete_actor_chips", ":id"), self.delete_actor_chips)
        self.route("DELETE", ("delete_actor_descriptors",), self.delete_actor_descriptors)
        self.route("POST", ("populate_MNLs",), self.populate_MNLs)
        self.route("POST", ("populate_linked_tracks", ":id"), self.populate_linked_tracks)
        self.route("POST", ("export_database", ":id"), self.export_database)
        self.route("POST", ("export_linked_tracks", ":id"), self.export_linked_tracks)
        self.route("GET", ("get_database_download", ":id"), self.get_database_download)
        self.route("POST", ("export_linked_track_images",), self.export_linked_track_images)
        self.route("POST", ("refine",), self.refine)
        self.route("POST", ("refine_actor_links",), self.refine_actor_links)
        self.route("POST", ("push_down",), self.push_down)
        self.route("POST", ("remove_track_link",), self.remove_track_link)
        self.route("POST", ("force_unload_user_cache",), self.force_unload_user_cache)
        self.route("POST", ("set_linking_rules",), self.set_linking_rules)
        self.route("GET", ("get_user_cache",), self.get_user_caches)

        # Create database engine and sessionmaker
        self.engine = create_engine(self.databaseUri, echo=True)
        self.Session = sessionmaker(bind=self.engine)

        # Directory that holds the camera model data
        self.camera_model_dir = Path('/data/camera_models')

    @access.admin
    @autoDescribeRoute(
        Description("Reset the database by deleted and recreating the tables")
    )
    def reset_database(self):
        Base.metadata.drop_all(self.engine)
        Base.metadata.create_all(self.engine)
        return {}

    @access.user
    @autoDescribeRoute(
        Description("Add tracks from proceesed video to database").modelParam(
            "folderId",
            description="folder id of video clip",
            model=Folder,
            paramType="query",
            required=True,
            level=AccessType.WRITE,
        )
    )
    def add_tracks(self, folder):
        detect_file = detections_file(folder)
        if detect_file is None:
            return {"status": "no tracks"}
        track_data = getTrackData(detect_file)

        count = len(track_data.keys())
        folder_uid = str(folder["_id"])
        folder_obj = Folder().load(
            folder_uid, user=self.getCurrentUser(), level=AccessType.READ
        )

        # Get camera model map
        camera_model_map_file = self.camera_model_dir / 'meva-clip-camera-and-time-table.txt'
        camera_model_map = getCameraModelMap(camera_model_map_file)

        # Path to camera model file
        cam_model_dir = self.camera_model_dir / 'camera-models'

        self._add_tracks(track_data, folder_uid, folder_obj,
                         camera_model_map, cam_model_dir)

        return {"count": count}

    def _add_tracks(self, track_data, folder_uid, folder_obj, camera_model_map, cam_model_dir):
        session = self.Session()
        # We need to remove any existing folder and tracks that are set
        session.query(Track).filter(Track.source_uid == folder_uid).delete()
        session.query(Source).filter(Source.uid == folder_uid).delete()
        session.query(Detection).filter(Detection.source_uid == folder_uid).delete()

        starttime = datetime.fromtimestamp(0)
        endtime = datetime.fromtimestamp(0)
        folder_meta = folder_obj.get("meta", None)
        camera_id = ""
        location = ""
        if folder_meta is not None:
            meva_meta = folder_meta.get('MEVA_Meta', None)
            if meva_meta is not None:
                date_time = meva_meta.get("date").split('-')
                starttimestr = meva_meta.get("startTime").split('-')
                endtimestr = meva_meta.get("endTime").split('-')
                starttime = datetime(
                    int(date_time[0]),
                    int(date_time[1]),
                    int(date_time[2]),
                    int(starttimestr[0]),
                    int(starttimestr[1]),
                    int(starttimestr[2]),
                )
                endtime = datetime(
                    int(date_time[0]),
                    int(date_time[1]),
                    int(date_time[2]),
                    int(endtimestr[0]),
                    int(endtimestr[1]),
                    int(endtimestr[2]),
                )
                camera_id = meva_meta.get("cameraId", "")
                location = meva_meta.get("location", "")

        folder_db = Source(
            uid=folder_uid,
            name=folder_obj["name"],
            start_time=starttime.timestamp(),
            end_time=endtime.timestamp(),
        )
        session.add(folder_db)

        # TODO: This probably shouldn't be hard-coded.
        localTZ = pytz.timezone('US/East-Indiana')

        srcName = cleanName(folder_obj["name"])

        # Get camera models if available for location determination
        camera_model_name = camera_model_map.get(cleanName(folder_obj["name"]),
                                                 'no-camera-model')
        if 'no-camera-model' not in camera_model_name:
            npy_file = (Path(cam_model_dir) /
                        Path(camera_model_name).stem /
                        'xyz_per_pixel.npy')
            try:
                cm_xyz = np.load(npy_file)
            except IOError:
                cm_xyz = None
                print('WARNING: camera model npy file missing for ',
                      folder_obj["name"])
        else:
            print('WARNING: camera model missing for ', folder_obj["name"])
            cm_xyz = None

        tracks = []
        detections = []
        for track in track_data:
            confidence_pairs = track_data[track]['confidencePairs']
            if confidence_pairs[0][0].startswith('person_') is False:
                continue
            uid = uuid.uuid4().hex
            trk_start_time = getFrameTime(srcName, localTZ,
                                          track_data[track]['features'][0]['frame'],
                                          folder_obj.get('fps', 30))
            trk_end_time = getFrameTime(srcName, localTZ,
                                        track_data[track]['features'][-1]['frame'],
                                        folder_obj.get('fps', 30))

            tracks.append(
                Track(
                    uid=uid,
                    source_uid=folder_uid,
                    video_track_id=track_data[track]['trackId'],
                    start_time=trk_start_time,
                    end_time=trk_end_time,
                )
            )

            for detection in track_data[track]['features']:
                detect_uid = uuid.uuid4().hex
                if 'attributes' not in detection:
                    continue
                if 'poses' not in detection['attributes']:
                    continue
                pose = detection['attributes']['poses']['points']
                bounding_box = np.array([int(v) for v in detection['bounds']])

                vidTime = getFrameTime(srcName, localTZ, detection['frame'],
                                       folder_obj.get('fps', 30))

                if cm_xyz is not None:
                    detect = {}
                    detect["bounds"] = bounding_box

                    # get the location and covariance
                    loc, cov = geolocate_detection(detect, cm_xyz)

                    location = np.array2string(loc)
                    location_cov = np.array2string(cov)
                else:
                    location = None
                    location_cov = None

                # All poses aren't guaranteed to exist
                detections.append(
                    Detection(
                        uid=detect_uid,
                        source_uid=folder_uid,
                        track_uid=uid,
                        frame_number=detection['frame'],
                        time=vidTime,
                        bounding_box=np.array2string(bounding_box),
                        location=location,
                        location_cov=location_cov,
                        nose_x=pose.get('nose', [-1, -1])[0],
                        nose_y=pose.get('nose', [-1, -1])[1],
                        left_eye_x=pose.get('leftEye', [-1, -1])[0],
                        left_eye_y=pose.get('leftEye', [-1, -1])[1],
                        right_eye_x=pose.get('rightEye', [-1, -1])[0],
                        right_eye_y=pose.get('rightEye', [-1, -1])[1],
                        left_ear_x=pose.get('leftEar', [-1, -1])[0],
                        left_ear_y=pose.get('leftEar', [-1, -1])[1],
                        right_ear_x=pose.get('rightEar', [-1, -1])[0],
                        right_ear_y=pose.get('rightEar', [-1, -1])[1],
                        left_shoulder_x=pose.get('leftShoulder', [-1, -1])[0],
                        left_shoulder_y=pose.get('leftShoulder', [-1, -1])[1],
                        right_shoulder_x=pose.get('rightShoulder', [-1, -1])[0],
                        right_shoulder_y=pose.get('rightShoulder', [-1, -1])[1],
                        left_elbow_x=pose.get('leftElbow', [-1, -1])[0],
                        left_elbow_y=pose.get('leftElbow', [-1, -1])[1],
                        right_elbow_x=pose.get('rightElbow', [-1, -1])[0],
                        right_elbow_y=pose.get('rightElbow', [-1, -1])[1],
                        left_wrist_x=pose.get('leftWrist', [-1, -1])[0],
                        left_wrist_y=pose.get('leftWrist', [-1, -1])[1],
                        right_wrist_x=pose.get('rightWrist', [-1, -1])[0],
                        right_wrist_y=pose.get('rightWrist', [-1, -1])[1],
                        left_hip_x=pose.get('leftHip', [-1, -1])[0],
                        left_hip_y=pose.get('leftHip', [-1, -1])[1],
                        right_hip_x=pose.get('rightHip', [-1, -1])[0],
                        right_hip_y=pose.get('rightHip', [-1, -1])[1],
                        left_knee_x=pose.get('leftKnee', [-1, -1])[0],
                        left_knee_y=pose.get('leftKnee', [-1, -1])[1],
                        right_knee_x=pose.get('rightKnee', [-1, -1])[0],
                        right_knee_y=pose.get('rightKnee', [-1, -1])[1],
                        left_ankle_x=pose.get('leftAnkle', [-1, -1])[0],
                        left_ankle_y=pose.get('leftAnkle', [-1, -1])[1],
                        right_ankle_x=pose.get('rightAnkle', [-1, -1])[0],
                        right_ankle_y=pose.get('rightAnkle', [-1, -1])[1],
                    )
                )

        session.add_all(tracks)
        session.add_all(detections)
        session.commit()

        # tag any tracks as exemplar if they have gps ids in the source
        self._GpsIdsToExemplarTracks()

        # now we need to calculate the mustNotLinks
        # self._populateMNLs()

    @access.user
    @autoDescribeRoute(
        Description("Add Actor Chips from actor folder after uploaded")
        .modelParam(
            "id",
            description="base folderID of the MEVA Data",
            model=Folder,
            required=True,
            level=AccessType.WRITE,
        )
    )
    def add_actor_chips(self, folder):
        user = self.getCurrentUser()
        actor_data = load_actor_json(user, folder)

        self._add_actor_chips(actor_data)

        return {}

    def _add_actor_chips(self, actor_data):
        session = self.Session()
        # TODO: Only remove actor chips from database
        # We need to remove any existing folder and tracks that are set
        # session.query(Source).delete()
        actor_chips = []
        actor_display_chips = []
        tracks = []
        tracks_added = set()
        existing = session.query(Track).filter(Track.exemplar == 1).all()
        existing_tracks = []
        for item in existing:
            existing_tracks.append(item.exemplar_image1[0:20])

        updatedlist = {}
        for item in actor_data.keys():
            if item in existing_tracks:
                continue
            updatedlist[item] = actor_data[item]
        for item in updatedlist.values():
            date_time_str = str(item['date'])
            starttime = datetime(
                int(date_time_str[0:4]),
                int(date_time_str[4:6]),
                int(date_time_str[6:8]),
            )
            endtime = datetime(
                int(date_time_str[0:4]),
                int(date_time_str[4:6]),
                int(date_time_str[6:8]) + 1,
            )
            uid = uuid.uuid4().hex
            thumbnail_front = ''
            thumbnail_back = ''
            full_front = ''
            full_back = ''
            if 'b' in item:
                thumbnail_back = item['b']['thumb']
                full_back = item['b']['full']
            if 'f' in item:
                thumbnail_front = item['f']['thumb']
                full_front = item['f']['full']
            actor_display_chip_db = session.query(ActorDisplayChip).filter(
                ActorDisplayChip.filename == item['filename']
            ).first()
            if actor_display_chip_db is None:
                actor_display_chip_db = ActorDisplayChip(
                    uid=uid,
                    filename=item['filename'],
                    date=starttime,
                    actor_id=item['actor'],
                    gps_id=item['gps'],
                    outfit_id=item['outfit'],
                    thumbnail_back=thumbnail_back,
                    thumbnail_front=thumbnail_front,
                    full_back=full_back,
                    full_front=full_front,
                )
                actor_display_chips.append(actor_display_chip_db)
            else:
                uid = actor_display_chip_db.uid
            actor_chip_db = session.query(Source).filter(
                Source.girder_id == item['actor']
            ).first()
            if actor_chip_db is None:
                actor_chip_db = Source(
                    uid=uid,
                    start_time=starttime.timestamp(),
                    end_time=endtime.timestamp(),
                    girder_id=item['actor'],
                    name=item['filename'],
                    gps_tracker_id=(
                        int(item['gps']) if item['gps'].isnumeric() else None
                    ),
                    outfit_id=item['outfit'],
                )
                actor_chips.append(actor_chip_db)
            if ( (item['gps'] is not None) and
                 ( (item['gps'], item['outfit'], item['actor']) not in tracks_added) ):
                track_db = session.query(Track).filter(
                    Track.source_uid == uid
                ).first()
                if track_db is None:
                    track_db = Track(
                        uid=uuid.uuid4().hex,
                        source_uid=uid,
                        start_time=starttime.timestamp(),
                        end_time=endtime.timestamp(),
                        video_track_id=item['actor'],
                        exemplar=True,
                        gps_ids=(
                            int(item['gps']) if item['gps'].isnumeric() else None
                        ),
                        
                    )
                    tracks.append(track_db)
                tracks_added.add( (item['gps'], item['outfit'], item['actor']) )

        session.add_all(actor_chips)
        session.add_all(actor_display_chips)
        session.add_all(tracks)
        session.commit()

        self._GpsIdsToExemplarTracks()


    @access.user
    @autoDescribeRoute(
        Description("Get list of existing folders in system")
    )
    def get_source_folders(self):
        return self._get_source_folders()

    def _get_source_folders(self):
        session = self.Session()
        # We need to remove any existing folder and tracks that are set
        query = session.query(Source).filter(Source.outfit_id == None).all()
        folder_list = []
        for item in query:
            folder_list.append(item.uid)
        return folder_list

    def _get_initial_potential_links(self, session, base_folders, offset, pageSize):
        return {}

    def _get_source_track(self, session, sourceUID=None, folder_id=None, track_id=None):
        source_track = {}
        result = []
        base_folders = []
        return source_track, base_folders

    def _processed_filtered_tracks(self, session, filters, offset, pageSize):
        query = (
            session.query(Track)
            .join(Track.source_uid)
            .options(subqueryload(Track.folder))
        )
        if filters is not None:
            if 'date' in filters.keys():
                date_split = filters['date'].split('-')
                start_time = datetime(
                    int(date_split[0]), int(date_split[1]), int(date_split[2])
                )
                end_time = start_time + timedelta(days=1)
                if 'startTime' in filters.keys():
                    starttimestr = filters["startTime"].split('-')
                    start_time = start_time.replace(
                        hour=int(starttimestr[0]), minute=int(starttimestr[1])
                    )
                    end_time = start_time + timedelta(minutes=3)
                print(f"STARTTIME: {start_time} ENDTIME: {end_time}")
                query = query.filter(
                    and_(
                        Source.start_time >= start_time,
                        Source.start_time <= end_time,
                    )
                )
            if 'location' in filters.keys():
                query = query.filter(Source.location == filters['location'])
            if 'cameraId' in filters.keys():
                query = query.filter(Source.camera_id == filters['cameraId'])
        return paginate(query, offset + 1, pageSize)

    @access.user
    @autoDescribeRoute(
        Description("get a list of all linked tracks")
        .param(
            "offset",
            description="Offset for pagination",
            paramType="query",
            dataType="number",
            required=False,
            default=0,
        )
        .param(
            "pageSize",
            description="pageSize for pagination",
            paramType="query",
            dataType="number",
            required=False,
            default=20,
        )
        .param(
            "confirmed",
            "True=Only Confirmed; False=Unconfirmed Only; None=Both",
            dataType="boolean",
            required=False,
            default=None

        ).param(
            "names",
            "True=include filenames",
            dataType="boolean",
            required=False,
            default=False
        )
    )
    def get_linked_tracks(self, offset=0, pageSize=20, confirmed=None, names=False):

        responses = self._getLinksFromResponses(names, confirmed)
        return_response = responses[int(offset*pageSize):int((offset+1)*pageSize):]
        pages = math.ceil(len(responses) / pageSize)

        return {
            'pageResults': return_response,
            'totalPages': pages,
        }

    @access.user
    @autoDescribeRoute(
        Description("get a list linked tracks to specific chip")
        .param(
            "actorId",
            description="ActorId",
            paramType="query",
            dataType="string",
            required=True,
        )
        .param(
            "datestr",
            description="Date for the actor",
            paramType="query",
            dataType="string",
            required=True,
        )
        .param(
            "outfitId",
            description="OutfitId",
            paramType="query",
            dataType="string",
            required=True,
        )
        .jsonParam(
            "filters",
            "filter using the filter for meva data",
            requireObject=False,
            required=False,
        )
    )
    def get_actor_links(self, actorId, datestr, outfitId, filters={}):

        responses = self._getActorLinks(actorId, datestr, outfitId)
        return {
            'results': responses,
            'totalPages': 1,
        }


    @access.user
    @autoDescribeRoute(
        Description("get info about linked tracks")
    )
    def get_linking_info(self):
        return self._getLinkingInfo()

    @access.user
    @autoDescribeRoute(
        Description("Get tracks not associated with uid")
        .param(
            "sourceUID",
            description="track uid",
            paramType="query",
            dataType="string",
            required=True,
        )
        .param(
            "offset",
            "offset in returned results",
            paramType="query",
            dataType="integer",
            default=0,
        )
        .param(
            "pageSize",
            "number of results to return",
            paramType="query",
            dataType="integer",
            default=20,
        )
        .jsonParam(
            "filter",
            "array of UIDs to filter out",
            requireObject=False,
            required=False,
        )
    )
    def get_page(self, sourceUID, offset, pageSize, filterIds=[]):
        session = self.Session()
        source_track, base_folders = self._get_source_track(session, sourceUID)

        results = self._get_potential_links(
            session, filterIds, base_folders, offset, pageSize
        )

        total_pages = results.pages
        track_results = {}
        for row in results.items:
            track_results[row.uid] = []
            for track in row.tracks:
                folder_obj = Folder().load(
                    track.folder_id, user=self.getCurrentUser(), level=AccessType.READ
                )
                track_results[row.uid].append(
                    {
                        "folderId": track.folder_id,
                        "trackId": track.track_id,
                        "name": folder_obj['name'],
                    }
                )

        return {
            "pageResults": track_results,
            "sourceTrack": source_track,
            "totalPages": total_pages,
        }

    @access.admin
    @autoDescribeRoute(Description('Remove User Linking Cache')
        .param(
            "userId",
            "userId you want to remove from cache",
            paramType="query",
            dataType="string",
            required=True,
        )
    )
    def force_unload_user_cache(self, userId):
        settingsFile = '/data/db/settings.json'
        settings = getSettings(settingsFile)
        removeUserData(settings, userId)
        saveSettings(settings, settingsFile)
        return


    @access.admin
    @autoDescribeRoute(Description('Remove User Linking Cache')
        .param(
            "superKeyFrameMod",
            "modulus value for superkey frames",
            paramType="query",
            dataType="number",
            default=300,
        )
        .param(
            "suggestedLimit",
            "filter IDK by greater than this number",
            paramType="query",
            dataType="number",
            default=0,
        )
    )
    def set_linking_rules(self, superKeyFrameMod, suggestedLimit):
        settingsFile = '/data/db/settings.json'
        settings = getSettings(settingsFile)
        settings["superKeyFrameMod"] = superKeyFrameMod
        settings["suggestedLimit"] = suggestedLimit
        saveSettings(settings, settingsFile)
        return    

    @access.admin
    @autoDescribeRoute(Description('Get User Caches')
    )
    def get_user_caches(self):
        settingsFile = '/data/db/settings.json'
        settings = getSettings(settingsFile)
        user_cache = {}
        if 'activeUsers' in settings.keys():
            for user in settings['activeUsers']:
                if user in settings.keys():
                    cache_user = User().load(user, level=AccessType.ADMIN, user=self.getCurrentUser())
                    name = cache_user['login']
                    count = len(settings[user])
                    user_cache[str(user)] = { "name": name, "count": count}
        print(user_cache)
        return user_cache


    @access.user
    @autoDescribeRoute(Description("Get Random assortment of tracks and actors with a random source person")
        .param(
            "offset",
            "offset in returned results",
            paramType="query",
            dataType="integer",
            default=0,)
        .param(
            "pageSize",
            "number of results to return",
            paramType="query",
            dataType="integer",
            default=50,)
        .param(
            "runReID",
            "run ReID on the source chip",
            paramType="query",
            dataType="boolean",
            default=False,)
        .param(
            "getAllActorsForTheDay",
            "Only ReID with actors for the day",
            paramType="query",
            dataType="boolean",
            default=False,
        )
        .jsonParam(
            "data",
            "array of UIDs to filter out",
            requireObject=False,
            required=False,
        )
        .param(
            "reloadDBDataIntoMemory",
            "forces the reloading of the database items into memory",
            paramType="query",
            dataType="boolean",
            default=False,
        )
        .param(
            "forceUnloadUserData",
            "returns a user's checked out items to the full cache",
            paramType="query",
            dataType="boolean",
            default=False,
        )
    )
    def get_random(self, offset, pageSize, runReID=False,
                   getAllActorsForTheDay=False, data=None,
                   reloadDBDataIntoMemory=False,
                   forceUnloadUserData=False):
        session = self.Session()
        if data is not None and 'source' in data:
            source = data['source']
        else:
            source = None
        return self._get_random(session, offset, pageSize,
                                source, runReID, getAllActorsForTheDay,
                                reloadDBDataIntoMemory, forceUnloadUserData)

    def _get_random(self, session, offset, pageSize, source=None,
                    runReID=False, getAllActorsForTheDay=False,
                    reloadDBDataIntoMemory=False, forceUnloadUserData=False):
        settingsFile = '/data/db/settings.json'
        getJustSource = False
        userID = str(self.getCurrentUser()['_id'])

        if runReID and getAllActorsForTheDay:
            getJustSource = True  # this makes the list of results return as uids
        sourceID, results = getPotentialLinks(session=session,
                                              settingsFile=settingsFile,
                                              source=source,
                                              getJustSource=getJustSource,
                                              forceDBReload=reloadDBDataIntoMemory,
                                              userID=userID,
                                              forceUnloadUserData=forceUnloadUserData)
        if runReID:
            # get the image from this source
            img = self._get_video_image(sourceID)


            if getAllActorsForTheDay:
                # to run this with all checkins for the day, set
                # this to empty list
                results = []
            sourceID, results = sortByReID(session,
                                           sourceID,
                                           results,
                                           img)

        # pageinate at the end so we get all the results sent to ReID:
        totalPages, results = paginateList(results, offset, pageSize)
        print(results)

        return {
            "source": sourceID,
            "results": results,
            "totalPages": totalPages,
        }

    @access.user
    @autoDescribeRoute(Description("Get unique values for actor filters"))
    def get_actor_filter(self):
        session = self.Session()
        return self._get_actor_filter(session)

    def _get_actor_filter(self, session):
        query = session.query(ActorDisplayChip.date).distinct()
        dates = []
        for item in query:
            print(item)
            dates.append(item[0].strftime('%Y%m%d'))
        query2 = session.query(ActorDisplayChip.actor_id).distinct()
        actors = []
        for item in query2:
            print(item)
            actors.append(item[0])

        return {
            'dates': dates,
            'actors': actors,
        }

    @access.user
    @autoDescribeRoute(
        Description("Get actor chips")
        .param(
            "actorId",
            "ActorId for ",
            paramType="query",
            dataType="string",
        )
        .param(
            "datestr",
            "%YYYY%MM%DD date for the actor",
            paramType="query",
            dataType="string",
        )
        .param(
            "outfitId",
            "outfitId",
            paramType="query",
            dataType="string",
        )

    )
    def get_actor_chip(self, actorId, datestr, outfitId):
        session = self.Session()
        return self._get_actor_chip(session, actorId, datestr, outfitId)

    def _get_actor_chip(self, session, actorId, datestr, outfitId):
        query = session.query(ActorDisplayChip)
        date_time_str = datestr
        date_time = datetime(
            int(date_time_str[0:4]),
            int(date_time_str[4:6]),
            int(date_time_str[6:8]),
        )
        query = query.filter(
            and_(
                ActorDisplayChip.date == date_time,
                ActorDisplayChip.actor_id == actorId,
                ActorDisplayChip.outfit_id == outfitId
            )
        )
        results = query.all()
        if len(results) == 0:
            return []
        else:
            for item in results:
                return {
                    "uid": item.uid,
                    "actor": item.actor_id,
                    "gps": item.gps_id,
                    "date": item.date.strftime('%Y%m%d'),
                    "outfit": item.outfit_id,
                    "front": item.thumbnail_front,
                    "back": item.thumbnail_back,
                }

    @access.user
    @autoDescribeRoute(
        Description("Get actor chips")
        .param(
            "offset",
            "offset in returned results",
            paramType="query",
            dataType="integer",
            default=0,
        )
        .param(
            "pageSize",
            "number of results to return",
            paramType="query",
            dataType="integer",
            default=20,
        )
        .jsonParam(
            "filters",
            "json filter for actors to display",
            requireObject=False,
            required=False,
        )
    )
    def get_actor_chips(self, offset, pageSize, filters={}):
        session = self.Session()
        return self._get_actor_chips(session, offset, pageSize, filters)

    def _get_actor_chips(self, session, offset=0, pageSize=20, filters={}):
        query = session.query(ActorDisplayChip)
        if filters is not None:
            if 'date' in filters.keys():
                date_time_str = filters['date']
                date_time = datetime(
                    int(date_time_str[0:4]),
                    int(date_time_str[4:6]),
                    int(date_time_str[6:8]),
                )
                query = query.filter(
                    and_(
                        ActorDisplayChip.date == date_time,
                    )
                )
            if 'actor' in filters.keys():
                query = query.filter(ActorDisplayChip.actor_id == filters['actor'])

        results = paginate(query, offset + 1, pageSize)
        total_pages = results.pages
        pageResults = []
        for item in results.items:
            pageResults.append(
                {
                    "uid": item.uid,
                    "actor": item.actor_id,
                    "gps": item.gps_id,
                    "date": item.date.strftime('%Y%m%d'),
                    "outfit": item.outfit_id,
                    "front": item.thumbnail_front,
                    "back": item.thumbnail_back,
                }
            )
        return {
            "pageResults": pageResults,
            "totalPages": total_pages,
        }

    @access.user
    @autoDescribeRoute(
        Description("Get actor images")
        .jsonParam(
            "data",
            "json actorData for retrieving",
            requireObject=True,
            required=True,
        )
        .param(
            "imageIdentifiers",
            "if more than front and back, image names to include in return",
            paramType="query",
            dataType="boolean",
            default=False,
        )

    )
    def get_actor_images(self, data={}, imageIdentifiers=False):
        session = self.Session()
        return self._get_actor_images(session, data, imageIdentifiers)

    def _get_actor_images(self, session, data={}, imageIdentifiers=False):
        query = session.query(ActorDisplayChip)
        if data is not None:
            if 'date' in data.keys():
                date_time_str = data['date']
                date_time = datetime(
                    int(date_time_str[0:4]),
                    int(date_time_str[4:6]),
                    int(date_time_str[6:8]),
                )
                query = query.filter(
                    and_(
                        ActorDisplayChip.date == date_time,
                    )
                )
            if 'actorId' in data.keys():
                query = query.filter(ActorDisplayChip.actor_id == data['actorId'])
            if 'outfitId' in data.keys():
                query = query.filter(ActorDisplayChip.outfit_id == data['outfitId'])

        results = query.all()
        if len(results) == 0:
            return []
        else:
            for item in results:
                return_val = [
                    item.thumbnail_front
                ]
                if item.thumbnail_back != '':
                    return_val.append(item.thumbnail_back)
                return return_val

    @access.user
    @autoDescribeRoute(
        Description("Push Down the current Source to the bottom of the list so they don't come up until later")
        .param(
            "pageSize",
            "number of results to return",
            paramType="query",
            dataType="integer",
            default=200,
        )
        .param(
            "runReID",
            "run ReID on the source chip",
            paramType="query",
            dataType="boolean",
            default=False,)
        .param(
            "getAllActorsForTheDay",
            "Only ReID with actors for the day",
            paramType="query",
            dataType="boolean",
            default=False,)

        .jsonParam(
            "data",
            "source person element and list of other persons and their state",
            requireObject=True,
            paramType="body",
        )
    )
    def push_down(self, pageSize=200, runReID=False, getAllActorsForTheDay=False, data={}):
        session = self.Session()
        self._push_down(data)
        # So we need to takes the tracks that are linked to the positive UIDs and update the link_uid to the source
        # then we delete/remove all Link with that UID
        print(f"runReID: {runReID} getAllActorsForTheDay: {getAllActorsForTheDay}")
        return self._get_random(session, 0, pageSize, None, runReID, getAllActorsForTheDay)

    def _push_down(self, data):
        session = self.Session()
        if 'params' in data:
            params = data['params']
            if 'source' in params:
                source = self._get_track_id(params['source'])
            else:
                raise Exception(
                    'Requires a source person in the refinement data')

        settingsFile = '/data/db/settings.json'
        settings = getSettings(settingsFile)
        suggestedLimit = 1
        if 'suggestedLimit' in settings:
            suggestedLimit = settings["suggestedLimit"]
        suggestions = []
        if (source is not None):
            results = session.query(Suggested).filter(Suggested.track1_uid == source).all()
            if len(results) == 0:
                suggested = Suggested(
                    uid=uuid.uuid4().hex,
                    track1_uid=source,
                    suggestion_count=suggestedLimit + 1
                )
                suggestions.append(suggested)
                session.add_all(suggestions)
            else:
                for item in results:
                    item.suggestion_count += suggestedLimit + 1
        session.commit()

    @access.user
    @autoDescribeRoute(
        Description("Post list of tracks with their states and the source track to update the system")
        .param(
            "pageSize",
            "number of results to return",
            paramType="query",
            dataType="integer",
            default=50,
        )
        .jsonParam(
            "data",
            "source person element and list of other persons and their state",
            requireObject=True,
            paramType="body",
        )
    )
    def refine_actor_links(self, pageSize=200, data={}):
        self._refineActorLinks(data)
        params = data['params']
        actorId = params['source']['actorId']
        datestr = params['source']['date']
        outfitId = params['source']['outfitId']
        responses = self._getActorLinks(actorId, datestr, outfitId)
        return {
            'results': responses,
            'totalPages': 1,
        }

    @access.user
    @autoDescribeRoute(
        Description("Post list of tracks with their states and the source track to update the system")
        .param(
            "pageSize",
            "number of results to return",
            paramType="query",
            dataType="integer",
            default=50,
        )
        .param(
            "runReID",
            "run ReID on the source chip",
            paramType="query",
            dataType="boolean",
            default=False,)
        .param(
            "getAllActorsForTheDay",
            "Only ReID with actors for the day",
            paramType="query",
            dataType="boolean",
            default=False,)

        .jsonParam(
            "data",
            "source person element and list of other persons and their state",
            requireObject=True,
            paramType="body",
        )
    )
    def refine(self, pageSize=200, runReID=False, getAllActorsForTheDay=False, data={}):
        session = self.Session()
        self._refine(data)
        # So we need to takes the tracks that are linked to the positive UIDs and update the link_uid to the source
        # then we delete/remove all Link with that UID
        return self._get_random(session, 0, pageSize, None, runReID, getAllActorsForTheDay)

    def _refine(self, data):
        session = self.Session()
        if 'params' in data:
            params = data['params']
            if 'refinements' in params:
                refinements = [(self._get_track_id(ref), ref['state'])
                               for ref in params['refinements']]
            if 'source' in params:
                source = self._get_track_id(params['source'])
            else:
                raise Exception('Requires a source person in the refinement data')

        print('RefineMents')
        print(refinements)
        print('source')
        print(source)
        responses = []
        for track_uid, response in refinements:
            if (source is not None and track_uid is not None):
                saveTime = time.time()
                track_time = self._get_track_start_time(track_uid)
                # Need to check for a positive response and if that positive response is on an Actor, if so we need to
                # iterate through all non-comitted responses not in the list and mark them as negative
                resp_dict = {
                    'negative': 0.0,
                    'positive': 1.0,
                    'unknown': 0.5
                }
                response1 = Response(
                    track1_uid=source,
                    track2_uid=track_uid,
                    response=resp_dict.get(response, 0.5),
                    start_time=track_time,
                    saved_time=saveTime
                )
                response2 = Response(
                    track2_uid=source,
                    track1_uid=track_uid,
                    response=resp_dict.get(response, 0.5),
                    start_time=track_time,
                    saved_time=saveTime
                )
                responses.append(response1)
                responses.append(response2)

        session.add_all(responses)
        session.commit()

    def _get_track_id(self, data):
        session = self.Session()
        if data['type'] == 'actor':
            source_ids = session.query(Source.uid).filter(and_(
                Source.girder_id == data['actorId'],
                Source.outfit_id == data['outfitId'],
                Source.start_time == datetime.strptime(data['date'], '%Y%m%d').timestamp()
            ))
            for source_id in source_ids:
                track_id = session.query(Track.uid).filter(
                    Track.source_uid == source_id[0]
                ).first()
                if track_id is None:
                    continue
        elif data['type'] == 'track':
            track_id = session.query(Track.uid).filter(and_(
                Track.source_uid == data['folderId'],
                Track.video_track_id == data['trackId']
            )).first()
        else:
            track_id = None

        return track_id if track_id is None else track_id[0]

    def _get_track_start_time(self, track_uid):
        session = self.Session()
        track_time = -99999  # this way we know if it returns a bad number

        track = session.query(Track).filter(
            Track.uid == track_uid
        ).first()

        if track is not None:
            track_time = track.start_time

        return track_time

    def _update_links(self, session, sourceUID, linkIds):
        session.commit()

    def _get_potential_links(self, session, filterIds, base_folders, offset, pageSize):
        return {}

    @access.user
    @autoDescribeRoute(
        Description("Remove a Track from the linked Tracks")
        .param(
            "sourceUID",
            description="sourceUID",
            paramType="query",
            dataType="string",
            required=True,
        )
        .param(
            "folderId",
            description="folder id of video clip to start linking on",
            paramType="query",
            dataType="string",
            required=True,
        )
        .param(
            "trackId",
            description="Start Linking on this trackId",
            paramType="query",
            dataType="string",
            required=True,
        )
    )
    def remove_track_link(self, sourceUID, folderId, trackId):
        return

    @access.user
    @autoDescribeRoute(
        Description("Add GPS tracks from gps folder after uploaded")
        .modelParam(
            "id",
            description="base folderID of the GPS data",
            model=Folder,
            required=True,
            level=AccessType.WRITE,
        )
    )
    def add_gps_tracks(self, folder):
        user = self.getCurrentUser()
        # Add each GPX file to the list in the root folder
        gpsItems = list(Folder().childItems(
            folder,
            filters={"lowerName": {"$regex": gpxRegex}}
        ))
        for item in gpsItems:
            file = Item().childFiles(item)[0]
            gpx_file = (b"".join(list(File().download(file, headers=False)())).decode())
            self._add_gps(gpx_file)
        return

    def _add_gps(self, fileObj):
        session = self.Session()
        gpx = gpxpy.parse(fileObj)

        gps_points = []
        for track in gpx.tracks:
            name = track.name
            for seg in track.segments:
                for pt in seg.points:
                    gps_points.append(
                        Gps(
                            uid=uuid.uuid4().hex,
                            name=name,
                            time=pt.time.timestamp(),
                            latitude=pt.latitude,
                            longitude=pt.longitude,
                            elevation=pt.elevation,
                        )
                    )

        session.add_all(gps_points)
        session.commit()

    @access.admin
    @autoDescribeRoute(
        Description("Add camera Models to storage and the database")
        .modelParam(
            "id",
            description="base folderID of the camera model zip file",
            model=Folder,
            required=True,
            level=AccessType.WRITE,
        )
    )
    def add_camera_models(self, folder):
        user = self.getCurrentUser()
        # Add each GPX file to the list in the root folder
        zipItems = Folder().childItems(
            folder,
            filters={"lowerName": {"$regex": zipRegex}}
        )
        if zipItems.count() >= 1:
            zip_file_obj = Item().childFiles(zipItems.next())[0]
            stream = File().download(zip_file_obj, headers=False)
            bytes = b''.join(stream())
            zip_file = self.camera_model_dir / "zipped.zip"
            with open(zip_file, "wb") as binary_file:
                # Write bytes to file
                binary_file.write(bytes)
            with ZipFile(zip_file, 'r') as zipObj:
                listOfFileNames = zipObj.namelist()
                for fileName in listOfFileNames:
                    zipObj.extract(fileName, self.camera_model_dir)
            krtd_files = list(self.camera_model_dir.glob('**/*.krtd'))

            for cam_f in krtd_files:
                print(cam_f)
                print('Loading krtd file ', cam_f.name)
                with open(cam_f) as f:
                    self._add_krtd_camera(cam_f.name, f.read())
        return {}

    @access.admin
    @autoDescribeRoute(
        Description("Add actor descriptor Zip File")
        .modelParam(
            "id",
            description="base folderID of the actor descriptor zip file",
            model=Folder,
            required=True,
            level=AccessType.WRITE,
        )
    )
    def add_actor_descriptors(self, folder):
        user = self.getCurrentUser()
        zipItems = Folder().childItems(
            folder,
            filters={"lowerName": {"$regex": zipRegex}}
        )
        if zipItems.count() >= 1:
            zipFile = Item().childFiles(zipItems.next())[0]
            token = Token().createToken(user=user, days=14)
            extract_actor_descriptors.delay(
                folder,
                str(zipFile["_id"]),
                girder_job_title=("extract Actor Descriptors {}".format(str(folder["_id"]))),
                girder_client_token=str(token["_id"]),
            )

        return

    def _add_krtd_camera(self, filename, krtd_data):
        session = self.Session()

        krtd_arr = krtd_data.split()
        k_mat = np.fromstring(' '.join(krtd_arr[:9]), dtype=float, sep=' ').reshape(
            (3, 3)
        )
        r_mat = np.fromstring(' '.join(krtd_arr[9:18]), dtype=float, sep=' ').reshape(
            (3, 3)
        )
        t_vec = np.fromstring(' '.join(krtd_arr[18:21]), dtype=float, sep=' ')
        d_vec = np.fromstring(' '.join(krtd_arr[21:]), dtype=float, sep=' ')

        camera_db = Camera(
            uid=uuid.uuid4().hex,
            name=filename,
            k=k_mat.tostring(),
            r=r_mat.tostring(),
            t=t_vec.tostring(),
            d=d_vec.tostring(),
        )

        session.add(camera_db)
        session.commit()

    @access.user
    @autoDescribeRoute(
        Description("Get Input File Status")
    )
    def get_input_status(self):
        session = self.Session()
        gps_length = session.query(Gps).count()
        camera_length = session.query(Camera).count()
        actor_length = session.query(ActorDisplayChip).count()
        descriptors_exist = 0
        if os.path.isdir('/data/db/reid'):
            descriptors_exist = len(os.listdir('/data/db/reid'))
        data = []
        return {
            "cameraModels": camera_length,
            "actorGPS": gps_length,
            "actorChips": actor_length,
            "actorDescriptors": descriptors_exist
        }


    @access.user
    @autoDescribeRoute(
        Description("Get limited gps tracks for testing")
    )
    def get_gps_tracks(self):
        session = self.Session()
        query = session.query(Gps)
        results = paginate(query, 1, 2000).items
        data = []
        for item in results:
            data.append({
                'uid': item.uid,
                'name': item.name,
                'time': item.time,
                'latitude': item.latitude,
                'longitude': item.longitude,
                'elevation': item.elevation,
            })
        return data

    @access.admin
    @autoDescribeRoute(
        Description("Delete all GPS elements")
    )
    def delete_gps_tracks(self):
        session = self.Session()
        session.query(Gps).delete()
        session.commit()
        return True

    @access.user
    @autoDescribeRoute(
        Description("Get limited camera models for viewing")
    )
    def get_camera_models(self):
        session = self.Session()
        query = session.query(Camera)
        results = paginate(query, 1, 2000).items
        data = []
        for item in results:
            data.append({
                'uid': item.uid,
                'name': item.name,
                'k': item.k,
                'r': item.r,
                'd': item.d,
                't': item.t,
            })
        return data

    def _get_video_image(self, sourceUID):
        # sourceUID  - either the
        if sourceUID['type'] == 'track':
            folderId = sourceUID['folderId']
            trackId = sourceUID['trackId']
            folder = Folder().load(folderId, user=self.getCurrentUser(), level=AccessType.READ)
            file = chips_file(folder)
            if file is None:
                return []
            json_tracks = getChipData(file)
            if json_tracks[trackId] is not None:
                frameIds = []
                bounds = []
                highest_score = 0.0
                highest_chip = None
                highest_bounds = None
                highest_frame = None
                for item in json_tracks[trackId]:
                    if 'chipId' in item:
                        highest_score = max(item['score'], highest_score)
                        if highest_score == item['score']:
                            highest_chip = item['frameId']
                            highest_bounds = item['bounds']
                        frameIds.append(item['frameId'])
                        bounds.append(item['bounds'])
                if highest_chip:
                    file = File().load(highest_chip, user=self.getCurrentUser(), level=AccessType.READ)
                    downloaded = (b"".join(list(File().download(file, headers=False)())))
                    img = Image.open(
                        BytesIO(downloaded)).crop((highest_bounds[0], highest_bounds[1], highest_bounds[2], highest_bounds[3]))
                    # img.save(f"/data/{highest_chip}.jpeg") test for viewing
                    return img
        if sourceUID['type'] == 'actor':
            actorId = sourceUID['actorId']
            date = sourceUID['date']
            outfitId = sourceUID['outfitId']
            # Get front image or back if no front and return it
        return None

    @access.admin
    @autoDescribeRoute(
        Description("Delete all Camera Models from database")
    )
    def delete_camera_models(self):
        session = self.Session()
        session.query(Camera).delete()
        session.commit()
        base_dir = '/data/camera_models'
        data = os.listdir(base_dir)
        for item in data:
            data_path = f'{base_dir}/{item}'
            try:
                shutil.rmtree(data_path)
            except OSError:
                os.remove(data_path)

        return True

    @access.admin
    @autoDescribeRoute(
        Description("Delete all linking tables")
    )
    def delete_linking_tables(self):
        session = self.Session()
        session.query(Afinity).delete()
        session.query(MustNotLink).delete()
        session.query(Response).delete()
        session.commit()

        return True

    @access.admin
    @autoDescribeRoute(
        Description("Reset all Response tables")
    )
    def reset_response_tables(self):
        session = self.Session()
        session.query(Response).delete()
        session.commit()

    @access.admin
    @autoDescribeRoute(
        Description("Delete all video tracks and sources")
    )
    def delete_video_tracks(self):
        session = self.Session()
        session.query(Source).filter(Source.gps_tracker_id == None).delete()
        session.query(Track).filter(Track.gps_ids == None).delete()
        session.query(Track).filter(Track.gps_ids == "GNone").delete()
        session.commit()

        return True

    @access.admin
    @autoDescribeRoute(
        Description("Delete  actor Chips and GPS")
        .modelParam(
            "id",
            description="base folderID of the MevaDatasetFolder",
            model=Folder,
            required=True,
            level=AccessType.WRITE,
        )
        .param('month', 'What month to delete data from', default='all', required=True, enum=['all', '03', '05'])
    )
    def delete_actor_chips(self, folder, month):
        user = self.getCurrentUser()
        session = self.Session()
        print(f'Month: {month}')
        if month == 'all':
            session.query(Source).filter(Source.gps_tracker_id != None).delete()
            session.query(Track).filter(Track.gps_ids != None).delete()
            session.query(ActorDisplayChip).delete()
            session.commit()
        if month == '05':
            start_time = datetime(2018, 5, 1)
            end_time = datetime(2018, 5, 31)
            unix_start = time.mktime(start_time.timetuple())
            unix_end = time.mktime(end_time.timetuple())
            session.query(Source).filter(and_(
                Source.gps_tracker_id != None,
                Source.start_time > unix_start,
                Source.start_time < unix_end
            )).delete()
            session.query(Track).filter(and_(
                Track.gps_ids != None,
                Track.start_time > unix_start,
                Track.start_time < unix_end
            )).delete()
            session.query(ActorDisplayChip).filter(and_(
                ActorDisplayChip.date >= '2018-05-01',
                ActorDisplayChip.date <= '2018-05-31',
                )
            ).delete()
            session.commit()
            # Next we need to kick off a job to remove actor chips from the main chip list
            actorJsonId = Item().findWithPermissions({
                'meta.actorJSON': {'$in': [True, 'true', 'True']},
                'meta.baseMevaFolder': str(folder["_id"]),
                },
                user=user,
            )
            print(actorJsonId)
            if month != 'all':
                deletion_filter = f'2018{month}'
                print(f'STarting  remove Actor Descriptors with {deletion_filter}')
                token = Token().createToken(user=user, days=4)
                remove_actor_images.delay(
                    folder,
                    str(actorJsonId[0]["_id"]),
                    deletion_filter,
                    girder_job_title=(f'Delete Actor Chips for month: {month}'),
                    girder_client_token=str(token["_id"]),
                )


        return True


    @access.admin
    @autoDescribeRoute(
        Description("Delete actor descriptors")
    )
    def delete_actor_descriptors(self):
        shutil.rmtree('/data/db/reid')
        return True

    def _get_detection(self, uid):
        session = self.Session()

        return session.query(Detection).filter(Detection.uid == uid).first()

    @access.admin
    @autoDescribeRoute(
        Description("Populate_linked_tracks")
        .modelParam(
            "id",
            description="base folderID MEVA folder",
            model=Folder,
            required=True,
            level=AccessType.WRITE,
        )
        .param(
            "forceAdd",
            "Whether to force all tracks to be added again",
            paramType="formData",
            dataType="boolean",
            default=False,
            required=True,
        )
        .param(
            "excludeMay",
            "Whether to force all tracks to be added again",
            paramType="formData",
            dataType="boolean",
            default=False,
            required=True,
        )

    )
    def populate_linked_tracks(self, folder, forceAdd, excludeMay):
        user = self.getCurrentUser()
        token = Token().createToken(user=user, days=4)
        add_linking_tracks_to_database.delay(
            folder=folder,
            forceAdd=forceAdd,
            excludeMay=excludeMay,
            girder_client_token=str(token["_id"]),
            girder_job_title=(
                f"Adding All New Link Ready Folders from {folder['_id']} ",
            ),
        )

    @access.admin
    @autoDescribeRoute(
        Description("Populate Must not Link")
    )
    def populate_MNLs(self):

        user = self.getCurrentUser()
        token = Token().createToken(user=user, days=4)
        local_must_not_link_task.delay(
            girder_client_token=str(token["_id"]),
            girder_job_title=(
                "Running Must Not Link Task",
            ),
        )

    @access.admin
    @autoDescribeRoute(
        Description("Export DB")
        .modelParam(
            "id",
            description="base folderID MEVA folder",
            model=Folder,
            required=True,
            level=AccessType.WRITE,
        )
    )
    def export_database(self, folder):

        user = self.getCurrentUser()
        db_folder = Folder().createFolder(
            folder,
            DBFolderName,
            description="Actor folder",
            parentType="folder",
            public=False,
            creator=user,
            reuseExisting=True,
        )
        zipItems = list(
            Folder().childItems(
                db_folder,
                filters={"lowerName": {"$regex": zipRegex}},
                sort=[("created", pymongo.DESCENDING)],
            )
        )
        for item in zipItems:
            Item().remove(item)
        token = Token().createToken(user=user, days=4)
        export_DB.delay(
            folder=db_folder,
            girder_client_token=str(token["_id"]),
            girder_job_title=(
            "Exporting Database",
            ),
        )

    @access.admin
    @autoDescribeRoute(
        Description('Export Linked Track Images')
    )
    def export_linked_track_images(self):
        second_query = {
            f'meta.{LinkedTracksExportFolder}': {'$in': TRUTHY_META_VALUES},
        }
        export_root_folder = list(
            Folder().findWithPermissions(second_query, user=self.getCurrentUser())
        )
        if len(export_root_folder) == 0:
            raise RestException(f"Didn't find a root folder with tag '{LinkedTracksExportFolder}'")
        user = self.getCurrentUser()
        token = Token().createToken(user=user, days=4)
        export_linked_track_images_task.delay(
            folder=str(export_root_folder[0]["_id"]),
            girder_client_token=str(token["_id"]),
            girder_job_title=(
            "Exporting Linked Chips",
            ),
        )

    @access.admin
    @autoDescribeRoute(
        Description("Export Linked Tracks")
        .modelParam(
            "id",
            description="base folderID MEVA folder",
            model=Folder,
            required=True,
            level=AccessType.WRITE,
        )
        .param(
            "confirmed",
            description="confirmed Links",
            paramType="query",
            dataType="boolean",
            default=False
        )
    )
    def export_linked_tracks(self, folder, confirmed):
        user = self.getCurrentUser()
        second_query = {
            f'meta.{FinalTrackDownloadFolder}': {'$in': TRUTHY_META_VALUES},
        }
        export_root_folder = list(
            Folder().findWithPermissions(second_query, user=self.getCurrentUser())
        )
        if len(export_root_folder) > 0:
            export_base_folder = export_root_folder[0]
        else:
            return
        export_folder = Folder().createFolder(
            export_base_folder,
            'Track Download Folder',
            description="Track Download Folder",
            parentType="folder",
            public=True,
            creator=user,
            reuseExisting=True,
        )
        existing = Item().findOne({"meta.exportTrackLinks": True})
        if existing is not None:
            Item().remove(existing)
        # Now we get the linking data
        responses = self._getLinksFromResponses(useFileName=True, confirmed=confirmed)
        # Now we Add the file to the base folder as a linked_tracks.json
        newResultItem = Item().createItem('trackLinks.json', user, export_folder, reuseExisting=True)
        Item().setMetadata(newResultItem, {"exportTrackLinks": True}, allowNull=True)

        json_bytes = json.dumps(responses).encode()
        byteIO = io.BytesIO(json_bytes)
        newfile = Upload().uploadFromFile(
            byteIO,
            len(json_bytes),
            'trackLinks.json',
            parentType="item",
            parent=newResultItem,
            user=user,
            mimeType="application/json",
        )

        url = f'/api/v1/item/{newResultItem["_id"]}/download'
        return url

    @access.admin
    @autoDescribeRoute(
        Description("Get the latest download URL for the DB file")
        .modelParam(
            "id",
            description="base folderID MEVA folder",
            model=Folder,
            required=True,
            level=AccessType.WRITE,
        )
    )
    def get_database_download(self, folder):

        user = self.getCurrentUser()
        db_folder = Folder().createFolder(
            folder,
            DBFolderName,
            description="Actor folder",
            parentType="folder",
            public=False,
            creator=user,
            reuseExisting=True,
        )
        zipItems = list(
            Folder().childItems(
                db_folder,
                filters={"lowerName": {"$regex": zipRegex}},
                sort=[("created", pymongo.DESCENDING)],
            )
        )
        for item in zipItems:
            val = f'/api/v1/item/{item["_id"]}/download'
            return val

        return 'Need to export the Database file first'


    def _GpsIdsToExemplarTracks(self):
        session = self.Session()

        print('Loading all tracks...')
        tracks = session.query(Track).all()
        print('Loading all sources...')
        sources = session.query(Source).all()

        print('removing non exemplar tracks')
        eTracks = []
        for track in tracks:
            if track.exemplar:
                eTracks.append(track)
        print(f"There are {len(eTracks)} exemplar tracks")

        print('adding in gps ids')
        # loop though all the tracks and get the gpsid
        for track in eTracks:
            for src in sources:
                if track.source_uid == src.uid:
                    if src.gps_tracker_id is not None:
                        track.gps_ids = 'G' + str(src.gps_tracker_id)
                    track.exemplar_image1 = src.name
                    # print(src.gps_tracker_id)
        session.commit()

    def _refineActorLinks(self, data):
        session = self.Session()
        if 'params' in data:
            params = data['params']
            if 'refinements' in params:
                refinements = [(self._get_track_id(ref), ref['state'])
                               for ref in params['refinements']]
            if 'source' in params:
                source = self._get_track_id(params['source'])
            else:
                raise Exception('Requires a source person in the refinement data')

        responses = []
        for track_uid, response in refinements:
            if (source is not None and track_uid is not None):
                # Need to check for a positive response and if that positive response is on an Actor, if so we need to
                # iterate through all non-comitted responses not in the list and mark them as negative
                resp_dict = {
                    'negative': 0.0,
                    'positive': 2.0,
                    'unknown': 0.5
                }
                if resp_dict.get(response, 0.0) == 2.0:
                    query = session.query(Response).filter(or_(
                        and_(Response.track1_uid == source, Response.track2_uid == track_uid),
                        and_(Response.track1_uid == track_uid, Response.track2_uid == source))
                    )
                    for item in query:
                        item.response = 2.0
                        item.saved_time = time.time()
                else: # negative response we remove it
                    query = session.query(Response).filter(or_(
                        and_(Response.track1_uid == source, Response.track2_uid == track_uid),
                        and_(Response.track1_uid == track_uid, Response.track2_uid == source))
                    )
                    for item in query:
                        item.response = 0.0
                        item.saved_time = time.time()
        session.commit()

    def _getActorLinks(self, actorId, datestr, outfitId):
        session = self.Session()
        unix_date = time.mktime(datetime.strptime(datestr, "%Y%m%d").timetuple())
        source_actors = session.query(Source).filter(Source.girder_id == actorId, Source.outfit_id == outfitId, Source.start_time == unix_date)
        tracks = []
        # now we need to get all postive matches for this
        source_track = None
        for source_actor in source_actors:
            if source_actor:
                source_track = session.query(Track).filter(Track.source_uid == source_actor.uid).first()
                if source_track is None:
                    continue
        if source_actor is not None and source_track is not None:
            responses = session.query(Response).filter(Response.track1_uid == source_track.uid, Response.response >= 1).all()
            # now we need to convert the responses back to video trackIds
            response_uids = []
            response_vals = {}
            for item in responses:
                response_uids.append(item.track2_uid)
                response_vals[item.track2_uid] = (item.response)
            # now lets get track info for these items
            track_list = session.query(Track).filter(Track.uid.in_(response_uids)).all()
            for item in track_list:
                if item.exemplar != True:
                    link = {}
                    link['folderId'] = item.source_uid
                    link['trackId'] = item.video_track_id
                    link['type'] = 'track'
                    link['response'] = response_vals[item.uid]
                    tracks.append(link)
        return tracks

    def _getLinksFromResponses(self, useFileName=False, confirmed=None):
        session = self.Session()
        # grab all tracks
        print('Loading all tracks...')
        allTracks = session.query(Track).all()
        print('Loading responses...')
        responses = session.query(Response).all()
        print('Loading sources...')
        sources = session.query(Source).all()

        print('Sorting source by uid...')
        sourceByUid = {}
        for src in sources:
            sourceByUid[src.uid] = src

        print('Loading exemplars...')
        exemplars = {}
        tracks = {}
        for track in allTracks:
            # make the track easily referenced by uid
            tracks[track.uid] = track
            if track.exemplar:
                exemplars[track.uid] = track

        # sort the responses
        print('Sorting responses...')
        yeses = []
        nos = []
        idks = []
        check_response_level = 1
        if confirmed is True:
            check_response_level = 2
        for resp in responses:
            if confirmed is False:
                if resp.response >= check_response_level and resp.response < 2:
                    yeses.append(resp)
            elif resp.response >= check_response_level:
                yeses.append(resp)
            elif resp.response == 0:
                nos.append(resp)
            else:
                idks.append(resp)

        print(f"Yes responses: {len(yeses)}")
        print(f"No responses: {len(nos)}")
        print(f"IDK responses: {len(idks)}")

        # loop through the yeses and get the details
        positiveLinks = {}
        cnt = 0
        actors = {}
        for yes in yeses:
            src = {}
            link = {}
            t1 = yes.track1_uid
            t2 = yes.track2_uid
            src1 = sourceByUid[tracks[t1].source_uid]
            src2 = sourceByUid[tracks[t2].source_uid]
            if t1 in exemplars:
                # t1 is actor
                src['type'] = 'actor'
                src['actorId'] = src1.girder_id
                src['date'] = src1.name.split('-')[1]
                src['outfitId'] = src1.outfit_id
                # next must be video
                link['folderId'] = src2.uid
                link['trackId'] = tracks[t2].video_track_id
                link['type'] = 'track'
                if useFileName is True:
                    link['name']  = src2.name
            elif t2 in exemplars:
                # t2 is actor
                src['type'] = 'actor'
                src['actorId'] = src2.girder_id
                src['date'] = src2.name.split('-')[1]
                src['outfitId'] = src2.outfit_id
                # next must be video
                link['folderId'] = src1.uid
                link['trackId'] = tracks[t1].video_track_id
                link['type'] = 'track'
                if useFileName is True:
                    link['name']  = src1.name

            else:
                # neither are actors
                src['folderId'] = src1.uid
                src['trackId'] = tracks[t1].video_track_id
                src['type'] = 'track'
                if useFileName is True:
                    src['name']  = src1.name
                link['folderId'] = src2.uid
                link['trackId'] = tracks[t2].video_track_id
                link['type'] = 'track'
                if useFileName is True:
                    link['name']  = src2.name

            positiveLinks[cnt] = {}
            positiveLinks[cnt]['source'] = src
            if src['type'] == 'actor':
                identity = f"{src['actorId']}-{src['date']}-{src['outfitId']}"
                if identity not in actors:
                    actors[identity] = {
                        "actorId": src['actorId'],
                        "date": src['date'],
                        "outfitId": src['outfitId'],
                        "type": 'actor',
                        "links": []
                    }
                if link not in actors[identity]['links']:
                    actors[identity]['links'].append(link)

            positiveLinks[cnt]['link'] = link
            cnt += 1

        return list(actors.values())

    def _getLinkingInfo(self):
        session = self.Session()
        total_links = len(session.query(Response).filter(Response.response >= 1.0).all())/2
        links = self._getLinksFromResponses()
        confirmed_links = len(session.query(Response).filter(Response.response >= 2.0).all())/2
        I_Dont_Knows = len(session.query(Suggested).all())

        may_links = 0

        actor_ids = {}
        total_outfit_links = 0
        locations_per_outfit = 0
        for outfit in links:
            if outfit["actorId"] not in actor_ids.keys():
                actor_ids[outfit["actorId"]] = True
            total_outfit_links += len(outfit["links"])
            outfit_vid_map = {}
            for outfitlinks in outfit["links"]:
                if outfitlinks['folderId'] not in outfit_vid_map:
                    outfit_vid_map[outfitlinks['folderId']] = True
            locations_per_outfit += len(outfit_vid_map.keys())

        total_actors = len(actor_ids.keys())

        frame_overlap_tracks = session.query(Track, Response, Detection)\
            .where(and_(Track.gps_ids == None, Track.uid == Response.track1_uid, Response.response >= 1, Detection.track_uid == Response.track1_uid, Detection.frame_number % 300 == 0))\
            .group_by(Detection.track_uid)\
            .all()
        start_time = datetime(2018, 5, 1)
        end_time = datetime(2018, 5, 31)
        unix_start = time.mktime(start_time.timetuple())
        unix_end = time.mktime(end_time.timetuple())
        may_link_query = session.query(Track, Response)\
            .where(and_(Track.gps_ids == None, Track.uid == Response.track1_uid, Response.response >= 1, Track.start_time > unix_start, Track.start_time < unix_end)).group_by(Track.uid).all()
        may_links = len(may_link_query)
        may_confirmed_query = session.query(Track, Response)\
            .where(and_(Track.gps_ids == None, Track.uid == Response.track1_uid, Response.response >= 2.0, Track.start_time > unix_start, Track.start_time < unix_end)).group_by(Track.uid).all()
        may_confirmed_links = len(may_confirmed_query)
        if len(links) == 0 or total_actors == 0:
            return {
                "Outfits Linked": len(links),
                "Total Links (Including Track to Track)": total_links,
                "Total Actors by id Linked": total_actors,
                "Total Outfit links": total_outfit_links,
                "AVG Tracks Per Outfit": 0,
                "AVG Locations Per Outfit": 0,
                "AVG Locations Per PersonId": 0,
                "Linked Tracks on Super Key Frames": (len(frame_overlap_tracks)),
                "I Don't Knows": I_Dont_Knows,
                "Confirmed Links": confirmed_links,
                "May Links": may_links,
                "May Confirmed Links": may_confirmed_links,

            }
        data = {
            "Outfits Linked": len(links),
            "Total Links (Including Track to Track)": total_links,
            "Total Actors by id Linked": total_actors,
            "Total Outfit links": total_outfit_links,
            "AVG Tracks Per Outfit": total_outfit_links / len(links),
            "AVG Locations Per Outfit": locations_per_outfit / len(links),
            "AVG Locations Per PersonId": locations_per_outfit / total_actors,
            "Linked Tracks on Super Key Frames": (len(frame_overlap_tracks)),
            "I Don't Knows": I_Dont_Knows,
            "Confirmed Links": confirmed_links,
            "May Links": may_links,
            "May Confirmed Links": may_confirmed_links,


        }
        return data
