from sqlalchemy import (
    Boolean,
    Column,
    Date,
    DateTime,
    Float,
    ForeignKey,
    ForeignKeyConstraint,
    Integer,
    PrimaryKeyConstraint,
    String,
)
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship

Base = declarative_base()


class Detection(Base):
    __tablename__ = 'detections'
    uid = Column(String, primary_key=True)
    source_uid = Column(String, ForeignKey('sources.uid'))
    track_uid = Column(String)
    frame_number = Column(Integer)
    time = Column(Float)
    locked = Column(Boolean)
    bounding_box = Column(String)
    nose_x = Column(Integer)
    nose_y = Column(Integer)
    left_eye_x = Column(Integer)
    left_eye_y = Column(Integer)
    right_eye_x = Column(Integer)
    right_eye_y = Column(Integer)
    left_ear_x = Column(Integer)
    left_ear_y = Column(Integer)
    right_ear_x = Column(Integer)
    right_ear_y = Column(Integer)
    left_shoulder_x = Column(Integer)
    left_shoulder_y = Column(Integer)
    right_shoulder_x = Column(Integer)
    right_shoulder_y = Column(Integer)
    left_elbow_x = Column(Integer)
    left_elbow_y = Column(Integer)
    right_elbow_x = Column(Integer)
    right_elbow_y = Column(Integer)
    left_wrist_x = Column(Integer)
    left_wrist_y = Column(Integer)
    right_wrist_x = Column(Integer)
    right_wrist_y = Column(Integer)
    left_hip_x = Column(Integer)
    left_hip_y = Column(Integer)
    right_hip_x = Column(Integer)
    right_hip_y = Column(Integer)
    left_knee_x = Column(Integer)
    left_knee_y = Column(Integer)
    right_knee_x = Column(Integer)
    right_knee_y = Column(Integer)
    left_ankle_x = Column(Integer)
    left_ankle_y = Column(Integer)
    right_ankle_x = Column(Integer)
    right_ankle_y = Column(Integer)
    appearance_descriptor = Column(String)
    location = Column(String)
    location_cov = Column(String)
    source = relationship("Source", back_populates='detection')

    def bb_upper(self):
        return (self.bounding_box_ul_x, self.bounding_box_ul_y)

    def bb_lower(self):
        return (self.bounding_box_lr_x, self.bounding_box_lr_y)


class Source(Base):
    __tablename__ = 'sources'
    uid = Column(String, primary_key=True)
    girder_id = Column(String)
    name = Column(String)
    start_time = Column(Float)
    end_time = Column(Float)
    archive_location = Column(String)
    gps_tracker_id = Column(Integer)
    outfit_id = Column(String)
    track = relationship('Track', back_populates='source')
    detection = relationship('Detection', back_populates='source')


class Track(Base):
    __tablename__ = 'tracks'
    uid = Column(String, primary_key=True)
    video_track_id = Column(String)
    source_uid = Column(String, ForeignKey('sources.uid'))
    start_time = Column(Float)
    end_time = Column(Float)
    gps_ids = Column(String)
    detecion_uids = Column(String)
    exemplar = Column(Boolean)
    locked = Column(Boolean)
    exemplar_detection = Column(String)
    exemplar_image1 = Column(String)
    exemplar_image2 = Column(String)
    exemplar_image3 = Column(String)
    source = relationship('Source', back_populates='track')


class ActorDisplayChip(Base):
    __tablename__ = 'actorchips'
    uid = Column(String, primary_key=True)
    filename = Column(String)
    actor_id = Column(String)
    gps_id = Column(String)
    outfit_id = Column(String)
    date = Column(DateTime)
    thumbnail_front = Column(String)
    thumbnail_back = Column(String)
    full_front = Column(String)
    full_back = Column(String)

class Response(Base):
    __tablename__ = 'responses'
    track1_uid = Column(String, primary_key=True)
    track2_uid = Column(String, primary_key=True)
    response = Column(Float)
    start_time = Column(Float)
    saved_time = Column(Float)
    ForeignKeyConstraint(['track1_uid', 'track2_uid'], ['tracks.uid', 'tracks.uid'])


class Afinity(Base):
    __tablename__ = 'afinities'
    track1_uid = Column(String, primary_key=True)
    track2_uid = Column(String, primary_key=True)
    score = Column(Float)
    ForeignKeyConstraint(['track1_uid', 'track2_uid'], ['tracks.uid', 'tracks.uid'])


class MustNotLink(Base):
    __tablename__ = 'must_not_link'
    track1_uid = Column(String, primary_key=True)
    track2_uid = Column(String, primary_key=True)
    ForeignKeyConstraint(['track1_uid', 'track2_uid'], ['tracks.uid', 'tracks.uid'])


class Gps(Base):
    __tablename__ = 'gps'
    uid = Column(String, primary_key=True)
    track_uids = Column(String)
    name = Column(String)
    time = Column(Float)
    latitude = Column(Float)
    longitude = Column(Float)
    elevation = Column(Float)


class Camera(Base):
    __tablename__ = 'cameras'
    uid = Column(String, primary_key=True)
    name = Column(String)
    k = Column(String)
    r = Column(String)
    t = Column(String)
    d = Column(String)


class Suggested(Base):
    __tablename__ = 'suggested'
    uid = Column(String, primary_key=True)
    track1_uid = Column(String)
    track2_uid = Column(String)
    suggestion_count = Column(Integer)
    t1_type = Column(String)
    t2_type = Column(String)
    json = Column(String)
