import datetime
import numpy as np
import pytz
import logging
import subprocess
import math
import sys
import pandas as pd
import time

from meva_server.meva_database.models import Source, Gps, Track, Afinity, \
    Response

# WGS84 constants
_a = 6378137
_f = 1/(298257223563/1000000000)
_e2 = _f*(2-_f)
_e2m = np.square(1-_f)
_e2a = abs(_e2)
_e4a = np.square(_e2)
epsilon = np.finfo(float).eps
_maxrad = 2 * _a / epsilon


def geolocate_detection(det, cm_xyz, pixels_err=20, meters_err=2):
    """Geolocate detection into east/north/up coordinates encoded by cm_xyz.

    :param det: Dictionary with key 'bounds' encoding the [left-x, top-y,
        right-x, bottom-y] coordinates of the bounding box in image
        coordinates.
    :type det: dict

    :param cm_xyz: Three-channel array with dimensions (height, width, 3) with
        height and width matching the resolution of the camera image. Each
        channel corresponds to the easting, northing, up coordinates (meters)
        respectively for the world surface visible by each pixel.
    :type cm_xyz:

    :param pixels_err: Estimate for the accuracy of 'cm_xyz' in
        units of pixels. For example, a value of 10 indicates that true
        localization associated with a particular pixel in the map is found
        within +/- 10 pixels of the sampling.
    :type pixels_err: float

    :param meters_err: Estimate for the global accuracy of 'cm_xyz' in meters.
    This value is used to increase the localization covariance to incorporate
    inperfections in the camera model.
    :type meters_err: float

    :return: Estimate for the mean (m) and covariance (m^2) of the normal
        distribution for the localization of the detection with the east,
        north, up coordinate system encoded by 'cm_xyz'.

    """
    lx, ty, rx, by = [int(v) for v in det['bounds']]
    lx -= pixels_err
    rx += pixels_err
    ty -= pixels_err
    by += pixels_err

    height, width = cm_xyz.shape[:2]
    l = max([0, lx])
    r = min([width, rx])
    t = max([0, ty])
    b = min([width, by])

    # We are after an estimation for where in the world the person in the
    # detection is standing. If we had an unobstructed view of their feet, and
    # the pose skeleton captures the feet locations properly, we could just
    # unproject the feet location by sampling the cm_xyz at that location.
    # However, we often do not have an unoccluded view of the person's feet.
    # Also, camera model inaccuracies coupled with sharp discontinuities in
    # depth can lead to innacuracies.

    # This operation draws localization estimates from the camera model for
    # every pixel in the bounding box. While generally an overestimation of
    # the localization uncertainty, it is a safe approach.
    xyz = np.array([cm_xyz[t:b, l:r, i].ravel() for i in range(3)])

    loc = np.mean(xyz, axis=1)
    cov = np.cov(xyz)

    cov = cov + np.identity(3) * meters_err ** 2

    return loc, cov


def getSource(session, source_uid) -> Source:
    return session.query(Source).filter(Source.uid == source_uid).first()


def cleanName(videoName: str) -> str:
    cln = ''
    vidSplit = videoName.split('.')
    if 'avi' in vidSplit[-1] or 'json' in vidSplit[-1]:
        vidSplit = vidSplit[:-1]
    if 'meva' in vidSplit[-1]:
        vidSplit = vidSplit[:-1]
    if 'r' in vidSplit[-1]:
        vidSplit = vidSplit[:-1]

    for spl in vidSplit:
        cln += spl + '.'

    cln = cln[:-1]

    return cln


def unixTimeFromVideoUTC(video_fname, localTime):
    date, tstart, tend, _, cam = video_fname.split('.')

    s = '-'.join([date, tstart])
    naive = datetime.datetime.strptime(s, "%Y-%m-%d-%H-%M-%S")
    local_dt = localTime.localize(naive, is_dst=None)
    utc_dt = local_dt.astimezone(pytz.utc)

    return utc_dt.timestamp()


def getFrameTime(src_name, localTime, frame_number, fps):
    frameTime = unixTimeFromVideoUTC(src_name, localTime)
    # add to the time based on the frame number
    frameTime += (float(frame_number)/fps)
    return frameTime


def getCameraName(fullName: str) -> str:
    cam = ''
    if '.' in fullName:
        camParts = fullName.split('.')
        for cp in camParts:
            if cp[0] == 'G':
                cam = cp
    return cam


def getCameraModelMap(fileName: str) -> dict:
    camera_model_map = {}
    with open(fileName) as f:
        for line in f.readlines():
            res = line.split(' ')
            camera_model_map[res[0]] = res[2]
    return camera_model_map


def getVecFromString(locStr: str, dtype=np.float64) -> np.array:
    prebb = locStr.replace('[', '').replace(']', '')
    return np.array(prebb.split(), dtype=dtype)


def getLocationFromString(locStr: str) -> np.array:
    """
    takes a Nx1 numpy.array2string str value and converts it
    into a numpy.array
    """
    preloc = locStr.replace('[', '').replace(']', '')
    return np.array(preloc.split(), dtype=np.float64)


def getCovarianceFromString(covStr: str) -> np.array:
    """
    takes a numpy.array2string str value and converts it
    into a numpy.array - expects to see 3x3 covarant matrix
    """
    precov = covStr.replace('[', '').replace(']', '')
    return np.array(precov.split(), dtype=np.float64).reshape((3, 3))


def getScoreForPositions(pos1, cov1, pos2, cov2):
    """
    takes 2 positions with their covariants and returns a
    probability of convergence/proximity

    Args:
        pos1 (numpy.array): 3x1 location
        cov1 (numpy.array): 3x3 covariance
        pos2 (numpy.array): 3x1 location
        cov2 (numpy.array): 3x3 covariance

    Returns:
        float: number from 0 to 1 that represents the probablity
           that these 2 positions converge or are in close proximity
    """
    dpos = pos1 - pos2
    sig_sum = cov1 + cov2
    return np.exp(-np.dot(dpos.T, np.dot(np.linalg.inv(sig_sum), dpos))/2)


def turnOffSqlAlchemyLogging():
    """
    disable the SQL logging to the console
    """
    # turn off sqlalchemy logs
    for name in logging.root.manager.loggerDict:
        if 'sqlalchemy' in name:
            logger = logging.getLogger(name)
            logger.disabled = True


def getBoundingBoxFromString(bbStr: str) -> np.array:
    """
    takes a np.array2string and converts back to np.array
    note: for bb it expects 4x1

    Args:
        bbStr (str): string for bounding box

    Returns:
        np.array: [description]
    """
    precov = bbStr.replace('[', '').replace(']', '')
    return precov.split()


def getLocationAndCovariantFromGPSLog(wp: Gps, meters_error,
                                      latitude, longitude, height):
    """
    Takes a gps location and that has been referenced from
    a central location and makes it more general to the ENU (east
    north up) world model

    Args:
        wp (Gps): gps waypoint
        meters_error ([type]): how far off the gps could be
        latitude ([type]): [description]
        longitude ([type]): [description]
        height ([type]): [description]

    Returns:
        [type]: [description]
    """
    east, north, up = llh_to_enu(wp.latitude, wp.longitude,
                                 wp.elevation, latitude,
                                 longitude, height,
                                 meters_error)
    loc = np.array([east, north, up])
    cov = np.identity(3)*meters_error**2
    # covariance below is elongated in elevation due to the gps
    # logger used not having good elevation results
    cov[2, 2] = 100 * 100  # extend the cylinder

    return loc, cov


def getLatLonHeightFromEastNorthUp(location: np.array,
                                   latitude, longitude, height):
    """
    Takes a location that has been referenced from east, north, up from
    a central location and makes it match latitude, longitude, and altitude

    Args:
        location: east north up
        latitude ([type]): [description]
        longitude ([type]): [description]
        height ([type]): [description]

    Returns:
        [type]: [description]
    """
    lat, lon, h = enu_to_llh(location[0], location[1],
                             location[2], latitude,
                             longitude, height)
    loc = np.array([lat, lon, h])
    return loc


def geocentric_rotation(sphi, cphi, slam, clam):
    """
    This rotation matrix is given by the following quaternion operations
    qrot(lam, [0,0,1]) * qrot(phi, [0,-1,0]) * [1,1,1,1]/2
    or
    qrot(pi/2 + lam, [0,0,1]) * qrot(-pi/2 + phi , [-1,0,0])
    where
    qrot(t,v) = [cos(t/2), sin(t/2)*v[1], sin(t/2)*v[2], sin(t/2)*v[3]]

    """
    M = np.zeros(9)
    # Local X axis (east) in geocentric coords
    M[0] = -slam
    M[3] = clam
    M[6] = 0
    # Local Y axis (north) in geocentric coords
    M[1] = -clam * sphi
    M[4] = -slam * sphi
    M[7] = cphi
    # Local Z axis (up) in geocentric coords
    M[2] = clam * cphi
    M[5] = slam * cphi
    M[8] = sphi
    return M


def sincosd(x):
    """
    * Evaluate the sine and cosine function with the argument in degrees
    *
    * @tparam T the type of the arguments.
    * @param[in] x in degrees.
    * @param[out] sinx sin(<i>x</i>).
    * @param[out] cosx cos(<i>x</i>).
    *
    * The results obey exactly the elementary properties of the trigonometric
    * functions, e.g., sin 9&deg; = cos 81&deg; = &minus; sin 123456789&deg;.
    * If x = &minus;0, then \e sinx = &minus;0; this is the only case where
    * &minus;0 is returned.

    """
    r = np.fmod(x, 360)
    q = int(np.floor(r / 90 + 0.5))
    r -= 90 * q
    # now abs(r) <= 45
    r *= np.pi / 180
    s = np.sin(r)
    c = np.cos(r)

    # python 2.7 on Windows 32-bit machines has problems dealingwith -0.0.
    if x == 0:
        s = x

    if np.uint8(q) & np.uint8(3) == np.uint(0):
        sinx = s
        cosx = c
    elif np.uint8(q) & np.uint8(3) == np.uint(1):
        sinx = c
        cosx = -s
    elif np.uint8(q) & np.uint8(3) == np.uint(2):
        sinx = -s
        cosx = -c
    else:
      sinx = -c
      cosx = s

    # Set sign of 0 results.  -0 only produced for sin(-0)
    if x:
        sinx += 0
        cosx += 0

    return sinx, cosx


def llh_to_enu(lat, lon, h, lat0, lon0, h0, in_degrees=True, pure_python=True):
    """
    Convert latitude, longitude, and height to east, north, up.

    East, north, and up are coordinates within a local level Cartesian
    coordinate system with origin at geodetic latitude lat0, longitude lon0,
    and height h0 above the WGS84 ellipsoid. Up is normal to the ellipsoid
    surface and north is in the direction of the true north.

    :param lat: Geodetic latitude of the point to convert.
    :type lat: float

    :param lon: Longitude of the point to convert.
    :type lon: float

    :param h: Height above WGS84 ellipsoid of the point to convert (meters).
    :type h: float

    :param lat0: Geodetic latitude of the local east/north/up coordinate
        system.
    :type lat: float

    :param lon0: Longitude of the local east/north/up coordinate system.
    :type lon: float

    :param h0: Height above the WGS84 ellipsoid of the local east/north/up
        coordinate system (meters).
    :type height: float

    :param in_degrees: Specify that all angles are in degrees.
    :type in_degrees: bool

    :param pure_python: Specify whether to use the pure-Python implementation
        (faster) or the reference implementation from GeographicLib's command
        line call to CartConvert.
    :type pure_python: bool

    :return: East, north, and up coordinates (meters) of the converted point.
    :rtype: list

    Example:
    lat = 35.906437
    lon = -79.056282
    h = 123
    lat0 = 35.905446
    lon0 = -79.060788
    h0 = 0

    """
    if not in_degrees:
        lat = lat*180/np.pi
        lon = lon*180/np.pi
        lat0 = lat0*180/np.pi
        lon0 = lon0*180/np.pi

    if pure_python:
        sphi, cphi = sincosd(lat0)
        slam, clam = sincosd(lon0)
        _r = geocentric_rotation(sphi, cphi, slam, clam)
        xc, yc, zc = llh_to_ecef(lat, lon, h, in_degrees=True)
        _x0, _y0, _z0 = llh_to_ecef(lat0, lon0, h0, in_degrees=True)
        xc -= _x0
        yc -= _y0
        zc -= _z0
        x = _r[0] * xc + _r[3] * yc + _r[6] * zc
        y = _r[1] * xc + _r[4] * yc + _r[7] * zc
        z = _r[2] * xc + _r[5] * yc + _r[8] * zc
        return [x, y, z]
    else:
        output = subprocess.check_output(['CartConvert', '-l',
                                          str(lat0), str(lon0),
                                          str(h0), '--input-string',
                                          ' '.join([str(lat), str(lon), str(h)])])
        return [float(s) for s in output.split('\n')[0].split(' ')]


def llh_to_ecef(lat, lon, h, in_degrees=True):
    """
    Ported from GeographicLib.

    :param lat: Geodetic latitude of the point to convert.
    :type lat: float

    :param lon: Longitude of the point to convert.
    :type lon: float

    :param h: Height above WGS84 ellipsoid of the point to convert (meters).
    :type h: float

    :param in_degrees: Specify that all angles are in degrees.
    :type in_degrees: bool

    :return: ECEF X,Y,Z coordinates (meters).
    :rtype: list

    """
    if not in_degrees:
        lat = lat*180/np.pi
        lon = lon*180/np.pi

    sphi, cphi = sincosd(lat)
    slam, clam = sincosd(lon)

    n = _a/np.sqrt(1-_e2*np.square(sphi))
    Z = (_e2m * n + h) * sphi
    X = (n + h) * cphi
    Y = X * slam
    X *= clam
    return [float(X), float(Y), float(Z)]


def enu_to_llh(east, north, up, lat0, lon0, h0, in_degrees=True,
               pure_python=True):
    """
    Convert east, north, and up to latitude, longitude, and altitude.

    East, north, and up are coordinates within a local level Cartesian
    coordinate system with origin at geodetic latitude lat0, longitude lon0,
    and height h0 above the WGS84 ellipsoid. Up is normal to the ellipsoid
    surface and north is in the direction of the true north.

    :param east: Easting coordinate (meters) of the point to convert.
    :type east: float

    :param north: Northing coordinate (meters) of the point to convert.
    :type north: float

    :param up: Up coordinate (meters) of the point to convert.
    :type up: float

    :param lat0: Geodetic latitude of the local east/north/up coordinate
        system.
    :type lat: float

    :param lon0: Longitude of the local east/north/up coordinate system.
    :type lon: float

    :param h0: Height above the WGS84 ellipsoid of the local east/north/up
        coordinate system (meters).
    :type height: float

    :param in_degrees: Specify that all angles are in degrees.
    :type in_degrees: bool

    :param pure_python: Specify whether to use the pure-Python implementation
        (faster) or the reference implementation from GeographicLib's command
        line call to CartConvert.
    :type pure_python: bool

    :return: Geodetic latitude, longitude, and height (meters) above the WGS84
        ellipsoid of the converted point.
    :rtype: list

    Example:
    lat = 35.906437
    lon = -79.056282
    h = 123
    lat0 = 35.905446
    lon0 = -79.060788
    h0 = 0

    """
    if not in_degrees:
        lat0 = lat0*180/np.pi
        lon0 = lon0*180/np.pi

    if pure_python:
        x, y, z = east, north, up
        sphi, cphi = sincosd(lat0)
        slam, clam = sincosd(lon0)
        _r = geocentric_rotation(sphi, cphi, slam, clam)
        _x0, _y0, _z0 = llh_to_ecef(lat0, lon0, h0, in_degrees=True)

        xc = _x0 + _r[0] * x + _r[1] * y + _r[2] * z,
        yc = _y0 + _r[3] * x + _r[4] * y + _r[5] * z,
        zc = _z0 + _r[6] * x + _r[7] * y + _r[8] * z
        lat, lon, h = ecef_to_llh(xc, yc, zc, in_degrees)
    else:
        output = subprocess.check_output(['CartConvert', '-r', '-l', str(lat0),
                                          str(lon0), str(h0), '--input-string',
                                          ' '.join([str(east), str(north),
                                                    str(up)])])

        lat, lon, h = [float(s) for s in output.split('\n')[0].split(' ')]

    if not in_degrees:
        lat = lat*180/np.pi
        lon = lon*180/np.pi

    return [lat, lon, h]


def ecef_to_llh(X, Y, Z, in_degrees=True):
    """
    Ported from GeographicLib.

    :param X: ECEF x coordinate (meters).
    :type X: float

    :param X: ECEF x coordinate (meters).
    :type X: float

    :param X: ECEF x coordinate (meters).
    :type X: float

    :param in_degrees: Specify that all angles are in degrees.
    :type in_degrees: bool

    :return: Geodetic latitude, longitude, and height (meters) above the WGS84
        ellipsoid of the converted point.
    :rtype: list

    Example:
    x = 5634247
    y = 2050698
    z = 2167698

    """
    R = np.hypot(X, Y)
    if R == 0:
        slam = 0
        clam = 1
    else:
        slam = Y / R
        clam = X / R

    h = np.hypot(R, Z)      # Distance to center of earth
    if (h > _maxrad):
        # We really far away (> 12 million light years) treat the earth as a
        # point and h, above, is an acceptable approximation to the height.
        # This avoids overflow, e.g., in the computation of disc below.  It's
        # possible that h has overflowed to inf but that's OK.
        #
        # Treat the case X, Y finite, but R overflows to +inf by scaling by 2.
        R = np.hypot(X/2, Y/2)

        if R == 0:
            slam = 0
            clam = 1
        else:
            slam = (Y/2) / R
            clam = (X/2) / R

        H = np.hypot(Z/2, R)
        sphi = (Z/2) / H
        cphi = R / H
    elif _e4a == 0:
        # Treat the spherical case.  Dealing with underflow in the general case
        # with _e2 = 0 is difficult.  Origin maps to N pole same as with
        # ellipsoid.
        if h == 0:
            H = np.hypot(1, R)
            sphi = 1 / H
        else:
            H = np.hypot(Z, R)
            sphi = Z / H

        cphi = R / H
        h -= _a
    else:
        # Treat prolate spheroids by swapping R and Z here and by switching
        # the arguments to phi = atan2(...) at the end.
        p = np.square(R / _a)
        q = _e2m * np.square(Z / _a)
        r = (p + q - _e4a) / 6
        if _f < 0:
            p, q = q, p

        if not (_e4a * q == 0 and r <= 0):
            # Avoid possible division by zero when r = 0 by multiplying
            # equations for s and t by r^3 and r, resp.
            S = _e4a * p * q / 4  # S = r^3 * s
            r2 = np.square(r)
            r3 = r * r2
            disc = S * (2 * r3 + S)
            u = r
            if (disc >= 0):
                T3 = S + r3
                # Pick the sign on the sqrt to maximize abs(T3).  This
                # minimizes loss of precision due to cancellation.  The result
                # is unchanged because of the way the T is used in definition
                # of u.
                if T3 < 0:  # T3 = (r * t)^3
                    T3 += -np.sqrt(disc)
                else:
                    T3 += np.sqrt(disc)

                # N.B. cbrt always returns the real root.  cbrt(-8) = -2.
                T = np.cbrt(T3)  # T = r * t
                # T can be zero but then r2 / T -> 0.
                if T != 0:
                    u += T + (r2 / T)

            else:
                # T is complex, but the way u is defined the result is real.
                ang = np.arctan2(np.sqrt(-disc), -(S + r3))
                # There are three possible cube roots.  We choose the root
                # which avoids cancellation.  Note that disc < 0 implies that
                # r < 0.
                u += 2 * r * np.cos(ang / 3)

            v = np.sqrt(np.square(u) + _e4a * q)  # guaranteed positive
            # Avoid loss of accuracy when u < 0.  Underflow doesn't occur in
            # e4 * q / (v - u) because u ~ e^4 when q is small and u < 0.
            if u < 0:  # u+v guaranteed positive
                uv = _e4a * q / (v - u)
            else:
                uv = u + v

            # Need to guard against w going negative due to roundoff in uv - q.
            w = np.maximum(0, _e2a * (uv - q) / (2 * v))
            # Rearrange expression for k to avoid loss of accuracy due to
            # subtraction.  Division by 0 not possible because uv > 0, w >= 0.
            k = uv / (np.sqrt(uv + np.square(w)) + w)
            if _f >= 0:
                k1 = k
                k2 = k + _e2
            else:
                k1 = k - _e2
                k2 = k

            d = k1 * R / k2
            H = np.hypot(Z/k1, R/k2)
            sphi = (Z/k1) / H
            cphi = (R/k2) / H
            h = (1 - _e2m/k1) * np.hypot(d, Z)

        else:    # e4 * q == 0 && r <= 0
            # This leads to k = 0 (oblate, equatorial plane) and k + e^2 = 0
            # (prolate, rotation axis) and the generation of 0/0 in the general
            # formulas for phi and h.  using the general formula and division by 0
            # in formula for h.  So handle this case by taking the limits:
            # f > 0: z -> 0, k      ->   e2 * sqrt(q)/sqrt(e4 - p)
            # f < 0: R -> 0, k + e2 -> - e2 * sqrt(q)/sqrt(e4 - p)
            if _f >= 0:
                zz = np.sqrt((_e4a - p) / _e2m)
            else:
                zz = np.sqrt(p / _e2m)

            if _f < 0:
                xx = np.sqrt(_e4a - p)
            else:
                xx = np.sqrt(p)

            H = np.hypot(zz, xx)
            sphi = zz / H
            cphi = xx / H
            if Z < 0:
                sphi = -sphi  # for tiny negative Z (not for prolate)

            if _f >= 0:
                h = - _a * (_e2m) * H / _e2a
            else:
                h = - _a * (1) * H / _e2a

    lat = float(np.arctan2(sphi, cphi)*180/np.pi)
    lon = float(np.arctan2(slam, clam)*180/np.pi)
    h = float(h)
    return lat, lon, h


def getDFWithNoDuplicates(df1: pd.DataFrame, df2: pd.DataFrame,
                          strColumns: list) -> pd.DataFrame:
    df3 = pd.merge(df1, df2, on=strColumns,
                   how='outer',
                   indicator=True).query("_merge != 'both'")
    df3 = df3.drop(['_merge'], axis=1).reset_index(drop=True)
    return df3


def getDF1WithoutAnyInDF2(df1: pd.DataFrame, df2: pd.DataFrame,
                          strColumns: list) -> pd.DataFrame:
    df3 = pd.merge(df1, df2, on=strColumns,
                   how='left',
                   indicator=True).query("_merge != 'both'")
    df3 = df3.drop(['_merge'], axis=1).reset_index(drop=True)
    return df3


def isTimeOverlap(track1: Track, track2: Track, detects: dict,
                  gpsSorted, timeSyncError) -> bool:
    """
    Checks the times of 2 tracks to see if there is any temporal overlap

    Args:
        track1 (Track) and track2 (Track) to compare.
        gpsSorted: dict of Gps items sorted by uid
        timeSyncError: possible error between the gps and camera times

    Returns:
        bool: is there temporal overalp
    """

    # handle exemplars different
    t1srcStart = None
    t1srcEnd = None
    t2srcStart = None
    t2srcEnd = None
    if track1.exemplar is True and track2.exemplar is True:
        return True  # this will be MNL in the next check
    elif track1.exemplar is True:
        if track2.uid not in detects.keys() or len(detects[track2.uid]) < 2:
            return False
        if track1.gps_ids not in gpsSorted.keys():
            return False
        t1srcStart = gpsSorted[track1.gps_ids][1].time
        t1srcEnd = gpsSorted[track1.gps_ids][-1].time
        t2srcStart = track2.start_time
        t2srcEnd = track2.end_time
    elif track2.exemplar is True:
        if track1.uid not in detects.keys() or len(detects[track1.uid]) < 2:
            return False
        if track2.gps_ids not in gpsSorted.keys():
            return False
        t1srcStart = track1.start_time
        t1srcEnd = track1.end_time
        t2srcStart = gpsSorted[track2.gps_ids][1].time
        t2srcEnd = gpsSorted[track2.gps_ids][-1].time
    else:
        # case for 2 inter video
        if len(detects[track1.uid]) < 2 or len(detects[track2.uid]) < 2:
            return False
        if track1.uid not in detects.keys() or track2.uid not in detects.keys():
            return False
        t1srcStart = track1.start_time
        t1srcEnd = track1.end_time
        t2srcStart = track2.start_time
        t2srcEnd = track2.end_time

    # make up for time sync errors
    t1srcStart -= timeSyncError
    t1srcEnd += timeSyncError
    t2srcStart -= timeSyncError
    t2srcEnd += timeSyncError

    # if the t2 start is in between the t1 star and end
    if t1srcStart <= t2srcStart <= t1srcEnd:
        return True
    # if t2 end is in between the t1 start and end
    elif t1srcStart <= t2srcEnd <= t1srcEnd:
        return True
    # ---- checking both ways since the exemplar could be longer time period
    if t2srcStart <= t1srcStart <= t2srcEnd:
        return True
    elif t2srcStart <= t1srcEnd <= t2srcEnd:
        return True
    else:
        return False


def checkForMustNotLinks(track1, track2, track_detections, track_gpslogs,
                         timeSyncError, detectMinTime, detectMaxTime,
                         infinity, meters_err, scoreThreshold,
                         latitude, longitude, height):
    """
    Check two tracks to see if they should not be linked. Once it is determined
    the two tracks should not be linked the function returns.

    :return: (MustNotLink (True/False), Affinity score (0 to 1 float))
    """

    mustNotLink = False
    detections1 = []
    detections2 = []
    gpsFlag = 0  # this indicates if any of the detections are of type gps
    t1TimeUID = track1.uid
    t2TimeUID = track2.uid

    # Two exemplars should never be linked
    if (track1.exemplar is True and track2.exemplar is True):
        mustNotLink = True
        return (mustNotLink, 0.0)
    # see if one of the two is an examplar
    elif track1.exemplar is True:
        if track1.uid in track_gpslogs.keys():
            detections1 = track_gpslogs[track1.uid]
        else:
            # no gps match so we need to ignore this one
            return (False, 0.0)
        detections2 = track_detections[track2.uid]
        gpsFlag = 1  # the first item is a gps log
        t1TimeUID = detections1[0].name
    elif track2.exemplar is True:
        detections1 = track_detections[track1.uid]
        if track2.uid in track_gpslogs:
            detections2 = track_gpslogs[track2.uid]
        else:
            # no gps match so we need to ignore this one
            return (False, 0.0)
        gpsFlag = 2  # the second item is a gps log
        t2TimeUID = detections2[0].name
    else:  # two video tracks
        detections1 = track_detections[track1.uid]
        detections2 = track_detections[track2.uid]

    win_start_time = max(detectMaxTime[t1TimeUID],
                         detectMaxTime[t2TimeUID]) - timeSyncError
    win_end_time = min(detectMinTime[t1TimeUID],
                       detectMinTime[t2TimeUID]) + timeSyncError

    if win_start_time >= win_end_time:
        return (False, 0.0)

    detections1 = [x for x in detections1 if x.time >=
                   win_start_time and x.time <= win_end_time]
    detections2 = [x for x in detections2 if x.time >=
                   win_start_time and x.time <= win_end_time]

    # Loop over all detections for each track
    score = infinity
    for det1 in detections1:
        # we only have somethign to say about affinity if there is ever
        # a pair of detections that "overlap, meaning they share states that
        # are close enough in time
        innerScore = 0  # inner score is meant to capture the smallest value
        anyOverlap = False
        for det2 in detections2:

            # Check if tracks appear in the same frame of a video
            if gpsFlag == 0:  # don't run this check for gps tracks
                if (track1.source_uid == track2.source_uid):  # same video source
                    if (det1.frame_number == det2.frame_number):  # same frame num
                        # these can't be the same
                        mustNotLink = True
                        return (mustNotLink, 0.0)

            # check the times for a close period
            if abs(det1.time - det2.time) > timeSyncError:
                continue

            anyOverlap = True

            # Check localization ellipsoids
            new_score = 0
            loc1 = None
            loc2 = None

            if gpsFlag == 0:
                if det1.location is not None and det2.location is not None:
                    loc1 = getLocationFromString(det1.location)
                    cov1 = getCovarianceFromString(det1.location_cov)
                    loc2 = getLocationFromString(det2.location)
                    cov2 = getCovarianceFromString(det2.location_cov)
            elif gpsFlag == 1:
                if det2.location is not None:
                    loc2 = getLocationFromString(det2.location)
                    cov2 = getCovarianceFromString(det2.location_cov)

                loc1, cov1 = getLocationAndCovariantFromGPSLog(det1,
                                                               meters_err,
                                                               latitude,
                                                               longitude,
                                                               height)
            elif gpsFlag == 2:
                if det1.location is not None:
                    loc1 = getLocationFromString(det1.location)
                    cov1 = getCovarianceFromString(det1.location_cov)

                loc2, cov2 = getLocationAndCovariantFromGPSLog(det2,
                                                               meters_err,
                                                               latitude,
                                                               longitude,
                                                               height)
            if loc1 is not None and loc2 is not None:
                new_score = getScoreForPositions(loc1, cov1, loc2, cov2)

            # store the highest match
            innerScore = max(innerScore, new_score)

        if anyOverlap:
            score = min(innerScore, score)

    # check the affinity score to see if the MustNotLink is true
    # if the score is less than the threshold then we set mnl=true
    if score < scoreThreshold and score != infinity:
        mustNotLink = True

    # send the final score
    return (mustNotLink, score)
