import json
from girder import constants
from girder.models.file import File
from girder.models.token import Token
import pymongo
from meva_server.constants import ExportMarker, FinalTrackDownloadFolder, StatsFolderName, jsonRegex
from meva_server.serializers.models import Track
from re import M
import math

from girder.exceptions import RestException
from meva_server.utils import (
    detections_file,
    detections_item,
    get_filtered_folders,
    getTrackData,
)
from meva_tasks.tasks import calculate_track_stats
from meva_utils import FALSY_META_VALUES, TRUTHY_META_VALUES
from girder.api import access
from girder.api.describe import Description, autoDescribeRoute
from girder.api.rest import Resource
from girder.constants import AccessType
from girder.models.folder import Folder
from girder.models.item import Item
from girder.models.upload import Upload
from girder.models.user import User
from meva_server.meva_state.models import MEVA_State
from girder.utility import ziputil
from girder.api.rest import getApiUrl


class MEVAState(Resource):
    def __init__(self):
        super(MEVAState, self).__init__()
        self.resourceName = "meva_state"
        self.route("PUT", (":id",), self.create_mevastate)
        self.route("PUT", (":id", "data_import"), self.mevastate_import)
        self.route("GET", (":id", "folder"), self.get_mevastate)
        self.route("GET", (":id", "all_meva"), self.get_allmevastate)
        self.route("GET", (":id", "get_state_info"), self.get_state_info)
        self.route(
            "POST", (":id", "export_video_and_tracks"), self.export_video_and_tracks
        )
        self.route('POST', (":id", "export_final_tracks"), self.export_final_tracks)
        self.route("DELETE", (), self.remove_all)
        self.route("DELETE", ("delete_state_id", ), self.delete_state_id)
        self.route("POST", (":id",), self.populate_folder)
        self.route("POST", (), self.populate)
        self.route("POST", (":id", "recalculate_tracks"), self.recalculate_tracks)
        self.route("POST", (":id", "calculate_stats"), self.calculate_stats)
        self.route("GET", (":id", "get_stats"), self.get_stats)
        self.route("GET", (":id", "meva_filter"), self.get_mevafilter)
        self.route("POST", (":id", "meva_metadata"), self.set_meva_meta)
        self.route("POST", (":id", "meva_thumbnails"), self.set_meva_thumbnails)
        self.route("POST", (":id", "imported_tracks"), self.imported_tracks)
        self.route("POST", (":id", "cleaning_complete"), self.cleaning_complete)
        self.route("POST", (":id", "reviewing_complete"), self.reviewing_complete)
        self.route("POST", (":id", "linking_complete"), self.linking_complete)
        self.route("POST", (":id", "set_assignment"), self.set_assignment)
        self.route("POST", (":id", "set_state"), self.set_state)
        self.route("PUT", (":id", "set_mevadata"), self.set_mevadata)
        self.route("GET", (":id", "get_assigned"), self.get_assigned)

    @access.user
    @autoDescribeRoute(
        Description("Set MEVA state for folder to imported_tracks")
        .modelParam(
            "id",
            description="Folder to apply state to",
            model=Folder,
            level=AccessType.WRITE,
        )
        .jsonParam(
            "data",
            "JSON Dict of settings",
            paramType="body",
            required=True,
        )
    )
    def mevastate_import(self, folder, data):
        # Change state for cleaning tracks
        query = {"folderId": str(folder["_id"])}
        meva_state = MEVA_State().findOne(query)
        if meva_state is None:
            raise RestException(
                f'No Existing state found for Meva Folder: {folder["_id"]}'
            )

        user = self.getCurrentUser()
        MEVA_State().setKeys(str(folder["_id"]), data)
        return

    @access.admin
    @autoDescribeRoute(
        Description("Delete MEVA State for specific ID")
        .param(
            "folder",
            description="FolderId of the state to delete",
            paramType="query",
            required=True,
        )
    )
    def delete_state_id(self, folder):
        # Change state for cleaning tracks
        query = {"folderId": str(folder)}
        meva_state = MEVA_State().findOne(query)
        if meva_state is None:
            raise RestException(
                f'No Existing state found for Meva Folder: {folder}'
            )

        user = self.getCurrentUser()
        MEVA_State().removeWithQuery(query)
        return

    @access.user
    @autoDescribeRoute(
        Description("Import Meva State data from a JSON location").modelParam(
            "id",
            description="Folder to apply state to",
            model=Folder,
            level=AccessType.WRITE,
        )
    )
    def create_mevastate(self, folder):
        # Create initial Meva State based off of information
        user = self.getCurrentUser()
        MEVA_State().createState(folder, str(folder["parentId"]), user)
        return

    @access.user
    @autoDescribeRoute(
        Description("Get MEVA state for folder to imported_tracks")
        .responseClass('MEVA_State', array=True)
        .modelParam(
            "id",
            description="FolderId to get state from",
            model=Folder,
            level=AccessType.WRITE,
        )
        .pagingParams(defaultSort="folderId")
    )
    def get_mevastate(self, folder, limit, offset, sort):
        # Create initial Meva State based off of information
        user = self.getCurrentUser()
        query = {"folderId": str(folder["_id"])}
        return MEVA_State().find(query, offset=offset, limit=limit, sort=sort)

    @access.user
    @autoDescribeRoute(
        Description("Get All MEVAState")
        .modelParam(
            "id",
            description="Base root Folder to apply state to",
            model=Folder,
            level=AccessType.READ,
        )
        .jsonParam(
            "filters",
            "JSON Settings for the filtering",
            required=False,
        )
        .pagingParams(defaultSort='created')
    )
    def get_allmevastate(self, folder, filters, limit, offset, sort):
        user = self.getCurrentUser()
        query = self.get_filter_query(folder, user, filters)
        results = MEVA_State().find(
            query, offset=offset, limit=limit, sort=sort, user=self.getCurrentUser()
        )
        if results is not None:
            print(f"Count: {results.count()}")
            total = results.count()
            pages = math.ceil(results.count() / limit)
            structured_results = {
                'totalPages': pages,
                'pageResults': list(results),
            }
            return structured_results

        return {'totalPages': 0, 'results': []}

    @access.user
    @autoDescribeRoute(
        Description("Get All MEVAState").modelParam(
            "id",
            description="Get Base Info about the state",
            model=Folder,
            level=AccessType.READ,
        )
    )
    def get_state_info(self, folder):
        user = self.getCurrentUser()
        results = list(MEVA_State().find(user=self.getCurrentUser()))
        accumulator = {
            '2018-05': {
                'totalVideos': 0,
                'totalCleanedVideos': 0,
                'totalCleanedTracks': 0,
                'totalVideosInCleaning': 0,
                'totalVideosInReview': 0,
                'totalVideosAwaitingAssignment': 0,
                'emptyVideos': 0,
                "lowPriority": 0
            },
            '2018-03': {
                'totalVideos': 0,
                'totalCleanedVideos': 0,
                'totalCleanedTracks': 0,
                'totalVideosInCleaning': 0,
                'totalVideosInReview': 0,
                'totalVideosAwaitingAssignment': 0,
                'emptyVideos': 0,
                "lowPriority": 0

            },
        }
        order = [
            "Imported",
            "Cleaning",
            "Reviewing",
            "Linking",
            "Complete",
        ]
        if results is not None:
            for item in results:
                if 'metadata' not in item:
                    MEVA_State().remove(item)
                    continue
                if str(folder['_id']) != item['root']:
                    continue
                if 'date' not in item['metadata']:
                    print("no Date found in item")
                    print(item)
                    MEVA_State().remove(item)
                    continue
                indexer = ''
                if item['metadata']['date'].startswith('2018-05'):
                    indexer = '2018-05'
                if item['metadata']['date'].startswith('2018-03'):
                    indexer = '2018-03'
                if indexer == '':
                    print('Indexer is still empty')
                    print(item)
                accumulator[indexer]['totalVideos'] += 1
                if 'MEVA_State' in item:
                    meva_state = item['MEVA_State']
                    if "EmptyVideo" in meva_state.keys():
                        if meva_state["EmptyVideo"] is True:
                            accumulator[indexer]["emptyVideos"] += 1
                            continue
                    if "LowPriority" in meva_state.keys():
                        if meva_state["LowPriority"] is True:
                            accumulator[indexer]["lowPriority"] += 1
                            continue
                    state_counter = 0
                    for state in order:
                        if meva_state[state] is True:
                            state_counter += 1
                    if order[state_counter - 1] == "Linking":
                        accumulator[indexer]["totalCleanedVideos"] += 1
                        accumulator[indexer]["totalCleanedTracks"] += item["MEVA_Data"][
                            "endTracks"
                        ]
                    if order[state_counter] == "Reviewing":
                        accumulator[indexer]["totalVideosInReview"] += 1
                    if order[state_counter] == "Cleaning":
                        accumulator[indexer]["totalVideosInCleaning"] += 1
                if 'MEVA_Assignment' in item:
                    if item['MEVA_Assignment']['Cleaner'] == None:
                        accumulator[indexer]["totalVideosAwaitingAssignment"] += 1

        return accumulator

    def convert_meva_state_filter(self, state_filter, query_and):
        # So we need to take the order of Meva_State and create a full filter based on a single value
        # order is Imported -> Cleaning -> Reviewing -> Linking -> Complete
        order = [
            "Imported",
            "Cleaning",
            "Reviewing",
            "Linking",
            "Complete",
        ]
        base_state = {
            "Imported": False,
            "Cleaning": False,
            "Reviewing": False,
            "Linking": False,
            "Complete": False,
        }

        # We need to do a check to make sure the video isn't marked as empty
        if state_filter == "EmptyVideo":
            query_and.append({'MEVA_State.EmptyVideo': {'$in': TRUTHY_META_VALUES}})
        elif state_filter == "LowPriority":
            query_and.append({'MEVA_State.LowPriority': {'$in': TRUTHY_META_VALUES}})
        else:
            query_and.append({'MEVA_State.EmptyVideo': {'$exists': False}})
            index = order.index(state_filter)
            for state in range(5):
                values = TRUTHY_META_VALUES
                if state > index:
                    values = FALSY_META_VALUES
                query_and.append({f'MEVA_State.{order[state]}': {'$in': values}})

    def get_filter_query(self, folder, user, filters):
        query = {'root': str(folder['_id'])}
        if filters is not None:
            query = {'$and': [query]}
            if 'date' in filters.keys():
                query["$and"].append({'metadata.date': {'$in': [filters['date']]}})
            if 'location' in filters.keys():
                query["$and"].append(
                    {'metadata.location': {'$in': [filters['location']]}}
                )
            if 'cameraId' in filters.keys():
                query["$and"].append(
                    {'metadata.cameraId': {'$in': [filters['cameraId']]}}
                )
            if 'startTime' in filters.keys():
                query["$and"].append(
                    {'metadata.startTime': {'$regex': filters['startTime']}}
                )
            if 'state' in filters.keys():
                self.convert_meva_state_filter(filters["state"], query["$and"])
                # query["$and"].append(
                #    {f'MEVA_State.{filters["state"]}': {'$in': TRUTHY_META_VALUES}}
                # )
            else:
                query["$and"].append({'MEVA_State.EmptyVideo': {'$exists': False}})
                query["$and"].append({'MEVA_State.LowPriority': {'$exists': False}})
            if 'assigned' in filters.keys():
                query['$or'] = []
                if filters['assigned'] == 'myself':
                    assignedName = user['login']
                elif filters['assigned'] == 'unassigned':
                    assignedName = None
                else:
                    assignedName = filters['assigned']
                query["$or"].append(
                    {
                        'MEVA_Assignment.Cleaner.name': assignedName,
                        'MEVA_State.Reviewing': {'$in': FALSY_META_VALUES},
                    }
                )
                query["$or"].append(
                    {
                        'MEVA_Assignment.Reviewer.name': assignedName,
                        'MEVA_State.Reviewing': {'$in': TRUTHY_META_VALUES},
                    }
                )
        return query

    @access.admin
    @autoDescribeRoute(
        Description("Export Current Filter to public folder for testing")
        .modelParam(
            "id",
            description="Base root Folder to apply state to",
            model=Folder,
            level=AccessType.READ,
        )
        .jsonParam(
            "filters",
            "JSON Settings for the filtering",
            required=False,
        )
    )
    def export_final_tracks(self, folder,):
        user = self.getCurrentUser()
        filter = {"state": "Linking"}
        query = self.get_filter_query(folder, user, filter)
        results = list(MEVA_State().find(query=query, user=user))
        second_query = {
            f'meta.{FinalTrackDownloadFolder}': {'$in': TRUTHY_META_VALUES},
        }
        export_root_folder = list(
            Folder().findWithPermissions(second_query, user=self.getCurrentUser())
        )
        if len(export_root_folder) > 0:
            export_base_folder = export_root_folder[0]
        else:
            return
        export_folder = Folder().createFolder(
            export_base_folder,
            'Track Download Folder',
            description="Track Download Folder",
            parentType="folder",
            public=True,
            creator=user,
            reuseExisting=True,
        )

        child_items = Folder().childItems(export_folder)

        existing_items = {}

        if child_items.count() > 0:
            for item in child_items:
                filename = item["name"].replace('.avi', '.json')
                existing_items[filename] = True            
        print(existing_items)

        if results is not None:
            for item in results:
                filename = item['filename'].replace('.avi', '.json')
                if filename in existing_items.keys(): # skip existing items
                    continue;
                girder_folder_id = item['folderId']
                # grab detection file and move it to the new folder
                file_item = detections_item(
                    Folder().load(girder_folder_id, user=user), strict=True
                )
                child_files  = Item().childFiles(file_item)

                new_item = Item().createItem(
                    filename,
                    user,
                    export_folder,
                )
                file = File().copyFile(
                    child_files[0],
                    user,
                    new_item
                )
                file["name"] = filename
                File().save(file)
        url = f'/api/v1/folder/{export_folder["_id"]}/download'
        return url

    @access.admin
    @autoDescribeRoute(
        Description("Export Current Filter to public folder for testing")
        .modelParam(
            "id",
            description="Base root Folder to apply state to",
            model=Folder,
            level=AccessType.READ,
        )
        .jsonParam(
            "filters",
            "JSON Settings for the filtering",
            required=False,
        )
        .pagingParams(defaultSort='created')
    )
    def export_video_and_tracks(self, folder, filters, limit, offset, sort):
        user = self.getCurrentUser()
        query = self.get_filter_query(folder, user, filters)
        results = list(MEVA_State().find(query=query, user=user))
        second_query = {
            f'meta.{ExportMarker}': {'$in': TRUTHY_META_VALUES},
        }
        export_root_folder = list(
            Folder().findWithPermissions(second_query, user=self.getCurrentUser())
        )
        if len(export_root_folder) > 0:
            export_base_folder = export_root_folder[0]
        else:
            return
        export_folder = Folder().createFolder(
            export_base_folder,
            'Export Folder',
            description="Export folder",
            parentType="folder",
            public=True,
            creator=user,
            reuseExisting=True,
        )
        if export_folder is not None:
            Folder().remove(export_folder)
        export_folder = Folder().createFolder(
            export_base_folder,
            'Export Folder',
            description="Actor folder",
            parentType="folder",
            public=True,
            creator=user,
            reuseExisting=True,
        )
        batch_list = []
        if results is not None:
            for item in results:
                item_json = {}
                # create JSON file and move it to the public folder
                filename = item['filename']
                print(f"FILENAME: {filename}")
                girder_folder_id = item['folderId']
                item_json['filename'] = filename
                end_time = item['metadata']['endTime'].split('-')[0]
                if item["metadata"]["date"].startswith('201803'):
                    item_json[
                        'videoURL'
                    ] = f's3://mevadata-public-01/drops-123-r13/{item["metadata"]["date"]}/{end_time}/{filename}'
                else:
                    item_json[
                        'videoURL'
                    ] = f's3://meva-etl-input/may-re-id-release-candidates/{item["metadata"]["date"]}/{end_time}/{filename}'

                # grab detection file and move it to the new folder
                file_item = detections_item(
                    Folder().load(girder_folder_id, user=user), strict=True
                )
                upload_item = Item().copyItem(
                    file_item,
                    user,
                    filename.replace('.avi', '.json'),
                    export_folder,
                    filename,
                )
                first_file = next(Item().childFiles(upload_item), None)
                item_json[
                    'tracksURL'
                ] = f"{getApiUrl().replace('/api','/girder/api')}/girder/file/{str(first_file['_id'])}/download"
                batch_list.append(item_json)
                print(batch_list)
        return batch_list

    @access.admin
    @autoDescribeRoute(Description("Delete All MEVA State"))
    def remove_all(self):
        meva_states = MEVA_State().find()
        for state in meva_states:
            MEVA_State().remove(state)
        return

    @access.user
    @autoDescribeRoute(
        Description("Populate a single folder based on files").modelParam(
            "id",
            description="Base root Folder to apply state to",
            model=Folder,
            level=AccessType.READ,
        )
    )
    def populate_folder(self, folder):
        admin_user = self.getCurrentUser()
        meva_items = list(
            Folder().findWithPermissions(
                {
                    '$and': [
                        {'meta.MEVA': {'$in': [True, 'true', 'True']}},
                        {'parentId': str(folder['_id'])},
                    ]
                },
                user=self.getCurrentUser(),
            )
        )
        for item in meva_items:
            videoFolder = item
            user = User().load(str(videoFolder["creatorId"]), force=True)
            track_dict = getTrackData(detections_file(videoFolder))
            track_count = len(track_dict.keys())
            meva_state = MEVA_State().createState(
                videoFolder,
                str(videoFolder['parentId']),
                user,
                str(videoFolder['created']),
            )
            file_name = videoFolder['name']
            base_file = file_name.split('/')
            base_file = base_file[len(base_file) - 1]
            info_list = base_file.split('.')
            print(f'Attempting to add file: {file_name}')
            if len(info_list) >= 5:
                meva_data = {
                    "date": info_list[0],
                    "startTime": info_list[1][0:5],
                    "endTime": info_list[2][0:5],
                    "location": info_list[3],
                    "cameraId": info_list[4],
                }
                MEVA_State().setMetadata(str(videoFolder['_id']), meva_data)
            meta = videoFolder.get('meta', None)
            if meta is not None:
                if 'thumbnails' in meta:
                    MEVA_State().setMetadataThumbnails(
                        str(videoFolder['_id']), meta['thumbnails']
                    )
            if track_count > 0:
                MEVA_State().setMevaData(
                    str(videoFolder['_id']), 'beginTracks', track_count
                )
                MEVA_State().setState(str(videoFolder['_id']), 'Imported', True)
        return

    @access.admin
    @autoDescribeRoute(Description("Populate All States"))
    def populate(self):
        admin_user = self.getCurrentUser()
        meva_items = list(
            Folder().findWithPermissions(
                {'meta.MEVA': {'$in': [True, 'true', 'True']}},
                user=self.getCurrentUser(),
            )
        )
        for item in meva_items:
            videoFolder = item
            user = User().load(str(videoFolder["creatorId"]), force=True)
            track_dict = getTrackData(detections_file(videoFolder))
            track_count = len(track_dict.keys())
            meva_state = MEVA_State().createState(
                videoFolder,
                str(videoFolder['parentId']),
                user,
                str(videoFolder['created']),
            )
            file_name = videoFolder['name']
            base_file = file_name.split('/')
            base_file = base_file[len(base_file) - 1]
            info_list = base_file.split('.')
            print(f'Attempting to add file: {file_name}')
            if len(info_list) >= 5:
                meva_data = {
                    "date": info_list[0],
                    "startTime": info_list[1][0:5],
                    "endTime": info_list[2][0:5],
                    "location": info_list[3],
                    "cameraId": info_list[4],
                }
                MEVA_State().setMetadata(str(videoFolder['_id']), meva_data)
            meta = videoFolder.get('meta', None)
            if meta is not None:
                if 'thumbnails' in meta:
                    MEVA_State().setMetadataThumbnails(
                        str(videoFolder['_id']), meta['thumbnails']
                    )
            if track_count > 0:
                MEVA_State().setMevaData(
                    str(videoFolder['_id']), 'beginTracks', track_count
                )
                MEVA_State().setState(str(videoFolder['_id']), 'Imported', True)
        return

    @access.admin
    @autoDescribeRoute(
        Description(
            "Recalculate the final tracks based on the ignore persons specification"
        ).modelParam(
            "id",
            description="Base root Folder to apply state to",
            model=Folder,
            level=AccessType.READ,
        )
    )
    def recalculate_tracks(self, folder):
        user = self.getCurrentUser()
        query = {'root': str(folder['_id'])}
        results = MEVA_State().find(query, user=self.getCurrentUser())
        if results is not None:
            for item in results:
                folderId = item['folderId']
                file = detections_file(
                    Folder().load(folderId, user=user, level=AccessType.READ)
                )
                if file is None:
                    continue
                track_dict = getTrackData(file)
                track_count = 0
                for t in track_dict.values():
                    track = Track(**t)
                    if (
                        track.confidencePairs[0][0].startswith('person') is True
                        or track.confidencePairs[0][0].startswith('Person') is True
                    ):
                        track_count = track_count + 1
                MEVA_State().setMevaData(str(folderId), 'endTracks', track_count)

    @access.admin
    @autoDescribeRoute(
        Description(
            "Recalculate the final tracks based on the ignore persons specification"
        )
        .modelParam(
            "id",
            description="Base root Folder to apply state to",
            model=Folder,
            level=AccessType.READ,
        )
        .param(
            "month",
            description="User ID to assign role",
            paramType="query",
            required=True,
            default="2018-03",
        )
    )
    def calculate_stats(self, folder, month):
        user = self.getCurrentUser()
        token = Token().createToken(user=user, days=4)
        statsFolder = Folder().createFolder(
            folder,
            StatsFolderName,
            description="Actor folder",
            parentType="folder",
            public=False,
            creator=user,
            reuseExisting=True,
        )
        jsonItems = list(
            Folder().childItems(
                statsFolder,
                filters={"lowerName": {"$regex": jsonRegex}},
                sort=[("created", pymongo.DESCENDING)],
            )
        )
        for item in jsonItems:
            if item["name"] == f"{month}.json":
                Item().remove(item)

        calculate_track_stats.delay(
            folder=folder,
            uploadFolder=statsFolder,
            month=month,
            girder_client_token=str(token["_id"]),
            girder_job_title=(f"Calculating Stats for month {month} ",),
        )

    @access.admin
    @autoDescribeRoute(
        Description(
            "Recalculate the final tracks based on the ignore persons specification"
        ).modelParam(
            "id",
            description="Base root Folder to apply state to",
            model=Folder,
            level=AccessType.READ,
        )
    )
    def get_stats(self, folder):
        user = self.getCurrentUser()
        statsFolder = Folder().createFolder(
            folder,
            StatsFolderName,
            description="Actor folder",
            parentType="folder",
            public=False,
            creator=user,
            reuseExisting=True,
        )
        # Read each file as a JSON and load it into a dict
        stats = {}
        jsonItems = list(
            Folder().childItems(
                statsFolder,
                filters={"lowerName": {"$regex": jsonRegex}},
                sort=[("created", pymongo.DESCENDING)],
            )
        )
        for item in jsonItems:
            first_file = next(Item().childFiles(item), None)
            data = json.loads(
                b"".join(list(File().download(first_file, headers=False)())).decode()
            )
            stats[item["name"].replace('.json', '')] = data
        return stats

    @access.user
    @autoDescribeRoute(
        Description("Get a list of possibile filter values for MEVA State").modelParam(
            "id",
            description="Base folder ID",
            model=Folder,
            level=AccessType.READ,
            required=True,
        )
    )
    def get_mevafilter(self, folder):
        # Create initial Meva State based off of information
        user = self.getCurrentUser()
        query = {'root': str(folder['_id'])}
        meva_items = list(
            MEVA_State().find(
                query,
                user=self.getCurrentUser(),
            )
        )
        meva_items.sort(key=lambda d: d["created"], reverse=True)
        # Compile a list of dates/ times and start/end Pairs
        results = {
            "dates": set(),
            "times": set(),
            "locations": set(),
            "cameraIds": set(),
            "users": set(),
        }
        if len(meva_items) > 0:
            for folder in meva_items:
                meta = folder['metadata']
                assignments = folder['MEVA_Assignment']
                if assignments:
                    cleaner = assignments['Cleaner']
                    reviewer = assignments['Reviewer']
                    if cleaner:
                        results['users'].add(cleaner['name'])
                    if reviewer:
                        results['users'].add(reviewer['name'])
                if meta:
                    MEVA_Meta = meta
                    if "date" in MEVA_Meta:
                        results["dates"].add(MEVA_Meta["date"])
                    if "startTime" in MEVA_Meta:
                        start = MEVA_Meta["startTime"]
                        start_split = start.split('-')
                        start = f"{start_split[0]}-{start_split[1]}"
                    if "endTime" in MEVA_Meta:
                        end = MEVA_Meta["endTime"]
                        end_split = end.split('-')
                        end = f"{end_split[0]}-{end_split[1]}"
                    results["times"].add(f'{start} to {end}')
                    if "location" in MEVA_Meta:
                        results["locations"].add(MEVA_Meta["location"])
                    if "cameraId" in MEVA_Meta:
                        results["cameraIds"].add(MEVA_Meta["cameraId"])

        print(results)
        results["locations"] = sorted(results["locations"])
        results["cameraIds"] = sorted(results["cameraIds"])
        results["dates"] = sorted(results["dates"])
        results["times"] = sorted(results["times"])
        results["users"] = sorted(results["users"])
        return results

    @access.user
    @autoDescribeRoute(
        Description("Set Metadata for MEVA folder")
        .modelParam(
            "id",
            description="Folder to apply state to",
            model=Folder,
            level=AccessType.WRITE,
        )
        .jsonParam(
            "metadata",
            "JSON with the metadata to set",
            requireObject=True,
            paramType="body",
        )
    )
    def set_meva_meta(self, folder, metadata):
        # Create initial Meva State based off of information
        user = self.getCurrentUser()
        MEVA_State().setMetadata(folder["_id"], metadata)
        return

    @access.user
    @autoDescribeRoute(
        Description("Set thumbnails for MEVA folder")
        .modelParam(
            "id",
            description="Folder to apply state to",
            model=Folder,
            level=AccessType.WRITE,
        )
        .jsonParam(
            "thumbnails",
            "JSON with the thumbnails to set",
            requireObject=True,
            paramType="body",
        )
    )
    def set_meva_thumbnails(self, folder, thumbnails):
        # Create initial Meva State based off of information
        user = self.getCurrentUser()
        if 'thumbnails' in thumbnails:
            MEVA_State().setMetadataThumbnails(folder["_id"], thumbnails['thumbnails'])
        return

    @access.user
    @autoDescribeRoute(
        Description("Set MEVA state for folder to imported_tracks").modelParam(
            "id",
            description="Folder to apply state to",
            model=Folder,
            level=AccessType.WRITE,
        )
    )
    def imported_tracks(self, folder):
        # Change state for imported tracks
        MEVA_State().setState(folder["_id"], "Imported", True)
        return

    @access.user
    @autoDescribeRoute(
        Description("Set MEVA state for folder to cleaning complete").modelParam(
            "id",
            description="Folder containing the items to process",
            model=Folder,
            level=AccessType.WRITE,
        )
    )
    def cleaning_complete(self, folder):
        # Change state for cleaning tracks
        return

    @access.user
    @autoDescribeRoute(
        Description("Set MEVA state for folder to reviewing complete").modelParam(
            "id",
            description="Folder containing the items to process",
            model=Folder,
            level=AccessType.WRITE,
        )
    )
    def reviewing_complete(self, folder):
        # Change state for cleaning tracks
        return

    @access.user
    @autoDescribeRoute(
        Description("Set MEVA state for folder to linking complete").modelParam(
            "id",
            description="Folder containing the items to process",
            model=Folder,
            level=AccessType.WRITE,
        )
    )
    def linking_complete(self, folder):
        # Change state for cleaning tracks
        return

    @access.user
    @autoDescribeRoute(
        Description("Set MEVA annotator user")
        .modelParam(
            "id",
            description="folderId",
            model=Folder,
            level=AccessType.WRITE,
        )
        .param(
            "userId",
            description="User ID to assign role",
            paramType="query",
            required=True,
        )
        .param(
            "role",
            description="Role to assign to user",
            paramType="query",
            required=True,
        )
    )
    def set_assignment(self, folder, userId, role):
        # Change state for cleaning tracks
        query = {"folderId": str(folder["_id"])}
        meva_state = MEVA_State().findOne(query)
        user = self.getCurrentUser()
        if role == 'owner':
            if (
                user['admin'] is False
                and meva_state['MEVA_Assignment']['Cleaner'] is not None
                and user['_id'] != meva_state['owner']
            ):
                raise RestException(
                    'Only the owner or an admin can changed ownership role'
                )
        if userId == 'none':
            MEVA_State().setAssignment(str(folder["_id"]), role, None)
            return

        role_user = User().load(str(userId), force=True)
        if role_user is None:
            raise RestException(f'There is no user with the ID {userId}')
        MEVA_State().setAssignment(str(folder["_id"]), role, role_user)
        return

    @access.user
    @autoDescribeRoute(
        Description("Set MEVA annotator user")
        .modelParam(
            "id",
            description="folderId",
            model=Folder,
            level=AccessType.WRITE,
        )
        .param(
            "state",
            description="State to tranistion",
            paramType="query",
            required=True,
        )
        .param(
            "time",
            description="Time it took to set State",
            paramType="query",
            required=True,
        )
        .param(
            "val",
            description="True or False state val",
            paramType="query",
            dataType="boolean",
            default=True,
        )
    )
    def set_state(self, folder, state, time, val):
        # Change state for cleaning tracks
        query = {"folderId": str(folder["_id"])}
        meva_state = MEVA_State().findOne(query)
        user = self.getCurrentUser()
        MEVA_State().setState(str(folder["_id"]), state, val, time)
        return

    @access.user
    @autoDescribeRoute(
        Description("Set MEVA annotator user")
        .modelParam(
            "id",
            description="folderId",
            model=Folder,
            level=AccessType.WRITE,
        )
        .jsonParam(
            "data",
            "JSON Settings for the filtering",
            requireObject=True,
            paramType="body",
            required=True,
        )
    )
    def set_mevadata(self, folder, data):
        # Change state for cleaning tracks
        query = {"folderId": str(folder["_id"])}
        meva_state = MEVA_State().findOne(query)
        user = self.getCurrentUser()
        for key in data.keys():
            if key in meva_state['MEVA_Data']:
                MEVA_State().setMevaData(str(folder["_id"]), key, data[key])
        return

    @access.user
    @autoDescribeRoute(
        Description("Set MEVA reviewer user").modelParam(
            "id",
            description="User ID to set",
            model=User,
            level=AccessType.WRITE,
        )
    )
    def set_reviewer(self, user):
        # Change state for cleaning tracks
        return

    @access.user
    @autoDescribeRoute(
        Description("Set MEVA linker user").modelParam(
            "id",
            description="User ID to set",
            model=User,
            level=AccessType.WRITE,
        )
    )
    def set_linker(self, user):
        # Change state for cleaning tracks
        return

    @access.admin
    @autoDescribeRoute(
        Description("Set MEVA ownder user").modelParam(
            "id",
            description="User ID to set",
            model=User,
            level=AccessType.WRITE,
        )
    )
    def set_owner(self, user):
        # Change state for cleaning tracks
        return

    @access.user
    @autoDescribeRoute(
        Description("Get Assigned tasks for a user").modelParam(
            "id",
            description="User ID toquery",
            model=User,
            level=AccessType.WRITE,
        )
    )
    def get_assigned(self, user):
        # Change state for cleaning tracks
        return
