import datetime
from girder.constants import SortDir
from girder.exceptions import ValidationException, RestException
from girder.models.model_base import Model
from girder import events
from dateutil import parser


class MEVA_State(Model):
    # This is NOT an access controlled model; it is expected that all endpoints
    # will be sensibly guarded instead.
    def __init__(self):
        # Do the bindings before calling __init__(), in case a derived class
        # wants to change things in initialize()
        events.bind('model.folder.remove', 'removeImport', self._cleanupDeletedEntity)
        super().__init__()

    def _cleanupDeletedEntity(self, event):
        # remove data if the folderId matches
        entityDoc = event.info
        folderId = entityDoc['_id']
        meva_state = self.findOne({'folderId': str(folderId)})
        if meva_state is not None:
            self.remove(meva_state)

    def initialize(self):
        self.name = 'MEVA_STATE'
        self.ensureIndices(
            [
                'folderId',
                'owner',
                'owner_name' 'locked',
                'created',
                (
                    [
                        ('created', SortDir.ASCENDING),
                    ],
                    {},
                ),
            ]
        )
        self.MEVA_State = {
            "Imported": False,
            "Cleaning": False,
            "Reviewing": False,
            "Linking": False,
            "Complete": False,
        }
        self.MEVA_Assignment = {
            "Cleaner": None,
            "Reviewer": None,
            "Linker": None,
        }
        self.MEVA_Data = {
            'beginTracks': 0,
            'endTracks': 0,
            'cleanTime': 0,
            'cleanDuration': 0,
            'reviewTime': 0,
            'reviewDuration': 0,
            'linkDuration': 0,
            'linkTime': 0,
        }
        self.locked = False
        self.metadata = {}

    def createState(self, folder, root, owner, created_date=None):
        meva_state = self.findOne({'folderId': str(folder['_id'])})
        if not meva_state:
            if created_date is None:
                created = datetime.datetime.utcnow()
            else:
                created = parser.parse(created_date)
            meva_state = dict(
                folderId=str(folder['_id']),
                filename=str(folder['name']),
                root=root,
                owner=str(owner['_id']),
                owner_name=owner['login'],
                locked=False,
                created=created,
                MEVA_State={
                    "Imported": False,
                    "Cleaning": False,
                    "Reviewing": False,
                    "Linking": False,
                    "Complete": False,
                },
                MEVA_Assignment={
                    "Cleaner": None,
                    "Reviewer": None,
                    "Linker": None,
                },
                MEVA_Data={
                    'beginTracks': 0,
                    'endTracks': 0,
                    'CleaningTime': 0,
                    'CleaningDuration': 0,
                    'ReviewingTime': 0,
                    'ReviewingDuration': 0,
                    'LinkingDuration': 0,
                    'LinkingTime': 0,
                    'CompleteDuration': 0,
                    'CompleteTime': 0,

                },
                metadata={},
            )
            meva_state = self.save(meva_state)
        return meva_state

    def setKeys(self, folderId, vals: dict):
        reserved = ['folderId', 'owner', 'ownername', 'filename', 'root', 'metadata']
        meva_state = self.findOne({'folderId': str(folderId)})
        if meva_state is None:
            raise RestException(f'No Meva_State found for {folderId}')

        for key in vals.keys():
            if key not in reserved:
                meva_state[key] = vals[key]
        self.save(meva_state)

    def setState(self, folderId, state, val, time=0):
        meva_state = self.findOne({'folderId': str(folderId)})
        if meva_state is None:
            raise RestException(f'No Meva_State found for {folderId}')
        locking = False
        if state == 'EmptyVideo':
            if val is True:
                meva_state['MEVA_State']["EmptyVideo"] = val
            elif state in meva_state['MEVA_State']:
                meva_state['MEVA_State'].pop(state)
        elif state == 'LowPriority':
            if val is True:
                meva_state['MEVA_State']["LowPriority"] = val
            elif state in meva_state['MEVA_State']:
                meva_state['MEVA_State'].pop(state)
        elif state in meva_state['MEVA_State']:
            meva_state['MEVA_State'][state] = val
            if f"{state}Duration" in meva_state['MEVA_Data']:
                time_state = state
                if state == 'Reviewing':
                    time_state = 'Cleaning'
                if state == 'Linking':
                    if val is True:
                        locking = True
                    meva_state['locked'] = val
                    time_state = 'Reviewing'
                if state == 'Complete':
                    time_state = 'Linking'
                if val is True:
                    meva_state['MEVA_Data'][f"{time_state}Duration"] = time
            if f"{state}Time" in meva_state['MEVA_Data'] and val is True:
                meva_state['MEVA_Data'][f"{state}Time"] = datetime.datetime.utcnow()
        else:
            raise RestException(f'No State: {state} defined in MEVA_State')
        self.save(meva_state)

    def setMetadata(self, folderId, metadata):
        meva_state = self.findOne({'folderId': str(folderId)})
        if meva_state is None:
            raise RestException(f'No Meva_State found for {folderId}')
        meva_state['metadata'].update(metadata)
        self.save(meva_state)

    def setMetadataThumbnails(self, folderId, thumbnails):
        meva_state = self.findOne({'folderId': str(folderId)})
        if meva_state is None:
            raise RestException(f'No Meva_State found for {folderId}')
        meva_state['metadata']['thumbnails'] = thumbnails
        self.save(meva_state)

    def setMevaData(self, folderId, data, value):
        meva_state = self.findOne({'folderId': str(folderId)})
        if meva_state is None:
            raise RestException(f'No Meva_State found for {folderId}')
        if data in meva_state['MEVA_Data']:
            meva_state['MEVA_Data'][data] = value
        else:
            raise RestException(f'No State: {data} defined in MEVA_Data')
        self.save(meva_state)

    def setAssignment(self, folderId, role, user):
        meva_state = self.findOne({'folderId': folderId})
        if meva_state is None:
            raise RestException(f'No Meva_State found for {folderId}')
        if role in meva_state['MEVA_Assignment']:
            if user is not None:
                meva_state['MEVA_Assignment'][role] = {
                    "name": user['login'],
                    "id": str(user["_id"]),
                }
            else:
                meva_state['MEVA_Assignment'][role] = None
        elif role == 'owner':
            meva_state['owner'] = str(user["_id"])
            meva_state['owner_name'] = user['login']
        else:
            raise RestException(f' No Role: {role} defined in assigments')
        self.save(meva_state)

    def validate(self, doc):
        if not doc.get('folderId') or not isinstance(doc['folderId'], str):
            raise ValidationException('folderId must be a string')
        if 'owner' not in doc or not isinstance(doc['owner'], str):
            raise ValidationException('owner must be a string')
        if not doc.get('root') or not isinstance(doc['root'], str):
            raise ValidationException('root must be a string')
        return doc
