import json
import argparse
import tqdm
from meva_server.serializers.models import Feature, Track, interpolate


coco_part_names = [
    'nose',
    'leftEye',
    'rightEye',
    'leftEar',
    'rightEar',
    'leftShoulder',
    'rightShoulder',
    'leftElbow',
    'rightElbow',
    'leftWrist',
    'rightWrist',
    'leftHip',
    'rightHip',
    'leftKnee',
    'rightKnee',
    'leftAnkle',
    'rightAnkle',
]
pairs = [
    [0, 1],
    [0, 2],
    [1, 3],
    [2, 4],
    [5, 6],
    [5, 7],
    [7, 9],
    [6, 8],
    [8, 10],
    [11, 12],
    [11, 13],
    [13, 15],
    [12, 14],
    [14, 16],
    [6, 12],
    [5, 11],
]


def convert_detections_to_tracks(file):
    file_obj = open(file, "r")
    json_data = json.loads(file_obj.read())

    # organize file by ID and frame Numbers
    tracks = {}
    track_count = 0
    frame_count = 0
    for frame_str in json_data.keys():
        for detection in json_data[frame_str]:
            if (
                float(detection["score"]) < 0.0
            ):  # if we are implementing a filter based on score
                continue
            frame = int(frame_str)
            if frame % 900 != 0:
                continue
            pose_x = detection["pose_x"]
            pose_y = detection["pose_y"]
            pose = {}
            pair_count = 0
            for part in coco_part_names:
                pose[coco_part_names[pair_count]] = [
                    float(pose_x[pair_count]),
                    float(pose_y[pair_count]),
                ]
                pair_count += 1
            score = detection['score']
            trackId = track_count
            if "track_id" in detection:
                trackId = detection["track_id"]
                if trackId in tracks:
                    track = tracks[trackId]
                else:
                    track = Track(begin=frame, end=frame, trackId=trackId)
            else:
                track = Track(begin=frame, end=frame, trackId=trackId)
            track.confidencePairs = [[f'person_{trackId}', float(detection["score"])]]
            tracks[trackId] = track
            track.begin = min(frame, track.begin)
            track.end = max(track.end, frame)
            attributes = {"poses": {"points": {}, "score": score}}
            x_min = float('inf')
            x_max = float('-inf')
            y_min = float('inf')
            y_max = float('-inf')
            for key in coco_part_names:
                x = round(pose[key][0])
                y = round(pose[key][1])
                attributes['poses']['points'][key] = [x, y]
                x_max = max(x_max, x)
                y_max = max(y_max, y)
                x_min = min(x_min, x)
                y_min = min(y_min, y)
            bounds = [x_min, y_min, x_max, y_max]
            feature: Feature = Feature(
                frame=frame, bounds=bounds, attributes=attributes or None
            )
            track.features.append(feature)
            track_count += 1
        frame_count += 1
        print(f'Frame: {frame_count} : frame: {frame} - Tracks: {track_count}')

    return {trackId: track.dict(exclude_none=True) for trackId, track in tracks.items()}


def main(args):
    tracks = convert_detections_to_tracks(args.json_file)
    json_file = args.json_file.replace('.json', '-output.json')
    with open(json_file, "w") as outfile:
        json.dump(tracks, outfile)


def parse_args():
    parser = argparse.ArgumentParser(description='Track filtering')

    parser.add_argument('--json-file', default='', help='JSON file')

    return parser.parse_args()


if __name__ == '__main__':
    args = parse_args()
    main(args)
