import io
import json
import os
import uuid
from datetime import datetime
from pathlib import Path
from typing import Dict, Optional

import pymongo
from girder.exceptions import RestException
from girder.models.file import File
from girder.models.folder import Folder
from girder.models.item import Item
from girder.models.upload import Upload
from pymongo.cursor import Cursor

from meva_server.constants import csvRegex, jsonRegex
from meva_server.serializers import viame
from meva_utils import TRUTHY_META_VALUES
from meva_utils.types import GirderModel


def all_detections_items(folder: Folder) -> Cursor:
    """caller is responsible for verifying access permissions"""
    return Item().find({"meta.detection": str(folder['_id'])}).sort([("created", -1)])


def all_chips_items(folder: Folder) -> Cursor:
    """caller is responsible for verifying access permissions"""
    return Item().find({"meta.chip_json": str(folder['_id'])}).sort([("created", -1)])


def chips_item(folder: Folder, strict=False) -> Optional[GirderModel]:
    all_items = all_chips_items(folder)
    first_item = next(all_items, None)
    if first_item is None and strict:
        raise RestException(f"No chips for folder {folder['name']}")
    return first_item


def detections_item(folder: Folder, strict=False) -> Optional[GirderModel]:
    all_items = all_detections_items(folder)
    first_item = next(all_items, None)
    if first_item is None and strict:
        raise RestException(f"No detections for folder {folder['name']}")
    return first_item


def chips_file(folder: Folder, strict=False) -> Optional[GirderModel]:
    item = chips_item(folder, strict)
    if item is None and not strict:
        return None
    first_file = next(Item().childFiles(item), None)
    if first_file is None and strict:
        raise RestException(f"No file associated with chip item {item}")
    return first_file


def detections_file(folder: Folder, strict=False) -> Optional[GirderModel]:
    item = detections_item(folder, strict)
    if item is None and not strict:
        return None
    first_file = next(Item().childFiles(item), None)
    if first_file is None and strict:
        raise RestException(f"No file associated with detection item {item}")
    return first_file


def get_static_pipelines_path() -> Path:
    pipeline_path = None

    env_pipelines_path = os.getenv("VIAME_PIPELINES_PATH")
    if env_pipelines_path is None:
        raise Exception(
            "No pipeline path specified. "
            "Please set the VIAME_PIPELINES_PATH environment variable.",
        )

    pipeline_path = Path(env_pipelines_path)
    if not pipeline_path.exists():
        raise Exception("Specified pipeline path does not exist!")

    return pipeline_path


def get_or_create_auxiliary_folder(folder, user):
    return Folder().createFolder(folder, "auxiliary", reuseExisting=True, creator=user)


def get_or_create_metadata_folder(folder, user):
    return Folder().createFolder(folder, "metadata", reuseExisting=True, creator=user)


def get_or_create_thumbnail_folder(folder, user):
    return Folder().createFolder(folder, "thumbnail", reuseExisting=True, creator=user)


def move_existing_result_to_auxiliary_folder(folder, user):
    auxiliary = get_or_create_auxiliary_folder(folder, user)
    for item in all_detections_items(folder):
        Item().move(item, auxiliary)


def itemIsWebsafeVideo(item: Item) -> bool:
    return item.get("meta", {}).get("codec") == "h264"


def getTrackData(file: Optional[File]) -> Dict[str, dict]:
    if file is None:
        return {}
    if "csv" in file["exts"]:
        return viame.load_csv_as_tracks(
            b"".join(list(File().download(file, headers=False)()))
            .decode("utf-8")
            .splitlines()
        )
    return json.loads(b"".join(list(File().download(file, headers=False)())).decode())


def getChipData(file: Optional[File]) -> Dict[str, dict]:
    if file is None:
        return {}
    return json.loads(b"".join(list(File().download(file, headers=False)())).decode())


def process_csv(folder: GirderModel, user: GirderModel):
    """If there's a CSV in the folder, process it as a detections object"""
    csvItems = Folder().childItems(
        folder,
        filters={"lowerName": {"$regex": csvRegex}},
        sort=[("created", pymongo.DESCENDING)],
    )
    if csvItems.count() >= 1:
        auxiliary = get_or_create_auxiliary_folder(folder, user)
        file = Item().childFiles(csvItems.next())[0]
        tracks = getTrackData(file)
        saveTracks(folder, tracks, user)
        csvItems.rewind()
        for item in csvItems:
            Item().move(item, auxiliary)
        return True
    return False


def process_json(folder: GirderModel, user: GirderModel):
    # Find a json if it exists
    jsonItems = list(
        Folder().childItems(
            folder,
            filters={"lowerName": {"$regex": jsonRegex}},
            sort=[("created", pymongo.DESCENDING)],
        )
    )
    for item in jsonItems:
        item['meta']['detection'] = str(folder['_id'])
        Item().save(item)
    if len(jsonItems):
        move_existing_result_to_auxiliary_folder(folder, user)
        return True
    return False


def saveTracks(folder, tracks, user):
    timestamp = datetime.now().strftime("%m-%d-%Y_%H:%M:%S")
    item_name = f"result_{timestamp}.json"

    move_existing_result_to_auxiliary_folder(folder, user)
    newResultItem = Item().createItem(item_name, user, folder)
    Item().setMetadata(newResultItem, {"detection": str(folder["_id"])}, allowNull=True)

    json_bytes = json.dumps(tracks).encode()
    byteIO = io.BytesIO(json_bytes)
    Upload().uploadFromFile(
        byteIO,
        len(json_bytes),
        item_name,
        parentType="item",
        parent=newResultItem,
        user=user,
        mimeType="application/json",
    )


def get_filtered_folders(self, filters):
    limit, offset, sort = self.getPagingParameters(filters)
    query = {
        'meta.MEVA': {'$in': TRUTHY_META_VALUES},
    }
    if filters is not None:
        query = {'$and': [query]}
        if 'date' in filters.keys():
            query["$and"].append({'meta.MEVA_Meta.date': {'$in': [filters['date']]}})
        if 'location' in filters.keys():
            query["$and"].append(
                {'meta.MEVA_Meta.location': {'$in': [filters['location']]}}
            )
        if 'cameraId' in filters.keys():
            query["$and"].append(
                {'meta.MEVA_Meta.cameraId': {'$in': [filters['cameraId']]}}
            )
        if 'startTime' in filters.keys():
            query["$and"].append(
                {'meta.MEVA_Meta.startTime': {'$regex': filters['startTime']}}
            )
        if 'state' in filters.keys():
            query["$and"].append(
                {f'meta.MEVA_State.{filters["state"]}': {'$in': TRUTHY_META_VALUES}}
            )
    return Folder().findWithPermissions(
        query,
        offset=offset,
        limit=limit,
        sort=sort,
        user=self.getCurrentUser(),
    )


def load_actor_json(user, folder):
    actorJSONList = list(
        Item().findWithPermissions(
            {
                'meta.actorJSON': {'$in': [True, 'true', 'True']},
                'meta.baseMevaFolder': str(folder["_id"]),
            },
            user=user,
        )
    )
    if len(actorJSONList) == 0:
        return {"status": "no tracks"}
    actorJSONFile = Item().childFiles(actorJSONList[0])[0]

    return json.loads(
        b"".join(list(File().download(actorJSONFile, headers=False)())).decode()
    )
