import argparse
import json
import os
import shutil
from pathlib import Path
from typing import Dict, List

import cv2
import numpy as np


def intersection(a, b):
    x = max(a[0], b[0])
    y = max(a[1], b[1])
    w = min(a[0] + a[2], b[0] + b[2]) - x
    h = min(a[1] + a[3], b[1] + b[3]) - y
    if w < 0 or h < 0:
        return ()  # or (0,0,0,0) ?
    return (x, y, w, h)


def calculate_diva_priors(jsonFile):
    """
    Read in JSON file and video file and extract bounds and images for the sectioned areas
    """
    file_obj = open(jsonFile, "r")
    json_data = json.loads(file_obj.read())
    # Lets resort the file by frame number to make it easier
    diva_detection_count = 0
    non_diva_person_count = 0
    diva_person_count = 0
    diva_detection_map = {}
    diva_person_map = {}
    person_track_count = 0
    contains_diva_count = 0
    ignore_below_size = 0
    total_detection_count = 0
    ignore_with_large_size = 0
    total_track_count = 0
    flatter_list: Dict[str, Dict] = {}
    for track in json_data.values():
        found_diva = False
        has_large_ignore_detection = False
        total_track_count += 1
        if "features" in track:
            features = track["features"]
            for feature in features:
                if "attributes" in feature:
                    total_detection_count+=1
                    if "diva_id" in feature["attributes"]:
                        diva_detection_count+=1
                        diva_detection_map[feature['attributes']['diva_id']] = True
                        found_diva = True
                        if 'confidencePairs' in track:
                            if 'ignore' not in track['confidencePairs'][0][0]:
                                diva_person_count+=1
                                diva_person_map[feature['attributes']['diva_id']] = True
                    if 'confidencePairs' in track:
                        if 'ignore' in track['confidencePairs'][0][0]:
                            if 'bounds' in feature:
                                bounds = feature['bounds']
                                width = bounds[2] - bounds[0]
                                height = bounds[3] - bounds[1]
                                if height < 74 or width < 25:
                                    ignore_below_size += 1
                                else:
                                    has_large_ignore_detection = True
                    

                    
        if has_large_ignore_detection is True:
            ignore_with_large_size += 1
        if found_diva is False:
            if 'confidencePairs' in track:
                if 'ignore' not in track['confidencePairs'][0][0]:
                    non_diva_person_count+=1
        if 'confidencePairs' in track:
            if 'ignore' not in track['confidencePairs'][0][0]:
                person_track_count+=1
                if found_diva is True:
                    contains_diva_count+=1


    print(f"File: {jsonFile} DivaPriors: {diva_detection_count} DivaPersonCount: {diva_person_count} IngoreBelowSize: {ignore_below_size}")
    unique_detection_ids = len(diva_detection_map.keys())
    unique_person_ids = len(diva_person_map.keys())
    return { "diva_detection_count": diva_detection_count, "diva_person_count":diva_person_count, 
    "unique_detection_ids":unique_detection_ids, "unique_person_ids": unique_person_ids, "contains_diva_count": contains_diva_count, "ignore_with_large_size": ignore_with_large_size, "total_track_count": total_track_count,
     "non_diva_person_count": non_diva_person_count, "person_track_count": person_track_count, "ignore_below_size": ignore_below_size, "total_detection_count": total_detection_count,  }


def main(args):
    data_dir = Path(args.folder)
    track_files = list(data_dir.glob('**/*.json'))
    total_count = { "diva_detection_count": 0, "diva_person_count": 0, "unique_detection_ids": 0, "unique_person_ids": 0, "ignore_with_large_size": 0,
    "non_diva_person_count": 0, "person_track_count": 0, "contains_diva_count": 0, "ignore_below_size": 0, "total_detection_count": 0, "total_track_count": 0}
    for json_file in track_files:
        if json_file.is_file():
            additive_count = calculate_diva_priors(json_file)
            total_count["diva_detection_count"] += additive_count["diva_detection_count"]
            total_count["diva_person_count"] += additive_count["diva_person_count"]
            total_count["unique_detection_ids"] += additive_count["unique_detection_ids"]
            total_count["unique_person_ids"] += additive_count["unique_person_ids"]
            total_count["non_diva_person_count"] += additive_count["non_diva_person_count"]
            total_count["person_track_count"] += additive_count["person_track_count"]
            total_count["contains_diva_count"] += additive_count["contains_diva_count"]
            total_count["ignore_below_size"] += additive_count["ignore_below_size"]
            total_count["ignore_with_large_size"] += additive_count["ignore_with_large_size"]
            total_count["total_track_count"] += additive_count["total_track_count"]

    print("Totals")
    print(total_count)
    with open('output.json', 'w') as f:
        json.dump(total_count, f)



def parse_args():
    parser = argparse.ArgumentParser(description='Track filtering')

    parser.add_argument('--folder', default='', help='folder')

    return parser.parse_args()


if __name__ == '__main__':
    args = parse_args()
    main(args)
