import argparse
import json
import os
import shutil
from typing import Dict, List

import cv2
import numpy as np


def intersection(a, b):
    x = max(a[0], b[0])
    y = max(a[1], b[1])
    w = min(a[0] + a[2], b[0] + b[2]) - x
    h = min(a[1] + a[3], b[1] + b[3]) - y
    if w < 0 or h < 0:
        return ()  # or (0,0,0,0) ?
    return (x, y, w, h)


def start_chip_extraction(jsonFile, videoFile):
    """
    Read in JSON file and video file and extract bounds and images for the sectioned areas
    """
    padding = 0.20
    file_obj = open(jsonFile, "r")
    json_data = json.loads(file_obj.read())
    # Lets resort the file by frame number to make it easier
    flatter_list: Dict[str, Dict] = {}
    os.mkdir('tracks')
    os.mkdir('frames')
    for key in json_data.keys():
        track: List[Dict] = json_data[key]
        os.mkdir(f'tracks/{key}')
        index = 0
        for data in track:
            frame = data["frame"]
            if str(frame) not in flatter_list:
                flatter_list[str(frame)] = []
            data['trackId'] = key
            data['index'] = index
            flatter_list[str(frame)].append(data)
            index = index + 1
    vidcap = cv2.VideoCapture(videoFile)
    success, image = vidcap.read()
    count = 0
    success = True
    while success:
        success, image = vidcap.read()
        if str(count) in flatter_list and image is not None:
            # write the frame first
            cv2.imwrite("frames/frame%d.jpg" % count, image)  # save frame as JPEG file
            original = image.copy()

            for track in flatter_list[str(count)]:
                top = round(
                    track['bounds'][3]
                    + (padding * (track['bounds'][3] - track['bounds'][1]))
                )
                bottom = round(
                    track['bounds'][1]
                    - (padding * (track['bounds'][3] - track['bounds'][1]))
                )
                left = round(
                    track['bounds'][0]
                    - (padding * (track['bounds'][2] - track['bounds'][0]))
                )
                right = round(
                    track['bounds'][2]
                    + (padding * (track['bounds'][2] - track['bounds'][0]))
                )
                # bounds need to be clamped to left/right/top/bottom or preferably padded
                height, width, channels = original.shape
                newbounds = intersection(
                    [left, bottom, right - left, top - bottom], [0, 0, width, height]
                )
                bounds = original[
                    newbounds[1] : newbounds[1] + newbounds[3],
                    newbounds[0] : newbounds[0] + newbounds[2],
                ]
                final = bounds
                cv2.imwrite(
                    f"tracks/{track['trackId']}/{track['index']}.jpg", final
                )  # save frame as JPEG file
        count += 1


def main(args):
    start_chip_extraction(args.json_file, args.vid_file)


def parse_args():
    parser = argparse.ArgumentParser(description='Track filtering')

    parser.add_argument('--json-file', default='', help='JSON file')
    parser.add_argument('--vid-file', default='', help='Video File')

    return parser.parse_args()


if __name__ == '__main__':
    args = parse_args()
    main(args)
