import contextlib
import json
import os
import shutil
import subprocess
import tempfile
from zipfile import ZIP_DEFLATED, ZipFile
from pathlib import Path
from subprocess import Popen
from sys import meta_path
from typing import Dict, List, Optional, Tuple
from urllib import request
from urllib.parse import urlparse

from celery.app import base
from girder_client import GirderClient
from girder_worker.app import app
from girder_worker.task import Task
from girder_worker.utils import JobManager, JobStatus
from GPUtil import getGPUs

from meva_tasks.utils import (
    download_source_media,
    get_video_filename,
    stream_subprocess,
)
from meva_utils.types import AvailableJobSchema, GirderModel, PipelineJob

EMPTY_JOB_SCHEMA: AvailableJobSchema = {
    'pipelines': {},
}


def get_gpu_environment() -> Dict[str, str]:
    """Get environment variables for using CUDA enabled GPUs."""
    env = os.environ.copy()

    gpu_uuid = env.get("WORKER_GPU_UUID")
    gpus = [gpu.id for gpu in getGPUs() if gpu.uuid == gpu_uuid]

    # Only set this env var if WORKER_GPU_UUID was supplied,
    # and it matches an installed GPU
    if gpus:
        env["CUDA_VISIBLE_DEVICES"] = str(gpus[0])

    return env


class Config:
    def __init__(self):
        self.gpu_process_env = get_gpu_environment()


@app.task(bind=True, acks_late=True)
def run_pipeline(self: Task, params: PipelineJob):
    conf = Config()
    manager: JobManager = self.job_manager
    gc: GirderClient = self.girder_client

    # Extract params
    pipeline = params["pipeline"]
    input_folder = params["input_folder"]
    input_type = params["input_type"]
    output_folder = params["output_folder"]

    # Create temporary files/folders, removed at the end of the function
    input_path = Path(tempfile.mkdtemp())
    track_output_path = str(input_path)

    # defer cleanup
    def cleanup():
        # Files
        # Folders
        shutil.rmtree(input_path, ignore_errors=True)

    # gc.downloadFolderRecursive(input_folder, input_path)

    # Load the input file and change video name

    video_file = download_source_media(gc, input_folder, input_path)[0]
    video_location = str(input_path / video_file)

    os.chdir('/home/AlphaPose/')

    command = [
        "./poseFlow.sh",
        "./configs/coco/resnet/256x192_res50_lr1e-3_1x.yaml",
        "./pretrained_models/fast_res50_256x192.pth",
        f"{video_location}" f" {track_output_path}",
    ]
    cmd = " ".join(command)
    manager.write(f"Running command: {cmd}\n", forceFlush=True)

    process_err_file = tempfile.TemporaryFile()
    process = Popen(
        cmd,
        stdout=subprocess.PIPE,
        stderr=process_err_file,
        shell=True,
        executable='/bin/bash',
        env=conf.gpu_process_env,
    )
    stream_subprocess(process, self, manager, process_err_file, cleanup=cleanup)
    if self.canceled:
        return

    manager.updateStatus(JobStatus.PUSHING_OUTPUT)
    alphajson = list(Path(track_output_path).glob("*tracks.json"))[0]
    newfile = gc.uploadFileToFolder(output_folder, alphajson)

    gc.addMetadataToItem(
        newfile["itemId"], {"pipeline": pipeline, "detection": output_folder}
    )
    gc.sendRestRequest('POST', f'meva/alphapose_convert?folderId={output_folder}')
    gc.sendRestRequest(
        'PUT',
        f'meva/metadata/{output_folder}',
        json={
            "MEVA_State": {
                "Imported": True,
                "Cleaning": True,
                "Reviewing": False,
                "Linking": False,
                "Complete": False,
            }
        },
    )


@app.task(bind=True, acks_late=True)
def create_track_chips(self: Task, folder, jsonDict, metadataFolderId):
    folderId = str(folder["_id"])
    input_path = Path(tempfile.mkdtemp())

    # write the json to a file
    json_file = str(input_path / 'input.json')
    with open(json_file, 'w') as outfile:
        json.dump(jsonDict, outfile)

    def cleanup():
        shutil.rmtree(input_path, ignore_errors=True)

    gc: GirderClient = self.girder_client
    manager: JobManager = self.job_manager

    # gc.downloadFolderRecursive(folderId, input_path)

    video_file = video_file = download_source_media(gc, folder, input_path)[0]
    video_location = str(input_path / video_file)

    process_err_file = tempfile.TemporaryFile()
    print(video_location)
    command = [
        f"cd {str(input_path)} &&" "python3",
        "/home/meva_girder/meva_tasks/chip_extraction.py",
        "--json-file",
        json_file,
        "--vid-file",
        video_location,
    ]
    cmd = " ".join(command)
    manager.write(f"Running command: {cmd}\n", forceFlush=True)
    process = Popen(
        cmd,
        stdout=subprocess.PIPE,
        stderr=process_err_file,
        shell=True,
        executable='/bin/bash',
    )
    process_err_file = tempfile.TemporaryFile()

    stream_subprocess(process, self, manager, process_err_file)
    if self.canceled:
        return

    # Now we upload the data and placed Item Ids in there to make it easier to load info for display

    manager.updateStatus(JobStatus.PUSHING_OUTPUT)
    print('Creating frame Folder')
    frame_folder = gc.createFolder(metadataFolderId, 'frames')
    print('Creating tracks folder')
    tracks_folder = gc.createFolder(metadataFolderId, 'tracks', reuseExisting=True)
    track_ids = {}
    for key in jsonDict.keys():
        track_list = jsonDict[key]
        counter = 0
        # print(f"Creating current tracks folder: {key}")
        current_folder = gc.createFolder(tracks_folder["_id"], key)
        track_ids[key] = current_folder["_id"]
        for detection in track_list:
            frame = detection['frame']
            frame_path = (input_path / 'frames') / f"frame{frame}.jpg"
            if os.path.exists(frame_path):
                # print(f"Creating current frame folder: {frame}")
                frameId = gc.uploadFileToFolder(frame_folder["_id"], frame_path)
                detection['frameId'] = frameId["_id"]
            detection_path = ((input_path / 'tracks') / key) / f"{counter}.jpg"
            if os.path.exists(detection_path):
                # print(f"Creating current detection folder: {counter}")
                detectionId = gc.uploadFileToFolder(
                    current_folder["_id"], detection_path
                )
                detection['chipId'] = detectionId["_id"]
            counter = counter + 1

    chip_file = str(input_path / 'chip.json')
    with open(chip_file, 'w') as outfile:
        json.dump(jsonDict, outfile)

    chipjson = gc.uploadFileToFolder(metadataFolderId, str(chip_file))

    gc.addMetadataToItem(chipjson["itemId"], {"chip_json": str(folderId)})
    gc.addMetadataToFolder(folderId, {"track_chips": track_ids})

    gc.sendRestRequest(
        'PUT',
        f'meva/metadata/{folderId}',
        json={
            "MEVA_State": {
                "Imported": True,
                "Cleaning": True,
                "Reviewing": True,
                "Linking": False,
                "Complete": False,
            }
        },
    )

    cleanup()


@app.task(bind=True, acks_late=True)
def create_thumbnails(self: Task, folder, thumbnailFolderId):
    folderId = str(folder["_id"])
    input_path = Path(tempfile.mkdtemp())

    def cleanup():
        shutil.rmtree(input_path, ignore_errors=True)

    gc: GirderClient = self.girder_client
    manager: JobManager = self.job_manager

    video_file = download_source_media(gc, folder, input_path)[0]
    video_location = str(input_path / video_file)

    process_err_file = tempfile.TemporaryFile()
    print(video_location)
    command = [
        "ffmpeg",
        "-ss 3",
        "-i",
        video_location,
        "-vf",
        "\"select=gt(scene\,0.4),scale=200:-1\"",
        "-frames:v",
        "10",
        "-vsync",
        "vfr",
        "-vf",
        "\"fps=fps=1/5,scale=200:-1\"",
        f"{input_path}/out%02d.jpg",
    ]
    cmd = " ".join(command)
    manager.write(f"Running command: {cmd}\n", forceFlush=True)
    process = Popen(
        cmd,
        stdout=subprocess.PIPE,
        stderr=process_err_file,
        shell=True,
        executable='/bin/bash',
    )
    process_err_file = tempfile.TemporaryFile()

    stream_subprocess(process, self, manager, process_err_file, cleanup=cleanup)
    if self.canceled:
        return

    manager.updateStatus(JobStatus.PUSHING_OUTPUT)
    output_jpgs = list(input_path.glob("out*.jpg"))
    thumbnail_list = []
    for item in output_jpgs:
        new_file = gc.uploadFileToFolder(thumbnailFolderId, item)
        thumbnail_list.append(new_file["itemId"])
    meta_data = {"thumbnails": thumbnail_list}
    gc.addMetadataToFolder(
        folderId,
        meta_data,
    )
    gc.sendRestRequest('POST', f'meva_state/{folderId}/meva_thumbnails', json=meta_data)


@app.task(bind=True, acks_late=True)
def convert_video(self: Task, path, folderId, auxiliaryFolderId, itemId, forceTranscode=False):
    # Delete is true, so the tempfile is deleted when the block closes.
    # We are only using this to get a name, and recreating it below.
    with tempfile.NamedTemporaryFile(suffix=".mp4", delete=True) as temp:
        output_path = temp.name

    def cleanup():
        with contextlib.suppress(FileNotFoundError):
            os.remove(output_path)

    gc: GirderClient = self.girder_client
    manager: JobManager = self.job_manager

    # Extract metadata
    print(path)
    file_name = os.path.join(path, os.listdir(path)[0])
    command = [
        "ffprobe",
        "-print_format",
        "json",
        "-v",
        "quiet",
        "-show_format",
        "-show_streams",
        file_name,
    ]

    process_err_file = tempfile.TemporaryFile()
    process = Popen(
        command,
        stdout=subprocess.PIPE,
        stderr=process_err_file,
    )
    stdout = stream_subprocess(
        process, self, manager, process_err_file, keep_stdout=True
    )
    if self.canceled:
        return

    jsoninfo = json.loads(stdout)
    videostream = list(
        filter(lambda x: x["codec_type"] == "video", jsoninfo["streams"])
    )
    if len(videostream) != 1:
        raise Exception('Expected 1 video stream, found {}'.format(len(videostream)))

    transcode = True
    convert = True
    if 'codec_name' in videostream[0]:
        codec_name = videostream[0]['codec_name']
        if codec_name == 'h264':
            transcode = False
    process_err_file = tempfile.TemporaryFile()
    if (transcode is True and convert is True) or forceTranscode is True:
        command = [
            "ffmpeg",
            "-i",
            file_name,
            "-c:v",
            "libx264",
            "-preset",
            "slow",
            "-crf",
            "26",
            # https://askubuntu.com/questions/1315697/could-not-find-tag-for-codec-pcm-s16le-in-stream-1-codec-not-currently-support
            "-c:a",
            "aac",
            # see native/<platform> code for a discussion of this option
            "-vf",
            "scale=ceil(iw*sar/2)*2:ceil(ih/2)*2,setsar=1",
            output_path,
        ]
    elif transcode is False and convert is True:
        command = [
            "ffmpeg",
            "-i",
            file_name,
            "-c:v",
            "copy",
            output_path,
        ]
    if transcode is False and convert is False:
        command = None
        output_path = file_name
    if command is not None:
        process = Popen(
            command,
            stdout=subprocess.PIPE,
            stderr=process_err_file,
        )
        stream_subprocess(process, self, manager, process_err_file, cleanup=cleanup)
        if self.canceled:
            return

    manager.updateStatus(JobStatus.PUSHING_OUTPUT)
    if command is not None:
        new_file = gc.uploadFileToFolder(folderId, output_path)
        gc.addMetadataToItem(
            new_file['itemId'],
            {
                "source_video": False,
                "transcoder": "ffmpeg",
                "codec": "h264",
                "transcoded": transcode or forceTranscode
            },
        )
        gc.addMetadataToFolder(folderId, {
            "transcoded": transcode or forceTranscode
        })
        gc.addMetadataToItem(
            itemId,
            {
                "source_video": True,
                "codec": videostream[0]["codec_name"],
            },
        )
    else:
        gc.addMetadataToItem(
            itemId,
            {
                "source_video": False,
                "codec": videostream[0]["codec_name"],
            },
        )
    frame_rate_str = videostream[0]["avg_frame_rate"]
    rate = frame_rate_str.split('/')
    if len(rate) == 1:
        frame_rate = float(rate[0])
    if len(rate) == 2:
        frame_rate = float(rate[0]) / float(rate[1])

    # Lets add metadata information to the metadata if it exists
    base_file = file_name.split('/')
    base_file = base_file[len(base_file) - 1]
    info_list = base_file.split('.')
    if len(info_list) >= 5:
        meva_data = {
            "date": info_list[0],
            "startTime": info_list[1],
            "endTime": info_list[2],
            "location": info_list[3],
            "cameraId": info_list[4],
        }
    else:
        meva_data = {
            "date": '2021-01-01',
            "startTime": '00-00-00',
            "endTime": '01-00-00',
            "location": base_file,
            "cameraId": '0',
        }

    meta_data = {
        "fps": frame_rate,  # TODO: current time system doesn't allow for non-int framerates
        "annotate": True,  # mark the parent folder as able to annotate.
        "ffprobe_info": videostream[0],
    }
    if meva_data is not None:
        meta_data["MEVA"] = True
        meta_data["MEVA_Meta"] = meva_data
        meta_data["MEVA_State"] = {
            "Imported": False,
            "Cleaning": False,
            "Reviewing": False,
            "Linking": False,
            "Complete": False,
        }
    gc.addMetadataToFolder(
        folderId,
        meta_data,
    )
    gc.sendRestRequest(
        'POST', f'meva_state/{folderId}/meva_metadata', json=meta_data["MEVA_Meta"]
    )
    cleanup()
    gc.sendRestRequest(
        'POST',
        f'meva/create_meva_thumbnails/{folderId}',
    )


@app.task(bind=True, acks_late=True)
def convert_images(self: Task, folderId):
    """
    Ensures that all images in a folder are in a web friendly format (png or jpeg).

    If conversions succeeds for an image, it will replace the image with an image
    of the same name, but in a web friendly extension.

    Returns the number of images successfully converted.
    """
    gc: GirderClient = self.girder_client
    manager: JobManager = self.job_manager

    items = gc.listItem(folderId)
    skip_item = (
        lambda item: item["name"].endswith(".png")
        or item["name"].endswith(".jpeg")
        or item["name"].endswith(".jpg")
    )
    items_to_convert = [item for item in items if not skip_item(item)]

    count = 0
    with tempfile.TemporaryDirectory() as temp:
        dest_dir = Path(temp)

        for item in items_to_convert:
            # Assumes 1 file per item
            gc.downloadItem(item["_id"], dest_dir, item["name"])

            item_path = dest_dir / item["name"]
            new_item_path = dest_dir / ".".join([*item["name"].split(".")[:-1], "png"])

            process_err_file = tempfile.TemporaryFile()
            process = Popen(
                ["ffmpeg", "-i", item_path, new_item_path],
                stdout=subprocess.PIPE,
                stderr=process_err_file,
            )
            stream_subprocess(process, self, manager, process_err_file)
            if self.canceled:
                return

            gc.uploadFileToFolder(folderId, new_item_path)
            gc.delete(f"item/{item['_id']}")
            count += 1

    gc.addMetadataToFolder(
        str(folderId),
        {"annotate": True},  # mark the parent folder as able to annotate.
    )

    return count

@app.task(bind=True, acks_late=True)
def add_linking_tracks_to_database(self: Task, folder, forceAdd=False, excludeMay=True):
    folderId = str(folder["_id"])
    gc: GirderClient = self.girder_client
    manager: JobManager = self.job_manager
    data = gc.sendRestRequest('GET', f'meva_state/{folderId}/all_meva?filters=%7B%22state%22%3A%22Linking%22%7D&limit=6000&sort=created&sortdir=1')
    source_folders = gc.sendRestRequest('GET', 'meva_database/get_source_folders')
    for item in data['pageResults']:
        item_folder_id = item['folderId']
        metadata = item['metadata']
        date = metadata['date']
        if excludeMay is True and '2018-05' in date:
            continue
        if item_folder_id not in source_folders or forceAdd is True:
            manager.write(f'Adding Folder Id: {item["folderId"]}\n')
            gc.sendRestRequest(
                'PUT',
                f'meva_database/add_tracks?folderId={item_folder_id}',
            )
    manager.write('All Tracks Should be added')


@app.task(bind=True, acks_late=True)
def calculate_track_stats(self: Task, folder, uploadFolder, month='2018-03'):
    folderId = str(folder["_id"])
    gc: GirderClient = self.girder_client
    manager: JobManager = self.job_manager
    input_path = Path(tempfile.mkdtemp())
    def cleanup():
        shutil.rmtree(input_path, ignore_errors=True)


    data = gc.sendRestRequest('GET', f'meva_state/{folderId}/all_meva?filters=%7B%22state%22%3A%22Linking%22%7D&limit=6000&sort=created&sortdir=1')
    manager.write("Getting meva state data")
    source_folders = gc.sendRestRequest('GET', 'meva_database/get_source_folders')
    for item in data['pageResults']:
        item_folder_id = item['folderId']
        metadata = item['metadata']
        date = metadata['date']
        name = item['filename']
        if month not in date:
            continue
        manager.write(f'Adding Folder Id: {item["folderId"]} with name: {name}\n')
        detection_data = gc.sendRestRequest(
            'GET',
            f'meva_detection?folderId={item_folder_id}&formatting=track_json',
        )
        # write the detection file so it can be read by the script
        json_name = f'{name}.json'
        with open(input_path / json_name, 'w') as f:
            json.dump(detection_data, f)
    # Now we call the JSON file calculator
    process_err_file = tempfile.TemporaryFile()
    command = [
        f"cd {str(input_path)} &&" "python3",
        "/home/meva_girder/meva_tasks/CalculateStats.py",
        "--folder",
        str(input_path),
    ]
    cmd = " ".join(command)
    manager.write(f"Running command: {cmd}\n", forceFlush=True)
    process = Popen(
        cmd,
        stdout=subprocess.PIPE,
        stderr=process_err_file,
        shell=True,
        executable='/bin/bash',
    )
    stream_subprocess(process, self, manager, process_err_file)
    if self.canceled:
        cleanup()
        return
    # Check for an output.json file and then place it in a specified location if found
    manager.write(f"Checking for file:{str(input_path)}/output.json")
    if os.path.exists(f"{str(input_path)}/output.json"):
        manager.write(f"Found output.json and copying", forceFlush=True)
        os.rename(f"{str(input_path)}/output.json", f"{str(input_path)}/{month}.json")
        gc.upload(f"{str(input_path)}/{month}.json", uploadFolder["_id"])
    manager.write('Data should be calculated with an output.json')
    cleanup()


@app.task(bind=True, acks_late=True)
def import_actor_images(self: Task, folder, zipFileId, existingActorJSON=None):
    input_path = Path(tempfile.mkdtemp())
    folderId = str(folder['_id'])
    # write the json to a file

    def cleanup():
        shutil.rmtree(input_path, ignore_errors=True)

    gc: GirderClient = self.girder_client
    manager: JobManager = self.job_manager
    zip_location = str(input_path / 'zipped.zip')
    gc.downloadFile(zipFileId, zip_location)
    if existingActorJSON is not None:
        gc.downloadFile(existingActorJSON, str(input_path / 'actors.json'))

    process_err_file = tempfile.TemporaryFile()
    command = []
    command.append(f"cd {str(input_path)} &&" "python3")
    command.append("/home/meva_girder/meva_tasks/actor_extraction.py")
    command.append("--zip-file")
    command.append(zip_location)
    if existingActorJSON is not None:
        command.append("--actorJSON")
        command.append(str(input_path / 'actors.json'))
    cmd = " ".join(command)
    manager.write(f"Running command: {cmd}\n", forceFlush=True)
    process = Popen(
        cmd,
        stdout=subprocess.PIPE,
        stderr=process_err_file,
        shell=True,
        executable='/bin/bash',
    )
    process_err_file = tempfile.TemporaryFile()

    stream_subprocess(process, self, manager, process_err_file)
    if self.canceled:
        return

    # Now we upload the data and placed Item Ids in there to make it easier to load info for display

    manager.updateStatus(JobStatus.PUSHING_OUTPUT)
    gc.upload(f'{input_path}/actors/**', folderId)
    gc.sendRestRequest(
        'POST',
        f'meva_checkin/import_actors/{folderId}',
    )
    gc.sendRestRequest(
        'PUT',
        f'meva_database/add_actor_chips/{folder["parentId"]}',
    )
    cleanup()

@app.task(bind=True, acks_late=True)
def remove_actor_images(self: Task, folder, actorJsonFileId, deletion):
    input_path = Path(tempfile.mkdtemp())
    folderId = str(folder['_id'])
    # write the json to a file

    def cleanup():
        shutil.rmtree(input_path, ignore_errors=True)

    gc: GirderClient = self.girder_client
    manager: JobManager = self.job_manager
    gc.downloadItem(actorJsonFileId, input_path)
    file_obj = open(input_path / 'actors.json', "r")
    actor_json_data = json.loads(file_obj.read())

    deletion_list = []
    updated_json = {}
    for key in actor_json_data.keys():
        item = actor_json_data[key]
        if item['date'].startswith(deletion):
            if 'b' in item.keys():
                for keyid in item['b'].keys():
                    deletion_list.append(item['b'][keyid])
            if 'f' in item.keys():
                for keyid in item['f'].keys():
                    deletion_list.append(item['f'][keyid])
        else:
            updated_json[key] = item
            
    print('Deletion List')
    print(deletion_list)

    for item in deletion_list:
        gc.delete(f'item/{item}')

    # Next we may need to get rid of some dangling empty folders?    # Now we read in the actor JSon file
    os.remove(input_path / 'actors.json')
    json_file = (input_path / 'actors.json')
    with open(json_file, 'w') as outfile:
        json.dump(updated_json, outfile)
    
    base_actor_item = gc.getItem(actorJsonFileId)
    manager.updateStatus(JobStatus.PUSHING_OUTPUT)

    metadata = base_actor_item['meta']
    baseFolderId = base_actor_item['folderId']
    gc.delete(f'item/{actorJsonFileId}')

    newActorJsonId = gc.uploadFileToFolder(baseFolderId, json_file)  
    gc.addMetadataToItem(str(newActorJsonId['itemId']), metadata)
    # Now we upload the data and placed Item Ids in there to make it easier to load info for display
    cleanup()

@app.task(bind=True, acks_late=True)
def import_gps_tracks(self: Task, folder, zipFileId):
    input_path = Path(tempfile.mkdtemp())
    folderId = str(folder['_id'])
    # write the json to a file

    def cleanup():
        shutil.rmtree(input_path, ignore_errors=True)

    gc: GirderClient = self.girder_client
    manager: JobManager = self.job_manager
    zip_location = str(input_path / 'zipped.zip')
    gc.downloadFile(zipFileId, zip_location)

    with ZipFile(zip_location, 'r') as zipObj:
        listOfFileNames = zipObj.namelist()
        for fileName in listOfFileNames:
            manager.write(f"Extracting: {fileName}\n", forceFlush=True)
            zipObj.extract(fileName, f'{input_path}/{fileName}')
    manager.updateStatus(JobStatus.PUSHING_OUTPUT)
    # push all gpx files to the proper folder
    gc.upload(f'{input_path}/**/*.gpx', folderId)
    gc.sendRestRequest(
        'PUT',
        f'meva_database/add_gps_tracks/{folder["_id"]}',
    )
    cleanup()

@app.task(bind=True, acks_late=True)
def extract_actor_descriptors(self: Task, folder, zipFileId):
    input_path = Path(tempfile.mkdtemp())
    folderId = str(folder['_id'])
    # write the json to a file
    if os.path.isdir('/data/db/reid'):
        shutil.rmtree('/data/db/reid')

    def cleanup():
        shutil.rmtree(input_path, ignore_errors=True)

    gc: GirderClient = self.girder_client
    manager: JobManager = self.job_manager
    zip_location = str(input_path / 'zipped.zip')
    gc.downloadFile(zipFileId, zip_location)

    with ZipFile(zip_location, 'r') as zipObj:
        listOfFileNames = zipObj.namelist()
        for fileName in listOfFileNames:
            manager.write(f"Extracting: {fileName}\n", forceFlush=True)
            zipObj.extract(fileName, f'/data/db/')
    manager.updateStatus(JobStatus.PUSHING_OUTPUT)
    # push all gpx files to the proper folder
    cleanup()


def zipdir(path, ziph):
    # ziph is zipfile handle
    for root, dirs, files in os.walk(path):
        for file in files:
            ziph.write(os.path.join(root, file),
                       os.path.relpath(os.path.join(root, file),
                                       os.path.join(path, '..')))


@app.task(bind=True, acks_late=True)
def export_DB(self: Task, folder):
    input_path = Path(tempfile.mkdtemp())
    folderId = str(folder['_id'])
    # write the json to a file

    def cleanup():
        shutil.rmtree(input_path, ignore_errors=True)

    gc: GirderClient = self.girder_client
    manager: JobManager = self.job_manager
    zip_file = f"{input_path}/database.zip"
    shutil.make_archive(f"{input_path}/database", 'zip', '/data/db', '/data/db')
    gc.upload(zip_file, folderId)
    manager.updateStatus(JobStatus.PUSHING_OUTPUT)
    cleanup()


@app.task(bind=True, acks_late=True)
def export_linked_track_images_task(self: Task, folder,):
    gc: GirderClient = self.girder_client
    manager: JobManager = self.job_manager
    limit = 10000
    actor_links = gc.sendRestRequest('GET', f'meva_database/get_linked_tracks?pageSize={limit}')['pageResults']
    count = 0
    os.chdir('/tmp')
    if (os.path.exists('/tmp/Links.zip')):
        os.remove('/tmp/Links.zip')
    if os.path.exists('./actors') is False:
        os.mkdir('./actors')
    for actor in actor_links:
        if actor['type'] != 'actor': #filter out track-to-track links for now
            continue
        actorId = actor['actorId']
        date = actor['date']
        outfitId = actor['outfitId']
        links = actor['links']
        print(f'Importing Actor {count} of {len(actor_links)}')
        # create the actorID/date/OutfitId structure
        if os.path.exists(f'./actors/{actorId}') is False:
            os.mkdir(f'./actors/{actorId}/')
        if os.path.exists(f'./actors/{actorId}/{date}') is False:
            os.mkdir(f'./actors/{actorId}/{date}')
        if os.path.exists(f'./actors/{actorId}/{date}/{outfitId}') is False:
            os.mkdir(f'./actors/{actorId}/{date}/{outfitId}')
        # Now that we have the links we take the folderIds get the video names download the json and the images
        root_path = f'./actors/{actorId}/{date}/{outfitId}'
        link_count = 0
        for link in links:
            print(f'Importing Link {link_count} of {len(links)}')
            if link['type'] != 'track': # only looking for tracks in here
                continue
            try:
                folder_name = gc.get(f'folder/{link["folderId"]}')['name']
            except:
                continue            
            if os.path.exists(f'{root_path}/{folder_name}') is False:
                os.mkdir(f'{root_path}/{folder_name}')
            # now we get the link data and download the individual chips
            link_data = gc.get(f'meva_detection/meva_chip?folderId={link["folderId"]}&trackId={link["trackId"]}')
            with open(f'{root_path}/{folder_name}/chips.json', 'w') as outfile:
                json.dump(link_data, outfile)
            for chip in link_data: # download chips and full frames
                if 'chipId' in chip.keys():
                    if os.path.exists( f'{root_path}/{folder_name}/chip_{chip["frame"]}.jpg') is False:
                        gc.downloadFile(chip['chipId'], f'{root_path}/{folder_name}/chip_{chip["frame"]}.jpg')
                        gc.downloadFile(chip['frameId'], f'{root_path}/{folder_name}/frame_{chip["frame"]}.jpg')
            link_count += 1
        count = count + 1 
    zipf = ZipFile('/tmp/Links.zip', 'w', ZIP_DEFLATED)
    zipdir('/tmp/actors', zipf)
    zipf.close()
    # upload Zip to export directory
    gc.uploadFileToFolder(folder, '/tmp/Links.zip')


@app.task(bind=True, acks_late=True)
def export_feature_links(self: Task, baseMevaFolder, baseLinkingFolder):
    gc: GirderClient = self.girder_client
    manager: JobManager = self.job_manager
    input_path = Path(tempfile.mkdtemp())
    folderId = str(baseMevaFolder['_id'])

    def cleanup():
        shutil.rmtree(input_path, ignore_errors=True)
    
    colorScale = ['rgb(165, 0, 38)', 'rgb(212, 50, 44)', 'rgb(241, 110, 67)', 'rgb(252, 172, 99)', 'rgb(254, 221, 141)', 'rgb(249, 247, 174)', 'rgb(215, 238, 142)', 'rgb(164, 216, 110)', 'rgb(100, 188, 97)', 'rgb(34, 150, 79)']
    feature_links = gc.sendRestRequest('GET', f'linked_feature/{folderId}/all_linkedFeatures')
    userId_map = {}
    linking_map = {}
    for item in feature_links:
        keys = item.keys()
        index = ''
        if 'sourceClipFilename' in keys and 'sourceTrackId' in keys:
            index = f'{item["sourceClipFilename"]}-{item["sourceTrackId"]}'
        if 'featureDrawings' in item.keys() and item['linked'] is True:
            featureDrawings = item['featureDrawings']
            userId = item['userId']
            if userId not in userId_map.keys():
                user_data = gc.sendRestRequest('GET', f'user/{userId}')
                userId_map[userId] = user_data

            user_login = userId_map[userId]['login']
            item['userLogin'] = user_login
            if os.path.exists(f'{input_path}/{index}') is False:
                os.mkdir(f'{input_path}/{index}/')
            if os.path.exists(f'{input_path}/{index}/{user_login}') is False:
                os.mkdir(f'{input_path}/{index}/{user_login}/')
            # Now lets get the girder items and download them
            for feature in featureDrawings:
                front_or_back = feature['identifier']
                name = f'{item["sourceClipFilename"]}-{item["sourceTrackId"]}-{item["linkedActor"]}-{front_or_back}.png'
                gc.downloadItem(feature['girderId'], f'{input_path}/{index}/{user_login}', name)
                feature['filename'] = f'./{index}/{user_login}/{name}'
            if index not in linking_map.keys():
                linking_map[index] = []
                splitActor = item["linkedActor"].split('-')
                actor_info = gc.sendRestRequest('GET', f'meva_database/get_actor_chip?actorId={splitActor[0]}&datestr={splitActor[1]}&outfitId={splitActor[2]}')
                if 'back' in actor_info.keys():
                    gc.downloadItem(actor_info['back'], f'{input_path}/{index}', f'{item["linkedActor"]}-back.jpg')
                if 'front' in actor_info.keys():
                    gc.downloadItem(actor_info['front'], f'{input_path}/{index}', f'{item["linkedActor"]}-front.jpg')
            linking_map[index].append(item)

    # now we just write the json _file
    baseJson = {
        'colorScale': colorScale,
        'links': linking_map
    }
            
    with open(f'{input_path}/feature_links.json', 'w') as outfile:
        json.dump(baseJson, outfile)

    zipf = ZipFile('/tmp/FeatureLinks.zip', 'w', ZIP_DEFLATED)
    zipdir(f'{input_path}', zipf)
    zipf.close()

    uploaded = gc.uploadFileToFolder(str(baseLinkingFolder['_id']), '/tmp/FeatureLinks.zip')
    gc.addMetadataToItem(uploaded["itemId"], {'FeatureExtractionZip': True})
    cleanup()
    os.remove('/tmp/FeatureLinks.zip')


@app.task(bind=True, acks_late=True)
def create_linking_list(self: Task, folder, limit=20, totalActors=20):
    input_path = Path(tempfile.mkdtemp())
    folderId = str(folder['_id'])
    # write the json to a file

    def cleanup():
        shutil.rmtree(input_path, ignore_errors=True)

    gc: GirderClient = self.girder_client
    manager: JobManager = self.job_manager
    # Get Actors.json
    links = gc.sendRestRequest('GET', 'meva_database/get_linked_tracks?offset=0&pageSize=9999&confirmed=false&names=true')
    print(links)
    with open(f'{input_path}/links.json', 'w') as outfile:
        json.dump(links, outfile)
    actors = gc.sendRestRequest('GET', 'meva_database/get_actor_chips?offset=0&pageSize=9999')
    print(actors)
    with open(f'{input_path}/actors.json', 'w') as outfile:
        json.dump(actors, outfile)

    process_err_file = tempfile.TemporaryFile()
    command = [
        f"cd {str(input_path)} && " "python3",
        "/home/meva_girder/meva_tasks/CreateLinkingList.py",
        f'{str(input_path)}/actors.json',
        f'{str(input_path)}/links.json',
        str(limit),
        str(totalActors),
    ]
    cmd = " ".join(command)
    manager.write(f"Running command: {cmd}\n", forceFlush=True)
    process = Popen(
        cmd,
        stdout=subprocess.PIPE,
        stderr=process_err_file,
        shell=True,
        executable='/bin/bash',
    )
    process_err_file = tempfile.TemporaryFile()

    stream_subprocess(process, self, manager, process_err_file)
    if self.canceled:
        return
    manager.updateStatus(JobStatus.PUSHING_OUTPUT)
    # now we grab the input_folder/linkingList.json and upload it
    file_obj = open(input_path / 'linkingList.json', "r")
    linkingList = json.loads(file_obj.read())
    data = {'params': linkingList}
    gc.sendRestRequest('POST', f'linked_feature/{folderId}/linkinglist', json=data )

    cleanup()
