import re
import shutil
import signal
from datetime import datetime
from pathlib import Path
from subprocess import Popen
from tempfile import mktemp
from typing import IO, Callable, Optional, Any, Dict, List, Union

from girder_client import GirderClient
from girder_worker.task import Task
from girder_worker.utils import JobManager, JobStatus


def stream_subprocess(
    process: Popen,
    task: Task,
    manager: JobManager,
    stderr_file: IO[bytes],
    keep_stdout: bool = False,
    cleanup: Optional[Callable] = None,
) -> str:
    """
    Stream live results from process to job manager

    :param process: Process to stream
    :param task: task to detect cancelation
    :param manager: job manager
    :param stderr_file: will log stderr to manager IF nonzero exit, else will close
    :param keep_stdout: will return stdout as a string if needed
    :param cleanup: a function to invoke if job errors or is canceled
    """
    start_time = datetime.now()
    stdout = ""

    if process.stdout is None:
        raise RuntimeError("Stdout must not be none")

    # call readline until it returns empty bytes
    base_count = -1
    for line in iter(process.stdout.readline, b''):
        line_str = line.decode('utf-8')
        if 'percent complete:' in line_str:
            match = re.search(r'percent complete\: ([0-9]+.[0-9])', line_str, re.I)
            if match:
                progress = float(match.group(1))
                manager.updateProgress(
                    total=100, current=progress, message="Updated", forceFlush=True
                )
        else:
            manager.write(line_str)
        if 'Rendering remaining' in line_str:
            match = re.search(r'Rendering remaining ([0-9]+) images', line_str, re.I)
            if match:
                if base_count == -1:
                    base_count = float(match.group(1))
                if base_count == 0:
                    base_count = 1
                progress = ((base_count - float(match.group(1))) / base_count) * 100
                manager.updateProgress(
                    total=100, current=progress, message="Updated", forceFlush=True
                )
        if keep_stdout:
            stdout += line_str
        if task.canceled:
            # Can never be sure what signal a process will respond to.
            process.send_signal(signal.SIGTERM)
            process.send_signal(signal.SIGKILL)
            break

    # flush logs
    manager._flush()
    # Wait for exit up to 30 seconds after kill
    code = process.wait(30)

    if task.canceled:
        manager.write('\nCanceled during subprocess run.\n')
        manager.updateStatus(JobStatus.CANCELED)
        stderr_file.close()
        if cleanup:
            cleanup()
        return stdout

    if code > 0:
        stderr_file.seek(0)
        stderr = stderr_file.read().decode()
        stderr_file.close()
        if cleanup:
            cleanup()
        raise RuntimeError(
            'Pipeline exited with nonzero status code {}: {}'.format(
                process.returncode, stderr
            )
        )
    else:
        end_time = datetime.now()
        manager.write(f"\nProcess completed in {str((end_time - start_time))}\n")

    stderr_file.close()

    return stdout


def get_video_filename(folderId: str, girder_client: GirderClient) -> Optional[str]:
    """
    Searches a folderId for videos that are compatible with training/pipelines

    * look for {"codec": 'h264', "source_video": False | None }, a transcoded video
    * then fall back to {"source_video": True}, the user uploaded video
    * If neither found it will return None

    :folderId: Current path to where the items sit
    :girder_client: girder_client used to request the data
    """
    folder_contents = girder_client.listItem(folderId)
    backup_converted_file = None
    for item in folder_contents:
        file_name = item.get("name")
        meta = item.get("meta", {})
        if meta.get("source_video") is True:
            backup_converted_file = file_name
        elif meta.get("codec") == "h264":
            return file_name
    return backup_converted_file


def fromMeta(obj, key: str, default=None, required=False) -> Any:
    """Safely get a property from girder metadata"""
    if not required:
        return obj.get("meta", {}).get(key, default)
    else:
        return obj["meta"][key]


def download_source_media(girder_client: GirderClient, folder, dest: Path) -> List[str]:
    """
    Download source media for folder from girder
    """
    if fromMeta(folder, 'type') == 'image-sequence':
        image_items = girder_client.get(
            'meva/valid_images', {'folderId': folder["_id"]}
        )
        for item in image_items:
            girder_client.downloadItem(str(item["_id"]), str(dest))
        return [str(dest / item['name']) for item in image_items]
    elif fromMeta(folder, 'type') == 'video':
        clip_meta = girder_client.get(
            "meva_detection/clip_meta", {'folderId': folder['_id']}
        )
        destination_path = str(dest / clip_meta['video']['name'])
        girder_client.downloadFile(str(clip_meta['video']['_id']), destination_path)
        return [destination_path]
    else:
        raise Exception(f"unexpected folder {str(folder)}")
